"""
A module containing functions for parsing the content of a commandline file.


This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: cmdparser.py 1071 2008-05-25 18:58:08Z ahuesgen $
"""

import re, os, shlex

ESCAPE_UNQUOTED = ["\\", "\"", "'", " ", "\r", "\n", "\t", "#"]
ESCAPE_QUOTED_DOUBLE = ['"']

def escChars(string, chars):
    """
    Escapes the all given characters in the string by prepending a
    backslash to them.
    """
    for char in chars:
        string = string.replace(char, "\\" + char)
    return string

def escape(string, rule):
    """
    Escapes all special characters in the given string.
    """
    return escChars(string, rule)

def replaceVar(varName, escaperule, env):
    """
    Searches for a substitution for the given environment variable in the
    passed environment and escapes the substitution. If no entry is found in
    the environment, an empty string is used for substitution.
    """
    if varName in env:
        substitution = env[varName]
    else:
        substitution = ""
    return escape(substitution, escaperule)


def findNextUnescapedChar(cmdline, char):
    """
    Returns the pos of the character behind the match.
    """
    regStr=r"(?P<pre>\\*)%c" % char
    regex = re.compile(regStr)
    found = False
    pos = 0
    match = regex.search(cmdline, pos)
    while match != None:
        pre = match.group("pre")
        if len(pre) % 2 == 0:
            return match.end()
        pos = match.end()
        match = regex.search(cmdline, pos)
    # not found... return pos after string
    return len(cmdline)+1

def findNextNewline(cmdline):
    pos1 = cmdline.find("\n")
    pos2 = cmdline.find("\r")
    if pos1 == -1 and pos2 == -1:
        return len(cmdline)
    elif pos1 == -1:
        return pos2 + 1
    elif pos2 == -1:
        return pos1 + 1
    else:
        return min(pos1, pos2) + 1

def splitInQuotedParts(cmdline):
    parts = []
    while cmdline != "":
        pos1a = findNextUnescapedChar(cmdline, "'")
        pos1b = findNextUnescapedChar(cmdline, '"')
        pos1c = findNextUnescapedChar(cmdline, "#")
        pos1 = min(pos1a, pos1b, pos1c)
        if pos1 > len(cmdline): # no token found
            # add rest of string
            parts.append(("NOT QUOTED", cmdline))
            cmdline = ""
        else:
            # add part of string preceeding the found token
            parts.append(("NOT QUOTED", cmdline[:pos1]))
            cmdline = cmdline[pos1:]
            if cmdline != "":
                if pos1 == pos1c: # comment found
                    pos2 = findNextNewline(cmdline)
                    parts.append(("NOT QUOTED", cmdline[:pos2]))
                    cmdline = cmdline[pos2:]
                else: # start of quotation found
                    if pos1 == pos1a: # ' found
                        char = "'"
                        # no escaping in '' strings
                        pos2 = cmdline.find("'") + 1
                        if pos2 == 0: # not found
                            pos2 == len(cmdline) + 1
                    else: # " found
                        char = '"'
                        # escaping in "" strings
                        pos2 = findNextUnescapedChar(cmdline, char)
                    if pos2 > len(cmdline): # incomplete commandline
                        raise ValueError, "No closing quotation"
                    parts.append(("QUOTED %c" % char, cmdline[:pos2]))
                    cmdline = cmdline[pos2:]
    return parts


def parseCommandlineForVars(cmdline, escaperule, env=None):
    """
    Parses a string containing a commandline for all occurances of environment
    variables (e.g. $VAR or ${VAR}) and replaces them by the values in the
    given environment. If no environment is passed, the system environment is
    used.  If a variable is used in the commandline but not declared in the
    environment, it is substituted by an empty string.
    """
    # generate regex. the same regex appeares 2 times, once with curly braces
    # (${FOO}), once without ($FOO). Group names are pre1, var1 (curly) or
    # pre2, var2 (non-curly) resp. Only one version will match!
    regexMask = r"((?P<pre%d>\\*)\$%s(?P<var%d>[A-Za-z_][A-Za-z0-9_]*)%s)"
    regex = (regexMask + "|" + regexMask) % (1, "\\{", 1, "\\}", 2, "", 2, "")
    
    varRegex = re.compile(regex)
    if env == None:
        env = os.environ
    match = varRegex.search(cmdline)
    newCmdline = ""
    lastPos = 0
    while match != None:
        # add non-variable part before match to commandline
        newCmdline += cmdline[lastPos:match.start()]
        lastPos = match.end()
        # either pre1/var1 or pre2/var2 are != None => select this one
        pre = match.group("pre1")
        if pre == None: pre = match.group("pre2")
        var = match.group("var1")
        if var == None: var = match.group("var2")
        # num(/) before variable odd => escaped, else => not escaped
        if len(pre) % 2 == 0:
            # var found
            newCmdline += pre
            newCmdline += replaceVar(var, escaperule, env)
        else:
            # not a var
            newCmdline += match.group()
        match = varRegex.search(cmdline, lastPos)
    newCmdline += cmdline[lastPos:]
    return newCmdline

def handleEnvVars(cmdline, env=None):
    parts = splitInQuotedParts(cmdline)
    newCmdline = ""
    escaperules = {"NOT QUOTED": ESCAPE_UNQUOTED,
            "QUOTED \"": ESCAPE_QUOTED_DOUBLE}
    for part in parts:
        if part[0] in escaperules:
            rule = escaperules[part[0]]
            newCmdline += parseCommandlineForVars(part[1], rule, env)
        else: # '' quoted string do not support envvars
            newCmdline += part[1]
    return newCmdline

def splitCommandline(cmdline, parseForVars = False, env=None):
    """
    Splits a string containing a commandline into single entries. If requested,
    environment variables are substituted by their contents first.
    """
    if parseForVars:
        cmdline = handleEnvVars(cmdline, env)
    return shlex.split(cmdline, comments=True)


def test():
    os.environ["COMMENT"] = "#A Comment???"
    oldCmdline = r"HELLO \${NOVAR} \"${TERM}\"\\${SHELL} $COMMENT $PWD $EMPTY '$FOO' $TERM"
    print "oldCmdline was:\n%s" % oldCmdline
    print ""
    cmdline = splitCommandline(oldCmdline, True)

    print "Tokens found:"
    for token in cmdline:
        print token

#test()

