import sys
import struct

"""
sources:
  http://wiki.tcl.tk/757
  http://en.wikipedia.org/wiki/JPEG
  http://de.wikipedia.org/wiki/JPEG_File_Interchange_Format
  http://www.64lines.com/jpeg-width-height
  http://nothings.org/
  http://www.jpeg.org/public/jfif.pdf

JPEG and JFIF files start with the "START OF IMAGE" marker 0xffd8,
followed by a number of blocks. Each block starts with a 2-bytes marker
(i.e. 0xffXY) and a 2-bytes big-endian length field. The length is
measured in bytes and comprises the length field, but not the block marker.
The "START OF FRAME" block (several variations, see below) starts with
the precision, number of components and image dimensions. That's the
information we need.
"""

SOI = "\xff\xd8" # start of image
JFIF = "\xff\xe0" # JFIF tag

# the following start of frame markers have been taken from
# http://de.wikipedia.org/wiki/JPEG_File_Interchange_Format
SOFs = ["\xff\xc0", # Baseline DCT
    "\xff\xc1", # Extended sequential DCT, NOT TESTED
    "\xff\xc2", # Progressive DCT
    "\xff\xc3", # Lossless (sequential), NOT TESTED
    "\xff\xc5", # differential sequential DCT, NOT TESTED
    "\xff\xc6", # differential progressive DCT, NOT TESTED
    "\xff\xc7", # differential lossless (sequential), NOT TESTED
    "\xff\xc9", # extended sequential DCT, NOT TESTED
    "\xff\xca", # Progressive DCT, NOT TESTED
    "\xff\xcb", # Lossless (sequential), NOT TESTED
    "\xff\xcc", # differential sequential DCT, NOT TESTED
    "\xff\xcd", # differential progressive DCT, NOT TESTED
    "\xff\xce" # differential lossless (sequential), NOT TESTED
]


class JpegError(Exception):
    pass


def _readSOI(f, filename):
    content = f.read(2)
    if content != SOI:
        raise JpegError("File '%s' is no JPEG image. The file does not start "
                "with the JPEG 'Start of Image'-marker" % filename)
    return


def _readMarker(f, filename, pos):
    marker = f.read(2)
    if len(marker) != 2 or marker[0] != "\xff":
        raise JpegError("Could not read JPEG image '%s'. No valid marker "
                "found at position %d." % (filename, pos))
    return marker


def _readLength(f, filename):
    lengthBytes = f.read(2)
    if len(lengthBytes) != 2:
        raise JpegError("Could not read JPEG image '%s'. Unexpected end of "
                "file." % filename)
    length = struct.unpack(">H", lengthBytes)[0]
    
    return length, lengthBytes


def _readBlockContents(f, filename, lengthBytes):
    length = struct.unpack(">H", lengthBytes)[0]
    block = lengthBytes + f.read(length-2)
    if len(block) != length:
        raise JpegError("Could not read JPEG image '%s'. Unexpected end of "
                "file." % filename)
    
    return block


def _readBlock(f, filename, pos):
    marker = _readMarker(f, filename, pos)
    length, lengthBytes = _readLength(f, filename)
    block = _readBlockContents(f, filename, lengthBytes)
    
    pos += 2 + length
    return (marker, block, pos)


def _readMetadata(sofBlock, filename, pos):
    if len(sofBlock) < 8: # 2 bytes length + 6 bytes metadata (BHHB)
        errorMsg = ("Could not read JPEG image '%s'. 'Start of Frame' "
                "block (at position x) is too short to contain the needed "
                "metadata.") % filename
        raise JpegError(errorMsg)
    precision, height, width, components = struct.unpack(">BHHB",
            sofBlock[2:8])
    return dict(precision=precision, width=width, height=height,
            components=components)


def readJpegMetadata(filename):
    f = open(filename, "rb")
    try:
        _readSOI(f, filename)
        p = 2
        marker = None
        while marker not in SOFs:
            marker, block, p = _readBlock(f, filename, p)
    
        return _readMetadata(block, filename, p)
    finally:
        f.close()


if __name__ == "__main__":
    metadata = readJpegMetadata(sys.argv[1])
    print metadata

