"""
Allows parsing of strings for formating and layout tokens.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: layoutlexer.py 1071 2008-05-25 18:58:08Z ahuesgen $
"""
import re
import lexer

def __genToken(match, id):
    token = {}
    token["ID"] = id
    return token

def __genMarkerToken(match, id):
    token = __genToken(match, id)
    token["VALUE"] = int(match.group("id"))
    return token

def __genTextTokenMarkers(match, id):
    token = __genToken(match, id)
    text = match.group(0)
    token["TEXT"] = text.replace("\\\\", "\\").replace("\\*", "*").replace("\\_", "_").replace("\\<", "<").replace("\\=", "=")
    return token

def __genTextTokenNoMarkers(match, id):
    token = __genToken(match, id)
    text = match.group(0)
    token["TEXT"] = text.replace("\\\\", "\\").replace("\\*", "*").replace("\\_", "_").replace("\\=", "=")
    return token

def __genFormatToken(match, id):
    token = __genToken(match, id)
    token["TYPE"] = match.group(0)
    return token

# format tokens (* and _)
ID_FORMAT = 1
# normal text
ID_TEXT = 2
# a newline
ID_NEWLINE = 3
# a marker <id:123> which marks a position within the text
ID_MARKER = 4

__LEXEXPRS_COMMON = [("\n", ID_NEWLINE, __genToken),
                    ("[_*=]", ID_FORMAT, __genFormatToken)]


# lexexpressions for lexer with marker support
LEXEXPRS_MARKERS = [("([^*\\\\_=<\n]|(\\\\\\*)|(\\\\_)|(\\\\=)|(\\\\<)|(\\\\\\\\))+", ID_TEXT, __genTextTokenMarkers),
                     ("<id:(?P<id>[0-9]+)>", ID_MARKER, __genMarkerToken)] \
                     + __LEXEXPRS_COMMON

# lexexpressions for lexer without marker support
LEXEXPRS_NO_MARKERS =[("([^*\\\\_=\n]|(\\\\\\*)|(\\\\_)|(\\\\=)|(\\\\\\\\))+", ID_TEXT, __genTextTokenNoMarkers)] \
        + __LEXEXPRS_COMMON


class LayoutLexer(lexer.Lexer):
    def __init__(self, supportMarkers):
        lexer.Lexer.__init__(self)
        if supportMarkers:
            expressions = LEXEXPRS_MARKERS
        else:
            expressions = LEXEXPRS_NO_MARKERS
        
        for expr in expressions:
            cregex = re.compile(expr[0], re.DOTALL)
            lexExpr = lexer.LexExpr(cregex, False, expr[2], expr[1])
            self.addLexExpr(lexExpr)
