"""
Module containing a variety of specialized PDF objects.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: pdfadvobjs.py 2689 2013-04-17 10:41:53Z ahuesgen $
"""

import pdfobjs, pdfutils, pdftypes, vhisttags, pdflayouter, layoutitem
import re

def validateObj(assertion, msg):
    if not assertion:
        raise pdfobjs.PDFObjectError, msg

class PDFFileSpecObject(pdfobjs.PDFStructObject):
    """
    A pdf object for a file specification.
    
    """
    def __init__(self, id, gen, filename, embeddedFile, desc = ""):
        """
        Creates a filespecification object.
        
        @type  id: int
        @param id: The id of the pdf object.
        
        @type  gen: int
        @param gen: The generation number of the pdf object.
        
        @type  filename: str
        @param filename: The name of the file, the filespec refers to. This
                         name must not correspond to any real filename.
        
        @type  embeddedFile: pdftypes.PDFRef
        @param embeddedFile: A reference to the embedded file object.
        
        @type  desc: str
        @param desc: A description of the file.
        
        """
        pdfobjs.PDFStructObject.__init__(self, id, gen)
        self.set("/Type", "/Filespec")
        self.set("/F", pdfutils.makePDFStringUTF8(filename))
        self.set("/Desc", pdfutils.makePDFString(desc))
        self.set("/EF/F", embeddedFile)


class PDFEmbeddedFileObject(pdfobjs.PDFStructObject):
    """
    A pdf object for an embedded file.
    
    """
    def __init__(self, id, gen, filename, type, desc, comment, content,
                compression, returnobj, md5 = None, cmd5 = None):
        """
        Creates a new embedded file object including a file stream and a
        vhist tag.
        
        """
        pdfobjs.PDFStructObject.__init__(self, id, gen)
        
        offsetPre = len("\nstream\n")
        offsetPost = len("\nendstream\n%")
        
        self.tag = vhisttags.EmbeddedFileTag(filename, type, desc, comment,
                content, compression, offsetPre, offsetPost, md5, cmd5)
        
        self.set("/Type", "/EmbeddedFile")
        self.set("/Length", self.tag.getStreamSize())
        if compression not in ("none", "flate"):
            raise pdfobjs.PDFObjectError, "invalid compression: " + compression
        if compression == "flate":
            self.set("/Filter", "/FlateDecode")
            returnobj.compressedSize = self.tag.getCompressedSize()
            returnobj.compressedMD5 = self.tag.getCompressedMD5()
        else:
            returnobj.compressedSize = None
            returnobj.compressedMD5 = None
        
        self.set("/Params/Size", self.tag.uncompressedSize)
        self.setAppendix("".join(["%", self.tag.getBeginTag(), "\nstream\n",
                self.tag.getCompressedContent(), "\nendstream\n%",
                self.tag.getEndTag(), "\n"]))


class PDFInfoObject(pdfobjs.PDFStructObject):
    def __init__(self, id, gen = 0, title = "", author = "", subject = "", \
            creator = "", keywords = "", producer = ""):
        
        makePDFString = pdfutils.makePDFString
        pdfobjs.PDFStructObject.__init__(self, id, gen)
        self.root = {}
        title    == "" or self.set("/Title", makePDFString(title))
        author   == "" or self.set("/Author", makePDFString(author))
        subject  == "" or self.set("/Subject", makePDFString(subject))
        creator  == "" or self.set("/Creator", makePDFString(creator))
        keywords == "" or self.set("/Keywords", makePDFString(keywords))
        producer == "" or self.set("/Producer", makePDFString(producer))


class PDFPagesObject(pdfobjs.PDFStructObject):
    def __init__(self, id, gen = 0):
        pdfobjs.PDFStructObject.__init__(self, id, gen)
        self.set("/Type", "/Pages")
        self.set("/Kids", [])
        self.set("/Count", 0)
    
    def addPage(self, page):
        self.set("/Kids[*]", page)
        self.set("/Count", len(self.get("/Kids")))
    
    def fromStructObject(self, structObj):
        validateObj(structObj.exists("/Type"),
                "Pages object has no /Type field")
        validateObj(structObj.exists("/Kids"),
                "Pages object has no /Kids field")
        validateObj(structObj.exists("/Count"),
                "Pages object has no /Count field")
        typef = structObj.get("/Type")
        kidsf = structObj.get("/Kids")
        countf = structObj.get("/Count")
        validateObj(typef == "/Pages",
                "Pages object's /Type is not /Pages, but %s" % typef)
        validateObj(len(kidsf) == countf,
                "Pages object's /Count does not match length of /Kids")
        # all tests passed, object is used
        self.root = structObj.root


class PDFPageObject(pdfobjs.PDFStructObject):
    def __init__(self, id, gen, parent, resources, contents = None):
        pdfobjs.PDFStructObject.__init__(self, id, gen)
        self.set("/Type", "/Page")
        self.set("/Parent", parent)
        self.set("/Resources", resources)
        # /Group is needed. otherwise, transparency generates weird artifacts
        self.set("/Group/Type", "/Group")
        self.set("/Group/S", "/Transparency")
        self.set("/Group/CS", "/DeviceRGB")
        self.set("/MediaBox", [0, 0, 595, 842])
        self.set("/Annots", [])
        if contents != None:
            self.set("/Contents", contents)
    
    def addAnnot(self, annot):
        self.set("/Annots[*]", annot)


class PDFPageLabelsObject(pdfobjs.PDFStructObject):
    def __init__(self, id, gen):
        pdfobjs.PDFStructObject.__init__(self, id, gen)
        self.set("/Nums", [])
    
    def addLabel(self, index, title, numbering=True):
        self.set("/Nums[*]", index)
        num = len(self.get("/Nums"))
        self.set("/Nums[%d]/P" % num, pdfutils.makePDFString(title))
        if numbering:
            self.set("/Nums[%d]/S" % num, "/D")
    
    def fromStructObject(self, structObj):
        validateObj(structObj.exists("/Nums"),
                "Pagelabels object has no /Nums field")
        # all tests passed, object is used
        self.root = structObj.root


class PDFOutlinesObject(pdfobjs.PDFStructObject):
    def __init__(self, id, gen):
        pdfobjs.PDFStructObject.__init__(self, id, gen)
        self.set("/Type", "/Outlines")
        self.set("/Count", 0)
    
    def addOutlineObject(self, obj):
        if self.exists("/Last"):
            last = self.get("/Last")
            last.setNext(obj)
            obj.setPrev(last)
            self.set("/Last", obj)
            self.set("/Count", self.get("/Count") + 1)
        else:
            # empty outlines
            self.set("/Last", obj)
            self.set("/First", obj)
            self.set("/Count", 1)
    
    def getLastObjectID(structObj):
        if structObj.exists("/Last"):
            return structObj.get("/Last")
        else:
            return None
    
    getLastObjectID = staticmethod(getLastObjectID)
    
    def fromStructObject(self, structObj, lastObj):
        validateObj(structObj.exists("/Type"),
                "Outlines object has no /Type field")
        typef = structObj.get("/Type")
        validateObj(typef == "/Outlines",
                "Outlines object's /Type is not /Outlines, but %s" % typef)
        validateObj(structObj.exists("/First") == structObj.exists("/Last"),
                "Outlines object only contains one of {/First, /Last}")
        # all tests passed, object is used
        if lastObj != None:
            structObj.set("/Last", lastObj)
        self.root = structObj.root

class PDFOutlineObject(pdfobjs.PDFStructObject):
    def __init__(self, id, gen, title, parent, dest):
        pdfobjs.PDFStructObject.__init__(self, id, gen)
        self.set("/Title", pdfutils.makePDFString(title))
        self.set("/Parent", parent)
        self.set("/Dest[0]", dest)
        self.set("/Dest[1]", "/FitH")
        #TODO: make position configurable
        self.set("/Dest[2]", 842)
    
    def setNext(self, obj):
        self.set("/Next", obj)
    
    def setPrev(self, obj):
        self.set("/Prev", obj)
    
    def fromStructObject(self, structObj):
        validateObj(structObj.exists("/Parent"),
                "Outline Object has no parent")
        # all tests passed, object is used
        self.root = structObj.root


class PDFAnnotObject(pdfobjs.PDFStructObject):
    def __init__(self, id, gen, filespec, parent, subj, contents, rect, author,
                appearance):
        pdfobjs.PDFStructObject.__init__(self, id, gen)
        self.set("/Type", "/Annot")
        self.set("/Subtype", "/FileAttachment")
        self.set("/P", parent)
        self.set("/Subj", pdfutils.makePDFString(subj))
        self.set("/FS", filespec)
        self.set("/T", pdfutils.makePDFString(author))
        for i in range(0, 4):
            self.set("/Rect[%d]" % i, rect[i])
        self.set("/Contents", pdfutils.makePDFString(contents))
        self.set("/AP/N", appearance[0])
        self.set("/AP/R", appearance[1])
        self.set("/AP/D", appearance[2])


class PDFPaperclipObject(pdfobjs.PDFStructObject):
    def __init__(self, id, gen, filename, parent, filespec, title, line,
            paperclips):
        pdfobjs.PDFStructObject.__init__(self, id, gen)
        self.set("/Type", "/Annot")
        self.set("/Subtype", "/FileAttachment")
        self.set("/Parent", parent)
        self.set("/F", 0) # set no flag
        self.set("/FS", filespec)
        self.set("/Contents", pdfutils.makePDFString("file: " + filename))
        self.set("/T", pdfutils.makePDFString(title))
        dimensions = [20, 775 - line * 11, 40, 755 - line * 11]
        self.set("/Rect", dimensions)
        self.set("/AP/N", paperclips[0])
        self.set("/AP/R", paperclips[1])
        self.set("/AP/D", paperclips[2])


class PDFContentObject(pdfobjs.PDFObject):
    def __init__(self, id, gen = 0):
        pdfobjs.PDFObject.__init__(self, id, gen)
        self.__texts = []
        self.__rawStream = ""
    
    def appendRawStream(self, rawData):
        self.__rawStream += rawData
    
    def addLine(self, x1_pt, y1_pt, x2_pt, y2_pt, width_pt):
        szl = "%% line begin\n%f %f m\n%f %f l\n%f w\nS\n%% line end\n" \
                % (x1_pt, y1_pt, x2_pt, y2_pt, width_pt)
        self.__rawStream += szl
    
    def addText(self, x, y, textStream, font = "/F1", size = 20):
        self.__texts.append((x, y, textStream, font, size))
    
    def addImage(self, x, y, width, height, imageName):
        rawData = "q\n1 0 0 1 %f %f cm\n%f 0 0 %f 0 0 cm\n%s Do\nQ"
        rawData = rawData % (x, y, width, height, imageName)
        self.appendRawStream(rawData)
    
    def addLayoutItems(self, layoutItems):
        for litem in layoutItems:
            x = litem.x
            y = litem.y
            item = litem.item
            
            if isinstance(item, layoutitem.Graphic):
                self.addImage(x, y, item.width, item.height, item.pdfname)
            else:
                item.addImage(x, y, item)

    def genStream(self):
        streamList = []
        for textblock in self.__texts:
            x = textblock[0]
            y = textblock[1]
            textStream = textblock[2]
            fonts = {None: textblock[3], "*": "/F3", "_": "/F4", "=": textblock[3]}
            blue = "0.30 0.30 0.86 rg"
            black = "0 0 0 rg"
            colors = {None: black, "*": black, "_": black, "=": blue}
            size = textblock[4]
            
            # stores the single parts of the stream string (assembled later)
            streamList.append("BT")
            streamList.append("%d %d Td" % (x, y))
            streamList.append("%d TL" % (size + 1))
            for token in textStream:
                if token["ID"] == pdflayouter.ID_TEXT:
                    font = fonts[token["FORMAT"]]
                    color = colors[token["FORMAT"]]
                    streamList.append(color)
                    streamList.append("%s %d Tf" % (font, size))
                    pdfString = pdfutils.makePDFString(token["TEXT"])
                    streamList.append("%s Tj" % pdfString)
                elif token["ID"] == pdflayouter.ID_NEWLINE:
                    y += size
                    streamList.append("T*")
                else:
                    raise Exception, "Unreachable passage reached"
            streamList.append("ET")
        stream = self.__rawStream + "\n".join(streamList)
        return stream
    
    def toStr(self):
        stream = self.genStream()
        objectBody = "<<\n   /Length %d\n>>\n" % len(stream)
        objectBody += "stream\n%s\nendstream\n" % stream
        return self.getObjectString(objectBody)


#obj = PDFEmbeddedFileObject(1,0, "lexer.py", "text/python", "python script",
#        "just a test", "none")
#print obj.toStr()

"""
ref1 = pdftypes.PDFRef(1, 0)
ref2 = pdftypes.PDFRef(2, 0)
ref3 = pdftypes.PDFRef(3, 0)

obj = PDFPageObject(10, 0, ref1, ref2, ref3)
annot = PDFAnnotObject(31, 0, ref1, obj, "Hello World", "file: FOO",
        (1, 2, 3, 4), "author", (ref1, ref2, ref3))
obj.addAnnot(annot)

print obj.toStr()
print annot.toStr()

obj2 = PDFPageLabelsObject(1, 0)
obj2.addLabel(0, "Titlepage", False)
obj2.addLabel(1, "Section 1")

print obj2.toStr()


outl = PDFOutlinesObject(123, 0)
o1 = PDFOutlineObject(124, 0, "Foo", outl, ref1)
o2 = PDFOutlineObject(125, 0, "Bar", outl, ref2)

outl.addOutlineObject(o1)
outl.addOutlineObject(o2)

print outl.toStr()
print o1.toStr()
print o2.toStr()


"""
