"""
Module containing a parser for parsing (uncompressed) pdf objects.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: pdfparser.py 2486 2012-04-23 09:50:08Z ahuesgen $
"""

from pdflexer import ID_BOOL, ID_NULL, ID_ARRAY_BEGIN, ID_ARRAY_END, \
ID_DICT_BEGIN, ID_DICT_END, ID_NUMBER, ID_NAME, ID_HEXSTR, ID_LITSTR, \
ID_STREAM, ID_OBJ, ID_REF, ID_ENDOBJ

import lexer, pdflexer, pdfobjs

class ParserError(RuntimeError):
    """
    Exception, which is thrown if an error occures during parsing.
    
    """
    pass



class TokenStream(object):
    """
    A stream containing tokens generated by a pdflexer.
    """
    def __init__(self, tokens):
        """
        Creates a new token stream.
        
        @type  tokens: list of tokens
        @param tokens: a list containing any number of tokens.
        
        """
        self.__tokens = tokens
        self.resetStream()
    
    def getNextToken(self):
        """
        Returns the next token in the tokenstream. If the stream is empty or no
        token is left in the stream, None is returned.
        
        @rtype:  token or None
        @return: the next token in the stream.
        
        """
        if self.tokensLeft():
            curToken = self.__tokens[self.__pos]
            self.__pos += 1
            return curToken
        else:
            return None
    
    def undoToken(self):
        """
        Reinserts the last removed token to the tokenstream.
        """
        if self.__pos > 0:
            self.__pos -= 1
    
    def getStream(self):
        """
        Returns the list containing the tokenstream.
        
        @rtype:  list of tokens
        @return: the tokenstream
        
        """
        return self.__tokens
    
    def getPos(self):
        """
        Returns the number of tokens removed from the tokenstream.
        
        @rtype: int
        @return: The current position within the tokenstream.
        
        """
        return self.__pos
    
    def resetStream(self):
        """
        Resets the stream so that a call to getStream will return the first
        token in the stream. The tokenstream itself is kept unchanged.
        
        """
        self.__pos = 0
    
    def tokensLeft(self):
        """
        Returns whether or not there are tokens left in the stream.
        
        @rtype:  bool
        @return: True if there are still tokens left in the tokenstream and
                 False otherwise.
        
        """
        return self.__pos < len(self.__tokens)
    
    def isEmpty(self):
        """
        Returns whether or not the tokenstream contains any tokens.
        
        @rtype:  bool
        @return: True if the stream contains tokens and False otherwise.
        
        """
        return len(self.__tokens) == 0




class PDFParser:
    """
    A Parser for creating pdf objects from a string.
    
    """
    
    __errorParse = "Parser error at position %d: %s"
    __errorLex   = "Lexer error after position %d: %s"
    __errorWrongStart = "Wrong token found at beginning of tokenstream. " + \
            "x y obj or << expected but %s found."
    __errorEmpty = "Tokenstream is empty."
    __errorWrongToken = "Wrong token ('%s') found." 
    __errorEnd = "Unexpected end of token stream"
    __errorEndExpected = "'endobj' or 'stream' expected but '%s' found"
    
    def __init__(self):
        """
        Creates a new PDFParser.
        
        """
        self.lex = pdflexer.PDFLexer()
    
    def parsePDF(self, string, byteOffset = 0):
        """
        Parses the given string and returns the created object.
        
        @type  string: str
        @param string: The string, which is parsed.
        
        @type  byteOffset: int
        @param byteOffset: An optional offset, which is added to position
                           information in exception messages. This is useful
                           in case that the given string is only a part a
                           longer string or file.
        
        @raise ParserError: This exception is raised when a lexing or parsing
                            error occures.
        
        """
        try:
            # try to lex the given string
            self.lex.lexString(string)
        except lexer.LexerError, e:
            # find out at which location the error occured
            tokens = self.lex.getResult()
            if len(tokens) > 0:
                pos = tokens[-1]["ENDPOS"]
            else:
                pos = 0
            pos += byteOffset
            
            # raise exception
            raise ParserError, self.__errorLex % (pos, e)
        
        tokens = self.lex.getResult()
        stream = TokenStream(tokens)
        try:
            # look at first token
            curToken = stream.getNextToken()
            if curToken != None:
                # first token must be an object beginning or a dictionary
                # (allowed to support the trailer dictionary)
                if curToken["ID"] == ID_OBJ:
                    newObj = pdfobjs.PDFStructObject(curToken["OBJID"],
                            curToken["OBJGEN"])
                    # parse object
                    self.__parseObject(newObj, stream)
                    return newObj
                elif curToken["ID"] == ID_DICT_BEGIN:
                    # parse dictionary (create fake object and return root)
                    fakeObj = pdfobjs.PDFStructObject(1, 0)
                    self.__parseDict(fakeObj, stream, "")
                    return fakeObj
                else:
                    raise ParserError, self.__errorWrongStart % \
                            (curToken["TEXT"])
            else:
                raise ParserError, self.__errorEmpty
        except ParserError, e:
            # catch all ParseErrors and rethrow them with additional information
            if not stream.isEmpty():
                errorToken = stream.getStream()[stream.getPos()-1]
                pos = errorToken["POS"]
            else:
                pos = 0
            
            pos += byteOffset
            raise ParserError, self.__errorParse % (pos, e)
    
    def __handleNextToken(self, pdfObj, stream, path):
        """
        Gets the next token from the tokenstream and tries to add it to the
        given pdf structured object at the given path. Bools, numbers, string,
        names, null objects, references, dictionaries and arrays are handled
        by this function.
        
        @type  pdfObj: pdfobjs.PDFStructObject
        @param pdfObj: A pdf object, to which the next token should be added.
        
        @type  stream: pdfparser.TokenStream
        @param stream: A tokenstream, from which the next token is removed.
        
        @type  path: str
        @param path: A valid pdf path at which the token should be inserted
                     into the pdf object.
        
        @note: Since the method also handles dictionaries and arrays, which
               themself may contain elements, the function may remove more than
               one token from the tokenstream.
        
        """
        validValueTypes = [ID_BOOL, ID_NUMBER, ID_HEXSTR, ID_LITSTR, \
                ID_NAME, ID_NULL, ID_REF, ID_DICT_BEGIN, ID_ARRAY_BEGIN]
        
        # get and handle next token
        curToken = stream.getNextToken()
        if curToken != None:
            if curToken["ID"] in validValueTypes:
                # add tokens value to object
                pdfObj.set(path, curToken["VALUE"])
            else:
                raise ParserError, self.__errorWrongToken % curToken["TEXT"]
            
            # handle datastructures (dicts and arrays)
            if curToken["ID"] == ID_DICT_BEGIN:
                self.__parseDict(pdfObj, stream, path)
            elif curToken["ID"] == ID_ARRAY_BEGIN:
                self.__parseArray(pdfObj, stream, path)
        else:
            raise ParserError, self.__errorEnd
    
    def __EndOfObjReached(self, stream):
        """
        Checks if the next token in the given tokenstream is an endobject or
        a stream token. After examination, the token is returned to the
        tokenstream. If there are no tokens left in the stream, a
        ParserError is raised.
        
        @type  stream: pdfparser.Tokenstream
        @param stream: A tokenstream, whose next token is examind for an
        endtoken.
        
        @rtype:  bool
        @return: True if the token is an endobject or stream token and False
                 otherwise.
        
        """
        # check if the object is finished
        curToken = stream.getNextToken()
        if curToken != None:
            if curToken["ID"] in [ID_ENDOBJ, ID_STREAM]:
                return True
            else:
                stream.undoToken()
                return False
        else:
            raise ParserError, self.__errorEnd
    
    def __parseObject(self, pdfObj, stream):
        """
        Parses the tokens in a tokenstream and creates a complete pdf object
        from them. The first token containing the object id and generation
        ("x y obj") must have been removed from the stream.
        
        @type  pdfObj: pdfobjs.PDFStructObject
        @param pdfObj: A pdf structured object, into which the parsed object
                       is written.
        
        @type  stream: pdfparser.TokenStream
        @param stream: A tokenstream, which is parsed to create the object.
        
        """
        validValueTypes = [ID_BOOL, ID_NUMBER, ID_HEXSTR, ID_LITSTR, \
                ID_NAME, ID_NULL, ID_REF]
        
        # check if the object does not contain any values
        if self.__EndOfObjReached(stream):
            return
        
        # handle value for object's root
        self.__handleNextToken(pdfObj, stream, "")
        
        # the object must end here
        if self.__EndOfObjReached(stream):
            return
        else:
            raise ParserError, self.__errorEndExpected % curToken["TEXT"]
    
    def __parseDict(self, pdfObj, stream, path):
        """
        Parses the tokens in a tokenstream and creates a pdf dictionary from
        them. The first token containing the dictionary beginning ("<<") must
        have been removed from the stream.
        
        @type  pdfObj: pdfobjs.PDFStructObject
        @param pdfObj: A pdf structured object, into which the parsed
                       dictionary is written.
        
        @type  stream: pdfparser.TokenStream
        @param stream: A tokenstream, which is parsed to create the dictionary.
        
        @type  path: str
        @param path: A valid pdf path at which the dictionary should be
                     inserted into the pdf object.
        
        """
        nextIsKey = True
        curKey = None
        finished = False
        
        while not finished:
            if nextIsKey:
                curToken = stream.getNextToken()
                if curToken != None:
                    # we need a key (or the end of the dictionary)
                    if curToken["ID"] == ID_DICT_END:
                        finished = True
                    elif curToken["ID"] == ID_NAME:
                        curKey = curToken["VALUE"]
                    else:
                        raise ParserError, self.__errorWrongToken % \
                                curToken["TEXT"]
                else:
                    raise ParserError, self.__errorEnd
            else:
                # we need a value
                newPath = path + curKey
                self.__handleNextToken(pdfObj, stream, newPath)
                curKey = None
            nextIsKey = not nextIsKey
    
    def __parseArray(self, pdfObj, stream, path):
        """
        Parses the tokens in a tokenstream and creates a pdf array from
        them. The first token containing the dictionary beginning ("[") must
        have been removed from the stream.
        
        @type  pdfObj: pdfobjs.PDFStructObject
        @param pdfObj: A pdf structured object, into which the parsed
                       array is written.
        
        @type  stream: pdfparser.TokenStream
        @param stream: A tokenstream, which is parsed to create the array.
        
        @type  path: str
        @param path: A valid pdf path at which the array should be
                     inserted into the pdf object.
        
        """
        finished = False
        nextIndex = 0
        
        while not finished:
            curToken = stream.getNextToken()
            if curToken != None:
                if curToken["ID"] == ID_ARRAY_END:
                    finished = True
                else:
                    stream.undoToken()
                    newPath = path + ("[%d]" % nextIndex)
                    self.__handleNextToken(pdfObj, stream, newPath)
            else:
                raise ParserError, "Unexpected end of token stream"
            nextIndex += 1


def test():
    print "PARSER TEST"
    print ""
    
    parse = PDFParser()
    parsestrings = []
    parsestrings.append("0 1 obj.....")
    parsestrings.append("0 1 obj << /Name 123 /FOO (bar) >> endobj")
    parsestrings.append("0 100 obj << /Name [1 2 3 4] /FOO /Bar >> endobj")
    parsestrings.append("5 0 obj << /Name <</FOO [()] /Bar <<>>>> >> endobj")
    parsestrings.append("0 1 R endobj")
    parsestrings.append("0 1 obj <<  ] endobj")
    parsestrings.append("0 0 obj << [ /Test >> endobj")
    parsestrings.append("0 0 obj << /Name [")
    parsestrings.append("trailer\n << /Size 8\n /Root 1 0 R >> startxref")

    for s in parsestrings:
        result = None
        try:
            print ""
            print "SEARCHSTRING:", repr(s)
            result = parse.parsePDF(s, 0)
        except ParserError, e:
            print str(e)
        
        if isinstance(result, pdfobjs.PDFStructObject):
            print "%d %d obj: %s" % (result.id, result.gen, str(result.root))
        elif isinstance(result, dict):
            print str(result)


#test()
