"""
Module containing various utility functions for creating and working with
pdf related objects in python.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: pdfutils.py 2601 2013-03-27 13:33:00Z ahuesgen $
"""

import datetime, re
import utilities, pdftypes, logger

class ConversionError(Exception):
    pass


def makePDFBool(boolean):
    """
    Converts a python bool object into a pdf boolean object.
    
    @type  boolean: bool or str
    @param boolean: True or False or "true" or "false".
    
    @rtype:  bool
    @return: A bool object representing the given value.
    
    """
    #TODO: document exceptions
    utilities.checkType(boolean, (bool, str))
    if isinstance(boolean, bool):
        return boolean
    else:
        if boolean in ["true", "false"]:
            dic = {"true": True, "false": False}
            return dic[boolean]
        else:
            raise ConversionError, "invalid boolean value '%s'" % boolean

def makePDFNumber(number):
    """
    Converts a python integer, float or string representation of a number into
    a pdf number object.
    
    @type  number: int, float or str
    @param number: A number or a string containing a number.
    
    @rtype:  pdf usable number (int or float)
    @return: A number, which can be used as a pdf number object.
    
    """
    #TODO: document exceptions
    utilities.checkType(number,(int, float, str))
    if isinstance(number, bool):
        raise TypeError, "parameter of type bool not supported"
    if isinstance(number, (int, float)):
        return number
    else:
        try:
            try:
                return int(number)
            except ValueError:
                return float(number)
        except ValueError:
            raise ConversionError, "string '%s' does not represent a number" % \
                    number

def makePDFStringUTF8(string):
    if type(string) is unicode:
        string = string.encode("utf-8")
    return makePDFString(string)

def makePDFString(string):
    """
    Converts a python string into a pdf compatible string representation.
    
    @type  string: str encoded as latin-1 or unicode. If the string is a
                   unicode string, it is encoded as latin-1.
    @param string: A python string, which should be converted into a pdf
                   string.
    
    @rtype:  str
    @return: A pdf compatible representation of a string.
    
    """
    #TODO: add tests, add support for unicode strings, document exceptions
    if type(string) is unicode:
        try:
            string = string.encode("latin-1")
            #return toPDFString(string)
        except UnicodeEncodeError, e:
            #hexstr = toPDFHexString(string)
            #print hexstr
            #return hexstr

            logger.error("user input contains non-latin-1 characters. The" + \
                    " current version of vhist does not support characters" + \
                    " outside the latin-1 character range.")
            raise ConversionError, "could not convert character to latin-1"
    return toPDFString(string)

    

def makePDFName(string):
    """
    Converts a python string into a pdf compatible name object representation.
    
    @type string: str
    @param string: A python string, which should be converted into a pdf
                   string.
    
    @rtype:  str
    @return: A pdf compatible representation of a name object.
    
    """
    #TODO: add tests, document exceptions
    return toPDFName(string)

def makePDFArray():
    """
    Creates a pdf compatible array representation.
    
    @rtype:  list
    @return: An empty list.
    
    """
    return []

def makePDFDict():
    """
    Creates a pdf compatible dictionary representation.
    
    @rtype:  dict
    @return: An empty dictionary.
    
    """
    return {}

def makePDFNull():
    """
    Creates a pdf compatible null object.
    
    @rtype:  pdftypes.PDFNull
    @return: A null object.
    
    """
    return pdftypes.PDFNull()

def makePDFRef(id, gen):
    """
    Creates a pdf reference.
    
    @type  id: int > 0
    @param id: The id of the object, to which the reference refers.
    
    @type  gen: int >= 0
    @param gen: The generation of the object, to which the reference refers.
    
    @rtype:  pdftypes.PDFRef
    @return: A pdf reference.
    
    """
    return pdftypes.PDFRef(id, gen)

def isPDFBool(obj):
    """
    Tests if the given object is a valid pdf boolean object.
    
    @type  obj: any
    @param obj: The object, which should be tested.
    
    @rtype:  bool
    @return: True if the given object is a pdf boolean and False otherwise.
    
    """
    return isinstance(obj, bool)

def isPDFNumber(obj):
    """
    Tests if the given object is a valid pdf number.
    
    @type  obj: any
    @param obj: The object, which should be tested.
    
    @rtype:  bool
    @return: True if the given object is a pdf number and False otherwise.
    
    """
    return isinstance(obj, (int, float)) and not isinstance(obj, bool)

def isPDFString(obj):
    """
    Tests if the given object is a valid pdf string.
    
    @type  obj: any
    @param obj: The object, which should be tested.
    
    @rtype:  bool
    @return: True if the given object is a pdf string and False otherwise.
    
    @attention: performs a shallow test, only. Programming errors may lead to
                invalid objects, which pass this test.
    
    """
    #TODO: add tests, write better implementation
    if isinstance(obj, str):
        first = obj[:1]
        last  = obj[-1:]
        if (first == "(" and last == ")") or (first == "<" and last == ">"):
            return True
    return False

def isPDFName(obj):
    """
    Tests if the given object is a valid pdf name object.
    
    @type  obj: any
    @param obj: The object, which should be tested.
    
    @rtype:  bool
    @return: True if the given object is a pdf name and False otherwise.
    
    """
    try:
        normalizePDFName(obj)
    except Exception, e:
        return False
    return True

def isPDFArray(obj):
    """
    Tests if the given object is a valid pdf array.
    
    @type  obj: any
    @param obj: The object, which should be tested.
    
    @rtype:  bool
    @return: True if the given object is a pdf array and False otherwise.
    
    @attention: performs a shallow test, only. Programming errors may lead to
                invalid objects, which pass this test.
    
    """
    #TODO: add tests
    return isinstance(obj, list)

def isPDFDict(obj):
    """
    Tests if the given object is a valid pdf dictionary.
    
    @type  obj: any
    @param obj: The object, which should be tested.
    
    @rtype:  bool
    @return: True if the given object is a pdf dictionary and False otherwise.
    
    @attention: performs a shallow test, only. Programming errors may lead to
                invalid objects, which pass this test.
    
    """
    #TODO: add tests
    return isinstance(obj, dict)

def isPDFNull(obj):
    """
    Tests if the given object is a valid pdf null object.
    
    @type  obj: any
    @param obj: The object, which should be tested.
    
    @rtype:  bool
    @return: True if the given object is a pdf null object and False otherwise.
    
    """
    #TODO: add tests
    return isinstance(obj, pdftypes.PDFNull)

def isPDFRef(obj):
    """
    Tests if the given object is a valid pdf reference.
    
    @type  obj: any
    @param obj: The object, which should be tested.
    
    @rtype:  bool
    @return: True if the given object is a pdf reference and False otherwise.
    
    """
    #TODO: add tests
    return isinstance(obj, pdftypes.PDFRef)



def serializePDFBool(obj):
    """
    Creates a string from a pdf compatible boolean object. The string can be
    included into a pdf file.
    
    @type  obj: bool
    @param obj: A pdf compatible boolean object.
    
    @rtype:  str
    @return: A string representing the given object.
    
    """
    #TODO: add tests
    dic = {True: "true", False: "false"}
    return dic[obj]

def serializePDFNumber(obj):
    """
    Creates a string from a pdf compatible number object. The string can be
    included into a pdf file.
    
    @type  obj: int or float
    @param obj: A pdf compatible number object.
    
    @rtype:  str
    @return: A string representing the given object.
    
    """
    #TODO: add tests
    return str(obj)

def serializePDFString(obj):
    """
    Creates a string from a pdf compatible string object. The string can be
    included into a pdf file.
    
    @type  obj: str
    @param obj: A string, which contains a valid literal or hexadecimal
                pdf string.
    
    @rtype:  str
    @return: A string representing the given object.
    
    """
    #TODO: add tests
    return obj

def serializePDFName(obj):
    """
    Creates a string from a pdf compatible name object. The string can be
    included into a pdf file.
    
    @type  obj: str
    @param obj: A string, which contains a valid pdf name.
    
    @rtype:  str
    @return: A string representing the given object.
    
    """
    #TODO: add tests
    return obj

def serializePDFNull():
    """
    Creates the string "null". The string can be included into a pdf file as
    a null object.
    
    @rtype:  str
    @return: A string representing the given object.
    
    """
    #TODO: add tests
    return "null"

def serializePDFRef(obj):
    """
    Creates a string from a pdf compatible reference. The string can be
    included into a pdf file.
    
    @type  obj: pdftypes.PDFRef
    @param obj: A reference to a pdf object.
    
    @rtype:  str
    @return: A string representing the given object.
    
    """
    return obj.getObjectRef()




def toPDFDate(date):
    """
    Converts a datatime object into a pdf date string, which can be included
    into a pdf file.
    
    @type  date: datetime.datetime
    @param date: a datetime object with the date, which should be converted.
                             Alternative, every object with year, month, day, hour,
                             minute and second members (all ints) is valid.
    
    @rtype:  str
    @return: a string in the pdf date format representing the given date.
    
    """
    args = (date.year, date.month, date.day,
                    date.hour, date.minute, date.second)
    return "(D:%04d%02d%02d%02d%02d%02d)" % args


def toPDFString(string):
    """
    Converts a python string into a pdf string, which can be included into a
    pdf file. The characters (, ) and \ are escaped.
    
    @type  string: str
    @param string: a python string, which should be converted into a pdf
    string.
    
    @rtype:  str
    @return: a string in the pdf literal string format representing the given
                     string.
    
    """
    lookup = {"\n":  "\\n",  "\r":  "\\r", "\t":  "\\t",
            "\f":  "\\f",  "\b":  "\\b", "(":   "\\(",
            ")": "\\)", "\\": "\\\\"}
    
    pdfStr = "("
    for char in string:
        if (char in lookup.iterkeys()):
            pdfStr += lookup[char]
        else:
            pdfStr += char
    
    return pdfStr + ")"


def toPDFHexString(string):
    utf16Str = utf16ifyString(string)
    print type(string), type(utf16Str)
    hexString = ["<feff"]
    for c in utf16Str:
        hexChar = "%02x" % ord(c)
        print hexChar, c
        hexString.append(hexChar)
    hexString.append(">")
    return "".join(hexString)

def toPDFArray(list):
    """
    Converts a list of strings into a pdf array, which can be included into a
    pdf file. The strings in the array are not converted to pdf strings.
    
    @type  list: list of str
    @param list: a list of strings, which is converted into a pdf array.
    
    @rtype:  str
    @return: a string containing the represetation of a  pdfarray with the
             passed elements.
    
    """
    return "[" + " ".join(list) + "]"


def toPDFName(string):
    """
    Converts a python string into a pdf name string.
    
    @type  string: str
    @param string: A normal python string, which should be converted into a
                   pdf name string.
    
    @rtype:  str
    @return: a pdf name string with all characters outside the range [33, 127]
             escaped.
    
    """
    pdfname = "/"
    
    for char in string:
        if (33 < ord(char) < 127) and (char not in "\n\f\r\b\t\0 ()<>[]{}/%#"):
            pdfname += char
        else:
            pdfname += ("#%02x" % ord(char))
    return pdfname


def fromPDFName(name):
    """
    Converts a pdf name string into a standard python string.
    
    @type  name: str
    @param name: A valid pdf name string.
    
    @rtype:  str
    @return: A standard python string.
    
    @raise RuntimeError: This exception is raised if the pdf name string is
                         malformed.
    
    """
    if name == "" or name[0] != "/":
        raise RuntimeError, "name must begin with /"
    
    string = []
    i = 1
    while i < len(name):
        if name[i] == "#":
            hexNum = name[i+1:i+3]
            if len(hexNum) != 2:
                raise RuntimeError, \
                        ("hex number length not correct in name %s (must be " +
                        "exactly 2 bytes long") % name
            try:
                string.append(chr(int(hexNum, 16)))
            except ValueError, e:
                raise RuntimeError, "invalid hexnumber in name %s" % name
            i += 2
        elif name[i] in "\n\f\r\b\t\0 ()<>[]{}/%":
            raise RuntimeError, "name %s contains invalid character %s" % \
                    (name, repr(name[i]))
        else:
            string.append(name[i])
        i += 1
        
    return "".join(string)


def normalizePDFName(name):
    """
    Converts a pdf name into a normalized form.
    
    @type  name: str
    @param name: the pdf name, which should be normalized.
    
    @rtype:  str
    @return: a normalized pdf name.
    
    @raise RuntimeError: This exception is raised if the pdf name is
                                             malformed.
    
    """
    return toPDFName(fromPDFName(name))

def normalizePDFString(string):
    """
    Converts a pdf string into a normalized form.
    
    @type  string: str
    @param string: the pdf string, which should be normalized.
    
    @rtype:  str
    @return: a normalized pdf string.
    
    @raise RuntimeError: This exception is raised if the pdf string is
                         malformed.
    
    """
    #TODO: implementation missing
    return string


def utfifyString(string, encoding):
    """
    Converts a string into a string encoded with the given encoding.
    
    @type  string: str or unicode
    @param string: Either a python unicode string or a normal utf8 encoded
                               string.
    
    @type  encoding: str
    @param encoding: The encoding used to encode the string.
    
    @rtype:  str
    @return: a string encoded with the given encoding.
    
    """
    decodedString = string
    if isinstance(decodedString, str):
            try:
                    decodedString = decodedString.decode("utf-8")
            except UnicodeDecodeError, e:
                    errorMsg = repr(string) + " is not utf-8 encoded: " + str(e)
                    raise UnicodeError, errorMsg
    
    if isinstance(decodedString, unicode):
            return decodedString.encode(encoding)
    else:
            raise TypeError, repr(string) + " must be an utf-8 encoded or unicode string"


def utf8ifyString(string):
    """
    Converts a string into a utf8 encoded string.
    
    @type  string: str or unicode
    @param string: Either a python unicode string or a normal utf8 encoded
                               string.
    
    @rtype:  str
    @return: a utf8-encoded string.
    
    """
    return utfifyString(string, "utf-8")


def utf16ifyString(string):
    """
    Converts a string into a utf16 encoded string.
    
    @type  string: str or unicode
    @param string: Either a python unicode string or a normal utf8 encoded
                               string.
    
    @rtype:  str
    @return: a utf16-encoded string.
    """
    return utfifyString(string, "utf-16-be")


def makeTagCompatibleString(string):
    """
    Converts a string into a tag compatible string and escapes the ] and \
    characters.
    
    @type  string: str
    @param string: The string, which should be made tag compatible.
    
    @rtype:  str
    @return: a tag compatible string.
    """
    return string.replace("\\", "\\\\").replace("]", "\\]")
