"""
Module containing unittests for pdfparser.py.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: test_pdfparser.py 1071 2008-05-25 18:58:08Z ahuesgen $
"""

import pdfparser, pdftypes, pdfobjs
import unittest, verbosetestcase

class TestParserError(verbosetestcase.VerboseTestCase):
    def setUp(self):
        pass
    
    def tearDown(self):
        pass
    
    def testThrow(self):
        def myThrow():
            raise pdfparser.ParserError, "error"
        
        # test if the correct exception type is thrown
        self.assertRaises(pdfparser.ParserError, myThrow)
        
        # test if the error message is transfered correctly
        try:
            raise pdfparser.ParserError, "error"
        except pdfparser.ParserError, e:
            self.assertEquals("error", str(e))


class TestTokenStream(verbosetestcase.VerboseTestCase):
    def setUp(self):
        self.tokens = [1, 2, 3, 4, 5]
        self.stream = pdfparser.TokenStream(self.tokens)
    
    def tearDown(self):
        del self.tokens
        del self.stream
    
    def testConstruction(self):
        self.assertEquals(self.stream.getStream(), self.tokens)
    
    def testNextToken(self):
        self.assertEquals(self.stream.getPos(), 0)
        for i in self.tokens:
            self.assertEquals(self.stream.getNextToken(), i)
            self.assertEquals(self.stream.getPos(), i)
    
    def testUndoToken(self):
        self.stream.getNextToken()
        self.stream.undoToken()
        self.assertEquals(self.stream.getNextToken(), 1)
    
    def testUndoToken2(self):
        self.stream.undoToken()
        self.assertEquals(self.stream.getNextToken(), 1)
    
    def testTokensLeft(self):
        self.assertEquals(self.stream.tokensLeft(), True)
        
        for i in self.tokens:
            self.stream.getNextToken()
        
        self.assertEquals(self.stream.tokensLeft(), False)
        self.assertEquals(self.stream.getNextToken(), None)
    
    def testIsEmpty(self):
        emptyStream = pdfparser.TokenStream([])
        self.assertEquals(emptyStream.isEmpty(), True)
        self.assertEquals(self.stream.isEmpty(), False)


class TestPDFParser(verbosetestcase.VerboseTestCase):
    def compareObjects(self, id, gen, root, returnedObj, msg):
        self.assertEquals(id, returnedObj.id, msg)
        self.assertEquals(gen, returnedObj.gen, msg)
        self.assertEquals(root, returnedObj.get(""), msg)
    
    def setUp(self):
        self.parser = pdfparser.PDFParser()
    
    def tearDown(self):
        del self.parser
    
    def testEmptyObject(self):
        obj = self.parser.parsePDF("1 0 obj endobj")
        self.assertEquals(1, obj.id)
        self.assertEquals(0, obj.gen)
        self.assertRaises(pdfobjs.PDFObjectError, obj.get, "")
    
    def testBasicTests(self):
        testData = [
                (1, 0, False, "1 0 obj false endobj"),
                (2, 0, True, "2 0 obj true endobj"),
                (25, 12, pdftypes.PDFNull(), "25  12 obj\n null\n endobj"),
                (10, 234, 23, "10 234 obj\t 23 endobj\n"),
                (1, 0, "(Hello)", "1 0 obj (Hello) endobj"),
                (1, 1, "/Foo", "1 1 obj /Foo\r\nendobj"),
                (1, 0, {}, "1 0 obj <<>> endobj"),
                (1, 0, [], "1 0 obj\r [  ] endobj"),
                (1, 0, pdftypes.PDFRef(2, 0), "1   0 obj\n 2 0 R stream\n"),
                (2, 0, pdftypes.PDFRef(12, 0), "2 0 obj\n 12   0   R endobj")
                   ]
        
        for id, gen, root, string in testData:
            msg = str((id, gen, root, string))
            obj = self.parser.parsePDF(string)
            self.compareObjects(id, gen, root, obj, msg)
    
    def testComplex1(self):
        objStr = "1 0 obj [1 2 3 4] endobj"
        obj = self.parser.parsePDF(objStr, 0)
        
        testData = [("[0]", 1), ("[1]", 2), ("[2]", 3), ("[3]", 4)]
        
        for path, expected in testData:
            msg = str((path, expected))
            self.assertEquals(obj.get(path), expected, msg)
    
    def testComplex2(self):
        objStr = "1 0 obj << /foo /bar /moo 1 /koo 2 /Hello /WORLD>> endobj"
        obj = self.parser.parsePDF(objStr, 0)
        
        testData = [("/foo", "/bar"), ("/moo", 1), ("/koo", 2),
                ("/Hello", "/WORLD")]
        
        for path, expected in testData:
            msg = str((path, expected))
            self.assertEquals(obj.get(path), expected, msg)
    
    def testComplex3(self):
        objStr = "1 0 obj [/first << /foo /bar /baz [1] >> ] endobj"
        obj = self.parser.parsePDF(objStr, 0)
        
        testData = [("[0]", "/first"), ("[1]/foo", "/bar"),
                ("[1]/baz[0]", 1)]
        
        for path, expected in testData:
            msg = str((path, expected))
            self.assertEquals(obj.get(path), expected, msg)

if __name__ == "__main__":
    unittest.main()
