"""
Module defining the commandline interface, also contains 
unittests for pdflexer.py.


This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: userinterface.py 2695 2013-04-18 14:21:11Z ahuesgen $
"""

import sys
from optparse import OptionParser

import logger
import vhistconsts

USAGE = """%prog [options]
- this file is part of the VHIST project, an effort to create a "meta-log"
  facility with tools to document a file's processing history. Use this tool
  to add information on a workflow step by adding a section to an existing
  VHIST file, or create a new one."""


DEFAULT_DOC_ATTRS = {"title": None, "author": None, "subject": None,
        "keywords": None, "producer": None, "creator": None}

TRANSLATE_STEP_ATTRS = {"title": "title", "description": "description",
    "comment": "comment", "command": "command", "tool": "tool",
    "toolversion": "tool", "toolpath": "toolpath", "host": "host",
    "user": "user"}

DEFAULT_STEP_ATTRS = {"title": None, "description": None, "comment": None,
        "command": None, "tool": None, "toolpath": None, "host": None,
        "user": None}

# "filename" is set when the file is created and is no real attribute
DEFAULT_FILE_ATTRS = {"filetype": None, "description": None, "comment": None,
        "md5file": None}

DEFAULT_FILE_FLAGS = {"compress": None, "embed": None, "automd5": None,
        "optional": None, "previewws": None, "preview": None,
                      "thumbnail": None, "thumbnailonly": None}

def dictKeysToString(dic):
    keys = dic.keys()
    s = ", ".join(keys)
    return s

DOC_ATTRS = dictKeysToString(DEFAULT_DOC_ATTRS)
STEP_ATTRS = dictKeysToString(DEFAULT_STEP_ATTRS)
FILE_ATTRS = dictKeysToString(DEFAULT_FILE_ATTRS)
FILE_FLAGS = dictKeysToString(DEFAULT_FILE_FLAGS)

class CommandlineError(Exception):
    pass

class CommandlineFile(object):
    def __init__(self):
        self.fileAttrs = DEFAULT_FILE_ATTRS.copy()
        self.fileFlags = DEFAULT_FILE_FLAGS.copy()
        self.userAttrs = {}
        self.userAttrsUnits = {}
        self.userAttrsOrder = []


class CommandlineArgs(object):
    def __init__(self):
        self.docAttrs = DEFAULT_DOC_ATTRS.copy()
        self.stepAttrs = DEFAULT_STEP_ATTRS.copy()
        self.stepUserAttrs = {}
        self.stepUserAttrsUnits = {}
        self.stepUserAttrsOrder = []
        self.vhistRootFile = None
        self.genVhistFiles = []
        self.infiles = []
        self.outfiles = []
        self.lastFile = None
    
    def addInfile(self, infile):
        self.infiles.append(infile)
        self.lastFile = infile
    
    def addOutfile(self, outfile):
        self.outfiles.append(outfile)
        self.lastFile = outfile


def handleDocAttr(option, opt_str, value, parser):
    validKeys = DEFAULT_DOC_ATTRS.keys()
    # value is a tuple (key, value)
    key, val = value
    if key in validKeys:
        parser.attributes.docAttrs[key] = val
    else:
        raise CommandlineError, "'%s' is no valid document attribute" % key

def handleStepAttr(option, opt_str, value, parser):
    validKeys = TRANSLATE_STEP_ATTRS.keys()
    # value is a tuple (key, value)
    key, val = value
    if key in validKeys:
        key = TRANSLATE_STEP_ATTRS[key]
        parser.attributes.stepAttrs[key] = val
    else:
        raise CommandlineError, "'%s' is no valid step attribute" % key

def handleStepUserAttr(option, opt_str, value, parser):
    # value is a tuple (key, value, [unit])
    key = value[0]
    val = value[1]
    unit = None
    if len(value) > 2:
        unit = value[2]
    
    parser.attributes.stepUserAttrs[key] = val
    parser.attributes.stepUserAttrsUnits[key] = unit
    if key in parser.attributes.stepUserAttrsOrder:
        parser.attributes.stepUserAttrsOrder.remove(key)
    parser.attributes.stepUserAttrsOrder.append(key)

def handleFileAttr(option, opt_str, value, parser):
    # value is a tuple (key, value)
    key, val = value
    lastFile = parser.attributes.lastFile
    if lastFile == None:
        raise CommandlineError, ("file attribute '%s' was specified prior "
            "first file") % key
    
    validKeys = DEFAULT_FILE_ATTRS
    
    if key in validKeys:
        lastFile.fileAttrs[key] = val
    else:
        raise CommandlineError, "'%s' is no valid file attribute" % key

def handleFileFlag(option, opt_str, value, parser):
    def parseFlag(flag):
        # extracts basename of flag and whether the flag is enabled/disabled
        if flag.startswith("no-"):
            return False, flag[3:]
        else:
            return True, flag
    
    lastFile = parser.attributes.lastFile
    if lastFile == None:
        raise CommandlineError, ("file flag '%s' was specified prior first "
            "file") % value
    
    state, flag = parseFlag(value)
    validKeys = DEFAULT_FILE_FLAGS.keys()
    if flag in validKeys:
        lastFile.fileFlags[flag] = state
    else:
        raise CommandlineError, "'%s' is no valid file flag" % value

def handleFileUserAttr(option, opt_str, value, parser):
    # value is a tuple (key, value, [unit])
    key = value[0]
    val = value[1]
    unit = None
    if len(value) > 2:
        unit = value[2]
    
    lastFile = parser.attributes.lastFile
    if lastFile == None:
        raise CommandlineError, ("file user attribute '%s' was specified "
            "prior first file") % key
    
    lastFile.userAttrs[key] = val
    lastFile.userAttrsUnits[key] = unit
    if key in lastFile.userAttrsOrder:
        lastFile.userAttrsOrder.remove(key)
    lastFile.userAttrsOrder.append(key)

def handleNewInfile(option, opt_str, value, parser):
    infile = CommandlineFile()
    infile.fileAttrs["filename"] = value
    parser.attributes.addInfile(infile)

def handleNewOutfile(option, opt_str, value, parser):
    outfile = CommandlineFile()
    outfile.fileAttrs["filename"] = value
    parser.attributes.addOutfile(outfile)

def handleAppend(option, opt_str, value, parser):
    parser.values.genVhistFiles.append(value)
    parser.values.optionalVhistRootFile = value

def handleVersion(option, opt_str, value, parser):
    print vhistconsts.getVhistaddVersion()
    sys.exit(0)

def parseCommandline(args, ignoreCmdFile = False):
    parser = OptionParser(usage = USAGE)
    parser.attributes = CommandlineArgs()
    parser.prog = args[0]
    del args[0]
    logger.log(4, "Commandline:")
    logger.log(4, repr(args))
    
    parser.add_option("-V", "--version", action="callback",
            callback=handleVersion,
            help="print program's version information and quit.")
   
    parser.add_option("-i", "--infile", type="string", action="callback",
            callback=handleNewInfile, metavar="FILENAME",
            help="argument file used as input in "
            "this step. All file (user) attributes and flags specified "
            "with -f -a or -u after this option refer to this file. "
            "Examples: -i '/path/to/file' or -i '../myfile' or "
            "-i 'C:\\MyFile'")
    
    parser.add_option("-o", "--outfile", type="string", action="callback",
            callback=handleNewOutfile, metavar="FILENAME",
            help="argument file used as output in "
            "this step. All file (user) attributes and flags specified with "
            "-f -a or -u after this option refer to this file. Examples:"
            " -o '/path/to/file' or -o '../myfile' or -o 'C:\\MyFile'")
    
    parser.add_option("-I", type="string", action="store",
            dest="vhistRootFile", default=None, metavar="VHISTFILE",
            help="the rootfile from which the generated output vhist file "
            "is derived. If the file does not exist, an error is emitted")
    
    parser.add_option("-J", type="string", action="store",
            dest="optionalVhistRootFile", default=None, metavar="VHISTFILE",
            help="same as -I, however, the rootfile is ignored if it does "
            "not exist")
    
    parser.add_option("-O", type="string", action="append",
            dest="genVhistFiles", default=[], metavar="VHISTFILE",
            help="add an output vhist file. At least one output "
            "vhistfile must be specified")
    
    parser.add_option("-A", "--append-to-vhistfile", type="string",
            action="callback", nargs=1, callback=handleAppend,
            metavar="VHISTFILE",
            help = "appends to the given vhist file. Is a shortcut for "
            "-J VHISTFILE -O VHISTFILE")
    
    parser.add_option("-d", "--doc", type="string", action="callback", nargs=2,
            callback=handleDocAttr, metavar="KEY VALUE", help=("an attribute "
            "(key-value pair) of the vhist document. If a step is appended to "
            "an existing vhist file, these attributes are ignored. Valid "
            "keys are: %s. Examples: -d 'title' 'MyDocument' or --doc 'author' "
            "'Myself'") % DOC_ATTRS)
    
    parser.add_option("-s", "--step", type="string", action="callback",
            nargs=2, callback=handleStepAttr, metavar="KEY VALUE",
            help=("an attribute (key-value pair) of the new workflow step. "
            "Valid keys are: %s. Examples: -s 'title' 'MyStep' or --step "
            "'comment' 'optional step'") % STEP_ATTRS)
    
    parser.add_option("-U", "--step-user", type="string", action="callback",
            nargs=2, callback=handleStepUserAttr, metavar="KEY VALUE",
            help="a user defined attribute (key-value pair) of the current "
            "workflowstep. Examples: TODO")
    
    parser.add_option("-W", "--step-user-with-unit", type="string",
            action="callback",
            nargs=3, callback=handleStepUserAttr, metavar="KEY VALUE UNIT",
            help="a user defined attribute (key-value-unit triple) of the "
            "current workflowstep. Examples: TODO")
    
    parser.add_option("-a", "--file-attr", type="string",
            action="callback", nargs=2, callback=handleFileAttr,
            metavar="KEY VALUE", help=("an attribute (key-value pair) of the "
            "current in- or outfile. Valid keys are: %s. Examples: -f "
            "'filetype' 'text/xml' or --file 'md5file' 'C:\\Path\\MyMD5File'") \
            % FILE_ATTRS)
    
    parser.add_option("-f", "--file-flag", type="string", action="callback",
            nargs=1, callback=handleFileFlag, metavar="FLAG", help=("a flag (a "
            "property which can be enabled/disabled) of the current in- or "
            "outfile. Valid flags are: %s and negated versions of these "
            "flags beginning with 'no-'. Examples: -f compress or -f "
            "no-automd5") % FILE_FLAGS)
    
    parser.add_option("-u", "--file-user", type="string", action="callback",
            nargs=2, callback=handleFileUserAttr, metavar="KEY VALUE",
            help="a user defined attribute (key-value pair) of the current "
            "in- or outfile. Examples: -u 'url' 'http://www.nf.mpg.de' or "
            "--file-user 'ref id' '19231212'")
    
    parser.add_option("-w", "--file-user-with-unit", type="string",
            action="callback",
            nargs=3, callback=handleFileUserAttr, metavar="KEY VALUE UNIT",
            help="a user defined attribute (key-value-unit triple) of the "
            "current in- or outfile. Examples: -w 'pixelsize' "
            "'[1.0, 1.0, 1.5]' mm or --file-user-with-unit "
            "'Measurement Time' 125.2 sec")
    
    parser.add_option("-r", "--readme", type="string", action="store",
            dest="readme", default="", metavar="FILENAME",
            help="file containing the readme embedded at the beginning of a "
            "newly created vhistfile (not viewed in a pdf viewer)")
    
    parser.add_option("-1", "--firstpage", type="string", action="store",
            dest="firstpage", default="", metavar="FILENAME",
            help="file containing the content of the first page of a newly "
            "created vhistfile viewed in a. The file must be encoded in utf8")
    """
    parser.add_option("-c", "--config", type="string", action="store",
        dest="xmlconfigfile", default="",
        help="read settings from an XML file. If this option is set, all "
        "other options except --firstpage and --readme are ignored")
    """
    if not ignoreCmdFile:
        parser.add_option("-c", "--cmdfile", type="string", action="append",
                dest="cmdfiles", default=[], metavar="CMDFILE",
                help="A file containing the commandline. Several files can be"
                "specified using several -c options")
    
    parser.add_option("--sync-cwd", action="store_true",
                dest="syncCwd", default=False, help="if this option "
                "is set, the path of the first argfile is used s the "
                "current working directory. This means that all relative "
                "pathes are relative to the argfile. WARNING: This option "
                "is experimental and will probably change in the future.")
    
    parser.add_option("-p", "--pretend", action="store_true", dest="pretend",
            default=False, metavar="PRETEND", help="if this option is set, "
            "only the commandline arguments are parsed and validated but no "
            "section is generated")
    
    parser.add_option("-q", "--quiet", action="store_true", dest="quiet",
            default=False, metavar="QUIET", help="if this option is set, "
            "only errors are logged. This option overrides the verbose option")
    
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose",
            default=False, metavar="VERBOSE", help="if this option is set, "
            "vhistadd will generate as much output as possible")
    
    if not ignoreCmdFile:
        parser.add_option("--stdin", action="store_true", dest="stdin",
                default=False, metavar="STDIN", help="if this option is "
                "set, stdin is used as an additional commandline file.")
    
    parser.add_option("--debug", action="store_true", dest="debug",
            default=False, metavar="DEBUG", help="if this option is set, "
            "debug mode is enabled, generating more complete information in "
            "the case of errors.")
    
    try:
        options, arguments = parser.parse_args(args)
        #print options
    except SystemExit, e:
        raise
    except Exception, e:
        parser.error(str(e))
    if options.vhistRootFile != None and options.optionalVhistRootFile != None:
        parser.error("options -I and -J/-A are mutually exclusive")
    if len(options.genVhistFiles) == 0 and \
            (ignoreCmdFile or 
            (len(options.cmdfiles) == 0 and not options.stdin)):
        parser.error("option '-O <vhistfile>' is mandatory,\n" \
                         "use \"vhistadd -h\" for more information.")
    if len(arguments) != 0:
        parser.error(("the argument '%s' was specified.\n"
            "Arguments (without option) are not allowed by vhistadd.\n"
            "Perhaps a commandline error (check if you have typed \n"
            "-title instead of -s title or something similar).") % arguments[0])
    
    parser.attributes.rootIsOptional = False
    parser.attributes.vhistRootFile = options.vhistRootFile
    if options.optionalVhistRootFile != None:
        parser.attributes.vhistRootFile = options.optionalVhistRootFile
        parser.attributes.rootIsOptional = True
    parser.attributes.genVhistFiles = options.genVhistFiles
    return parser.attributes, options

def test():
    import sys
    result, options = parseCommandline(sys.argv)
    print "docAttrs:", result.docAttrs
    print "stepAttrs:", result.stepAttrs
    print "stepUserAttrs:", result.stepUserAttrs
    print "vhistfile:", result.vhistfile
    print "options:", options.__dict__
    print ""
    print "infiles"
    for i in result.infiles:
        print "fileAttrs:", i.fileAttrs
        print "fileFlags:", i.fileFlags
        print "userAttrs:", i.userAttrs
        print ""
    
    print "outfiles"
    for o in result.outfiles:
        print "fileAttrs:", o.fileAttrs
        print "fileFlags:", o.fileFlags
        print "userAttrs:", o.userAttrs
        print ""
    
    import workflow
    wf = workflow.Workflow()
    wf.genFromUI(result)
    print "\n"
    wf.toStr()

#test()
