"""
A module containing various general purpose helper functions.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: utilities.py 1895 2009-11-16 18:32:45Z ahuesgen $
"""

import sys, os
import checksums

def genMD5FromFile(filename):
    if os.path.isdir(filename):
        raise IOError("'%s' is a directory, you need to specify a file instead"
                % filename)
    f = open(filename, "rb")
    KBYTES = 1024
    BLOCKSIZE = 1024*KBYTES
    md5sum = checksums.md5("")
    try:
        data = f.read(BLOCKSIZE)
        while data != "":
            md5sum.update(data)
            data = f.read(BLOCKSIZE)
    finally:
        f.close()
    return md5sum.hexdigest()

def readFile(filename):
    if os.path.isdir(filename):
        raise IOError("'%s' is a directory, you need to specify a file instead"
                % filename)
    f = open(filename, "rb")
    try:
        content = f.read()
    finally:
        f.close()
    return content

def abort(string):
    print string
    print "abort"
    sys.exit(1)

def assertion(expr, errorMessage):
    if not expr:
        raise AssertionError, errorMessage

def checkType(obj, types, objName = None):
    """
    Checks whether or not an object's type is in a given list.
    
    @type  obj: any
    @param obj: the object, which is validated.
    
    @type  types: type or tuple of types
    @param types: a type or a tuple of valid types. If obj is an instance of at
                  least one of these types, it passes the validation.
    
    @type  objName: str or NoneType
    @param objName: an optional object name for to make object identification
                    easier.
    
    @raise TypeError: This exception is raised when the validation test fails
                      for this object.
    
    """
    if not isinstance(obj, types):
        typeList = [str(element) for element in types]
        typeStr  = ", ".join(typeList)
        if objName is None:
            objName = "object"
        errorMsg = "%s's type must be one of [%s] but was of type %s" % \
                (objName, typeStr, str(type(obj)))
        
        raise TypeError, errorMsg

def indexed(iterator, start = 0, step = 1):
    """
    An iterator which iterates over elements el of another iterator and
    yielding (el, i) element where i is an index.
    
    @type  alist: iterator
    @param alist: An iterator which is used as base for the new iterator.
    
    @type  start: int
    @param start: the start value of the indices (as in range).
    
    @type  step: int
    @param step: the increment of the indices (as in range).
    """
    i = start
    for element in iterator:
        yield element, i
        i += step

