"""
Module containing classes to create vhisttags.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: vhisttags.py 2695 2013-04-18 14:21:11Z ahuesgen $
"""

# number of digits in size field of section tag
NUM_DIGITS = 12
LARGEST_SIZE = 10**(NUM_DIGITS)-1

SECTION_TAG_START = "<--! $VHIST_SECTION"

SECTION_TAG = SECTION_TAG_START + " [version:%s][creator:%s][title:%s]" + \
              "[left:%d][size:%0*d][index:%d][md5section:%s][previousmd5:%s]" + \
              "[previousmarker:%s]-->"

import checksums, zlib
import vhistconsts

from tagutils import stringToTagstring

class TagError(Exception):
    pass

class SectionTag(object):
    """
    A class which describes a vhisttag for one section in a vhistfile.
    
    """
    def __init__(self, title, sectionPrefix, index, prevMD5, prevMarker):
        # initialize tag parameters
        self.version = stringToTagstring(vhistconsts.VHIST_VERSION)
        self.creator = stringToTagstring(vhistconsts.getVhistCreatorVersion())
        self.title   = stringToTagstring(title)
        self.prefixSize = len(sectionPrefix)
        self.index = index
        self.sectionMD5 = "0" * 32
        self.prefixMD5 = checksums.md5(sectionPrefix)
        self.postfixSize = 0
        
        if prevMD5 != None:
            self.prevMD5 = prevMD5.lower()
        else:
            self.prevMD5 = ""
        
        if prevMarker != None: 
            self.prevMarker = prevMarker
        else:
            self.prevMarker = ""
        
        self.blocksize = 0
    
    def getTag(self):
        return SECTION_TAG %  (self.version, self.creator, self.title, \
                               self.prefixSize, NUM_DIGITS, \
                               self.blocksize, self.index, self.sectionMD5, \
                               self.prevMD5, self.prevMarker)
    
    def getTagSize(self):
        return len(self.getTag())
    
    def setSectionPostfix(self, sectionPostfix):
        # calculate section size
        tagSize = self.getTagSize()
        self.postfixSize = len(sectionPostfix)
        self.blocksize = self.prefixSize + tagSize + self.postfixSize
        
        # validate section size
        if self.blocksize > LARGEST_SIZE:
            raise RuntimeError, ("Section too large. Maximum size: %d bytes. " +
                                "Size: %d") % (LARGEST_SIZE, self.blocksize)
        else:
            # generate section md5 sum
            self.sectionMD5 = "0" * 32
            DUMMY_TAG = self.getTag()
            correctMD5 = self.prefixMD5.copy()
            correctMD5.update(DUMMY_TAG)
            correctMD5.update(sectionPostfix)
            self.sectionMD5 = correctMD5.hexdigest()



class EmbeddedFileTag(object):
    def __init__(self, filename, type, desc, comment, content,
            compression = "flate", offsetPre = 8, offsetPost = 11, md5 = None,
            cmd5 = None):
        """
        Creates a pair of VHIST_EMBEDDEDFILE_BEGIN and VHIST_EMBEDDEDFILE_END
        tags.
        
        @type  filename: str
        @param filename: The name of the file.
        
        @type  type: str
        @param type: The vhist mime-like filetype of the file.
        
        @type  desc: str
        @param desc: A description of the file.
        
        @type  comment: str
        @param comment: Comments refering to the file.
        
        @type  compression: str
        @param compression: either "flate" for flate compression or "none" for
                            no compression.
        
        @type  offsetPre: int
        @param offsetPre: The number of bytes between the end of the
                          EmbeddedFileBegin tag (->) and the begin of the
                          filestream's content.
        
        @type  offsetPost: int
        @param offsetPost: The number of bytes between the end of the
                           filestream's content and the begin of the
                           EmbeddedFileEnd tag (<-).
        
        @type  md5: str or None
        @param md5: The md5 sum of the (uncompressed) file. If md5 is None,
                    the md5 sum is automatically generated from the files
                    content.
        
        @type  cmd5: str or None
        @param cmd5: The md5 sum of the compressed file. If cmd5 is None,
                    the md5 sum is automatically generated from the compressed
                    files content.
        
        @type  content: str
        @param content: The uncompressed content of the file.
        
        """
        self.filename = filename
        self.type = type
        self.desc = desc
        self.comment = comment
        if compression not in ["flate", "none"]:
            raise TagError, "invalid compression mode: %s" % compression
        else:
            self.compression = compression
        
        self.offsetPre = offsetPre
        self.offsetPost = offsetPost
        if md5 == None:
            self.md5 = self.__calcMD5(content)
        else:
            self.md5 = md5
        
        self.uncompressedSize = len(content)
        if self.compression == "flate":
            self.content = zlib.compress(content)
        else:
            self.content = content
        if self.compression != "none":
            self.compressedSize = len(self.content)
        else:
            self.compressedSize = ""
        
        self.streamSize = len(self.content)
        self.blocksize = self.offsetPre + self.streamSize + self.offsetPost
        
        if cmd5 == None:
            if self.compression != "none":
                self.cmd5 = self.__calcMD5(self.content)
            else:
                self.cmd5 = ""
        else:
            self.cmd5 = cmd5
    
    def __calcMD5(self, content):
        m = checksums.md5()
        m.update(content)
        return m.hexdigest()
    
    def __makeTagBody(self):
        stts = stringToTagstring
        return ("[filetype:%s][filename:%s][desc:%s][comment:%s]" + \
                "[compression:%s][filesize:%d][cfilesize:%s][blocksize:%d]" + \
                "[offset:%d][md5file:%s][md5cfile:%s]") % (stts(self.type),
                stts(self.filename), stts(self.desc), stts(self.comment),
                stts(self.compression), self.uncompressedSize,
                self.compressedSize, self.blocksize, self.offsetPre,
                self.md5, self.cmd5)
    
    def getBeginTag(self):
        """
        Returns the EMBEDDEDFILE_BEGIN tag.
        
        @rtype:  str
        @return: The EMBEDDEDFILE_BEGIN tag.
        
        """
        return "<--! $VHIST_EMBEDDEDFILE_BEGIN %s-->" % self.__makeTagBody()
    
    def getEndTag(self):
        """
        Returns the EMBEDDEDFILE_END tag.
        
        @rtype:  str
        @return: The EMBEDDEDFILE_END tag.
        
        """
        return "<--! $VHIST_EMBEDDEDFILE_END %s-->" % self.__makeTagBody()
    
    def getCompressedContent(self):
        """
        Returns the compressed content of the file.
        
        @rtype:  str
        @return: The compressed content of the file.
        
        """
        return self.content
    
    def getStreamSize(self):
        return self.streamSize
    
    def getCompressedSize(self):
        return self.compressedSize

    def getCompressedMD5(self):
        return self.cmd5

"""   
tag = EmbeddedFileTag("test.txt", "type", "desc", "comment", "content", "flate",
        123, 456, "md5", "cmd5")

print tag.getBeginTag()
print tag.getEndTag()
"""
