#!/usr/bin/env python
"""
This is the vhistxl tool for extracting and validating the contents of
VHIST files. This file can be either used as a standalone application
by calling it from the commandline (see --help option for information
on how to use it) or as a python module from another application by
calling the processFile function.

This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

Author: Andreas Huesgen
$Id: vhistxl.py 908 2008-02-13 12:34:32Z ahuesgen $
"""
import sys, os, time
from optparse import OptionParser

from vhist import validatetag, logger, vhistconsts

class StatusObject(object):
    def __init__(self):
        self.sections = []
        self.files = []
        self.general = {}

class OptionObject(object):
    def __init__(self, listFiles = True, validateSections = True,
            extractAll = True, outfiles = [], pretend = True, dir = "."):
        self.listFiles = listFiles
        self.validateSections = validateSections
        self.extractAll = extractAll
        self.outfiles = outfiles[:]
        self.pretend = pretend
        self.dir = dir


def makeListStr(alist):
    """
    Generates a ", " separated string from a list.
    
    @type  alist: list
    @param alist: A list of elements, which can be converted into strings
                  using str().
    
    @rtype:  str
    @return: The string containing the list elements.
    """
    return ", ".join([str(item) for item in alist])


def checkSectionIndices(sectionids):
    """
    Check if the same id appeares several times in 'sectionids' and
    if ids are missing in 'sectionids'.
    
    @type  sectionids: list of int.
    @param sectionids: A list of section ids. The list must be sorted.
    
    @rtype:  2 lists of ints
    @return: A list containing the ids missing in the input list and
             a list containing duplicate ids from the input list.
    """
    curpos = 0
    duplicated = []
    missing = []
    sectionids = sectionids[:]
    sectionids.sort()
    
    # a valid sectionids list contains all numbers from 1 to n once
    for i in xrange(1, sectionids[-1]):
        if i < sectionids[curpos]:
            missing.append(i)
        elif i > sectionids[curpos]:
            duplicated.append(sectionids[curpos])
            curpos += 1
        else:
            curpos += 1
    
    return missing, duplicated


def findAllSections(data):
    endpos = len(data)
    section = validatetag.findLastSection(data, endpos)
    
    # section located after the current section in file (parsed previously)
    nextSection = None
    
    # stores tupels (section, nextSection)
    sections = []
    sectionids = []
    
    while section != None:
        sections.insert(0, (section, nextSection))
        sectionids.insert(0, section["index"])
        nextSection = section
        if section["prevmark"] != 0:
            # find and parse previous section (further up in file)
            prevPos = section["tagstart"] - section["prevmark"]
            section = validatetag.parseSection(data, prevPos)
        else:
            section = None
    return sections, sectionids


def validateAllSections(data, sections):
    sectionResults = []
    for section in sections:
        result = validatetag.validateSection(data, section[0], section[1])
        sectionResults.append(result)
    return sectionResults


def warnWrongSectionOrder(sectionids):
    # create sorted copy of sectionids
    sortsectionids = sectionids[:]
    sortsectionids.sort()
    
    # find wrong order of sections
    if sortsectionids != sectionids:
        logger.warning("sections not in correct order")
        logger.warning("wrong order is: %s", makeListStr(sectionids))
        return False
    else:
        return True


def warnDuplicateMissingSectionIDs(sectionids):
    missing, duplicated = checkSectionIndices(sectionids)
    if len(missing) != 0:
        logger.warning("missing sections: %s", makeListStr(missing))
    if len(duplicated) != 0:
        logger.warning("duplicated sections: %s", makeListStr(duplicated))
    return missing, duplicated


def warnSectionlessFile(sectionids):
    if len(sectionids) == 0:
        logger.warning("no section found in file")
        return False
    else:
        return True


def validateSections(data):
    """
    Validates all sections in a given VHIST file.

    @type  data: str
    @param data: The body of the VHIST file as a string.
    """
    result = {"correctOrder": False, "hasSections": False, "duplicates": [],
            "missing": []}
    timestamp = time.asctime()
    logger.log(2, "scanning sections, started at %s.", timestamp)
    
    # start timer
    fStartClock = 1.0 * time.clock()
    
    sections, sectionids = findAllSections(data)
    sectionResults = validateAllSections(data, sections)
    
    result["hasSections"] = warnSectionlessFile(sectionids)
    result["correctOrder"] = warnWrongSectionOrder(sectionids)
    dup, miss = warnDuplicateMissingSectionIDs(sectionids)
    result["duplicates"] = dup
    result["missing"] = miss
    
    # stop timer
    fSecs = 1.0 * time.clock() - fStartClock
    logger.log(1,"number of sections found: %d, completed in %1.2fs." \
            % (len(sectionids), fSecs))
    
    logger.log(1, "")
    return result, sectionResults


def validateFiles(data, dir, extract, pretend):
    """
    Validates all embedded files in a VHIST file.
    
    @type  data: str
    @param data: The body of the VHIST file as a string.
    
    @type  dir: str
    @param dir: The directory into which files will be extracted.
    
    @type  extract: True or list of ints
    @param extract: A list of fileids of files, which should be extracted.
                    If extract is True, all files will be extracted.
    
    @type  pretend: bool
    @param pretend: True if files should only be decompressed/verified
                    but not stored on hard disk.
    """
    # start timer
    fileResults = []
    timestamp = time.asctime()
    logger.log(2, "scanning files, started at %s.", timestamp)
    
    fStartClock = 1.0 * time.clock()
    
    found = validatetag.parseFileBegin(data, 0)
    numFiles = 0
    
    while found != None:
        numFiles += 1
        found["index"] = numFiles
        validatetag.validateFile(data, found)
        # should the file be extracted?
        if extract == True or found["index"] in extract:
            result = validatetag.extractFile(data, found, not pretend, dir)
            result["filename"] = found["filename"]
            fileResults.append(result)
        logger.log(1, "")
        found = validatetag.parseFileBegin(data, found["blockend"])
    
    fSecs = 1.0 * time.clock() - fStartClock
    logger.log(1, "number of files found: %d, completed in %1.2fs." \
            % (numFiles, fSecs))
    logger.log(1, "")
    return fileResults




def handleVersion(option, opt_str, value, parser):
    print vhistconsts.getVhistXLVersion()
    parser.exit(0)


def parseArgs(args):
    DESCRIPTION = """\
    %prog - this file is part of the VHIST project, an effort to create
    a "meta-log" facility with tools to document a file's (usually
    an image volume for diagnostic imaging) processing history. Use this
    tool to extract and validate all information from a VHIST file."""
    
    USAGE = "%prog [options] infile"
    
    PROG = args[0]
    args = args[1:]
    
    # init option parser
    parser = OptionParser(description = DESCRIPTION,
                          usage = USAGE,
                          prog = PROG)
    
    parser.attributeDest = parser
    
    parser.add_option("-V", "--version", action="callback",
        callback=handleVersion,
        help="print program's version information and quit.")
    
    parser.add_option("-t", "--list", action="store_true",
        dest="listFiles", default=False,
        help="only list all embedded files, do not extract data")
    
    parser.add_option("-v", "--validate", action="store_true",
        dest="validateSections", default=False,
        help="validate sections")
    
    parser.add_option("-x", "--extract", action="store_true",
        dest="extractAll", default=False,
        help="extract all files to disk (implies -t)")
    
    parser.add_option("-f", "--extract-file", action="append",
        dest="outfiles", default=[], type="int", metavar="FILEID",
        help="extract the file with the given id to disk (implies -t)")
    
    parser.add_option("-p", "--pretend", action="store_true",
        dest="pretend", default=False,
        help="only decompress and test md5sums of files, but do not write "
             "to disk")
    
    parser.add_option("-d", "--dir", action="store",
        dest="dir", default=".", type="string",
        help="extract files to the given directory")
    
    (options, args) = parser.parse_args()
    return options, args


def processFile(options, args, enableLogger = False):
    if enableLogger:
        logger.setLogLevel(logger.LOG_LEVEL_MAX)
    else:
        logger.setLogLevel(logger.LOG_LEVEL_OFF)
    
    # postprocess options
    if options.extractAll:
        options.extract = True
    else:
        options.extract = options.outfiles
    
    # option "extract files" implies "list files"
    options.listFiles = options.listFiles or options.extract
    
    logger.log(1, "This is %s." % vhistconsts.getVhistXLVersion())
    
    # check infile
    if len(args) < 1:
        raise RuntimeError("no filename given")
    
    # print options and arguments
    filename = args[0]
    logger.log(1, "infile: \"%s\"", filename)
    
    logger.log(2, "options:")
    if options.listFiles:
        logger.log(4, "list files: yes")
    else:
        logger.log(4, "list files: no")
    
    if options.validateSections:
        logger.log(4, "validate sections: yes")
    else:
        logger.log(4, "validate sections: no")
    
    if options.extractAll:
        logger.log(4, "extract files: yes")
    elif options.outfiles != []:
        idlist = makeListStr(options.outfiles)
        logger.log(4, "extract files with indices: %s", idlist)
    else:
        logger.log(4, "extract files: no")
    
    if options.extract:
        # create and display directory
        logger.log(4, "extract files to directory: %s", options.dir)
        if os.path.exists(options.dir) and not os.path.isdir(options.dir):
            raise IOError("%s does already exist but is no directory" % options.dir)
        elif not os.path.exists(options.dir):
            try:
                os.mkdir(options.dir)
            except Exception, e:
                raise IOError("could not create directory: %s", str(e))
    
    # print pretend option
    if options.pretend:
        logger.log(4, "only pretend to extract files: yes")
    else:
        logger.log(4, "only pretend to extract files: no")
    
    logger.log(1, "")
    
    # open infile
    try:
        histfile = file(filename, "rb")
        data = histfile.read()
        histfile.close()
    except Exception, e:
        raise IOError("could not open file: %s", str(e))
    
    result = StatusObject()
    timestamp = time.asctime()
    logger.log(2, "scanning, started at %s.", timestamp)
    
    fStartClock = 1.0 * time.clock ()
    
    # parse and validate sections
    if options.validateSections:
        result.general, result.sections = validateSections(data)
    
    if options.listFiles:
        result.files = validateFiles(data, options.dir, options.extract, options.pretend)
    
    fSecs = 1.0 * time.clock() - fStartClock
    logger.log(1, "validation completed in %1.2fs." % fSecs)
    return result


def main():
    options, args = parseArgs(sys.argv)
    try:
        result = processFile(options, args, True)
    except Exception, e:
        logger.error(str(e))
        sys.exit(1)


if __name__ == "__main__":
    main()

