/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: main.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
*/

#include <QObject>
#include <QApplication>
#include <QtDebug>
#include <QStringList>

#include <iostream>

#include "clilogger.h"
#include "miscslots.h"
#include <unvhistlib/vhistparser.h>
#include <unvhistlib/validationqueue.h>
#include <unvhistlib/extractionqueue.h>


int main(int argc, char* argv[])
{
    // init application
    QCoreApplication app(argc, argv);
    CliLogger logger;
    MiscSlots miscSlots;
    
    // read commandline arguments
    QStringList args = app.arguments();
    if (args.count() <= 1)
    {
        logger.error("no vhist file specified");
        return -1;
    }
    QString filename = args[1];
    
    // init vhist parser
    vhist::VhistParser parser(vhist::MODE_ROBUST, &logger);
    parser.open(filename);
    
    // read and print section and file names
    vhist::SectionTagList sections = parser.getSections();
    vhist::FileTagList files = parser.getFiles();
    logger.log(QString("%1 sections").arg(sections.count()));
    logger.log(QString("%2 files").arg(files.count()));
    
    // validating everything
    vhist::ValidationQueue vqueue(&parser);
    QObject::connect(&vqueue,
            SIGNAL(percentFinishedChanged(int, const QString)),
            &miscSlots, SLOT(printPercentFinished(int, const QString)));
    
    logger.log("\nValidating files and sections...");
    while (vqueue.process());
    
    
    // extracting everything
    vhist::ExtractionQueue equeue(&parser);
    QObject::connect(&equeue,
            SIGNAL(percentFinishedChanged(int, const QString)),
            &miscSlots, SLOT(printPercentFinished(int, const QString)));
    
    // create a list with all files to extract
    QHash<int, QString> toExtract;
    for (int i = 0; i < files.count(); i++)
    {
        int fileid = files[i].id;
        // add a unique number to every filename
        QString filename = QString("%1-%2").arg(fileid).arg(files[i].filename);
        toExtract.insert(fileid, filename);
    }
    equeue.addExtractees(toExtract);
    logger.log("\nExtracting files");
    while(equeue.process());
    
    return 0;
}


