
/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: fileiterator.h 2791 2013-05-06 11:24:28Z michael $
 */

#ifndef FILEITERATOR_INCLUDED
#define FILEITERATOR_INCLUDED

#include "../shared/misc_defines.h"

#include <QString>
#include <QByteArray>
#include <QList>
#include <QSharedPointer>
#include <QFileInfo>


//#include "../shared/maybe.tpl"

class QFile;

namespace vhist
{

typedef QList<QByteArray> QByteArrays;
typedef QSharedPointer<QString> MaybeString;
typedef QSharedPointer<QByteArray> MaybeArray;

/**
 * This class makes it possible to iterate linewise over a file in both
 * directions without keeping the whole file inside memory. The class
 * maintains a cursor which points to the current line. The cursor can either
 * be moved up, down, to the beginning or end of the document or placed at
 * an arbitrary position inside the file.
 */
class VHIST_EXTERN FileIterator
{
    public: // construction, destruction and assignment
        FileIterator();
        ~FileIterator();
        
    public:
        bool open(const QString& filename);
        bool nextLine();
        bool prevLine();
        bool firstLine();
        bool lastLine();
        
        bool reachedStart();
        bool reachedEnd();
        bool onLine();
        bool empty();
        bool gotoPos(int pos);
        
        MaybeArray curLine();
        MaybeString curLineAsString();
        int curPos();
        QByteArray getBytes(int startPos, int numBytes);
        
    public:
        bool curLineStartsWithBytes(const char* str);

    private:
        bool containsNulls(const QByteArray& line);
        void closeFile();
        int getFileSize();

        bool getFirstBuffer();
        bool getLastBuffer();
        bool getNextBuffer();
        bool getPrevBuffer();
        bool getBuffer(int pos);
        
        bool searchForwards();
        bool searchBackwards();
    
    private:
        QFile* m_file;
        QFileInfo m_finfo;
        QByteArray m_buffer;
        QByteArray m_curLine;
        int m_startOfLine;
        int m_bufferStart;
        int m_posInBuffer;
};


} // namespace


#endif // FILEITERATOR_INCLUDED

