/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: validationqueue.h 2792 2013-05-06 11:25:27Z michael $
 */

#ifndef VALIDATIONQUEUE_INCLUDED
#define VALIDATIONQUEUE_INCLUDED

#include "../shared/misc_defines.h"

#include <QObject>
#include <QString>
#include <QQueue>
#include <QFile>

#include "validator.h"
#include "../shared/stopwatch.h"
#include "../shared/logsource.h"

namespace vhist
{

class VhistParser;
struct FileTag;
struct SectionTag;

enum ValidateeType
{
    VALIDATEE_SECTION,
    VALIDATEE_FILE,
    VALIDATEE_SAVEDFILE
};


struct VHIST_EXTERN Validatee
{
    ValidateeType type;
    int id;
    int start;
    int numBytes;
    int tagPos;
    QByteArray replacementTag;
    QString md5;
    QString cmd5;
    QString name; // file- or sectionname
};


struct VHIST_EXTERN SavedFile
{
    QString filename;
    QString md5;
    int filesize;
};


/**
 * A queue datastructure, which validates parts of a VHIST file.
 * Sections and files can be enqueued at the end of the queue and are
 * validated one after another. Entries in the queue are not validated in
 * one single (blocking) step but continuously. Each call to the process
 * method will validate a small fraction of data and inform whether all
 * files and sections in the queue have already been validated. Progress
 * information is published by a number of signals. Each instance of the
 * class uses one vhist::VhistParser instance as the source for embedded
 * files and sections.
 */
class VHIST_EXTERN ValidationQueue : public LogSource
{
    Q_OBJECT
    public: // construction, destruction, assignment
        ValidationQueue(VhistParser* parser = NULL, int chunksize = 100000);
        virtual ~ValidationQueue();
        void reset(VhistParser* parser);
        void repopulate(VhistParser* parser);
        void addValidatees(QList<int> files, QList<int> sections);
        void addValidatees(QList<SavedFile> files);
    
    public slots:
        void cancelValidation();
    
    public:
        bool process();
    
    signals:
        void fileProcessed(int id, bool valid);
        void savedFileProcessed(const QString filename, bool valid);
        void sectionProcessed(int id, bool valid);
        void percentFinishedChanged(int, const QString name);
    
    private:
        void reinit(VhistParser* parser);
        void nextEntry();
        Validatee validateeFromFile(const FileTag& tag);
        Validatee validateeFromSection(const SectionTag& tag);
        Validatee validateeFromSavedFile(const SavedFile& file);
    
    private: // member variables
        VhistParser* m_parser;
        Validator m_validator;
        StopWatch m_stopwatch;
        QQueue<Validatee> m_validatees;
        // the number of processed bytes of the current validatee
        int m_bytesHandled;
        // the total number of bytes to handle since the last restart
        qint64 m_totalBytes;
        // the total number of bytes handled since the last restart
        qint64 m_totalHandled;
        QFile m_currentFile;
        int m_chunksize;
};

}

#endif // VALIDATIONQUEUE_INCLUDED

