/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: argfileparser.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#include "../../include/arglib/argfileparser.h"

#include <QString>
#include <QList>
#include <QRegExp>
#include <QtDebug>


using namespace vhist;

/**
 * Creates a new ArgfileParser object.
 */
ArgfileParser::ArgfileParser()
{
}


/**
 * Destroys the ArfileParser object.
 */
ArgfileParser::~ArgfileParser()
{
}


/**
 * Parses the given argfile and returns a list of string arguments
 * similar to \a argv.
 *
 * @param content The content of an argfile, which should be parsed.
 * @param env An optional Environment containing variable names and
 * values. Variables used inside the argfile are replaced by the values
 * from the Environment. If a variable does not appear within the
 * Environment, it is replaced by an empty string. If \a env is \c NULL,
 * no variables are replaced.
 *
 * @return A list containing the single arguments as strings.
 * @note Since \c $ signs inside an argfile are used to indicate variables,
 * literal \c $ signs within the argfile are represented by two \c $ signs
 * in the argument list returned by parseArgfile().
 */
QStringList ArgfileParser::parseArgfile(const QString& content,
        const Environment* env)
{
    TokenList tokens = splitIntoTokens(content);
    //qDebug() << "1" << tokensToStrings(tokens) << "\n";
    tokens = removeDoubleSeparators(tokens);
    //qDebug() << "2" << tokensToStrings(tokens) << "\n";
    tokens = handleVariables(tokens, env);
    //qDebug() << "3" << tokensToStrings(tokens) << "\n";
    tokens = unescapeTokens(tokens);
    //qDebug() << "4" << tokensToStrings(tokens) << "\n";
    tokens = joinTokens(tokens);
    return tokensToStrings(tokens);
}


TokenList ArgfileParser::getHighlighterList(const QString& content)
{
    TokenList tokens = splitIntoTokens(content);
    tokens = unescapeTokens(tokens);
    return tokens;
}

TokenList ArgfileParser::splitIntoTokens(const QString& content)
{
    // regular expressions
    QRegExp whitespaceRegex("[\r\n\t ]+");
    QRegExp commentRegex("#[^\\r\\n]*.?");
    QRegExp quotedBackslashRegex("([^\\\\'\"#\r\n\t ]|\\\\.)+");
    QRegExp quotedSingleRegex("('[^']*')");
    QRegExp quotedDoubleRegex("(\"([^\"\\\\]|\\\\.)*\")");
    
    int pos = 0;
    QString rest = content;
    TokenList result;
    
    while (rest != "")
    {
        QString found("");
        if (whitespaceRegex.indexIn(rest) == 0)
        {
            // remove whitespace
            found = whitespaceRegex.cap(0);
            result << ArgfileToken(ArgfileToken::SEPARATOR, pos, found);
            pos += found.length();
        }
        else if (commentRegex.indexIn(rest) == 0)
        {
            // remove comment
            found = commentRegex.cap(0);
            result << ArgfileToken(ArgfileToken::SEPARATOR, pos, found);
            pos += found.length();
        }
        else if (quotedBackslashRegex.indexIn(rest) == 0)
        {
            found = quotedBackslashRegex.cap(0);
            result << ArgfileToken(found, pos, found);
            pos += found.length();
        }
        else if (quotedSingleRegex.indexIn(rest) == 0)
        {
            found = quotedSingleRegex.cap(0);
            result << ArgfileToken(found, pos, found);
            pos += found.length();
        }
        else if (quotedDoubleRegex.indexIn(rest) == 0)
        {
            found = quotedDoubleRegex.cap(0);
            result << ArgfileToken(found, pos, found);
            pos += found.length();
        }
        else
        {
            // error
            return TokenList();
        }
        
        rest = rest.mid(found.length());
    }
    return result;
}


TokenList ArgfileParser::removeDoubleSeparators(const TokenList& tokens)
{
    TokenList result;
    bool lastWasASeparator = false;
    
    for (int i = 0; i < tokens.size(); i++)
    {
        if (tokens[i].isSeparator() && !lastWasASeparator)
        {
            result << tokens[i];
            lastWasASeparator = true;
        }
        else if (!tokens[i].isSeparator())
        {
            result << tokens[i];
            lastWasASeparator = false;
        }
    }
    
    // remove leading and trailing sepatators
    if (result.size() > 0)
    {
        if (result[0].isSeparator())
            result.removeAt(0);
    }
    
    if (result.size() > 0)
    {
        if (result[result.size()-1].isSeparator())
            result.removeAt(result.size()-1);
    }

    return result;
}


TokenList ArgfileParser::unescapeTokens(const TokenList& tokens)
{
    TokenList result;
    for (int i = 0; i < tokens.size(); i++)
    {
        const ArgfileToken& curToken = tokens.at(i);
        if (curToken.isSeparator())
            result << curToken;
        else
        {
            QString token = curToken.str();
            QString token2("");
            if (token.startsWith('\''))
                token2 = unescapeSingleQuoted(token);
            else if (token.startsWith('\"'))
                token2 = unescapeDoubleQuoted(token);
            else
                token2 = unescapeBackslashed(token);
            
            result << ArgfileToken(token2, curToken.position(),
                    curToken.originalText());
        }
    }
    return result;
}


TokenList ArgfileParser::handleVariables(const TokenList& tokens,
        const Environment* env)
{
    TokenList result;
    for (int i = 0; i < tokens.size(); i++)
    {
        // ignore separators
        if (tokens[i].isSeparator())
            result << tokens[i];
        else
        {
            QString token = tokens[i].str();
            if (token.startsWith('\"')) // double quotes
            {
                QStringList subtokens = findVariablesDoubleQuoted(token);
                QString s = replaceVariables(subtokens, env).str();
                result << ArgfileToken("\"" + s + "\"");
            }
            else if (token.startsWith('\'')) // single quotes
                // variables cannot appear in single quoted text
                result << ArgfileToken(token);
            else // no quotes
            {
                QStringList subtokens = findVariablesBackslashed(token);
                result << replaceVariables(subtokens, env);
            }
        }
    }
    return result;
}


QStringList ArgfileParser::findVariablesInToken(const ArgfileToken& token)
{
    QRegExp regex("\\\\*\\$([A-Za-z0-9]+|\\{[A-Za-z0-9]+\\})");
    QStringList result;
    QString rest = token.str();
    //qDebug() << "token:" << rest;
    int pos = 0;
    
    while (rest != "")
    {
        if (regex.indexIn(rest, pos) == -1)
        {
            //qDebug() << "blahh";
            result << rest;
            rest = "";
        }
        else
        {
            // get variable string (+ preceeding backslashes)
            QString candidate = regex.cap(0);
            int posOfVar = candidate.indexOf("$");
            int lenOfVar = candidate.size() - posOfVar;
            // odd number of backslashes => $ is escaped => no variable
            if (posOfVar % 2 == 1)
                pos = regex.pos() + candidate.size();
            else
            {
                QString preVar = rest.left(posOfVar);
                if (preVar.size() != 0)
                    result << preVar;
                
                QString var = rest.mid(posOfVar, lenOfVar);
                result << var;
                
                rest = rest.mid(candidate.size());
                pos = 0;
            }
        }
    }
    //qDebug() << result;
    return result;
}


ArgfileToken ArgfileParser::replaceVariables(const QStringList& list,
        const Environment* env)
{
    QStringList result;
    for (int i = 0; i < list.size(); i++)
        if (list[i].startsWith("$") && env != NULL)
        {
            QString var = getVariableName(list[i]);
            result << escapeEverything(env->value(var, ""));
        }
        else
            result << list[i];
    
    return result.join("");
}


QString ArgfileParser::getVariableName(const QString& var)
{
    if (var.startsWith("${") && var.endsWith("}"))
        return var.mid(2, var.size() - 3);
    else if (var.startsWith("$"))
        return var.mid(1);
    else
        return "";
}


QStringList ArgfileParser::findVariablesBackslashed(const ArgfileToken& token)
{
    return findVariablesInToken(token);
}


QStringList ArgfileParser::findVariablesDoubleQuoted(const ArgfileToken& token)
{
    // remove quotes (first and last character)
    QString tmp = token.str().mid(1);
    tmp.chop(1);
    return findVariablesInToken(tmp);
}


QString ArgfileParser::unescapeDoubleQuoted(const QString& text)
{
    QString result = text.mid(1);
    result.chop(1);
    return unescapeBackslashed(result);
}


QString ArgfileParser::unescapeSingleQuoted(const QString& text)
{
    QString result = text.mid(1);
    result.chop(1);
    result.replace("$", "$$");
    return result;
}


QString ArgfileParser::unescapeBackslashed(const QString& text)
{
    QString result;
    result.reserve(text.size());
    
    for (int i = 0; i < text.size(); i++)
    {
        if ((text[i] == '\\') && (i < text.size() - 1))
        {
            if (text[i+1] == '$')
                result.append("$$");
            else
                // skip backslash
                result.append(text[i+1]);
            i++;
        }
        else
            result.append(text[i]);
    }
    return result;
}


QString ArgfileParser::escapeEverything(const QString& text)
{
    QString result;
    result.reserve(text.size() * 2);
    for (int i = 0; i < text.size(); i++)
        result.append("\\").append(text[i]);
    return result;
}


TokenList ArgfileParser::joinTokens(const TokenList& tokens)
{
    TokenList result;
    QString curToken;
    bool lastWasSeparator = true;
    
    for (int i = 0; i < tokens.size(); i++)
    {
        if (tokens[i].isSeparator())
        {
            result << ArgfileToken(curToken);
            curToken = "";
            lastWasSeparator = true;
        }
        else
        {
            curToken += tokens[i].str();
            lastWasSeparator = false;
        }
    }
    
    if (!lastWasSeparator)
        result << ArgfileToken(curToken);
    return result;
}


QStringList ArgfileParser::tokensToStrings(const TokenList& tokens)
{
    QStringList result;
    for (int i = 0; i < tokens.size(); i++)
        result << tokens[i].str();
    return result;
}

