
#include "../../include/arglib/argfilereader.h"
#include "../../include/arglib/argfileparser.h"

#include <QStringList>
#include <QHash>

#include <QtDebug>

using namespace vhist;


/**
 * Creates a new ArgfileReader.
 */
ArgfileReader::ArgfileReader()
{
}


/**
 * Destroys the ArgfileReader.
 */
ArgfileReader::~ArgfileReader()
{
}


/**
 * Parses the given argfile and generates an ArgFile object from
 * it. An optional Environment containing environment variables and their
 * values can be passed to the method. If \a env is \c NULL,
 * variables inside the argfile will not be replaced.
 * 
 * @param content The content of the argfile, which should be read.
 * @param env An optional hashtable containing environment variables.
 * If a variable is used in the argfile but is not part of the environment,
 * it is replaced by an empty string. If \a env is \c NULL, variables are
 * not replaced.
 *
 * @return An ArgFile object representing the input argfile.
 */
ArgFile ArgfileReader::readArgfile(const QString& content,
        const Environment* env)
{
    ArgfileParser parser;
    QStringList args = parser.parseArgfile(content, env);
    
    args = checkOptionSignatures(args);
    return parseCommandline(args);
}


void ArgfileReader::saveDocAttr(const QString& option,
        const QStringList& args, int pos, ArgFile& argfile)
{
    QString key = args[pos];;
    QString value = args[pos+1];
    if (key == "producer")
        argfile.setPdfProducer(value);
    else if (key == "creator")
        argfile.setPdfCreator(value);
    else if (key == "author")
        argfile.setPdfAuthor(value);
    else if (key == "keywords")
        argfile.setPdfKeywords(value);
    else if (key == "title")
        argfile.setPdfTitle(value);
    else if (key == "subject")
        argfile.setPdfSubject(value);
    else
    {
        QString err("first parameter (%1) of option %2 is invalid");
        emit log(err.arg(value).arg(option), LogSource::LOG_ERROR);
    }
}


void ArgfileReader::saveStepAttr(const QString& option,
        const QStringList& args, int pos, ArgFile& argfile)
{
    QString key = args[pos];
    QString value = args[pos+1];
    if (key == "comment")
        argfile.setComment(value);
    else if (key == "description")
        argfile.setDescription(value);
    else if (key == "title")
        argfile.setTitle(value);
    else if (key == "tool" || key == "toolversion")
        argfile.setTool(value);
    else if (key == "toolpath")
        argfile.setToolpath(value);
    else if (key == "host")
        argfile.setHost(value);
    else if (key == "command")
        argfile.setCommand(value);
    else if (key == "user")
        argfile.setUser(value);
    else
    {
        QString err("first parameter (%1) of option %2 is invalid");
        emit log(err.arg(value).arg(option), LogSource::LOG_ERROR);
    }
}


void ArgfileReader::saveFileAttr(const QString& option,
        const QStringList& args, int pos, WSFile& wsFile)
{
    QString key = args[pos];
    QString value = args[pos+1];
    if (key == "comment")
        wsFile.setComment(value);
    else if (key == "md5file")
        wsFile.setMD5File(value);
    else if (key == "filetype")
        wsFile.setFiletype(value);
    else if (key == "description")
        wsFile.setDescription(value);
    else
    {
        QString err("first parameter (%1) of option %2 is invalid");
        emit log(err.arg(value).arg(option), LogSource::LOG_ERROR);
    }
}


void ArgfileReader::saveFileFlag(const QString& option,
        const QStringList& args, int pos, WSFile& wsFile)
{
    QString baseOption(args[pos]);
    bool value = true;
    if (option.startsWith("no-"))
    {
        value = false;
        baseOption = baseOption.mid(3);
    }
    
    if (baseOption == "optional")
        wsFile.setOptional(value);
    if (baseOption == "compress")
        wsFile.setCompress(value);
    if (baseOption == "thumbnailonly")
        wsFile.setThumbnailOnly(value);
    if (baseOption == "previewws")
        wsFile.setPreviewWS(value);
    if (baseOption == "embed")
        wsFile.setEmbed(value);
    if (baseOption == "preview")
        wsFile.setPreview(value);
    if (baseOption == "automd5")
        wsFile.setAutoMD5(value);
    if (baseOption == "thumbnail")
        wsFile.setThumbnail(value);
    else
    {
        QString err("\"%1\" is no valid parameter for option %2");
        emit log(err.arg(value).arg(option), LogSource::LOG_ERROR);
    }
}


QStringList ArgfileReader::checkOptionSignatures(const QStringList& args)
{
    QStringList result;
    
    QHash<QString, int> numOfParams;
    QStringList zeroParams;
    zeroParams << "-h" << "--help" << "-V" << "--version" <<
            "--sync-cwd" << "-p" << "--pretend" << "-q" << "--quiet" <<
            "-v" << "--verbose" << "--stdin" << "--debug";
    QStringList oneParam;
    oneParam << "-i" << "--infile" << "-o" << "--outfile" <<
            "-I" << "-J" << "-O" << "-A" << "--append-to-vhistfile" <<
            "-f" << "--file-flag" << "-r" << "-1" << "-c";
    QStringList twoParams;
    twoParams << "-d" << "--doc" << "-s" << "--step" <<
            "-U" << "--step-user" << "-a" << "--file-attr" << "-u" <<
            "--file-user";
    
    for (int i = 0; i < zeroParams.size(); i++)
        numOfParams.insert(zeroParams[i], 0);
    
    for (int i = 0; i < oneParam.size(); i++)
        numOfParams.insert(oneParam[i], 1);
    
    for (int i = 0; i < twoParams.size(); i++)
        numOfParams.insert(twoParams[i], 2);
    
    
    for (int i = 0; i < args.size(); i++)
    {
        QString curParam = args[i];
        if (!numOfParams.contains(curParam))
        {
            log(QString("unknown option found: %1").arg(curParam),
                    LogSource::LOG_ERROR);
            return result;
        }
        else if (i + numOfParams[curParam] >= args.size())
        {
            log(QString("not enough parameters for option: %1").arg(curParam),
                    LogSource::LOG_ERROR);
            return result;
        }
        
        for (int j = 0; j < numOfParams[curParam] + 1; j++)
            result << args[i+j];
        i += numOfParams[curParam];
    }
    
    return result;
}


ArgFile ArgfileReader::parseCommandline(const QStringList& args)
{
    WSFile* curWSFile = NULL;
    ArgFile af;
    
    for (int i = 0; i < args.size(); i++)
    {
        QString curArg = args[i];
        
        /* vhistfile options */
        if (curArg == "-I")
        {
            af.setRootfile(args[i+1]);
            af.setRootIsOptional(false);
            i += 1;
        }
        else if (curArg == "-J")
        {
            af.setRootfile(args[i+1]);
            af.setRootIsOptional(true);
            i += 1;
        }
        else if (curArg == "-O")
        {
            af.addOutputVhistfile(args[i+1]);
            i += 1;
        }
        else if (curArg == "-A" || curArg == "--append-to-vhistfile")
        {
            af.addOutputVhistfile(args[i+1]);
            af.setRootfile(args[i+1]);
            af.setRootIsOptional(true);
            i += 1;
        }
        /* document and workflowstep options */
        else if (curArg == "-d" || curArg == "--doc")
        {
            saveDocAttr(curArg, args, i+1, af);
            i += 2;
        }
        else if (curArg == "-s" || curArg == "--step")
        {
            saveStepAttr(curArg, args, i+1, af);
            i += 2;
        }
        else if (curArg == "-U" || curArg == "--step-user")
        {
            af.addUserAttribute(args[i+1], args[i+2]);
            i += 2;
        }
        /* new input/output file options*/
        else if (curArg == "-i" || curArg == "--infile" || curArg == "-o" ||
                curArg == "--outfile")
        {
            WSFile f;
            WSFile::FilePurpose purpose = WSFile::OUTPUT_FILE;
            if (curArg == "-i" || curArg == "--infile")
                purpose = WSFile::INPUT_FILE;
            
            f.setFilePurpose(purpose);
            f.setFilenameAndPath(args[i+1]);
            curWSFile = af.addFile(f);
            i += 1;
        }
        /* file-flags and -attributes options */
        else if (curArg == "-a" || curArg == "--file-attr")
        {
            if (curWSFile != NULL)
                saveFileAttr(curArg, args, i+1, *curWSFile);
            i += 2;
        }
        else if (curArg == "-f" || curArg == "--file-flag")
        {
            if (curWSFile != NULL)
                saveFileFlag(curArg, args, i+1, *curWSFile);
            i += 1;
        }
        else if (curArg == "-u" || curArg == "--file-user")
        {
            if (curWSFile != NULL)
                curWSFile->addUserAttribute(args[i+1], args[i+2]);
            i += 2;
        }
        /* misc options */
        else if (curArg == "-p" || curArg == "--pretend")
            af.setPretend(true);
        else if (curArg == "-v" || curArg == "--verbose")
            af.setVerbosity(ArgFile::VERBOSE);
        else if (curArg == "-q" || curArg == "--quiet")
            af.setVerbosity(ArgFile::QUIET);
        else if (curArg == "-r" || curArg == "--readme")
        {
            af.setReadme(args[i+1]);
            i += 1;
        }
        else if (curArg == "-1" || curArg == "--firstpage")
        {
            af.setFirstpage(args[i+1]);
            i += 1;
        }
        else if (curArg == "--stdin")
        {
            // ignore --stdin (makes no sense in argfiles)
        }
        else if (curArg == "--debug")
            af.setDebugMode(true);
        else if (curArg == "-h" || curArg == "--help")
            af.setShowHelp(true);
        else if (curArg == "-V" || curArg == "--version")
            af.setShowVersion(true);
        else
        {
            QString err("invalid option (%1) found");
            emit log(err.arg(curArg));
        }
    }
    return af;
}

