/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: stopwatch.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#include "../../include/shared/stopwatch.h"
#include <time.h>

using namespace vhist;

/**
 * Creates a new StopWatch object and adds \a pLogTarget as
 * a listener. The listener will receive all timing results
 * generated by this object.
 *
 * @param[in] pLogTarget An object which receives timing events
 * via a slot.
 *
 * @see vhist::LogSource::publishToLogSlot
 */
StopWatch::StopWatch(QObject *pLogTarget)
  : LogSource(pLogTarget), m_running(false), m_startTime(0)
{
}


/**
 * Destroys the StopWatch object.
 */
StopWatch::~StopWatch()
{
    stop();
}


/**
 * Starts a new timing sequence. The StopWatch will count the time elapsed
 * till the next call to stop(). If another timing sequence is already
 * running, it will be stopped before starting the new one. It's results
 * will be published to all listeners of the StopWatch unless
 * \a abortPrevious is set to \c true in which case no results are published.
 *
 * @param[in] title The title of the new timing sequence.
 * @param[in] abortPrevious controls whether the currently running
 * timing sequence should be aborted or stopped and results published.
 */
void StopWatch::start(const QString& title, bool abortPrevious)
{
    stop(!abortPrevious);
    m_startTime = clock();
    m_running = true;
    m_title = title;
}


/**
 * Stops the currently running timing sequence. If no sequence is running,
 * calling this function is a \e NoOp.
 *
 * @param[in] generateMessage If \c true, a log message is generated
 * and will be published to all listeners. Otherwise, the timing results
 * will be silently discarded
 */
void StopWatch::stop(bool generateMessage)
{
    if (m_running && generateMessage)
    {
        double time = 1.0 * (clock() - m_startTime) / CLOCKS_PER_SEC;
        QString timeStr = QString::number(time, 'f', 2);
        emit log(m_title + ": " + timeStr + "s\n");
    }
    m_running = false;
}

