#include "../../include/unvhistlib/fileiterator.h"

#include <QFile>
#include <QFileInfo>
#include <QDebug>

using namespace vhist;

const int BUFFER_SIZE = 1024 * 10;

///////////////////////////////////////////////////////////////////
// Helper Functions

void appendToBuffer(QByteArray& dst, const QByteArray& src, int pos,
        int numBytes = -1);

/**
 * Appends numBytes of src to the end of dst, starting at position pos.
 * If numBytes is omitted, the complete buffer (starting at position pos)
 * is appended.
 */
void appendToBuffer(QByteArray& dst, const QByteArray& src, int pos,
        int numBytes)
{
    if (numBytes == -1)
        numBytes = src.size() - pos;
    numBytes = qBound(0, numBytes, src.size());
    const char* data = src.constData() + pos;
    dst.append(data, numBytes);
}


void prependToBuffer(QByteArray& dst, const QByteArray& src, int end,
        int numBytes = -1);

/**
 * Prepends numBytes of src to the start of dst, ending (excluding) with
 * position end.
 * If numBytes is omitted, the first end bytes are prepended.
 */
void prependToBuffer(QByteArray& dst, const QByteArray& src, int end,
        int numBytes)
{
    if (numBytes == -1)
        numBytes = end;
    numBytes = qBound(0, numBytes, src.size());
    int pos = end - numBytes;
    const char* data = src.constData() + pos;
    dst.prepend(QByteArray(data, numBytes));
}


/**
 * The first n bytes of buffer for the last occurence of ch,
 * This means that if the from is 0, no search occures.
 */
int findLastIndexOf(const QByteArray& buffer, char ch, int from)
{
    from--;
    if (from < 0)
        return -1;
    else
        return buffer.lastIndexOf(ch, from);
}



//////////////////////////////////////////////////////////////////

/**
 * Creates a new FileIterator object.
 */
FileIterator::FileIterator() : m_file(NULL),
        m_startOfLine(-1), m_bufferStart(-1), m_posInBuffer(0)
{
}


/**
 * Destroys the FileIterator object.
 */
FileIterator::~FileIterator()
{
    closeFile();
}


/**
 * Closes the currently open file.
 */
void FileIterator::closeFile()
{
    if (m_file)
        delete m_file;
    m_buffer.clear();
    m_bufferStart = -1;
    m_posInBuffer = 0;
    m_startOfLine = -1;
    m_file = NULL;
    m_finfo = QFileInfo();
}


/**
 * Returns the size of the file.
 *
 * @return the filesize in bytes.
 */
int FileIterator::getFileSize()
{
    if (m_file)
    {
        return m_finfo.size();
    }
    return 0;
}


/**
 * Opens the given file for iterating. If the file cannot be opened,
 * \c false is returned.
 *
 * @param[in] filename The name of the file.
 *
 * @return \c true if opening the file was successful and \c false otherwise.
 */
bool FileIterator::open(const QString& filename)
{
    closeFile();
    m_file = new QFile(filename);
    if (m_file->open(QIODevice::ReadOnly | QIODevice::Unbuffered))
    {
        m_finfo = QFileInfo(*m_file);
        firstLine();
        return true;
    }
    return false;
}


/**
 * Moves the cursor to the next line. If the cursor is already located at the
 * end of the file, \c false is returned.
 *
 * @return \c true if the cursor was moved to the next line and \c false if
 * the cursor has already reached the end of the file.
 */
bool FileIterator::nextLine()
{
    return searchForwards();
}


/**
 * Moves the cursor to the previous line. If the cursor is already located
 * at the end of the file, \c false is returned.
 *
 * @return \c true if the cursor was moved to the previous line and \c false
 * if the cursor has already reached the beginning of the file.
 */
bool FileIterator::prevLine()
{
    return searchBackwards();
}


/**
 * Moves the cursor to the first line of the file. Returns \c false if the
 * file is empty.
 * 
 * @return \c true on success and \c false if the document is empty.
 */
bool FileIterator::firstLine()
{
    return getFirstBuffer() && searchForwards();
}


/**
 * Move the cursor to the last line of the file. Returns \c false if the file
 * is empty.
 *
 * @return \c true on success and \c false if the document is empty.
 */
bool FileIterator::lastLine()
{
    return getLastBuffer() && searchBackwards();
}


/**
 * Returns \c true if the cursor has reached the start of the file.
 *
 * @return \c true if the cursor has reached the start of the file and
 * \c false otherwise.
 */
bool FileIterator::reachedStart()
{
    return m_bufferStart == 0 && m_posInBuffer == 0;
}


/**
 * Returns \c true if the cursor has reached the end of the file.
 *
 * @return \c true if the cursor has reached the end of the file and
 * \c false otherwise.
 */
bool FileIterator::reachedEnd()
{
    return getFileSize() <= m_bufferStart + m_posInBuffer;
}


/**
 * Not implemented, yet.
 */
bool FileIterator::onLine()
{
    return false;
}


/**
 * Returns whether the file is empty or not.
 *
 * @return \c true if the file is empty and \c false otherwise.
 */
bool FileIterator::empty()
{
    return getFileSize() <= 0;
}


bool FileIterator::containsNulls(const QByteArray& line)
{
    return line.contains('\0');
}


/**
 * Returns the current line as a QByteArray. If there is no current line,
 * e.g. because the file is empty or the cursor has reached the start
 * or end of the file, MaybeArray::nothing() is returned.
 *
 * @return The current line as a QByteArray.
 */
MaybeArray FileIterator::curLine()
{
    if (m_file)
        return MaybeArray(new QByteArray(m_curLine));
    return MaybeArray(NULL);
}


/**
 * Returns the current line as a QString. The line is therefore
 * interpreted as beeing encoded in UTF8. If there is no current line
 * or the line cannot be decoded from UTF8, MaybeString::nothing() is
 * returned.
 *
 * @return The current line as a QString.
 */
MaybeString FileIterator::curLineAsString()
{
    MaybeArray maybe = curLine();
    if (maybe != NULL)
        if (!containsNulls(*maybe))
        {
            QString lineStr = QString::fromUtf8(*maybe);
            return MaybeString(new QString(lineStr));
        }
    return MaybeString(NULL);
}


/**
 * Moves the cursor to the given location. Note that this position does not
 * have to be the beginning of a line. In this case, only the part of the
 * line following the cursor position is returned when calling curLine or
 * curLineAsString after gotoPos.
 * 
 * @param[in] pos The position to which the cursor should be moved. It's
 * measured in bytes and is relative to the beginning of the file.
 * 
 * @return \c true if moving the cursor was successful and \c false otherwise.
 */
bool FileIterator::gotoPos(int pos)
{
    if (pos >= 0 && pos < getFileSize())
    {
        m_bufferStart = qMax(0, pos - BUFFER_SIZE / 2);
        m_posInBuffer = pos - m_bufferStart;
        
        return getBuffer(m_bufferStart) && searchForwards();
    }
    return false;
}


/**
 * Returns the current position of the cursor in bytes relative to the
 * beginning of the document.
 *
 * @return The current position of the cursor.
 */
int FileIterator::curPos()
{
    return m_startOfLine;
}


/**
 * Returns a chunk of data from the file.
 *
 * @param[in] startPos The position of the first byte in the
 * requested chunk.
 *
 * @param[in] numBytes The number of bytes of the requested chunk.
 *
 * @return The chunk of data or an empty buffer if the requested chunk is
 * not completely inside the file.
 */
QByteArray FileIterator::getBytes(int startPos, int numBytes)
{
    if (startPos >= 0 && startPos < getFileSize() && numBytes >= 0
            && startPos + numBytes <= getFileSize() && m_file)
    {
        m_file->seek(startPos);
        QByteArray result = m_file->read(numBytes);
        return result;
    }
    return QByteArray();
}


bool FileIterator::curLineStartsWithBytes(const char* str)
{
    return m_curLine.startsWith(str);
}


////////////////////////////////////////////////
// internals

bool FileIterator::searchForwards()
{
    bool atEnd = reachedEnd();
    m_startOfLine = m_bufferStart + m_posInBuffer;
    if (!atEnd)
    {
        m_curLine.clear();
        int nextNewline = m_buffer.indexOf('\n', m_posInBuffer);
        while (nextNewline == -1)
        {
            appendToBuffer(m_curLine, m_buffer, m_posInBuffer);
            bool success = getNextBuffer();
            if (!success)
                break;
            nextNewline = m_buffer.indexOf('\n', m_posInBuffer);
        }
        
        if (nextNewline != -1)
        {
            // make sure the newline is in the buffer
            int numBytes = nextNewline - m_posInBuffer + 1;
            appendToBuffer(m_curLine, m_buffer, m_posInBuffer, numBytes);
            m_posInBuffer = nextNewline + 1;
        }
    }
    return !atEnd;
}


bool FileIterator::searchBackwards()
{
    bool atStart = reachedStart();
    if (!atStart)
    {
        m_curLine.clear();
        // find previous newline, start before the current character (newline)
        int nextNewline = findLastIndexOf(m_buffer, '\n', m_posInBuffer-1);
        while (nextNewline == -1)
        {
            prependToBuffer(m_curLine, m_buffer, m_posInBuffer);
            bool success = getPrevBuffer();
            if (!success)
                break;
            // don't skip the last character here (new buffer)
            nextNewline = findLastIndexOf(m_buffer, '\n', m_posInBuffer);
        }
        
        if (nextNewline != -1)
        {
            int numBytes = m_posInBuffer - nextNewline - 1;
            prependToBuffer(m_curLine, m_buffer, m_posInBuffer, numBytes);
            m_posInBuffer = nextNewline + 1;
        }
    }
    m_startOfLine = m_bufferStart + m_posInBuffer;
    return !atStart;
}


bool FileIterator::getFirstBuffer()
{
    m_bufferStart = 0;
    m_posInBuffer = 0;
    return getBuffer(m_bufferStart);
}


bool FileIterator::getLastBuffer()
{
    m_bufferStart = getFileSize();
    return getPrevBuffer();
}


bool FileIterator::getNextBuffer()
{
    m_bufferStart += m_buffer.size();
    m_posInBuffer = 0;
    return getBuffer(m_bufferStart);
}


bool FileIterator::getPrevBuffer()
{
    int oldBufferStart = m_bufferStart;
    if (m_bufferStart > 0)
    {
        m_bufferStart = qMax(0, m_bufferStart - BUFFER_SIZE);
        if (getBuffer(m_bufferStart))
        {
            m_posInBuffer = oldBufferStart - m_bufferStart;
            return true;
        }
    }
    m_posInBuffer = 0;
    return false;
}


bool FileIterator::getBuffer(int pos)
{
    if (m_file)
    {
        bool seekSuccess = m_file->seek(pos);
        bool endReached = m_file->atEnd();
        m_buffer = m_file->read(BUFFER_SIZE);
        //QByteArray b2(m_buffer);
        //b2.replace("\n", "\\n");
        //qDebug() << b2;
        return seekSuccess && (!endReached || !m_buffer.isEmpty());
    }
    return false;
}

