/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: md5generator.cpp 2793 2013-05-06 11:26:17Z michael $
 */

#include "../../include/unvhistlib/md5generator.h"

#include <QByteArray>

using namespace vhist;

/**
 * Creates a new MD5Generator instance and calculates the
 * MD5 sum for the given chunk of data. The calculation of the
 * checksum can be continued by calling update.
 * 
 * @param[in] data The data for which an MD5 sum should be calculated.
 */
MD5Generator::MD5Generator(const QByteArray& data)
{
    reset();
    update(data);
}


/**
 * Creates a new MD5Generator instance.
 */
MD5Generator::MD5Generator()
{
    reset();
}


/**
 * Resets the MD5Generator.
 */
void MD5Generator::reset()
{
    md5_init(&m_state);
    m_hexDigest = "";
}


/**
 * Destroys the MD5Generator instance.
 */
MD5Generator::~MD5Generator()
{
}


/**
 * Feeds more data to the MD5Generator. By using this function, calculating 
 * checksums can be done incrementally.
 *
 * @param[in] data The data over which the MD5 sum is calculated.
 */
void MD5Generator::update(const QByteArray& data)
{
    md5_append(&m_state, reinterpret_cast<const md5_byte_t*>(data.constData()),
            data.count());
    //md5_state_t *pms, const md5_byte_t *data, int nbytes);
}


/**
 * Returns the current checksum in hexdigest form.
 *
 * @return The MD5 sum.
 */
QString MD5Generator::getHexDigest()
{
    if (m_hexDigest != "")
        return m_hexDigest;
    
    char lut[] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b',
            'c', 'd', 'e', 'f'};
    
    md5_byte_t md5sum[16];
    md5_finish(&m_state, md5sum);
    
    QString hexdigest;
    hexdigest.reserve(32);
    for (int i = 0; i < 16; i++)
    {
        unsigned char b1 = md5sum[i] >> 4;
        unsigned char b2 = md5sum[i] & 0xf;
        hexdigest.append(lut[b1]);
        hexdigest.append(lut[b2]);
    }
    
    m_hexDigest = hexdigest;
    return hexdigest;
}

