/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: vhistparser.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#include <QRegExp>
#include <QStringList>
#include <QFile>
#include <QtDebug>
#include <QtEndian>
#include <QtDebug>
#include "time.h"
#include <QSet>

#include "../../include/unvhistlib/vhistparser.h"
#include "../../include/unvhistlib/md5generator.h"
#include "../../include/unvhistlib/inflater.h"

using namespace vhist;

// (any character but \ and ]) OR (\\ or \] or \n)
QString strRegex("((?:[^\\]\\\\]|(?:\\\\\\\\)|(?:\\\\\\])|(?:\\\\n))*)");
QString intRegex("([0-9]*)");
QString md5Regex("((?:)|(?:[a-f0-9]{32}))");
QString sectionRegexStr(QString("^%<--! \\$VHIST_SECTION \\[version:%1\\]"
        "\\[creator:%1\\]\\[title:%1\\]\\[left:%2\\]"
        "\\[size:%2\\]\\[index:%2\\]\\[md5section:%3\\]"
        "\\[previousmd5:%3\\]\\[previousmarker:%2\\]-->")
        .arg(strRegex, intRegex, md5Regex));
QString fileRegexStr(QString("^%<--! \\$VHIST_EMBEDDEDFILE_BEGIN \\[filetype:%1\\]"
        "\\[filename:%1\\]\\[desc:%1\\]\\[comment:%1\\]"
        "\\[compression:%1\\]\\[filesize:%2\\]\\[cfilesize:%2\\]"
        "\\[blocksize:%2\\]\\[offset:%2\\]\\[md5file:%3\\]"
        "\\[md5cfile:%3\\]-->").arg(strRegex, intRegex, md5Regex));

const int SecVersionPos = 1;
const int SecCreatorPos = 2;
const int SecTitlePos = 3;
const int SecLeftPos = 4;
const int SecSizePos = 5;
const int SecIndexPos = 6;
const int SecMd5sectionPos = 7;
const int SecPreviousmd5Pos = 8;
const int SecPreviousmarkerPos = 9;

const int FileFiletypePos = 1;
const int FileFilenamePos = 2;
const int FileDescriptionPos = 3;
const int FileCommentPos = 4;
const int FileCompressionPos = 5;
const int FileFilesizePos = 6;
const int FileCFilesizePos = 7;
const int FileBlocksizePos = 8;
const int FileOffsetPos = 9;
const int FileMd5filePos = 10;
const int FileMd5cfilePos = 11;

int makeInt(const QString& str)
{
    if (str == "")
        return -1;
    else
        return str.toInt();
}


void printSectionTag(const SectionTag& sec)
{
    qDebug() << "id" << sec.id;
    qDebug() << "position" << sec.position;
    qDebug() << "taglen" << sec.taglen;
    qDebug() << "version" << sec.version;
    qDebug() << "creator" << sec.creator;
    qDebug() << "title" << sec.title;
    qDebug() << "left" << sec.left;
    qDebug() << "size" << sec.size;
    qDebug() << "index" << sec.index;
    qDebug() << "md5section" << sec.md5section;
    qDebug() << "previousmd5" << sec.previousmd5;
    qDebug() << "previousmarker" << sec.previousmarker;
    qDebug() << "validationTag" << sec.validationTag;
}


void printFileTag(const FileTag& file)
{
    qDebug() << "id" << file.id;
    qDebug() << "position" << file.position;
    qDebug() << "taglen" << file.taglen;
    qDebug() << "filetype" << file.filetype;
    qDebug() << "filename" << file.filename;
    qDebug() << "description" << file.description;
    qDebug() << "comment" << file.comment;
    qDebug() << "compression" << file.compression;
    qDebug() << "filesize" << file.filesize;
    qDebug() << "cfilesize" << file.cfilesize;
    qDebug() << "blocksize" << file.blocksize;
    qDebug() << "offset" << file.offset;
    qDebug() << "md5file" << file.md5file;
    qDebug() << "md5cfile" << file.md5cfile;
}


/**
 * Creates a new parser for VHIST files.
 * 
 * @param[in] mode Specifies whether the parser should follow the
 * vhist spec strictly or provide more robust parsing.
 * @param[in] pLogTarget An optional object with a void log(const QString,
 * const vhist::LogType) slot.
 * The parser will send it's output to this object.
 */
VhistParser::VhistParser(ParseMode mode, QObject *pLogTarget) : 
        LogSource(pLogTarget), m_mode(mode), m_fileRegex(fileRegexStr),
        m_sectionRegex(sectionRegexStr)
{
}


/**
 * Destroys a VHIST parser instance.
 */
VhistParser::~VhistParser()
{
}


/**
 * Opens a VHIST file and parses it. If the file could not be opened
 * or detecting the VHIST header failed, \c false is returned.
 * 
 * @param[in] filename The name of the file, which should be opened.
 *
 * @return \c true if opening and parsing the file succeeded and
 * \c false otherwise.
 * 
 * @note The open function will parse the \e complete VHIST file. This
 * operation is very time efficient and uses only a constant amount
 * of memory, however, for large files, it still takes a noticable
 * amount of time to complete. Therefore, opening the same VHIST file
 * several times should be avoided.
 */
bool VhistParser::open(const QString& filename)
{
    m_sections.clear();
    m_files.clear();
    bool successful = m_iter.open(filename);
    if (!successful)
        return false;
    
    if (!headerFound())
        return false;
    
    m_sections = readSections();
    m_files = findFiles();
    return true;
}


/**
 * Returns a list of all sections in the current VHIST file. If no file
 * has been opened yet or if opening and parsing the last VHIST file
 * failed, an empty list is returned.
 * 
 * @return A list of all sections.
 */
SectionTagList VhistParser::getSections()
{
    return m_sections;
}


/**
 * Returns the number of sections in the current VHIST file.
 * 
 * @return The number of sections.
 */
int VhistParser::getNumSections()
{
    return m_sections.count();
}


/**
 * Returns a list of all embedded files in the current VHIST file.
 *
 * @return A list of all embedded files.
 */
FileTagList VhistParser::getFiles()
{
    return m_files;
}


/**
 * Returns the number of embedded files in the current VHIST file.
 *
 * @return The number of embedded files.
 */
int VhistParser::getNumFiles()
{
    return m_files.count();
}


/**
 * Returns a list of embedded files which are part of a specific
 * section. If no VHIST file is currently open or the section does
 * not exist, an empty list is returned.
 *
 * @param[in] sectionId The id of the section, which's files should be
 * returned.
 *
 * @return A list of all embedded files inside the given section.
 */
FileTagList VhistParser::getFilesInSection(int sectionId)
{
    FileTagList result;
    MaybeSectionTag maybe = getSection(sectionId);
    if (maybe != NULL)
    {
        SectionTag section = *maybe;
        FileTagList::iterator it;
        for (it = m_files.begin(); it != m_files.end(); ++it)
        {
            FileTag emFile = *it;
            if (fileInSection(emFile, section))
                result.append(emFile);
        }
    }
    return result;
}


/**
 * Returns a list of embedded files, which are not part of any section.
 * 
 * @note Sectionless files are not valid according to the VHIST standard,
 * they may, however, appear in VHIST files, which are slightly corrupted.
 * 
 * @return A list of all embedded sectionless files.
 */
FileTagList VhistParser::getSectionlessFiles()
{
    // store all file ids in this set and remove ids which
    // belong to one section
    QSet<int> fileSet;
    FileTagList fileList = getFiles();
    for (int i = 0; i < fileList.count(); i++)
        fileSet.insert(fileList[i].id);
    
    SectionTagList sections = getSections();
    for (int i = 0; i < sections.count(); i++)
    {
        FileTagList sectionFiles = getFilesInSection(sections[i].id);
        for (int j = 0; j < sectionFiles.count(); j++)
            fileSet.remove(sectionFiles[j].id);
    }
    
    // return all files which ids are still in the set
    for (int i = fileList.count() - 1; i >= 0; i--)
        if (!fileSet.contains(fileList[i].id))
            fileList.removeAt(i);
    
    return fileList;
}


// FileTagList VhistParser::getFilesInNoSection()
// {
//     return getSectionlessFiles();
// }


/**
 * Returns the section with the given id. If the id is not valid
 * or no VHIST file is currently opened, a \c NULL pointer
 * is returned.
 * 
 * @param[in] id The id of the section, which should be returned.
 * 
 * @return The requested section or \c NULL if an
 * error occured.
 */
MaybeSectionTag VhistParser::getSection(int id)
{
    if (id >= 0 && id < m_sections.count())
        return MaybeSectionTag(new SectionTag(m_sections[id]));
    else
        return MaybeSectionTag(NULL);
}


/**
 * Returnes the embedded file with the given id. If the id is not
 * valid or no VHIST file is currently opened, a \c NULL pointer
 * if returned.
 *
 * @param[in] id The id of the requested file.
 *
 * @return The requested embedded file or \c NULL if
 * and error occured.
 */
MaybeFileTag VhistParser::getFile(int id)
{
    if (id >= 0 && id < m_files.count())
        return MaybeFileTag(new FileTag(m_files[id]));
    else
        return MaybeFileTag(NULL);
}


/**
 * Returns the content of a given embedded file. The content is returned
 * as a QByteArray. If the file is compressed, it is automatically inflated.
 * In the case of an error, a \c NULL pointer is returned. This way,
 * it's clear if the retrieval of a \c 0 bytes large file succeeded or
 * failed.
 * 
 * @param[in] id The id of the requested file.
 * 
 * @return The content of the requested file on success or
 * a \c NULL pointer if an error occured.
 */
MaybeBytes VhistParser::getFileContent(int id)
{
    MaybeFileTag maybe = getFile(id);
    MaybeBytes maybeData = getCompressedFile(id);
    if (maybe != NULL && maybeData != NULL)
    {
        FileTag emFile = *maybe;
        QByteArray data = *maybeData;
        
        if (emFile.compression == "flate")
        {
            quint32 filesize = emFile.filesize;
            QByteArray decompressed = decompressBytes(data, filesize);
            if (decompressed.count() == emFile.filesize)
                return MaybeBytes(new QByteArray(decompressed));
        }
        else if (emFile.compression == "none")
        {
            return MaybeBytes(new QByteArray(data));
        }
    }
    
    return MaybeBytes(NULL);
}


/**
 * Writes an embedded file from the current VHIST file to harddisk.
 *
 * @param[in] id The id of the embedded file.
 * @param[in] filename The location on the harddisk, to which the file
 * should be written.
 * 
 * @return \c true if the operation succeeded and \c false otherwise.
 *
 * @note Depending on the size of the embedded file, this operation may take
 * a long time and block io during this period. In a GUI application, it's
 * more favorable to use vhist::ExtractionQueue instead. The function
 * only consumes a constant amount of memory.
 *
 * @see vhist::ExtractionQueue
 */
bool VhistParser::writeFileToDisk(int id, const QString& filename)
{
    unsigned int iStartClock = clock();
    
    log(QString("[writeFileToDisk] file: %1").arg(filename));
    MaybeFileTag maybe = getFile(id);
    if (maybe == NULL)
        return false;
    
    QFile file(filename);
    FileTag tag = *maybe;
    
    if (!file.open(QIODevice::WriteOnly | QIODevice::Unbuffered))
        return false;
    
    qint64 start = tag.position + tag.taglen + tag.offset;
    
    qint64 len = 0;
    if (tag.compression == "flate")
        len = tag.cfilesize;
    else if (tag.compression == "none")
        len = tag.filesize;
    else {
        log("*** ERROR: invalid compression method.", LogSource::LOG_ERROR);
        return false;
    }

    log(QString("original/embedded file size: %1/%2 MB (%3/%4 bytes)")
        .arg(tag.filesize / 1024.0 / 1024.0, 0, 'f', 2)
        .arg(tag.cfilesize / 1024.0 / 1024.0, 0, 'f', 2)
        .arg(tag.filesize)
        .arg(tag.cfilesize));
    
    qint64 pos = 0;
    qint64 blocksize =  4 * 1024 * 1024;
    Inflater inflate;
    
    qDebug() << "unextract" << tag.filename << "(" << start << len << ")";
    QByteArray data_raw, data_uncompressed;
    qint64 bytesWritten = 0, dataCount = 0;

    
    while (len > 0)
    {
        qint64 bytesToRead = qMin(len, blocksize);
        data_raw = m_iter.getBytes(start + pos, bytesToRead);
        
        if (tag.compression == "flate")
        {
            data_uncompressed = inflate.inflateData(data_raw);
            bytesWritten = file.write(data_uncompressed);    
            dataCount = data_uncompressed.count();
        } 
        else
        {
            bytesWritten = file.write(data_raw);
            dataCount = data_raw.count();
        }
        
        if (bytesWritten != dataCount)
        {
            log(QString("*** ERROR: IO Error."), LogSource::LOG_ERROR);
            return false;
        }
        len -= bytesToRead;
        pos += bytesToRead;
    }
    
    double tTotal = 1.0 * (clock() - iStartClock) / CLOCKS_PER_SEC;
    log(QString("[writeFileToDisk] total: %1s.")
        .arg(tTotal));
    
    return true;
}


/**
 * Calculates the MD5 checksum for an embedded file.
 * The method will calculate a pair of checksums, one checksum for the
 * compressed and one for the uncompressed data. If the embedded file
 * has not been compressed, the compressed MD5 will be empty. The checksums
 * are returned as hex digests. In the case of an error, 2 empty checksums
 * are returned.
 * 
 * @param[in] id The id of the embedded file, for which the checksums should
 * be calculated.
 * 
 * @return The checksums for the specified file.
 * 
 * @warning The method will load the complete file into memory at once.
 * For large files, this is not a good idea. Additionally, the function
 * will block until the checksums have been calculated. For more
 * sophisticated purposes, using vhist::ValidationQueue is the better
 * alternative.
 * 
 * @see vhist::ValidationQueue
 */
MD5Sums VhistParser::calcChecksumsForFile(int id)
{
    MaybeFileTag maybe = getFile(id);
    MaybeBytes maybeData = getCompressedFile(id);
    if (maybe != NULL && maybeData != NULL)
    {
        FileTag emFile = *maybe;
        QByteArray data = *maybeData;
        
        MD5Generator md5(data);
        QString compMD5 = md5.getHexDigest();
        
        QString decompMD5("");
        if (emFile.compression == "flate")
        {
            QByteArray decompData = decompressBytes(data, emFile.filesize);
            MD5Generator md5decomp(decompData);
            decompMD5 = md5decomp.getHexDigest();
        }
        return MD5Sums(compMD5, decompMD5);
    }
    return MD5Sums("", "");
}


/**
 * Returns a part of the current VHIST file. If the bytes could not be
 * read, an array of length \c 0 is returned.
 *
 * @param[in] startPos The index of the first byte, which is returned.
 * @param[in] numBytes The number of bytes to return.
 *
 * @return A bytearray containing the requested bytes.
 */
QByteArray VhistParser::getBytes(int startPos, int numBytes)
{
    return m_iter.getBytes(startPos, numBytes);
}


////////////////////////////////////////////////////////////////
// private methods

QByteArray VhistParser::makeValidationTag(const QString& md5ToReplace)
{
    QByteArray line = *m_iter.curLine();
    QByteArray toFind("[md5section:");
    toFind.append(md5ToReplace);
    toFind.append("]");
    QByteArray replaceWith("[md5section:00000000000000000000000000000000]");
    QByteArray modLine = line.replace(toFind, replaceWith);
    return modLine.right(modLine.size() - 1);
}


MaybeSectionTag VhistParser::parseSection(const QString& line)
{
    if (m_sectionRegex.indexIn(line) == 0)
    {
        QStringList matches = m_sectionRegex.capturedTexts();
        if (matches.size() > SecPreviousmarkerPos)
        {
            SectionTag section;
            // +/- 1 because of %-sign
            section.position = m_iter.curPos() + 1;
            section.taglen = matches.at(0).length() - 1;
            section.version = matches[SecVersionPos];
            section.creator = matches[SecCreatorPos];
            section.title = matches[SecTitlePos];
            section.left = makeInt(matches[SecLeftPos]);
            section.size = makeInt(matches[SecSizePos]);
            section.index = makeInt(matches[SecIndexPos]);
            section.md5section = matches[SecMd5sectionPos];
            section.previousmd5 = matches[SecPreviousmd5Pos];
            section.previousmarker = makeInt(matches[SecPreviousmarkerPos]);
            section.validationTag = makeValidationTag(section.md5section);
            return MaybeSectionTag(new SectionTag(section));
        }
    }
    return MaybeSectionTag(NULL);
}


MaybeSectionTag VhistParser::findLastSection()
{
    bool moreLines = m_iter.lastLine();
    while (moreLines)
    {
        MaybeString maybe = m_iter.curLineAsString();
        if (maybe != NULL)
        {
            if (maybe->startsWith("%<--! $VHIST_SECTION"))
            {
                MaybeSectionTag maybeSec = parseSection(*maybe);
                if (maybeSec != NULL)
                    return maybeSec;
            }
            else if (m_mode == MODE_ROBUST)
            {
                int pos = maybe->indexOf("%<--! $VHIST_SECTION");
                if (pos >= 0)
                {
                    MaybeSectionTag maybeSec = parseSection(
                            maybe->mid(pos));
                    if (maybeSec != NULL)
                        return maybeSec;
                }
            }
        }
        moreLines = m_iter.prevLine();
    }
    return MaybeSectionTag(NULL);
}


MaybeSectionTag VhistParser::findPrevSection(const SectionTag& sec)
{
    if (sec.previousmarker >= 0)
    {
        // -1 because we also want the % sign before the tag
        int prevPos = sec.position - sec.previousmarker - 1;
        m_iter.gotoPos(prevPos);
        MaybeString maybe = m_iter.curLineAsString();
        if (maybe != NULL)
        {
            MaybeSectionTag maybeTag = parseSection(*maybe);
            if (maybeTag != NULL)
                return maybeTag;
        }
        
    }
    else
        return MaybeSectionTag(NULL);
    
    if (m_mode == MODE_ROBUST)
    {
        for (int i = 0; i < 5; i++)
            m_iter.prevLine();
        
        for (int i = 0; i < 10; i++)
        {
            MaybeString maybe = m_iter.curLineAsString();
            if (maybe != NULL)
            {
                int pos = maybe->indexOf("%<--! $VHIST_SECTION");
                if (pos >= 0)
                {
                    QString line = maybe->mid(pos);
                    MaybeSectionTag maybeTag = parseSection(line);
                    if (maybeTag != NULL)
                        return maybeTag;
                }
            }
            m_iter.nextLine();
        }
    }
    
    return MaybeSectionTag(NULL);
}


SectionTagList VhistParser::readSections()
{
    SectionTagList sections;
    MaybeSectionTag maybe = findLastSection();
    int id = 0;
    while (maybe != NULL)
    {
        //printSectionTag(maybe.value);
        maybe->id = id;
        sections.append(*maybe);
        maybe = findPrevSection(*maybe);
        id++;
    }
    
    return sections;
}


MaybeFileTag VhistParser::parseFile(const QString& line)
{
    if (m_fileRegex.indexIn(line) == 0)
    {
        QStringList matches = m_fileRegex.capturedTexts();
        if (matches.size() > FileMd5cfilePos)
        {
            FileTag emFile;
            // +/- 1 because of %-sign
            emFile.position = m_iter.curPos() + 1;
            emFile.taglen = matches.at(0).length() - 1;
            emFile.filetype = matches[FileFiletypePos];
            emFile.filename = matches[FileFilenamePos];
            emFile.description = matches[FileDescriptionPos];
            emFile.comment = matches[FileCommentPos];
            emFile.compression = matches[FileCompressionPos];
            emFile.filesize = makeInt(matches[FileFilesizePos]);
            emFile.cfilesize = makeInt(matches[FileCFilesizePos]);
            emFile.blocksize = makeInt(matches[FileBlocksizePos]);
            emFile.offset = makeInt(matches[FileOffsetPos]);
            emFile.md5file = matches[FileMd5filePos];
            emFile.md5cfile = matches[FileMd5cfilePos];
            
            return MaybeFileTag(new FileTag(emFile));
        }
    }
    
    return MaybeFileTag(NULL);
}


FileTagList VhistParser::findFiles()
{
    FileTagList files;
    bool moreLines = m_iter.firstLine();
    
    int id = 0;
    while (moreLines)
    {
        if (m_iter.curLineStartsWithBytes("%<--! $VHIST_EMBEDDEDFILE_BEGIN"))
        {
            MaybeString maybe = m_iter.curLineAsString();
            if (maybe != NULL)
            {
                MaybeFileTag maybeFile = parseFile(*maybe);
                if (maybeFile != NULL)
                {
                    FileTag emFile = *maybeFile;
                    emFile.id = id;
                    id++;
                    files.append(emFile);
                    int afterFile = emFile.position + emFile.blocksize;
                    moreLines = m_iter.gotoPos(afterFile);
                }
            }
        }
        moreLines = m_iter.nextLine();
    }
    return files;
}


MaybeBytes VhistParser::getCompressedFile(int id)
{
    MaybeFileTag maybe = getFile(id);
    if (maybe != NULL)
    {
        FileTag emFile = *maybe;
        int start = emFile.position + emFile.taglen + emFile.offset;
        int len = emFile.cfilesize;
        QByteArray data = m_iter.getBytes(start, len);
        return MaybeBytes(new QByteArray(data));
    }
    return MaybeBytes(NULL);
}


QByteArray VhistParser::decompressBytes(QByteArray& data,
        quint32 expectedSize)
{
    quint32 filesize = qToBigEndian<quint32>(expectedSize);
    char* prefixData = reinterpret_cast<char*>(&filesize);
    QByteArray prefix(prefixData, sizeof(qint32));
    data.prepend(prefix);
    return qUncompress(data);
}





bool VhistParser::fileInSection(const FileTag& file,
        const SectionTag& section)
{
    int secLeft = section.position - section.left;
    int secRight = secLeft + section.size;
    int fileLeft = file.position;
    int fileRight = fileLeft + file.taglen + file.blocksize;
    return secLeft <= fileLeft && secRight >= fileRight;
}




bool VhistParser::headerFound()
{
    bool moreLines = m_iter.firstLine();
    for (int i = 0; i < 10 && moreLines; i++)
    {
        MaybeString maybe = m_iter.curLineAsString();
        if (maybe != NULL)
        {
            if (maybe->startsWith("%<--! $VHIST_PDF [version:1.0]"))
                return true;
            moreLines = m_iter.nextLine();
        }
    }
    return false;
}


