/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: vhistdelrunner.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#include "../../include/vhistaddlib/vhistdelrunner.h"
#include "../../include/shared/utils.h"

#include <QProcess>
#include <QString>
#include <QStringList>
#include <QDir>
#include <QRegExp>

#include <QtDebug>

using namespace vhist;

/**
 * Creates a new VhistdelRunner object.
 */
VhistdelRunner::VhistdelRunner(QObject* listener)
    : VhistRunner("vhistdel",listener)
{
}


/**
 * Destroys the VhistdelRunner object.
 */
VhistdelRunner::~VhistdelRunner()
{
}


/**
 * Runs \e m_vhistrun with the given argfile. The output of
 * \e m_vhistrun is parsed and reported on a line by line basis
 * using the vhist::LogSource interface.
 * 
 * @param[in] argfile The content of an argfile. \e m_vhistrun
 * will be called with the --stdin option and the argfile
 * if passed to \e m_vhistrun via the stdin pipe.
 * 
 * While executing "m_vhistrun", VhistRunner will regularly emit an
 * intercept signal, which can be used to update the user interface
 * of the calling application and react to user inputs.
 * 
 * @return The returncode of "m_vhistrun" if it started successfully
 * or one of the values from VHISTADD_ERROR.
 */
int VhistdelRunner::runProcess(const QString& szVHistFIle)
{
    // reset everything
    clearOutputRest();
    
    // create process
    QProcess proc;
    
    // find executables;
    QString python = findPython();
    QString vhistrun = findVhistrun();
    
    if (vhistrun == "")
    {
        emit log(QString("could not find any %1 executable\n").arg(vhistrun),
                LogSource::LOG_ERROR);
        return VHISTADD_NOT_FOUND;
    }
    
    // generate call
    QString program = python;
    QStringList argList;
    if (getUsePy2Exe())
        program = vhistrun;
    else
        argList << vhistrun;
    argList << szVHistFIle;
    
    // merge output channels
    proc.setProcessChannelMode(QProcess::MergedChannels);
    
    // start process
    proc.start(program, argList);
    
    // check if programm started correctly
    if (!proc.waitForStarted())
        if (proc.error() == QProcess::FailedToStart)
        {
            QString err=QString("%1 failed to start\n").arg(vhistrun);
            err = err.arg(proc.error());
            emit log(err, LogSource::LOG_ERROR);
            return VHISTADD_FAILED_TO_START;
        }
    
    // write argfile to stdin
    proc.closeWriteChannel();
    
    // wait for process to finish
    while (proc.state() != QProcess::NotRunning)
    {
        if (proc.waitForReadyRead(1000))
        {
            // get new data
            QString result = QString(proc.readAllStandardOutput());
            processOutput(result);
            emit intercept();
        }
    }
    
    QString rest = QString(proc.readAllStandardOutput());
    processOutput(rest, true);
    
    QString exitMessage(QObject::tr("Programm exited with error code: %1\n\n"));
    emit log(exitMessage.arg(proc.exitCode()), LogSource::LOG_MESSAGE);
    return proc.exitCode();
}


