/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: vhistrunner.cpp 2739 2013-04-24 08:54:08Z michael $
 */

#include "../../include/vhistaddlib/vhistrunner.h"
#include "../../include/shared/utils.h"

#include <QProcess>
#include <QString>
#include <QStringList>
#include <QDir>
#include <QRegExp>

#include <QtDebug>

using namespace vhist;

/**
 * Creates a new VhistRunner object.
 */
VhistRunner::VhistRunner(const QString &vhistrunstump, QObject* listener)
    : LogSource(listener)
{
    m_usePy2Exe=false;
	m_vhistrunstump=vhistrunstump;
	setUsePy2Exe(true);

	m_searchPaths << ".";
}


/**
 * Destroys the VhistRunner object.
 */
VhistRunner::~VhistRunner()
{
}


/**
 * Activates or deactivates \e py2exe support. If \e py2exe support is
 * activated, VhistRunner will search for a file named \e m_vhistrunstump.exe
 * inside the search path. Otherwise, it will search for \e m_vhistrunstump.py.
 * 
 * @param[in] state If \c true, \e py2exe support is activated.
 */
void VhistRunner::setUsePy2Exe(bool state)
{
#ifdef Q_OS_WIN
    m_usePy2Exe = state;
#else
    // ignore py2exe on not-Windows platforms
    state = state;
#endif

	if (m_usePy2Exe)
		m_vhistrun=m_vhistrunstump+".exe";
	else
		m_vhistrun=m_vhistrunstump+".py";
}


/**
 * Sets the search path to \a paths. VhistRunner searches the
 * given paths in the given order for a vhist executable. The
 * paths can be either absolute or relative to the executable,
 * which uses this library.
 * 
 * VhistRunner will use the first executable it finds in the
 * search path. By default, VhistRunner will only search in
 * the directory of the binary.
 * 
 * @param[in] paths A list of paths, which should be searched for
 * a "m_vhistrun" binary.
 * 
 * @note On Mac OS X, the reference executable refers to the binary
 * under Contents/MacOSX in the application bundle, not the bundle
 * itself.
 */
void VhistRunner::setSearchPaths(const QStringList& paths)
{
    m_searchPaths = paths;
}


// private methods

void VhistRunner::processOutputLine(const QString& line)
{
	QString szLine;
	LogSource::LogType enType;
    // parses the line and emits a log signal
	if (line.startsWith("***ERROR")) {
		szLine=line.mid(8);
		enType=LogSource::LOG_ERROR;
	}
	else if (line.startsWith("***WARNING")) {
		szLine=line.mid(10);
		enType=LogSource::LOG_WARNING;
	}
	else {
		szLine=line;
		enType=LogSource::LOG_MESSAGE;
	}
	//emit log(szLine, enType);
	emit log(szLine, (int) enType);
}

void VhistRunner::processOutput(const QString& output, bool finish)
{
    // add output to rest-buffer
    m_outputRest += output;
    
    // split output into lines
    QStringList lines = m_outputRest.split(QRegExp("(\r\n)|[\n\r]"));
    
    // process first n-1 lines (last line may be incomplete)
    for (int i = 0; i < lines.size() - 1 ; i++)
        processOutputLine(lines.at(i)+QLatin1String("\n"));
    
    // store last line for next iteration
    if (lines.size() > 0)
        m_outputRest = lines.at(lines.size() - 1);
    
    // application has finished => handle last line
    if (finish && m_outputRest != "")
        processOutputLine(m_outputRest+QLatin1String("\n"));
}

QString VhistRunner::searchInPath(const QString& path, const QString& file)
{
    QString filepath = getBinaryDir() + QDir::separator() + path;
    QFileInfo finfo(QDir(filepath), file);
    if (finfo.isFile())
        return finfo.filePath();
    else
        return "";
}

QString VhistRunner::findPython()
{
#ifdef Q_OS_WIN
    if (m_usePy2Exe)
        return "";
    else
        return "python.exe";
#else
    return "python";
#endif
}

QString VhistRunner::findVhistrun()
{
    QString fullVhistrun = "";
    for (int i = 0; i < m_searchPaths.size() && fullVhistrun == ""; i++)
        fullVhistrun = searchInPath(m_searchPaths.at(i), m_vhistrun);
    

    return fullVhistrun;
}

/**
 * @fn void vhist::VhistRunner::intercept()
 * This signal is emitted regularly while "m_vhistrun" is running.
 * It can be used to handle events in the user interface.
 */

