/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: path.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#include "../../include/vhistdifflib/path.h"

#include "../../include/vhistdifflib/pathexpander.h"
#include "../../include/vhistdifflib/priorityqueue.h"
#include "../../include/vhistdifflib/workflowtree.h"

#include <QtDebug>

using namespace vhist;


Path::Path(PathExpander* expander)
        : m_expander(expander), m_curCosts(0)
{
    m_associations.insert(NODE_ABOVE_ROOT, NODE_ABOVE_ROOT);
    m_revAssociations.insert(NODE_ABOVE_ROOT, NODE_ABOVE_ROOT);
    m_freeNodes1 = m_expander->nodeIds(FIRST_TREE);
    m_freeNodes2 = m_expander->nodeIds(SECOND_TREE);
}


Path::Path(const Path& rhs)
{
    *this = rhs;
}


Path::~Path()
{
}

Path& Path::operator=(const Path& rhs)
{
    if (this != &rhs)
    {
        m_expander = rhs.m_expander;
        m_associations = rhs.m_associations;
        m_revAssociations = rhs.m_revAssociations;
        m_calcParentCostsLater = rhs.m_calcParentCostsLater;
        m_freeNodes1 = rhs.m_freeNodes1;
        m_freeNodes2 = rhs.m_freeNodes2;
        m_curCosts = rhs.m_curCosts;
    }
    return *this;
}

void Path::associate(NodeId node1, NodeId node2)
{
    m_associations.insert(node1, node2);
    m_revAssociations.insert(node2, node1);
    m_freeNodes1.removeOne(node1);
    m_freeNodes2.removeOne(node2);
    m_curCosts += m_expander->costForAssociation(node1, node2,
            m_associations, m_revAssociations);
    bool calcParentLater = m_expander->calcParentCostsLater(node1, node2,
            m_associations, m_revAssociations);
    m_calcParentCostsLater.insert(node1, calcParentLater);

    if (m_freeNodes1.isEmpty())
    {
        m_curCosts += m_expander->costForUnmatchedNodes(m_freeNodes2.size());
        m_curCosts += m_expander->calcMissingParentCosts(m_associations,
                m_revAssociations, m_calcParentCostsLater);

        for (int i = 0; i < m_freeNodes2.size(); ++i)
        {
            m_revAssociations.insert(m_freeNodes2.first(), NULL_NODE);
            m_freeNodes2.removeFirst();
        }
    }
}


bool Path::reachedGoal()
{
    return m_freeNodes1.isEmpty();
}


void Path::expand(PriorityQueue<QSharedPointer<Path> > & m_queue)
{
    int node1 = m_freeNodes1.first();
    for (QLinkedList<int>::Iterator it = m_freeNodes2.begin();
            it != m_freeNodes2.end(); ++it)
    {
        int node2 = *it;
        QSharedPointer<Path> n = QSharedPointer<Path>(new Path(*this));
        n->associate(node1, node2);
        m_queue.push(n, (int) n->getCosts());
    }

    // association with null node
    QSharedPointer<Path> n = QSharedPointer<Path>(new Path(*this));
    n->associate(node1, NULL_NODE);
    m_queue.push(n, (int) n->getCosts());
}


float Path::getCosts()
{
    return m_curCosts + m_expander->heuristicCosts(m_freeNodes1, m_freeNodes2);
}


NodeId Path::association(NodeId nodeId, TreeId treeId) const
{
    if (treeId == FIRST_TREE)
        return m_associations.value(nodeId, INVALID_NODE);
    else if (treeId == SECOND_TREE)
        return m_revAssociations.value(nodeId, INVALID_NODE);
    else
        return INVALID_NODE;
}


bool Path::associated(NodeId nodeId, TreeId treeId) const
{
    if (treeId == FIRST_TREE)
        return m_associations.value(nodeId, NULL_NODE) == NULL_NODE;
    else if (treeId == SECOND_TREE)
        return m_revAssociations.value(nodeId, NULL_NODE) == NULL_NODE;
    else
        return false;
}


bool Path::parentChanged(NodeId nodeId, TreeId treeId) const
{
    if (treeId == FIRST_TREE && m_associations.contains(nodeId))
        return m_expander->parentChanged(nodeId, m_associations,
                m_revAssociations);
    else if (treeId == SECOND_TREE && m_revAssociations.contains(nodeId))
    {
        NodeId node1 = m_revAssociations[nodeId];
        return m_expander->parentChanged(node1, m_associations,
                m_revAssociations);
    }
    else
        return false;
}


void Path::printMe()
{
    QList<NodeId> keys = m_associations.keys();
    for (int i = 0; i < keys.size(); ++i)
        qDebug() << keys[i] << "->" << m_associations[keys[i]];
    qDebug() << "Costs:" << m_curCosts;
}

