/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: vhistdiff.cpp 2179 2011-06-24 09:03:24Z ahuesgen $
 */

#include "../../include/vhistdifflib/vhistdiff.h"

#include "../../include/vhistdifflib/pathexpander.h"
#include "../../include/vhistdifflib/costmatrix.h"
#include "../../include/vhistdifflib/dijkstra.h"


using namespace vhist;

typedef Dijkstra<Path> PathFinder;

VhistDiff::VhistDiff(const WorkflowTree& tree1, const WorkflowTree& tree2,
        WorkflowstepCosts* wfsCosts)
        : m_tree1(tree1), m_tree2(tree2), m_wfsCosts(wfsCosts)
{
}


VhistDiff::~VhistDiff()
{
}


VhistDiffResult VhistDiff::convertPath(const Path& path) const
{
    Associations associations = path.m_associations;
    Associations revAssociations = path.m_revAssociations;
    QHash<NodeId, bool> nodeMovedTree1, nodeMovedTree2;

    for (Associations::const_iterator it = associations.constBegin();
            it != associations.constEnd(); ++it)
    {
        NodeId key = it.key();
        nodeMovedTree1.insert(key, path.parentChanged(key, FIRST_TREE));
    }

    for (Associations::const_iterator it = revAssociations.constBegin();
            it != revAssociations.constEnd(); ++it)
    {
        NodeId key = it.key();
        nodeMovedTree2.insert(key, path.parentChanged(key, SECOND_TREE));
    }

    return VhistDiffResult(associations, revAssociations, nodeMovedTree1,
            nodeMovedTree2);
}


QSharedPointer<VhistDiffResult> VhistDiff::findBestMatch() const
{
    VhistDiffResults results = findBestMatches(1);
    if (results.size() >= 1)
        return QSharedPointer<VhistDiffResult>(
                new VhistDiffResult(results.at(0)));
    else
        return QSharedPointer<VhistDiffResult>(NULL);
}


VhistDiffResults VhistDiff::findBestMatches(int numMatches) const
{
    PathFinder pathfinder;
    CostMatrix matrix(m_tree1, m_tree2, m_wfsCosts);
    PathExpander expander(&m_tree1, &m_tree2, &matrix);
    QSharedPointer<Path> startNode(new Path(&expander));
    QList<QSharedPointer<Path> > paths = pathfinder.findPaths(startNode,
            numMatches);

    VhistDiffResults results;
    for (int i = 0; i < paths.size(); ++i)
        results.append(convertPath(*paths.at(i)));
    return results;
}


void VhistDiff::setTree1(const WorkflowTree& tree1)
{
    m_tree1 = tree1;
}


void VhistDiff::setTree2(const WorkflowTree& tree2)
{
    m_tree2 = tree2;
}


void VhistDiff::setCosts(WorkflowstepCosts* wfsCosts)
{
    m_wfsCosts = wfsCosts;
}


WorkflowTree* VhistDiff::tree1()
{
    return &m_tree1;
}


WorkflowTree* VhistDiff::tree2()
{
    return &m_tree2;
}

WorkflowstepCosts* VhistDiff::costs()
{
    return m_wfsCosts;
}
