/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: evaloperators.cpp 2477 2012-04-02 14:16:23Z michael $
 */

#include "../../include/vhistfindlib/evaloperators.h"
#include "../../include/vhistfindlib/returnvalue.h"

#include <cmath>

#include <QString>
#include <QRegExp>

using namespace vhist;

ReturnValue vhist::evaluateEq(const ReturnValue& v0,
        const ReturnValue& v1)
{
    if (v0.isString() && v1.isString())
        return ReturnValue::fromBool(v0.toString().toLower() == v1.toString().toLower());
    else if (!v0.isValid())
        return v0;
    else if (!v1.isValid())
        return v1;
    else // mixed types (string and bool or bool and string)...
        return ReturnValue(RETURN_VALUE_INVALID_TYPE, "");
}


ReturnValue vhist::evaluateNEq(const ReturnValue& v0,
        const ReturnValue& v1)
{
    ReturnValue res = vhist::evaluateEq(v0, v1);
    if (res.isBool())
        return ReturnValue::fromBool(!res.toBool());
    return res;
}


ReturnValue vhist::evaluateNot(const ReturnValue& v0)
{
    if (v0.isBool())
        return ReturnValue::fromBool(!v0.toBool());
    else if (v0.isString())
        return ReturnValue(RETURN_VALUE_INVALID_TYPE, "");
    else
        return v0;
}


ReturnValue vhist::evaluateAnd(const ReturnValue& v0, const ReturnValue& v1)
{
    if (v0.isString() || v1.isString())
        return ReturnValue(RETURN_VALUE_INVALID_TYPE, "");

    if (v0.isBool())
    {
        if (!v0.toBool())
            return v0;
        else
            return v1;
    }
    else
        return v0;
}


ReturnValue vhist::evaluateOr(const ReturnValue& v0, const ReturnValue& v1)
{
    if (v0.isBool() && v0.toBool())
        return v0;
    else if (v1.isString())
        return ReturnValue(RETURN_VALUE_INVALID_TYPE, "");
    else
        return v1;
}


ReturnValue vhist::evaluateLess(const ReturnValue& v0, const ReturnValue& v1)
{
    if (v0.isNumber() && v1.isNumber())
        return ReturnValue::fromBool(v0.toNumber() < v1.toNumber());
    else if (v0.isDate() && v1.isDate())
        return ReturnValue::fromBool(v0.toDate() < v1.toDate());
    else if (!v0.isValid())
        return v0;
    else if (!v1.isValid())
        return v1;
    else
        return ReturnValue(RETURN_VALUE_INVALID_TYPE, "");
}


ReturnValue vhist::evaluateLessEq(const ReturnValue& v0, const ReturnValue& v1)
{
    if (v0.isNumber() && v1.isNumber())
        return ReturnValue::fromBool(v0.toNumber() <= v1.toNumber());
    else if (v0.isDate() && v1.isDate())
        return ReturnValue::fromBool(v0.toDate() <= v1.toDate());
    else if (!v0.isValid())
        return v0;
    else if (!v1.isValid())
        return v1;
    else
        return ReturnValue(RETURN_VALUE_INVALID_TYPE, "");
}


ReturnValue vhist::evaluateGreater(const ReturnValue& v0, const ReturnValue& v1)
{
    if (v0.isNumber() && v1.isNumber())
        return ReturnValue::fromBool(v0.toNumber() > v1.toNumber());
    else if (v0.isDate() && v1.isDate())
        return ReturnValue::fromBool(v0.toDate() > v1.toDate());
    else if (!v0.isValid())
        return v0;
    else if (!v1.isValid())
        return v1;
    else
        return ReturnValue(RETURN_VALUE_INVALID_TYPE, "");
}


ReturnValue vhist::evaluateGreaterEq(const ReturnValue& v0, const ReturnValue& v1)
{
    if (v0.isNumber() && v1.isNumber())
        return ReturnValue::fromBool(v0.toNumber() >= v1.toNumber());
    else if (v0.isDate() && v1.isDate())
        return ReturnValue::fromBool(v0.toDate() >= v1.toDate());
    else if (!v0.isValid())
        return v0;
    else if (!v1.isValid())
        return v1;
    else
        return ReturnValue(RETURN_VALUE_INVALID_TYPE, "");
}


ReturnValue vhist::evaluateIsApprox(const ReturnValue& v0, const ReturnValue& v1)
{
    if (v0.isNumber() && v1.isNumber())
    {
        float x = v0.toNumber();
        float y = v1.toNumber();
        float thresh = 0.01f;
        bool result = (std::fabs(x - y) / std::max(x, y)) < thresh;
        return ReturnValue::fromBool(result);
    }
    else if (!v0.isValid())
        return v0;
    else if (!v1.isValid())
        return v1;
    else
        return ReturnValue(RETURN_VALUE_INVALID_TYPE, "");
}


ReturnValue vhist::evaluateContains(const ReturnValue& v0, const ReturnValue& v1)
{
    if (v0.isString() && v1.isString())
        return ReturnValue::fromBool(v0.toString().toLower().contains(v1.toString().toLower()));
    else if (!v0.isValid())
        return v0;
    else if (!v1.isValid())
        return v1;
    else
        return ReturnValue(RETURN_VALUE_INVALID_TYPE, "");
}


ReturnValue vhist::evaluateMatches(const ReturnValue& v0, const ReturnValue& v1)
{
    if (v0.isString() && v1.isString())
    {
        QRegExp reg(v1.toString().toLower(), Qt::CaseSensitive, QRegExp::Wildcard);
        QString text = v0.toString().toLower();
        int pos = reg.indexIn(v0.toString());
        return ReturnValue::fromBool(pos != -1 && reg.matchedLength() == text.length());
    }
    else if (!v0.isValid())
        return v0;
    else if (!v1.isValid())
        return v1;
    else
        return ReturnValue(RETURN_VALUE_INVALID_TYPE, "");
}
