/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: expressionvalue.cpp 2169 2011-05-30 09:11:44Z ahuesgen $
 */

#include "../../include/vhistfindlib/expressionvalue.h"
#include "../../include/vhistfindlib/operatorinfo.h"

#include <QStringList>


using namespace vhist;

ExpressionValue::ExpressionValue() : m_pos(-1)
{
}


bool ExpressionValue::isVariable() const
{
    return m_type == VALUE_TYPE_VARIABLE;
}


bool ExpressionValue::isString() const
{
    return m_type == VALUE_TYPE_STRING;
}


bool ExpressionValue::isOperator() const
{
    return m_type == VALUE_TYPE_OPERATOR;
}


QString ExpressionValue::string() const
{
    if (isString())
        return m_string;
    else
        return "";
}


QString ExpressionValue::originalText () const
{
    return m_originalText;
}


int ExpressionValue::position() const
{
    return m_pos;
}


int ExpressionValue::length() const
{
    return m_originalText.length();
}


QString ExpressionValue::operatorName() const
{
    if (isOperator())
        return m_operator.operatorValue;
    else
        return "";
}


int ExpressionValue::precedence() const
{
    if (isOperator())
        return m_operator.precedence;
    else
        return -1;
}

void ExpressionValue::setPrecedence(int newPrecedence)
{
    if (isOperator())
        m_operator.precedence = newPrecedence;
}


int ExpressionValue::numArgs() const
{
    if (isOperator())
        return m_operator.numArguments;
    else
        return -1;
}


char ExpressionValue::association() const
{
    if (isOperator())
        return m_operator.association;
    else
        return 'x';
}


QString ExpressionValue::variableName() const
{
    if (isVariable())
        return m_var.name;
    else
        return "";
}


QString ExpressionValue::variableAttribute() const
{
    if (isVariable())
        return m_var.attribute;
    else
        return "";
}


void ExpressionValue::setVariableName(const QString& name)
{
    if (m_type == VALUE_TYPE_VARIABLE)
        m_var.name = name;
}


ExpressionValue ExpressionValue::fromString(const QString& string, int pos,
        const QString& originalText)
{
    ExpressionValue v;
    v.m_type = VALUE_TYPE_STRING;
    v.m_string = string;

    v.m_pos = pos;
    v.m_originalText = originalText;

    return v;
}


ExpressionValue ExpressionValue::fromVariable(const QString& variableName,
        const QString attribute, int pos, const QString& originalText)
{
    ExpressionValue v;
    v.m_type = VALUE_TYPE_VARIABLE;
    v.m_var.name = variableName;
    v.m_var.attribute = attribute;

    v.m_pos = pos;
    v.m_originalText = originalText;

    return v;
}


ExpressionValue ExpressionValue::fromVariable(const QString& variable, int pos,
        const QString& originalText)
{
    QString varname;
    QString attribute;
    QString startString(":user{");

    if (variable.endsWith("}"))
    {
        int attrStart = variable.indexOf(startString);
        if (attrStart == -1)
            return fromVariable("", "");
        // extract userAttr name
        attribute = variable.mid(attrStart + startString.length(), -1);
        attribute.chop(1);
        attribute.prepend("user:");
        varname = variable.left(attrStart);
        return fromVariable(varname, attribute, pos, originalText);
    }

    QStringList vsplit = variable.split(":");

    if (vsplit.length() > 0)
        varname = vsplit[0];

    if (vsplit.length() > 1)
    {
        vsplit.removeFirst();
        QString attr = vsplit.join(":");
        attribute = attr;
    }

    return fromVariable(varname, attribute, pos, originalText);
}


ExpressionValue ExpressionValue::fromOperator(const QString& operatorValue,
        int pos, const QString& originalText)
{
    ExpressionValue v;
    v.m_type = VALUE_TYPE_OPERATOR;
    v.m_operator.operatorValue = operatorValue;
    v.m_operator.precedence = getPrecedence(operatorValue);
    v.m_operator.association = getAssociation(operatorValue);
    v.m_operator.numArguments = getNumArguments(operatorValue);

    v.m_pos = pos;
    v.m_originalText = originalText;

    return v;
}


/*QDebug operator<<(QDebug dbg, const ExpressionValue &v)
{
    if (v.isVariable())
        dbg << "Variable:" << v.variableName() << v.variableAttribute()
                << v.position() << v.originalText();
    else if (v.isString())
        dbg << "String:" << v.string() << v.position() << v.originalText();
    else if (v.isOperator())
        dbg << "Operator:" << v.precedence() << v.operatorName()
                << v.position() << v.originalText();
    return dbg;
}*/

QDebug operator<<(QDebug dbg, const ExpressionValue &v)
{
    if (v.isVariable())
        dbg << v.variableName() << v.variableAttribute();
    else if (v.isString())
        dbg << v.string();
    else if (v.isOperator())
        dbg << v.operatorName();
    return dbg;
}
