/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: returnvalue.cpp 2169 2011-05-30 09:11:44Z ahuesgen $
 */

#include "../../include/vhistfindlib/returnvalue.h"

#include <QRegExp>
#include <QStringList>

using namespace vhist;

QString dateRegstr("(?:([0-9]{2}|[0-9]{4})-([0-9]{1,2})-([0-9]{1,2}))");
QString timeRegstr("(?:([0-9]{1,2}):([0-9]{1,2})(?::([0-9]{1,2}))?)");

QList<int> strs2ints(const QStringList& strs)
{
    QList<int> ints;
    for (int i = 0; i < strs.length(); ++i)
    {
        bool ok;
        const QString& s = strs.at(i);
        int num = s.toInt(&ok);

        if (ok)
            ints.append(num);
        else
            ints.append(0);
    }
    return ints;
}


ReturnValue::ReturnValue(ReturnValueType t, QString v)
{
    m_type = t;
    m_value = v;
}


ReturnValue ReturnValue::fromString(const QString string)
{
    return ReturnValue(RETURN_VALUE_STRING, string);
}


ReturnValue ReturnValue::fromBool(bool boolean)
{
    QString v = boolean ? "true" : "false";
    return ReturnValue(RETURN_VALUE_BOOL, v);
}


ReturnValue ReturnValue::fromNumber(float number)
{
    QString n = QString::number(number);
    return ReturnValue(RETURN_VALUE_STRING, n);
}


bool ReturnValue::isNumber() const
{
    bool ok;
    toNumber(&ok);
    return ok;
}

bool ReturnValue::isDate() const
{
    bool ok;
    toDate(&ok);
    return ok;
}

bool ReturnValue::isString() const
{
    return m_type == RETURN_VALUE_STRING;
}


bool ReturnValue::isBool() const
{
    return m_type == RETURN_VALUE_BOOL;
}


bool ReturnValue::isValid() const
{
    return (isString() || isBool());
}


bool ReturnValue::isUnassignedVariable() const
{
    return m_type == RETURN_VALUE_NOT_ASSIGNED_YET;
}


float ReturnValue::toNumber(bool* success) const
{
    QString v = m_value.toLower();
    QString postFix;
    if (v.endsWith("kb") || v.endsWith("mb") || v.endsWith("gb"))
    {
        postFix = v.right(2);
        v.chop(2);
    }
    
    float scale = 1;
    if (postFix == "kb")
        scale = 1024;
    else if (postFix == "mb")
        scale = 1024 * 1024;
    else if (postFix == "gb")
        scale = 1024 * 1024 * 1024;

    QRegExp regexp("[+-]?[0-9.,]+(e[+-]?[0-9]+)?");
    if (regexp.indexIn(v) != 0)
    {
        success = false;
        return 0;
    }
    else
    {
        v = regexp.cap(0);
        return v.toFloat(success) * scale;
    }
}


QDateTime ReturnValue::toDate(bool* success) const
{
    // version 1: manditory date + manditory time
    QString regstr1 = QString("%1 +%2").arg(dateRegstr).arg(timeRegstr);
    QRegExp regex1 = QRegExp(regstr1);
    // version 2: manditory date + no time
    QString regstr2 = QString("%1").arg(dateRegstr);
    QRegExp regex2 = QRegExp(regstr2);
    // version 3: no date + manditory time
    QString regstr3 = QString("%1").arg(timeRegstr);
    QRegExp regex3 = QRegExp(regstr3);


    if (regex1.indexIn(m_value) == 0)
    {
        QList<int> ints = strs2ints(regex1.capturedTexts());
        QDateTime d = QDateTime(QDate(ints[1], ints[2], ints[3]), QTime(ints[4], ints[5], ints[6]));
        if (success != NULL)
            *success = d.isValid();
        return d;
    }
    else if (regex2.indexIn(m_value) == 0)
    {
        QList<int> ints = strs2ints(regex2.capturedTexts());
        QDateTime d = QDateTime(QDate(ints[1], ints[2], ints[3]));
        if (success != NULL)
            *success = d.isValid();
        return d;
    }
    else if (regex3.indexIn(m_value) == 0)
    {
        QList<int> ints = strs2ints(regex3.capturedTexts());
        QDateTime d = QDateTime(QDate::currentDate(), QTime(ints[1], ints[2], ints[3]));
        if (success != NULL)
            *success = d.isValid();
        return d;
    }

    if (success != NULL)
        *success = false;
    return QDateTime();
}


QString ReturnValue::toString(bool* success) const
{
    bool ok = isString();
    if (success != NULL)
        *success = ok;

    if (ok)
        return m_value;
    else
        return "";
}


bool ReturnValue::toBool(bool* success) const
{
    bool ok = isBool();
    if (success != NULL)
        *success = ok;

    if (ok)
        return (m_value == "true");
    else
        return false;
}


bool ReturnValue::operator ==(const ReturnValue& rhs)
{
    return (m_type == rhs.m_type) && (m_value == rhs.m_value);
}
