/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: argfilehighlighter.cpp 2706 2013-04-22 09:40:24Z ahuesgen $
 */

#include "argfilehighlighter.h"
#include <arglib/argfileparser.h>
#include <QSet>



ArgFileHighlighter::ArgFileHighlighter() : QSyntaxHighlighter((QObject*) NULL)
{
}


ArgFileHighlighter::~ArgFileHighlighter()
{
}


void ArgFileHighlighter::handleComment(const vhist::ArgfileToken& token,
        const QColor& color)
{
    highlightString(token.position(), token.originalLength(), color,
            ARG_TOKEN_STYLE_NORMAL);
}


void ArgFileHighlighter::handleArgument(const vhist::ArgfileToken& token,
        const QList<QSet<QString> >& tokenLists,
        const QList<QColor>& colorList, const QColor& defaultColor)
{
    int p = token.position();
    int l = token.originalLength();

    for (int i = 0; i < tokenLists.length(); ++i)
        if (tokenLists.at(i).contains(token.str()))
        {
            highlightString(p, l, colorList.at(i), ARG_TOKEN_STYLE_BOLD);
            return;
        }

    // default highlighting
    highlightString(p, l, defaultColor, ARG_TOKEN_STYLE_BOLD);
}


void ArgFileHighlighter::handleRest(const vhist::ArgfileToken& token,
        const QColor& color, int tokensSinceLastArgument)
{

    ArgTokenStyle style = tokensSinceLastArgument == 1 ?
            ARG_TOKEN_STYLE_ITALIC : ARG_TOKEN_STYLE_NORMAL;

    highlightString(token.position(), token.originalLength(), color, style);
}


void ArgFileHighlighter::highlightBlock(const QString& text)
{
    QSet<QString> tokens1, tokens2, tokens3, tokens4, tokens5, tokens6;
    QColor color1, color2, color3, color4, color5, color6;

    // init colors for arguments
    tokens1 << "-I" << "-J" << "-O" << "-A" << "--append-to-vhistfile";
    color1 = QColor(255, 90, 90);
    tokens2 << "-o" << "--outfile" << "-i" << "--infile";
    color2 = QColor(21, 9, 181);
    tokens3 << "-a" << "--file-attr" << "-f" << "--file-flag";
    color3 = QColor(85, 177, 230);
    tokens4 << "-u" << "--file-user" << "-w" << "--file-user-with-unit";
    color4 = QColor(165, 222, 255);
    tokens5 << "-d" << "--doc" << "-s" << "--step";
    color5 = QColor(189, 132, 21);
    tokens6 << "-U" << "--step-user" << "-W" << "--step-user-with-unit";
    color6 = QColor(220, 173, 81);

    QList<QSet<QString> > tokenLists;
    QList<QColor> colorList;

    tokenLists << tokens1 << tokens2 << tokens3 << tokens4 << tokens5 << tokens6;
    colorList << color1 << color2 << color3 << color4 << color5 << color6;

    QColor colorOtherArguments = QColor(0, 0, 0);
    QColor colorComment = QColor(108, 156, 87);
    QColor colorRest = QColor(0, 0, 0);

    // parse argfile
    vhist::ArgfileParser parser;
    vhist::TokenList tokens = parser.getHighlighterList(text);

    int tokensSinceLastArgument = 0;
    for (int i = 0; i < tokens.length(); ++i)
    {
        const vhist::ArgfileToken& curToken = tokens.at(i);

        if (curToken.originalText().startsWith("#"))
            handleComment(curToken, colorComment);
        else if (curToken.str().startsWith("-"))
        {
            handleArgument(curToken, tokenLists, colorList, colorOtherArguments);
            tokensSinceLastArgument = 0;
        }
        else if (!(curToken.str().trimmed() == ""))
        {
            tokensSinceLastArgument++;
            handleRest(curToken, colorRest, tokensSinceLastArgument);
        }
    }
}


void ArgFileHighlighter::highlightString(int pos, int length,
        const QColor& color, ArgTokenStyle style)
{
    QTextCharFormat format;
    QBrush foreground(color);
    format.setForeground(foreground);
    if (style == ARG_TOKEN_STYLE_BOLD)
        format.setFontWeight(QFont::DemiBold);
    else if (style == ARG_TOKEN_STYLE_ITALIC)
        format.setFontItalic(true);
    else
        format.setFontWeight(QFont::Normal);
    setFormat(pos, length, format);
}
