/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: findtextbrowser.cpp 2681 2013-04-16 14:09:31Z ahuesgen $
 */

#include "findtextbrowser.h"

#include <QVBoxLayout>
#include <QTextBlock>
#include <QAction>
#include <QtDebug>
#include <QApplication>

FindTextBrowser::FindTextBrowser(QWidget *parent, QTextBrowser* browser)
    : QWidget(parent), m_browser(browser)
{
    if (m_browser == NULL)
        m_browser = new QTextBrowser(this);

    m_highlighter = new SearchHighlighter(m_browser->document());

    initWidget();
    initSlots();
}


FindTextBrowser::~FindTextBrowser()
{

}


void FindTextBrowser::setHighlighterEnabled(bool active)
{
    if (active && m_highlighter == NULL)
        m_highlighter = new SearchHighlighter(m_browser->document());
    else if (!active && m_highlighter != NULL)
    {
        delete m_highlighter;
        m_highlighter = NULL;
    }
}


void FindTextBrowser::setTextBrowser(QTextBrowser* browser)
{
    m_layout->removeWidget(m_browser);

    delete m_highlighter;
    delete m_browser;

    m_browser = browser;
    m_highlighter = new SearchHighlighter(m_browser->document());
    m_layout->insertWidget(0, m_browser);
}

void FindTextBrowser::initWidget()
{
    m_layout = new QVBoxLayout(this);
    m_layout->setContentsMargins(0, 0, 0, 0);
    m_layout->setSpacing(0);
    setLayout(m_layout);
    m_layout->addWidget(m_browser);

    m_findBar = new FindBar(this);
    m_layout->addWidget(m_findBar);

    m_openBarAction = new QAction(this);
    this->addAction(m_openBarAction);
    m_openBarAction->setShortcut(QKeySequence::Find);
    m_openBarAction->setShortcutContext(Qt::WidgetWithChildrenShortcut);
    m_openBarAction->setEnabled(false);

    m_closeBarAction = new QAction(this);
    this->addAction(m_closeBarAction);
    m_closeBarAction->setShortcut(QString("Escape"));
    m_closeBarAction->setShortcutContext(Qt::WidgetWithChildrenShortcut);
}


void FindTextBrowser::initSlots()
{
    connect(m_findBar, SIGNAL(findStringChanged(QString)), this,
            SLOT(findStringChanged(QString)));
    connect(m_findBar, SIGNAL(caseSensitivityChanged(bool)), this,
            SLOT(caseSensitivityChanged(bool)));
    connect(m_findBar, SIGNAL(nextPressed()), this, SLOT(findNext()));
    connect(m_findBar, SIGNAL(previousPressed()), this, SLOT(findPrev()));
    connect(m_openBarAction, SIGNAL(triggered()), m_findBar, SLOT(showFindBar()));
    connect(m_closeBarAction, SIGNAL(triggered()), m_findBar, SLOT(hideFindBar()));
    connect(m_findBar, SIGNAL(findBarHidden()), this, SLOT(handleFindbarHiding()));
}

void FindTextBrowser::setShortcutsActive(bool active)
{
    m_openBarAction->setEnabled(active);
}


void FindTextBrowser::findStringChanged(const QString& findString)
{
    QTextCursor oldCursor = m_browser->textCursor();
    int position = qMin(oldCursor.anchor(), oldCursor.position());
    if (m_highlighter)
        m_highlighter->setFindString(findString);
    findNext(position);
}


void FindTextBrowser::caseSensitivityChanged(bool caseSensitive)
{
    QTextCursor oldCursor = m_browser->textCursor();
    int position = qMin(oldCursor.anchor(), oldCursor.position());
    if (m_highlighter)
        m_highlighter->setCaseSensitive(caseSensitive);
    findNext(position);
}


void FindTextBrowser::findNext(int position)
{
    showFindBar();
    QString findString = m_findBar->findString();
    bool caseSensitive = m_findBar->caseSensitive();
    QTextDocument::FindFlags flags = 0;
    if (caseSensitive)
        flags = QTextDocument::FindCaseSensitively;

    QTextCursor cursor = m_browser->textCursor();

    if (position == -1)
        position = qMax(0, qMax(cursor.anchor(), cursor.position()));

    QTextCursor c = m_browser->document()->find(findString, position, flags);

    // if nothing was found, start searching at the beginning of the document
    if (c.selectedText() == "")
        c = m_browser->document()->find(findString, 0, flags);

    if (c.isNull())
    {
        c = m_browser->textCursor();
        c.setPosition(qMin(c.anchor(), c.position()));
    }

    m_browser->setTextCursor(c);
}


void FindTextBrowser::findPrev(int position)
{
    showFindBar();
    QString findString = m_findBar->findString();
    bool caseSensitive = m_findBar->caseSensitive();
    QTextDocument::FindFlags flags = 0;
    if (caseSensitive)
        flags = QTextDocument::FindCaseSensitively;

    QTextCursor cursor = m_browser->textCursor();

    if (position == -1)
        position = qMax(0, qMin(cursor.anchor(), cursor.position()));

    QTextCursor c = m_browser->document()->find(findString,
            position, QTextDocument::FindBackward | flags);

    // if nothing was found, start searching at the end of the document
    if (c.selectedText() == "")
    {
        QTextBlock block = m_browser->document()->lastBlock();
        int endPos = block.position() + block.length();
        c = m_browser->document()->find(findString, endPos,
                QTextDocument::FindBackward);
    }

    if (c.isNull())
    {
        c = m_browser->textCursor();
        c.setPosition(qMin(c.anchor(), c.position()));
    }

    m_browser->setTextCursor(c);
}


QTextBrowser* FindTextBrowser::textBrowser()
{
    return m_browser;
}


void FindTextBrowser::hideFindBar()
{
    m_findBar->hideFindBar();
}


void FindTextBrowser::showFindBar()
{
    m_findBar->showFindBar();
}


FindTextBrowser* FindTextBrowser::getBrowserWithFocus(FindTextBrowser* def)
{
    QWidget* w = QApplication::focusWidget();

    while (w != NULL)
    {
        FindTextBrowser* browser = dynamic_cast<FindTextBrowser*>(w);
        if (browser != NULL)
            return browser;
        else
            w = w->parentWidget();
    }
    return def;
}


void FindTextBrowser::showFindBarForActiveWidget(FindTextBrowser* def)
{
    FindTextBrowser* browser = getBrowserWithFocus(def);
    if (browser != NULL)
        browser->showFindBar();
}


void FindTextBrowser::findNextForActiveWidget(FindTextBrowser* def)
{
    FindTextBrowser* browser = getBrowserWithFocus(def);
    if (browser != NULL)
        browser->findNext();
}


void FindTextBrowser::findPrevForActiveWidget(FindTextBrowser* def)
{
    FindTextBrowser* browser = getBrowserWithFocus(def);
    if (browser != NULL)
        browser->findPrev();
}


void FindTextBrowser::handleFindbarHiding()
{
    m_browser->setFocus();
}
