/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: recentfiles.cpp 2807 2013-05-07 15:55:54Z ahuesgen $
 */

#include "recentfiles.h"
#include "vsettings.h"

#include <QSettings>

#include <QtDebug>

const int NUMBER_OF_RECENT_FILES = 5;

RecentFiles::RecentFiles(const QString& configName, QMenu& menu, int index,
        QObject* parent)
        : QObject(parent), m_configOption(configName),
        m_mapper(this)
{
    QSettings settings(VSettings::getOrganization(),
            VSettings::getApplication());
    m_recentFiles = settings.value(m_configOption).toStringList();

    truncateRecentFilesList();

    QList<QAction*> actionsInMenu = menu.actions();
    // if index is negativ (-1), recent files actions are appended to the
    // end of the menu (otherwise, they are inserted somewhere in the middle)
    QAction* insertBefore = NULL;
    if (index >= 0 && index < actionsInMenu.length())
        insertBefore = actionsInMenu.at(index);

    for (int i = 0; i < NUMBER_OF_RECENT_FILES; i++)
    {
        QAction* nextAction = new QAction(&menu);
        m_actions.append(nextAction);
        connect(nextAction, SIGNAL(triggered()), &m_mapper, SLOT(map()));
        m_mapper.setMapping(nextAction, i);

        menu.insertAction(insertBefore, nextAction);
        updateAction(nextAction, i);
    }

    connect(&m_mapper, SIGNAL(mapped(int)),
            this, SLOT(emitRecentFileAction(int)));
}


void RecentFiles::updateAction(QAction* action, int index)
{
    action->setVisible(index < m_recentFiles.length());
    if (index < m_recentFiles.length())
    {
        QString entryname("&%1 %2");
        if (index >= 9)
            entryname = "%1 %2";
        entryname = entryname.arg(index+1).arg(m_recentFiles.at(index));
        action->setText(entryname);
    }
    else
        action->setText("");
}


void RecentFiles::truncateRecentFilesList()
{
    while (m_recentFiles.length() > qMax(0, NUMBER_OF_RECENT_FILES))
        m_recentFiles.removeLast();
}


void RecentFiles::addRecentFile(const QString& filename)
{
    int index = m_recentFiles.indexOf(filename);

    if (index >= 0)
        m_recentFiles.removeAt(index);
    m_recentFiles.prepend(filename);

    truncateRecentFilesList();

    for (int i = 0; i < m_actions.length(); ++i)
    {
        QAction* curAction = m_actions.at(i);
        updateAction(curAction, i);
    }

    QSettings settings(VSettings::getOrganization(),
            VSettings::getApplication());
    settings.setValue(m_configOption, m_recentFiles);
    settings.sync();
}


void RecentFiles::emitRecentFileAction(int index)
{
    if (index >= 0 && index < m_recentFiles.length())
    {
        QString filename = m_recentFiles.at(index);
        qDebug() << "recent file" << filename;
        emit recentFileSelected(filename);
    }
}
