/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: sharedmenu.cpp 2807 2013-05-07 15:55:54Z ahuesgen $
 */

// header with declaration of this class
#include "sharedmenu.h"

// Qt headers
#include <QApplication>
#include <QMainWindow>
#include <QMenuBar>
#include <QAction>
#include <QMenu>
#include <QList>
#include <QSignalMapper>
#include <QSettings>

#include <QMessageBox>
#include <QDateTime>
#include <QLatin1Char>

// debug header
#include <QtDebug>

// miscellaneous
#include "windowlist.h"
#include "constants.h"
#include "vsettings.h"
#include "recentfiles.h"


SharedMenu::SharedMenu() : m_recentFiles(NULL)
{
}

/**
 * Finds and returns a menu with a given \p title in the windows
 * menubar.
 */
QMenu* SharedMenu::findMenu(QMainWindow& win, const QString& title)
{
    QMenuBar* menuBar = win.menuBar();
    // no menu bar?
    if (menuBar == NULL)
        return NULL;
    
    QList<QAction*> menus = menuBar->actions();
    QList<QAction*>::iterator it;
    for (it = menus.begin(); it != menus.end(); ++it)
        if ((*it)->text() == title)
            return (*it)->menu();
    
    // menu not found
    return NULL;
}


QMenu* SharedMenu::createMenuBefore(QMainWindow& win, const QString& title,
        const QString& before)
{
    QMenuBar* menuBar = win.menuBar();

    if (menuBar != NULL)
    {
        QMenu* beforeMenu = findMenu(win, before);
        if (beforeMenu != NULL)
        {
            QMenu* newMenu = new QMenu(title, menuBar);
            menuBar->insertMenu(beforeMenu->menuAction(), newMenu);
            return newMenu;
        }
        else
            return menuBar->addMenu(title);
    }
    return NULL;
}


/**
 * Initializes the shared part of the windows menu.
 */
QMenu* SharedMenu::initMenu(QMainWindow& win,  const QString& recentFilesConfig)
{
    // init file menu
    QMenu* fileMenu = findMenu(win, tr("&File"));
    if (fileMenu != NULL)
        initFileMenu(*fileMenu, win, recentFilesConfig);
    
    QMenu* toolsMenu = createMenuBefore(win, tr("&Tools"), tr("&Help"));
    if (toolsMenu != NULL)
        initToolsMenu(*toolsMenu);

    QMenu* windowsMenu = createMenuBefore(win, tr("&Windows"), tr("&Help"));

    // init help menu
    QMenu* helpMenu = findMenu(win, tr("&Help"));
    if (helpMenu != NULL)
        initHelpMenu(*helpMenu);

    return windowsMenu;
}


/**
 * Initializes the file menu of the given window.
 */
void SharedMenu::initFileMenu(QMenu& fileMenu, QMainWindow& win,
    const QString& recentFilesOption)
{
    // find first action in the file menu
    QList<QAction*> actions = fileMenu.actions();
    QAction* firstAction = NULL;
    if (!actions.empty())
        firstAction = actions.first();
    
    // append entry to the end of the file menu
    fileMenu.addSeparator();
    
    if (recentFilesOption != "")
    {
        m_recentFiles = new RecentFiles(recentFilesOption, fileMenu, -1, this);
        connect(m_recentFiles, SIGNAL(recentFileSelected(QString)), this,
                SIGNAL(recentFileSelected(QString)));

        fileMenu.addSeparator();
    }

    QAction* close = fileMenu.addAction(tr("&Close Window"));
    close->setShortcut(QKeySequence::Close);
    connect(close, SIGNAL(triggered(bool)), &win,
            SLOT(close()));
    
    QApplication* app = qApp;
    QAction* quit = fileMenu.addAction(tr("&Quit"));
    quit->setShortcut(tr("Ctrl+Q"));
    connect(quit, SIGNAL(triggered(bool)), app,
            SLOT(closeAllWindows()));
}


void SharedMenu::removeRecentFiles()
{
    if (m_recentFiles != NULL)
    {
        delete m_recentFiles;
        m_recentFiles = NULL;
    }
}


void SharedMenu::addRecentFile(QString filename)
{
    if (m_recentFiles)
        m_recentFiles->addRecentFile(filename);
    else
        qDebug("Warning: addRecentFile called for shared menu without recent files.");
}


QStringList SharedMenu::getRecentFiles(const QString& configname)
{
    QSettings settings(VSettings::getOrganization(),
            VSettings::getApplication());

    return settings.value(configname).toStringList();
}


void SharedMenu::initToolsMenu(QMenu& toolsMenu)
{
    WindowList* winlist = WindowList::globalInstance();
    if (winlist != NULL)
    {
        QSignalMapper* mapper = new QSignalMapper(&toolsMenu);
        QHash<QString, WindowType> winTypes = winlist->getRegistratedWindowTypes();
        QList<QString> keys = winTypes.uniqueKeys();
        for (int i = 0; i < keys.length(); ++i)
        {
            // create action for current window type
            const WindowType& type = winTypes[keys.at(i)];
            QAction* action = new QAction(type.name(), &toolsMenu);
            action->setShortcut(type.shortcut());
            toolsMenu.addAction(action);

            // add action to signal mapper
            mapper->setMapping(action, type.name());
            connect(action, SIGNAL(triggered(bool)), mapper, SLOT(map()));
        }

        // forward signals to window list
        connect(mapper, SIGNAL(mapped(QString)), winlist,
                SLOT(createWindow(QString)));
    }
    else
        qDebug() << "WindowList not set";

    // add entries at the beginning of the file menu
    /*QAction* vhistez = new QAction(tr("Create VHIST Arg-File"), &toolsMenu);
        vhistez->setShortcut(tr("Ctrl+E"));
    toolsMenu.addAction(vhistez);
    connect(vhistez, SIGNAL(triggered(bool)), this,
            SLOT(createVHISTezWindow()));

    QAction* vhistit = new QAction(tr("Run VHIST Arg-File"), &toolsMenu);
        vhistit->setShortcut(tr("Ctrl+R"));
    toolsMenu.addAction(vhistit);
    connect(vhistit, SIGNAL(triggered(bool)), this,
            SLOT(createVHISTitWindow()));

    QAction* unvhist = new QAction(tr("Extract from VHIST File"), &toolsMenu);
        unvhist->setShortcut(tr("Ctrl+T"));
    toolsMenu.addAction(unvhist);
    connect(unvhist, SIGNAL(triggered(bool)), this,
            SLOT(createUnVHISTWindow()));*/
}


/**
 * Initializes the given help menu.
 */
void SharedMenu::initHelpMenu(QMenu& helpMenu)
{
    QAction* about = helpMenu.addAction(tr("About"));
    connect(about, SIGNAL(triggered(bool)), this,
            SLOT(showAboutDialog()));
}


/**
 * Updates the window menu of all registered windows.
 */
void SharedMenu::updateWindowMenu()
{
    WindowList* winlist = WindowList::globalInstance();
    if (winlist != NULL)
        winlist->updateAllMenus();
    else
        qDebug() << "WindowList not set";
}


void SharedMenu::createWindow(const QString& type, const QString& filename)
{
    WindowList* winlist = WindowList::globalInstance();
    if (winlist != NULL)
        winlist->createWindow(type, filename);
    else
        qDebug() << "WindowList not set";
}


/**
 * Displays an about dialog for vhistzard.
 */
void SharedMenu::showAboutDialog()
{
    int r = 0, g = 0, b = 0;
    // find current year
    QDateTime date = QDateTime::currentDateTime();
    QString curYear = date.toString("yyyy");
    
    // create message string
    QString msg("<body bgcolor=\"#%1%2%3\">"
		"<h2>%4</h2>"
		"<p>Copyright &copy; 2007-%6 MPIfnF</p>"
		"<p>see <a href=\"http://www.nf.mpg.de/vhist\">VHIST "
                "online</a> for details</p>"
		"<p>please send messages to "
                "<a href=\"mailto:vhist@nf.mpg.de\">vhist@nf.mpg.de</a></p>"
		"<p/><p/>"
		"</body>");
    msg = msg.arg(r, 2, 16, QLatin1Char('0'))
             .arg(g, 2, 16, QLatin1Char('0'))
             .arg(b, 2, 16, QLatin1Char('0'))
             .arg(VHISTZARD_FULLNAME)
             .arg(curYear);
    
    // show dialog
    QMessageBox::about(NULL, tr("About ") + VHISTZARD_NAME,
            msg);
}

