/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: utilities.cpp 2794 2013-05-06 11:26:53Z michael $
 */

#include "utilities.h"

#include <QCoreApplication>
#include <QApplication>
#include <QString>
#include <QDir>
#include <QFileInfo>
#include <QtDebug>
#include <QStringList>
#include <QDateTime>
#include <QProcess>

#if defined(Q_OS_WIN)
#include <windows.h>
#else
#include <string.h>
#include <unistd.h>
#endif

#ifdef Q_OS_MAC
#include <Carbon/Carbon.h>
#endif


QString joinPaths(const QString& path1, const QString& path2)
{
    return path1 + QDir::separator() + path2;
}


/**
 * The directory in which the application binary is located. On Mac OS X,
 * the MacOS/ directory within the bundle is returned. On all other
 * platforms, the bin directory is returned.
 * 
 * @return The path to the directory in which the binary is located.
 */
QString getBinaryDir()
{
    QString path(QCoreApplication::applicationDirPath());
    QDir dirPath(path);
    return dirPath.canonicalPath();
}


/**
 * The directory of the application. On Mac OS X, this is the directory
 * which contains the bundle, on all other platforms, \c getApplicationDir()
 * is the same as \c getBinaryDir().
 * 
 * @return The path to the application directory.
 */
QString getApplicationDir()
{
    QString dirStr = getBinaryDir();
    
#ifdef Q_OS_MAC    
    // remove <appname>.app/Contents/MacOS from path
    if (dirStr.endsWith(QString("/Contents/MacOS")))
    {
        QDir appDir(dirStr);
        appDir.cdUp();
        appDir.cdUp();
        appDir.cdUp();
        dirStr = appDir.absolutePath();
    }
#endif // end of Mac specific code
    
    return dirStr;
}


/**
 * Returns the root directory of the vhist core distribution is located.
 * On Mac OS X, this is the vhist-src directory within the package,
 * on other platforms, it is the root directory of the installation.
 *
 * @return The root directory of the vhist core distribution.
 */
QString getCoreDir()
{
    QString dirStr = getBinaryDir();
    
#ifdef Q_OS_MAC
    QDir appDir(dirStr + "/../vhist-src/");
    dirStr = appDir.absolutePath();
#else
    QDir appDir(dirStr);
    appDir.cdUp();
    dirStr = appDir.absolutePath();
#endif
    return dirStr;
}


/**
 * Returns the doc directory of the vhist distribution.
 * 
 * @return The doc directory.
 */
QString getDocDir()
{
    QDir appDir(getCoreDir() + "/doc/usage/");
    qDebug("%s",appDir.absolutePath().toLatin1().data());

    // !ms: added in parallel to change in GuiWindow::GuiWindow
    if (QDir(appDir.absolutePath() + "vhist/").exists())
        appDir = appDir.absolutePath() + "vhist";
    qDebug("%s",appDir.absolutePath().toLatin1().data());

    if (!appDir.exists())
    {
    // possibly we are debugging and running in the VHIST-subversion tree layout
        QDir appDir2(getApplicationDir());
        qDebug() << appDir2.absolutePath();
        appDir2.cdUp();
        appDir2.cdUp();
        appDir2.cdUp();
        qDebug() << appDir2.absolutePath();
        appDir2.cd("doc/generated/usage");
        qDebug("%s",appDir2.absolutePath().toLatin1().data());

        // !ms: added in parallel to change in GuiWindow::GuiWindow
        if (QDir(appDir2.absolutePath() + "vhist/").exists())
            appDir = appDir2.absolutePath() + "vhist";
        qDebug("%s",appDir2.absolutePath().toLatin1().data());

        return appDir2.absolutePath();
    }
    
    return appDir.absolutePath();
}

/**
 * Returns the bin directory of the distribution.
 *
 * @return The bin directory.
 */
QString getBinDir()
{
    QDir appDir(getCoreDir() + "/bin/");
    return appDir.absolutePath();
}


/**
 * Returns the examples directory of the distribution.
 *
 * @return The examples directory.
 */
QString getExamplesDir()
{
    QDir appDir(getCoreDir() + "/examples/");
    if (!appDir.exists())
    {
        QDir appDir2(getApplicationDir() + "/../examples/");
        return appDir2.absolutePath();
    }
    return appDir.absolutePath();
}


/**
 * Returns the templates directory of the distribution.
 *
 * @return The template directory.
 */
QString getTemplatesDir()
{
    QDir appDir(getCoreDir() + "/res/templates/");
    if (appDir.exists())
        return appDir.absolutePath();
    else
    {
        QDir appDir2(getApplicationDir() + "/../res/templates/");
        if (appDir2.exists())
            return appDir2.absolutePath();
        else {
            // debug dir
            QDir appDir3(getApplicationDir() + "/../templates/");
            if (appDir3.exists())
                return appDir3.absolutePath();
            else {
                QDir appDir4(getApplicationDir() + "/../vhistzard/templates/");
                if (appDir4.exists())
                    return appDir4.absolutePath();
            }            
        }
    }
    
    return appDir.absolutePath();
}

QString escapeHtml(const QString& input)
{
    QString s = input;
    s = s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;");
    return s.replace("\"", "&quot;").replace("\n", "<br />");
}

QString makeBold(const QString& input)
{
    QString bold("<b>%1</b>");
    return bold.arg(escapeHtml(input));
}

QString makeRed(const QString& input)
{
    QString red("<font color=\"#EB0014\">%1</font>");
    return red.arg(escapeHtml(input));
}

QString makeOrange(const QString& input)
{
    QString orange("<font color=\"#FEBF1C\">%1</font>");
    return orange.arg(escapeHtml(input));
}

/**
 * Makes the given path absolute if it is relative and removes ".." and "."
 * entries. The given path must not exist.
 * @param[in] path The path to make absolute.
 * @return The absolute path.
 */
QString makePathAbsolute(const QString& path)
{
    QFileInfo pinfo(path);
    return pinfo.absoluteFilePath();
}

/**
 * Makes the given path absolute if it is relative, removes ".." and "."
 * entries and follows all symlinks in the path. The given path must exist.
 * @param[i] path The path to make canonical.
 * @return The canonical path.
 */
QString makePathCanonical(const QString& path)
{
    QFileInfo pinfo(path);
    return pinfo.canonicalFilePath();
}

/* DOES NOT WORK CORRECTLY, YET.
 * movePathUp("/foo/bar", 3) => "/.."
  QString movePathUp(QString path, int count)
{
    for (int i = 0; i < count; i++)
        path += "/..";
    QFileInfo pinfo(path);
    return pinfo.absoluteFilePath();

}*/


/**
 * Searches a list of relative filepathes, having a common \a basePath, for
 * existing files. The first file in the list, which does exist, is returned.
 * If non of the files does exist, an empty string is returned.
 * 
 * @param[in] basePath A common prefix of all files in \a files.
 * @param[in] files A list of files relative to \a basePath.
 * @return The relative version of the first file in the list, which does
 * exist. If no file does exist, an empty string is returned.
 */
QString firstExistingFile(const QString& basePath, const QStringList& files)
{
    QStringList::const_iterator it;
    for (it = files.begin(); it != files.end(); ++it)
    {
        QString file(basePath + QString("/") + *it);
        QFileInfo finfo(file);
        if (finfo.exists() && finfo.isFile())
            return file;
    }
    return QString("");
}


/**
 * Searches a list of relative pathes, having a common \a basePath, for
 * existing pathes. The first path in the list, which does exist, is returned.
 * If non of the pathes does exist, an empty string is returned.
 * 
 * @param[in] basePath A common prefix of all pathes in \a files.
 * @param[in] files A list of pathes relative to \a basePath.
 * @return The relative version of the first path in the list, which does
 * exist. If no path does exist, an empty string is returned.
 */
QString firstExistingPath(const QString& basePath, const QStringList& files)
{
    QStringList::const_iterator it;
    for (it = files.begin(); it != files.end(); ++it)
    {
        QString file(basePath + QDir::separator() + *it);
        QFileInfo finfo(file);
        if (finfo.exists())
            return file;
    }
    return QString("");
}


/**
 * Recursively copies the content of a directory into another directory.
 * Symlinks to directories are not copied to prevent infinit recursions.
 * The destination directory must already exist before copyDirectory is
 * called.
 * 
 * @param[in] srcDir The directory which is copied.
 * @param[in] destDir The directory into which the content of \a srcDir is
 * copied.
 * @return \c true if copying was successful and \c false otherwise. If
 * the directory was copied partially, the copied files are not removed.
 */
bool copyDirectory(const QDir& srcDir, const QDir& destDir)
{
    QString srcPath = srcDir.canonicalPath();
    QString destPath = destDir.canonicalPath();
    if (srcPath == destPath)
        return false;

    QDir srcBase(srcDir);
    QDir destBase(destDir);
    
    QDir::Filters filters = QDir::Files | QDir::Dirs | QDir::NoDotAndDotDot
            | QDir::NoSymLinks;
    QStringList entries = srcBase.entryList(filters);
    QStringList::iterator it;
    
    bool successTillNow = true;
    
    for (it = entries.begin(); it != entries.end() && successTillNow; it++)
    {
        QString curSrc = srcBase.absoluteFilePath(*it);
        QString curDest = destBase.absoluteFilePath(*it);
        qDebug() << "copying:" << curSrc << " -> " << curDest;
        
        QFileInfo srcInfo(curSrc);
        
        if (srcInfo.isFile())
            successTillNow = QFile::copy(curSrc, curDest);
        else if (srcInfo.isDir() && !srcInfo.isSymLink())
        {
            successTillNow = (destBase.mkdir(*it) && 
                    copyDirectory(curSrc, curDest));
        }
    }
    return successTillNow;
}


/**
 * Recursively removes a directory and it's content. Symlinks to
 * directories will only be removed, but not entered by this function.
 * 
 * @param[in] dir The directory which should be deleted.
 * @return true if the directory was successfully deleted and false otherwise.
 */

bool removeDirectory(const QDir& dir)
{
    QDir::Filters filters = QDir::Files | QDir::Dirs | QDir::NoDotAndDotDot;
    QStringList entries = dir.entryList(filters);
    QStringList::iterator it;
    
    bool successTillNow = true;
    
    for (it = entries.begin(); it != entries.end() && successTillNow; it++)
    {
        QString curEntry = dir.absoluteFilePath(*it);
        QFileInfo entryInfo(curEntry);
        qDebug() << "delete.. " << curEntry;
        if (entryInfo.isFile() || entryInfo.isSymLink())
            successTillNow = QFile::remove(curEntry);
        else if (entryInfo.isDir())
            successTillNow = removeDirectory(QDir(curEntry));
    }
    if (successTillNow)
        successTillNow = dir.rmdir(dir.absolutePath());
    return successTillNow;
}

//    ***********************************************************************

static const QString monthStr[]= {
    QLatin1String("???"),
    QLatin1String("Jan"),
    QLatin1String("Feb"),
    QLatin1String("Mar"),
    QLatin1String("Apr"),
    QLatin1String("May"),
    QLatin1String("Jun"),
    QLatin1String("Jul"),
    QLatin1String("Aug"),
    QLatin1String("Sep"),
    QLatin1String("Oct"),
    QLatin1String("Nov"),
    QLatin1String("Dec"),
    QLatin1String("")
};

static const QString dayStr[]= {
    QLatin1String("Son"),
    QLatin1String("Mon"),
    QLatin1String("Tue"),
    QLatin1String("Wed"),
    QLatin1String("Thu"),
    QLatin1String("Fri"),
    QLatin1String("Sat"),
    QLatin1String("Son"),
    QLatin1String("")
};

//! non-localized months
QString getMonthStr (const int iMonth)
{
    if (iMonth >= 1 && iMonth <= 12)
        return monthStr[iMonth];
    return monthStr[0];
}

//! non-localized days
QString getDayStr(const int iDay)
{
    if (iDay >= 0 && iDay <= 7)
        return dayStr[iDay];
    return QLatin1String("???");
}

QString getDateTimeAsString(const QDateTime &dt, 
    const bool bComma /*= true*/)
{
    //date non-localized months
    return QString::fromLatin1("%1 %2")
        .arg(getMonthStr(dt.date().month()))
        .arg(dt.toString(bComma ? 
            QLatin1String("dd yyyy, hh:mm:ss") : 
            QLatin1String("dd yyyy hh:mm:ss")));
}

QString getCurrentDateTimeAsString() {
    return getDateTimeAsString(QDateTime::currentDateTime(), false);
}

VWaitCursor::VWaitCursor()
{QApplication::setOverrideCursor(QCursor(Qt::WaitCursor)); }

VWaitCursor::~VWaitCursor()
{QApplication::restoreOverrideCursor(); }

QString getComputerName()
{
#if defined(Q_OS_WIN)
    wchar_t szUserName[1024];
    DWORD iSize=sizeof(szUserName);
    ::GetComputerName(szUserName,&iSize);
    return QString::fromUtf16((ushort*)szUserName);
#else
    QProcess hn;
    hn.start("hostname", QStringList());
    if (!hn.waitForFinished())
        return "unknown";
    QByteArray result = hn.readAll();
    return QString::fromLocal8Bit(result).trimmed();
/*#elif defined(Q_OS_UNIX) // !ah does not work on all platforms (castor, u1)
    QByteArray b2 = qgetenv("PATH");
    QByteArray env=qgetenv("HOSTNAME");
    return QString::fromLocal8Bit(env.data());*/
#endif
}
