/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: windowlist.cpp 2809 2013-05-08 06:52:15Z michael $
 */

// header with declaration of this class
#include "windowlist.h"

// Qt headers
#include <QApplication>
#include <QSignalMapper>
#include <QPoint>
#include <QMenu>
#include <QAction>
#include <QMenuBar>

// debug header
#include <QtDebug>

// headers containing window definitions
#include "../vhistit/vhistitwindow.h"
#include "../vhistez/mainwindow.h"
#include "../unvhist/unvhistwindow.h"

// miscellaneous headers
#include "managedwindow.h"
#include "../shared/utilities.h"

//#ifdef Q_OS_MAC
//#include <MacWindows.h>
//#endif

WindowList* WindowList::M_globalList = NULL;


WindowList::WindowList(QString szTemplatesDir) :
        m_lastActiveWindow(NULL)
{
    if (szTemplatesDir != "")
        m_szTemplatesDir = szTemplatesDir;
    else m_szTemplatesDir = getTemplatesDir();
}


WindowList::~WindowList()
{
}


void WindowList::registerWindowType(const WindowType& descriptor)
{
    m_windowTypes.insert(descriptor.name(), descriptor);
}


void WindowList::setDefaultWindowType(const QString typeName)
{
    m_defaultWindowType = typeName;
}


QHash<QString, WindowType> WindowList::getRegistratedWindowTypes()
{
    return m_windowTypes;
}


void WindowList::createWindow(const QString& typeName, const QString& filename,
        const QVariantHash& parameters)
{
    if (m_windowTypes.contains(typeName))
    {
        const WindowType& desc = m_windowTypes[typeName];
        WindowCreatorFn creator = desc.windowCreator();
        ManagedWindow* win = creator(filename, parameters);
        initWindow(*win, desc);
        showWindow(*win);
    }
    else
        qDebug() << "No such window type: " << typeName;
}


void WindowList::addWindow(ManagedWindow& win, const WindowType& type)
{
    m_windowList.append(WindowListItem(&win, type));
    updateAllMenus();
}


void WindowList::removeAllWindows(void)
{
        QList<WindowListItem>::iterator i = m_windowList.begin();
	while (i != m_windowList.end())
	{
                i->first->close();
		++i;
	}
}

void WindowList::removeWindow(QObject* win)
{
    bool changed = false;
    QList<WindowListItem>::iterator i = m_windowList.begin();
    while (i != m_windowList.end())
        if (i->first == win)
        {
            i = m_windowList.erase(i);
            changed = true;
        }
        else
            ++i;
    
    if (changed)
        updateAllMenus();
}


void WindowList::initWindow(ManagedWindow& win, const WindowType& type)
{
    win.setAttribute(Qt::WA_DeleteOnClose);
    connect(&win, SIGNAL(destroyed(QObject*)), this,
            SLOT(removeWindow(QObject*)));
    placeWindow(win);
    addWindow(win, type);

#ifdef VHAVE_VHISTLIB
    win.removeMainMenu();
#endif
}


void WindowList::showWindow(ManagedWindow& win)
{
    win.show();
    win.activate();
}


bool WindowList::letWMPlaceWindow()
{
    return false;
}


void WindowList::placeWindow(ManagedWindow& win)
{
    if (!letWMPlaceWindow())
    {
        if (m_windowList.size() > 0)
        {
            QList<WindowListItem>::iterator i = m_windowList.end();
            ManagedWindow* lastWindow = (--i)->first;
            QPoint p(lastWindow->pos());
            // take framesize of visible window, since invisible x11 windows
            // do not have a border, yet
            int framesize = lastWindow->geometry().y() - lastWindow->y();
            QPoint offset(framesize, framesize);
            win.move(p.x() + offset.x(), p.y() + offset.y());
        }
    }
}


void WindowList::createFirstWindow()
{
    QWidgetList l = QApplication::topLevelWidgets();
    // no window opened, yet (list always contains an empty widget)
    if (QApplication::topLevelWidgets().size() <= 1)
        createWindow(m_defaultWindowType);
}


void WindowList::focusChanged()
{
    QWidget* activeWin = QApplication::activeWindow();
    if (activeWin != m_lastActiveWindow)
    {
        m_lastActiveWindow = activeWin;
        updateAllMenus();
    }
}


void WindowList::updateAllMenus()
{
#ifdef VHAVE_VHISTLIB
    return;
#endif

    QList<WindowListItem>::iterator it = m_windowList.begin();
    while (it != m_windowList.end())
    {
        ManagedWindow* win = it->first;
        QMenu* menu = win->getWindowMenu();
        if (menu != NULL)
        {
            updateWindowMenu(*menu);
        }
        ++it;
    }
}


void WindowList::updateWindowMenu(QMenu& windowMenu)
{
    windowMenu.clear();
    
    QAction* nextAction = windowMenu.addAction(tr("next Window"));
    connect(nextAction, SIGNAL(triggered()), this,
            SLOT(makeNextWindowActive()));
    QAction* prevAction = windowMenu.addAction(tr("previous Window"));
    connect(prevAction, SIGNAL(triggered()), this,
            SLOT(makePrevWindowActive()));
    windowMenu.addSeparator();
    
    QList<WindowListItem>::iterator i;
    for (i = m_windowList.begin(); i != m_windowList.end(); ++i)
    {
        QAction* action = windowMenu.addAction(i->first->windowTitle());
        connect(action, SIGNAL(triggered()), i->first, SLOT(activate()));
        // AH: workaround! if window is changed, updateAllMenus is call twice
        // once explicitly, once for during focus change
        connect(action, SIGNAL(triggered()), this, SLOT(updateAllMenus()));
        action->setCheckable(true);
        if (i->first == m_lastActiveWindow)
            action->setChecked(true);
        else
            action->setChecked(false);
    }
}


QList<WindowListItem>::iterator WindowList::findWindow(QWidget* win)
{
    QList<WindowListItem>::iterator it = m_windowList.begin();
    while (it != m_windowList.end() && it->first != win)
        ++it;
    
    return it;
}


void WindowList::makeNextWindowActive()
{
//#ifdef Q_OS_MAC
    /*
     * Test - does not work, yet
     * Problem - The default code also moves to windows on other spaces
     * under mac os x 10.5
     *
     WindowRef cwin = GetUserFocusWindow();
    WindowRef nwin = GetNextWindow(cwin);
    SelectWindow(nwin);
    BringToFront(nwin);*/
//#else
    if (m_windowList.empty())
        return;
    
    QWidget* activeWin = QApplication::activeWindow();
    QList<WindowListItem>::iterator it = findWindow(activeWin);
    
    // window found?
    if (it != m_windowList.end())
    {
        ++it;
        if (it == m_windowList.end())
            // wrap around to the start
            it = m_windowList.begin();
        it->first->activate();
    }
//#endif
}


void WindowList::makePrevWindowActive()
{
    if (m_windowList.empty())
        return;
    
    QWidget* activeWin = QApplication::activeWindow();
    QList<WindowListItem>::iterator it = findWindow(activeWin);
    
    // window found?
    if (it != m_windowList.end())
    {
        if (it != m_windowList.begin())
            --it;
        else
        {
            // wrap around to the end
            it = m_windowList.end();
            --it;
        }
        it->first->activate();
    }
}


void WindowList::setGlobalInstance(WindowList* list)
{
    WindowList::M_globalList = list;
}


WindowList* WindowList::globalInstance()
{
    return WindowList::M_globalList;
}


void WindowList::openFile(const QString& filename)
{
    QFileInfo finfo(filename);
    QString extension = finfo.suffix();

    QHash<QString, WindowType> types = getRegistratedWindowTypes();
    QList<QString> keys = types.uniqueKeys();

	QVariantHash parameters;
#ifdef VHAVE_VHISTLIB
	parameters.insert("TemplateDir",m_szTemplatesDir);
#endif

    for (int i = 0; i < keys.length(); ++i)
    {
        const WindowType& type = types.value(keys.at(i));
        QStringList associatedExtensions = type.associatedTypes();
        if (associatedExtensions.contains(extension))
        {
            createWindow(type.name(), filename, parameters);
            return;
        }
    }
    qDebug() << "error: no window type is associated with extension of file"
            << filename;
}
