/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: extractdialog.cpp 1603 2009-02-28 18:20:33Z ahuesgen $
 */

#include "extractdialog.h"
#include "../shared/helpviewer.h"
#include <QRegExp>
#include <QFileInfo>
#include <QFileDialog>
#include <QtDebug>

#include <iostream>
#include <cmath>

ExtractDialog::ExtractDialog(QWidget* parent, const QString& directory, int numFiles) : QDialog(parent)
{
    setupUi(this);
    QString plural = numFiles == 1 ? "" : "s";
    setWindowTitle(tr("Extract %1 File").arg(numFiles) + plural + "...");
    m_extractDirectoryEdit->setText(QFileInfo(directory).absoluteFilePath());
    //m_hiddenLegend = "<a href=\"toggle\">&gt;&gt; Legend</a>";
    //m_visibleLegend = m_legendLabel->text();
    //m_legendLabel->setText(m_hiddenLegend);
    initSignalSlots();
    uiChanged();
    layout()->setSizeConstraint(QLayout::SetFixedSize);
    adjustSize();
}

void ExtractDialog::initSignalSlots()
{
    connect(m_extractButton, SIGNAL(clicked()), this, SLOT(accept()));
    connect(m_cancelButton, SIGNAL(clicked()), this, SLOT(reject()));
    connect(m_adjustNamesCheck, SIGNAL(stateChanged(int)), this,
            SLOT(uiChanged()));
    connect(m_extractDirectoryEdit, SIGNAL(textChanged(const QString&)), this,
            SLOT(uiChanged()));
    connect(m_filenameMaskEdit, SIGNAL(textChanged(const QString&)), this,
            SLOT(uiChanged()));
    connect(m_openButton, SIGNAL(clicked()), this, SLOT(browseForDirectory()));
    connect(m_helpButton, SIGNAL(clicked()), this, SLOT(showLegendHelp()));
}


ExtractDialog::~ExtractDialog()
{
}


QString ExtractDialog::getDirectory()
{
    return m_extractDirectoryEdit->text();
}


bool ExtractDialog::getAskBeforeOverwrite()
{
    return m_askCheck->isChecked();
}


bool ExtractDialog::getValidateFiles()
{
    return m_validateCheck->isChecked();
}


void ExtractDialog::setAdjustFilenamesActive(int state)
{
    m_adjustNamesCheck->setChecked(state);
    uiChanged();
}


QString ExtractDialog::replacePlaceholders(const QString mask, const QString filename,
      int fileNumber, int numDigits)
{
    QString result;
    QRegExp regex("(%d)|(%f)|(%\\*f)|(%%)|(%\\*d)|(%[0-9]+d)|([^%]+)");
    QString rest(mask);
    while (rest != "")
    {
        int pos = regex.indexIn(rest);
        if (pos != 0)
            return "";
        else
        {
            QString found = regex.cap(0);
            if (found == "%f")
                result += filename;
            else if (found == "%*f")
                result += QString(filename).replace(QChar(' '), QChar('_'));
            else if (found == "%d")
                result += QString::number(fileNumber);
            else if (found == "%*d")
                result += QString("%1").arg(fileNumber, numDigits, 10, QChar('0'));
            else if (found.startsWith("%") && found.endsWith("d")) // %[number]d
            {
                QString numAsString = found.mid(1);
                numAsString.chop(1);
                int digits = numAsString.toInt();
                qDebug() << "digits" << digits;
                result += QString("%1").arg(fileNumber, digits, 10, QChar('0'));
            }
            else
                result += found;
            rest = rest.mid(found.count());
        }
    }
    qDebug() << "filename" << result;
    return result;
}

QString ExtractDialog::adjustFilename(const QString& filename, int numFiles, int fileNum)
{
    if (m_adjustNamesCheck->isChecked() && m_filenameMaskEdit->text() != "")
    {
        int numDigits = (int) std::ceil(std::log10((float)(numFiles + 1)));
        return replacePlaceholders(m_filenameMaskEdit->text(), filename,
                fileNum, numDigits);
    }
    return filename;
}

QStringList ExtractDialog::adjustFilenames(const QStringList& filenames)
{
    QStringList result;
    for (int i = 0; i < filenames.count(); i++)
    {
        QString newName = adjustFilename(filenames[i], filenames.count(), i + 1);
        result.append(newName);
    }
    return result;
}

bool ExtractDialog::filenameMaskValid()
{
    if (!m_adjustNamesCheck->isChecked())
        return true;
    if (replacePlaceholders(m_filenameMaskEdit->text(), "fakename", 1, 0) != "")
        return true;
    return false;
}


bool ExtractDialog::directoryValid()
{
    QFileInfo finfo(m_extractDirectoryEdit->text());
    return finfo.isDir();
}


void ExtractDialog::uiChanged()
{
    m_filenameMaskEdit->setEnabled(m_adjustNamesCheck->checkState());
    m_errorLabel->setText("");

    bool inputValid = true;

    if (!directoryValid())
    {
        inputValid = false;
        m_errorLabel->setText(tr("<span style=\"color:#ff481e;\">The selected directory does not exist</span>"));
    }

    if (!filenameMaskValid())
    {
        inputValid = false;
        m_errorLabel->setText(tr("<span style=\"color:#ff481e;\">The entered filemask is not valid</span>"));
    }

    m_extractButton->setEnabled(inputValid);
}


void ExtractDialog::browseForDirectory()
{
    QString directory = QFileDialog::getExistingDirectory(this,
            tr("Select a directory to extract"), m_extractDirectoryEdit->text());
    if (directory != "")
        m_extractDirectoryEdit->setText(QFileInfo(directory).absoluteFilePath());
}


void ExtractDialog::showLegendHelp()
{
    HelpViewer::globalOpenPage("vhist-usage.htm", "sec:unvhist-legend");
}

