/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: argfile.h 2054 2010-12-17 13:29:30Z michael $
 */

#include "maketree.h"

#include <QFileInfo>
#include <QDir>
#include <QtDebug>

QRegExp PATH_SEP("[^\\\\/]*((\\\\*)|(/*))");


QTreeWidgetItem* makeOutlineItem(const QString& title,
        TagType type, int id, const QString& anchor)
{
    QStringList strings;
    if (title == "")
    {
        if (type == FILE_TAG_TYPE)
            strings << "<no filename>";
        else if (type == SECTION_TAG_TYPE)
            strings << "<no title>";
    }
    else
        strings << title;

    //qDebug() << "outline id" << type << id;
    QTreeWidgetItem* item = new QTreeWidgetItem(strings);
    item->setData(0, TAG_TYPE_ROLE, type);
    item->setData(0, TAG_ID_ROLE, id);
    item->setData(0, TAG_ANCHOR_ROLE, anchor);


    QString iconName;
    if (id >= 0 && type == FILE_TAG_TYPE) // means file is embedded
        iconName = ":/resources/embedded.png";
    else
        iconName = ":/resources/not-embedded.png";

    // set icon
    QIcon icon(iconName);
    item->setIcon(0, icon);
    return item;
}


QStringList splitPath(const QString& path)
{
    QStringList result;
    QRegExp pathSep(PATH_SEP);

    int found = pathSep.indexIn(path);
    QString foundText = pathSep.cap();

    while (found != -1 && foundText != "")
    {
        result << foundText;
        //qDebug() << foundText;
        found = pathSep.indexIn(path, found + foundText.length());
        foundText = pathSep.cap();
    }

    return result;
}


QString lastPartOfPath(const QString& path)
{
    QStringList split(splitPath(path));
    if (split.length() > 0)
        return split.at(split.length() - 1);
    else
        return "";
}


QString makePathValid(const QString& path)
{
    QStringList split = splitPath(path);
    if (split.length() > 0)
        split[0].replace(":", "_");

    return split.join("");
}


QHash<QString, TagType> prefixesFromFiles(const QList<QStringList> splitFiles)
{
    QHash<QString, TagType> prefixes;

    for (int i = 0; i < splitFiles.length(); i++)
    {
        const QStringList& f  = splitFiles.at(i);
        if (f.length() == 1)
            prefixes.insert(f.at(0), FILE_TAG_TYPE);
        else if (f.length() > 1)
            prefixes.insert(f.at(0), DIR_TAG_TYPE);
    }
    return prefixes;
}


QTreeWidgetItem* createTree(const QList<QStringList>& splitFiles, const QList<FileData>& files, const QString& root)
{
    QHash<QString, TagType> prefixes = prefixesFromFiles(splitFiles);
    QStringList prefixList = prefixes.keys();

    // just one prefix => all files have the same prefix
    if (prefixList.length() == 1)
    {
        QString prefix = prefixList.at(0);
        // prefix is directory => recurse
        if (prefixes.value(prefix) == DIR_TAG_TYPE)
        {
            QString newRoot = root + prefix;
            if (root == "<files>")
                newRoot = prefix;

            // strip prefixes of all files
            QList<QStringList> newSplitFiles = splitFiles;
            for (int i = 0; i < newSplitFiles.length(); ++i)
                newSplitFiles[i].removeFirst();
            // repeat with truncated splitfiles and extendend root
            return createTree(newSplitFiles, files, newRoot);
        }
        else if (prefixes.value(prefix) == FILE_TAG_TYPE)
        {
            const FileData& f = files.at(0);
            QString filename = splitFiles.at(0).at(0);
            QTreeWidgetItem* treeNode = makeOutlineItem(root, DIR_TAG_TYPE, -1, "");
            treeNode->addChild(makeOutlineItem(filename, FILE_TAG_TYPE, f.id, f.anchor));
            return treeNode;
        }
    }
    else if (prefixList.length() > 0)
    {
        QTreeWidgetItem* treeNode = makeOutlineItem(root, DIR_TAG_TYPE, -1, "");

        for (int i = 0; i < prefixList.length(); i++)
        {
            const QString& prefix = prefixList.at(i);
            if (prefixes.value(prefix) == DIR_TAG_TYPE)
            {

                QList<FileData> subFiles;
                QList<QStringList> subSplits;
                for (int j = 0; j < files.count(); j++)
                {
                    if (splitFiles.at(j).at(0) == prefix)
                    {
                        FileData d = files.at(j);
                        subFiles << d;
                        QStringList rest = splitFiles.at(j).mid(1);
                        subSplits << rest;
                    }
                }
                treeNode->addChild(createTree(subSplits, subFiles, prefix));
            }
            else if (prefixes.value(prefix) == FILE_TAG_TYPE)
            {
                for (int j = 0; j < files.count(); j++)
                {
                    if (splitFiles.at(j).at(0) == prefix)
                    {
                        FileData f = files.at(j);
                        QString filename = splitFiles.at(j).at(0);
                        treeNode->addChild(makeOutlineItem(filename, FILE_TAG_TYPE, f.id, f.anchor));
                    }
                }
            }
        }
        return treeNode;
    }
    return NULL;

    /*// only one item and just a filename => create a leave, ignore root and
    if (files.length() == 1 && splitFiles.at(0).length() == 1)
    {
        const FileData& d = files.at(0);
        QString filename = splitFiles.at(0).at(0);
        return makeOutlineItem(filename, FILE_TAG_TYPE, d.id, d.anchor);
    }


    // more than one entry => create inner node
    QTreeWidgetItem* treeNode = makeOutlineItem(root, DIR_TAG_TYPE, -1, "");

    for (int i = 0; i < prefixes.length(); i++)
    {
        const QString& prefix = prefixes.at(i);

        QList<FileData> subFiles;
        QList<QStringList> subSplits;
        for (int j = 0; j < files.count(); j++)
        {
            if (splitFiles.count() > 0 && splitFiles.at(j).at(0) == prefix)
            {
                FileData d = files.at(j);
                subFiles << d;
                QStringList rest = splitFiles.at(j).mid(1);
                subSplits << rest;
            }
        }
        treeNode->addChild(createTree(subSplits, subFiles, prefix));
    }
*/
    //return treeNode;
}


 bool caseInsensitiveLessThan(const FileData &f1, const FileData &f2)
 {
     return f1.filename.toLower() < f2.filename.toLower();
 }

QTreeWidgetItem* treeFromFileList(const QList<FileData>& files)
{
    // split paths into components
    QList<QStringList> splitPaths;

    QList<FileData> sortedFiles = files;

    qSort(sortedFiles.begin(), sortedFiles.end(), caseInsensitiveLessThan);

    for (int i = 0; i < sortedFiles.length(); i++)
    {
        QStringList parts = splitPath(sortedFiles[i].filename);
        splitPaths << parts;
    }

    return createTree(splitPaths, sortedFiles, "<files>");
}


QString getCommonPrefix(QStringList filenames)
{
    if (filenames.length() == 0)
        return "";

    QList<QStringList> splitPaths;
    for (int i = 0; i < filenames.length(); ++i)
        splitPaths << splitPath(filenames.at(i));

    QString commonPrefix = "";
    bool stripMore = true;

    while (stripMore)
    {
        QString curStrip = splitPaths.at(0).at(0);
        for (int i = 0; i < splitPaths.length(); ++i)
        {
            stripMore = (stripMore && splitPaths.at(i).length() > 0 &&
                    curStrip == splitPaths.at(i).at(0));
            splitPaths[i].removeFirst();
        }

        if (stripMore)
            commonPrefix += curStrip;
    }
    return commonPrefix;
}

