/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: parsetemplate.cpp 2747 2013-04-24 12:35:23Z ahuesgen $
 */

#include "parsetemplate.h"
#include "../shared/utilities.h"
#include "../shared/constants.h"
#include <QtDebug>
#include <QFile>
#include <QFileInfo>
#include <QStringList>
#include <QRegExp>
#include "time.h"

using namespace vhist;

const QString ParseTemplate::m_szMarker = "<!-- #";
const QString ParseTemplate::m_szIFDEF = "ifdef";
const QString ParseTemplate::m_szIFNDEF = "ifndef";
const QString ParseTemplate::m_szIFEQUAL = "ifequal";
const QString ParseTemplate::m_szELSE = "else";
const QString ParseTemplate::m_szENDIF = "endif";
const QString ParseTemplate::m_szBEGINBLOCK = "begin";
const QString ParseTemplate::m_szENDBLOCK = "end";
const QString ParseTemplate::m_szEVENODDBEGIN = "evenodd begin";
const QString ParseTemplate::m_szEVENODDEND = "evenodd end";
const QString ParseTemplate::m_szEVEN = "even";
const QString ParseTemplate::m_szODD = "odd";
const QString ParseTemplate::m_szMEND = " -->";
const QString ParseTemplate::m_szBracketOpen = "(";
const QString ParseTemplate::m_szBracketClose = ")";
const QString ParseTemplate::m_szMAINBLOCKNAME = "__main__";
const QString ParseTemplate::m_szFILESBLOCKNAME = "__files__";
const QString ParseTemplate::m_szUSERATTRBLOCKNAME = "__userattributes__";
const QString ParseTemplate::m_szUSERFILEATTRBLOCKNAME = "__userfileattributes__";
const QString ParseTemplate::m_szPREVIEWWSBLOCKNAME = "__previewsWS__";
const QString ParseTemplate::m_szPREVIEWFILEBLOCKNAME = "__previewsFile__";
const QString ParseTemplate::m_szTOPBLOCKNAME = "__top__";
const QString ParseTemplate::m_szBOTTOMBLOCKNAME = "__bottom__";
const QString ParseTemplate::m_szDUMPALLBLOCKNAME = "__dumpall__";
const QString ParseTemplate::m_szDUMPENTRYBLOCKNAME = "__dumpentry__";
const int ParseTemplate::m_iLenMEND = 4;
const int ParseTemplate::m_iLenMarker = 6;

const QString ParseTemplate::m_szKeywordRegex = "(\\$[A-Za-z_0-9#\\[\\*\\]]+)";

const QString ParseTemplate::m_szKeywordEvenOdd = "$EVENODD";
const QString ParseTemplate::m_szKeywordUserKey = "$USRKEY";


QString makeSectionAnchor(int sectionID)
{
    return QString("sec:%1").arg(sectionID);
}


QString makeFileAnchor(int sectionID, int posInSection)
{
    return QString("file:%1:%2").arg(sectionID).arg(posInSection);
}


QString int2str(int value)
{
    if (value >= 0)
        return QString::number(value);
    return QString();
}


/**
 * Takes a string of the form \c "$FOO[*][**][***]" and replaces the
 * \c "[*...*]" placeholder with the replacement strings in \a replacements.
 */
QString expandKeyword(const QString& keyword, const QHash<QString, QString> indexVars)
{
    // [ n*asterisk ]
    QRegExp placeholderRegExp("\\[\\*+\\]");

    // replace all placeholders ([*...*]) by the values in indexVars
    int searchPos = 0;
    int offset = 0;
    QString result(keyword);
    while (-1 != (searchPos = placeholderRegExp.indexIn(keyword, searchPos)))
    {
        QString key = placeholderRegExp.cap();
        // generate value
        QString defaultValue = QString("%1 NOT FOUND").arg(key);
        QString value = indexVars.value(key, defaultValue);
        value = QString("[%1]").arg(value);

        // replace key by value
        result.replace(searchPos + offset, key.length(), value);

        // if the length of key and value differ, pos will diverge for keyword
        // and result => offset
        searchPos += key.length();
        offset += value.length() - key.length();
    }
    return result;
}


class HNode {
public:
    HNode() {
        m_ifLevel = 0;
        m_iBlockStartMarker = -1;
        m_iBlockEndMarker = -1;
        m_iContentsStart = -1;
        m_iContentsEnd = -1;
    }
    int m_ifLevel;
    QString m_szCommand; // #ifdef, #show...; empty for simple nodes
    QString m_szContents; // without surrounding comments
    int m_iBlockStartMarker, m_iBlockEndMarker; // including comment lines
    int m_iContentsStart, m_iContentsEnd; // without surrounding comments
};


class Block {
public:
    Block(QString szBlockName) {
        m_szBlockName = szBlockName;
        m_ifLevel = 0;
     }
    
    QList<HNode *>m_Nodes;
    QString m_szBlockName;
    int m_ifLevel;
};

/**
 * Finds out the name of the next block and its content's start
 * position. Returns the contents start position or -1 if some
 * error occured.
 */
int ParseTemplate::addNamedBlock(Block *pb, int markerpos)
{
    HNode *n = new HNode;
    pb->m_Nodes.append(n);
    
    n->m_iBlockStartMarker = markerpos;
    int iBracketOpen = m_szTemplate.indexOf(m_szBracketOpen, markerpos);
    if (iBracketOpen < 0) {
        qDebug() << "[addNamedBlock] szBracketOpen not found after pos: "
                 << markerpos;
        return -1;
    }
    int iBracketClose = m_szTemplate.indexOf(m_szBracketClose, iBracketOpen);
    if (iBracketClose < 0) {
        qDebug() << "[addNamedBlock] szBracketClose not found after pos: "
                 << iBracketOpen;
        return -1;
    }

    pb->m_szBlockName = m_szTemplate.mid(iBracketOpen + 1,
                                        iBracketClose - iBracketOpen - 1);
    m_HNamedBlocks.insert(pb->m_szBlockName, pb);

    n->m_iContentsStart = m_szTemplate.indexOf(m_szMEND, iBracketClose)
        + m_iLenMEND;
    if (n->m_iContentsStart < 0) {
        qDebug() << "[addNamedBlock] \"" << m_szMEND << "\" "
                 << "not found in szTemplate";
        return -1;
    }
    
    return n->m_iContentsStart;
}


/**
 * Adds a new HNode to the given block for the marker at position
 * \c markerpos. Returns the position after the marker or -1 in case
 * of an error.
 */
int ParseTemplate::addMarker(Block *pb, int markerpos)
{
    HNode *n = new HNode;
    pb->m_Nodes.append(n);
    n->m_ifLevel = pb->m_ifLevel;
    
    n->m_iBlockStartMarker = markerpos;
    n->m_iBlockEndMarker = m_szTemplate.indexOf(m_szMEND, markerpos + 4)
        + m_iLenMEND;
    if (n->m_iBlockEndMarker < 0) {
        qDebug() << "[addMarker] \"" << m_szMEND
                 << "\" not found after pos:" << markerpos;
        return -1;
    }
    n->m_szCommand = m_szTemplate.mid(markerpos + m_iLenMarker,
                                    n->m_iBlockEndMarker
                                    - markerpos - m_iLenMarker - m_iLenMEND);
    //qDebug() << "[addMarker] command:" << n->m_szCommand;
    return n->m_iBlockEndMarker;
}


/**
 * Add part between iStart and iEnd of m_szTemplate to block pb.
 */
void ParseTemplate::addContents(Block *pb, int iStart, int iEnd)
{
    HNode *n = new HNode;
    pb->m_Nodes.append(n);
    n->m_ifLevel = pb->m_ifLevel;    
    n->m_iBlockStartMarker = iStart;
    n->m_iBlockEndMarker = iEnd;
    n->m_iContentsStart = iStart;
    n->m_iContentsEnd = iEnd;
    n->m_szContents =
        m_szTemplate.mid(n->m_iContentsStart,
                         n->m_iContentsEnd - n->m_iContentsStart).trimmed();
    if (!n->m_szContents.isEmpty())
        n->m_szContents += "\n";
}

void ParseTemplate::cleanup(void)
{
    QHash<QString, Block*>::const_iterator i;    
    for (i = m_HNamedBlocks.begin(); i != m_HNamedBlocks.end(); ++i) {
        QList<HNode*>::const_iterator n;
        QList<HNode*> *pNodes = &(i.value()->m_Nodes);
        for (n = pNodes->begin(); n != pNodes->end(); n++)
            delete *n;
        pNodes->clear();
    }
    m_HNamedBlocks.clear();
}

bool ParseTemplate::parser(void)
{
    cleanup(); // m_LNamedBlocks, m_HNamedBlocks
    
    // create first block
    m_LNamedBlocks.append(new Block(m_szMAINBLOCKNAME));
    Block *pb = m_LNamedBlocks.at(0);
    m_HNamedBlocks.insert(pb->m_szBlockName, pb); // insert into Hash
    
    int markerpos = 0;
    int ipos = 0;
    int iblock = 0;
    while (ipos < m_maxpos) {
        markerpos = m_szTemplate.indexOf(m_szMarker, ipos);
        
        if (markerpos >= 0) {
            // add part before markerpos to block
            addContents(pb, ipos, markerpos);

            // found an IFDEF marker
            if (m_szTemplate.indexOf(m_szIFDEF, markerpos) - markerpos
                == m_iLenMarker) {
                ipos = addMarker(pb, markerpos);
                pb->m_ifLevel++;
            }
            // found an IFNDEF marker
            else if (m_szTemplate.indexOf(m_szIFNDEF, markerpos) - markerpos
                == m_iLenMarker) {
                ipos = addMarker(pb, markerpos);
                pb->m_ifLevel++;
            }
            // found an IFEQUAL marker
            else if (m_szTemplate.indexOf(m_szIFEQUAL, markerpos) - markerpos
                == m_iLenMarker) {
                ipos = addMarker(pb, markerpos);
                pb->m_ifLevel++;
            }
            // found an ELSE marker
            else if (m_szTemplate.indexOf(m_szELSE, markerpos) - markerpos
                == m_iLenMarker) {
                ipos = addMarker(pb, markerpos);
            }
            // found an ENDIF marker
            else if(m_szTemplate.indexOf(m_szENDIF, markerpos) - markerpos
                    == m_iLenMarker) {
                ipos = addMarker(pb, markerpos);
                pb->m_ifLevel--; // !sv todo check < 0: syntax error
            }
            // found a BEGINBLOCK marker
            else if(m_szTemplate.indexOf(m_szBEGINBLOCK, markerpos) - markerpos
                    == m_iLenMarker) {
                iblock++;
                // add a new block to the blocklist and find out about its name
                // pb now points to the new block
                m_LNamedBlocks.insert(iblock, new Block("unknown"));
                pb = m_LNamedBlocks.at(iblock);
                ipos = addNamedBlock(pb, markerpos);
            }

            // found an ENDBLOCK marker
            else if(m_szTemplate.indexOf(m_szENDBLOCK, markerpos) - markerpos
                    == m_iLenMarker) {
                ipos = addMarker(pb, markerpos);
                iblock--;
                if (iblock < 0) {
                    qDebug() << "[parser] iblock < 0: syntax error with named Groups"
                             << "around markerpos:" << markerpos;
                    return false;
                }
                pb = m_LNamedBlocks.at(iblock);
            }
            else { // some other marker
                ipos = addMarker(pb, markerpos);
            }
            
            if (ipos < 0) return false;
        } // marker pos >= 0
        else {
            // no further marker found;
            // everything up to end of template covered
            addContents(pb, ipos, m_maxpos);
            return true;
        }
    } // while ipos
    
    qDebug() << "[parser] should not be reached.";
    return false; 
}

ParseTemplate::ParseTemplate(QObject *pLogTarget)
{
    if (pLogTarget != NULL) {
        connect(this, SIGNAL(log(const QString)),
                pLogTarget, SLOT(log(const QString)));
    }
}

ParseTemplate::~ParseTemplate()
{
    cleanup();
}

/**
 * Replaces all keywords within a piece of html code by their actual contents.
 *
 * @param[in] szContents The html string, in which the keywords should be replaced.
 * @param[in] indexVars The values for all index placeholders. Placeholders are usually of the form
 * [*], [**], etc.
 *
 * @returns The expanded html code.
 */
QString ParseTemplate::expandContents(const QString szContents, const QHash<QString, QString> indexVars)
{
    QRegExp rx(m_szKeywordRegex);
    QString result = szContents; // expanded content

    int pos = 0;
    while (-1 != (pos = rx.indexIn(result, pos)))
    {
        QString keyword = rx.cap(1);
        QString expandedKeyword = expandKeyword(keyword, indexVars);
        qDebug() << "!!!!!" << keyword << expandedKeyword;
        QString replacement;

        // replace even-odd markers
        if (m_bEvenOddHandling && expandedKeyword == m_szKeywordEvenOdd)
        {
            replacement = m_nextEvenOddIsOdd ? m_szODD : m_szEVEN;
            m_nextEvenOddIsOdd = !m_nextEvenOddIsOdd;
        }
        else // handle everything else
        {
            QString defaultValue = QString("<i>%1</i>").arg(expandedKeyword);
            replacement = m_HOneSection.value(expandedKeyword,
                    defaultValue);
        }
        // replace keyword by replacement and advance
        result.replace(pos, keyword.length(), replacement);
        pos += replacement.length();
    }
    return result;
}


QString ParseTemplate::generateOneBlock(const QString szBlockName,
                                        QHash<QString, QString> indexVars,
                                        bool bKeepEvenOdd)
{
    Block *pBlock = m_HNamedBlocks[szBlockName];
    if (pBlock == NULL) {
        return "*** ERROR: Block \"" + szBlockName + "\" not found.";
    }

    QString szHTML;
    QList<HNode*>::const_iterator n;
    QList<HNode*> Nodes = pBlock->m_Nodes;
    bool bCondition = true;
    bool bFollow = false;
    int iPassLevel = 0;

    if (!bKeepEvenOdd)
        m_bEvenOddHandling = false;

    for (n = Nodes.begin(); n != Nodes.end(); n++)
    {
        if ((*n)->m_ifLevel == 0) iPassLevel = 0;
        bFollow = (iPassLevel >= (*n)->m_ifLevel);
        qDebug() << (*n)->m_szCommand << "if/pass:" << (*n)->m_ifLevel << iPassLevel;
        
        if (bFollow && (*n)->m_szCommand.startsWith(m_szENDIF))
        {
            iPassLevel--;
        }        
        else if (bFollow && (*n)->m_szCommand.isEmpty())
        {
            szHTML.append(expandContents((*n)->m_szContents, indexVars));
        }
        else if (bFollow && (*n)->m_szCommand.startsWith(m_szIFDEF))
        {
            QStringList parts = (*n)->m_szCommand.split(" ");
            if (parts.size() != 2)
            {
                qDebug() << "*** ERROR: [ifdef] needs one parameter: $keyword";
                return  "*** ERROR: [ifdef] needs one parameter: $keyword\n";
            }

            bCondition = !(m_HOneSection.value(expandKeyword(parts[1], indexVars), "")
                          .isEmpty());
            if (bCondition) iPassLevel++;
        }
        else if (bFollow && (*n)->m_szCommand.startsWith(m_szIFNDEF))
        {
            QStringList parts = (*n)->m_szCommand.split(" ");
            if (parts.size() != 2)
            {
                qDebug() << "*** ERROR: [ifndef] needs one parameter: $keyword";
                return  "*** ERROR: [ifndef] needs one parameter: $keyword\n";
            }            
            bCondition = (m_HOneSection.value(expandKeyword(parts[1], indexVars), "")
                          .isEmpty());
            if (bCondition) iPassLevel++;
        }
        else if (bFollow && (*n)->m_szCommand.startsWith(m_szIFEQUAL))
        {
            QStringList parts = (*n)->m_szCommand.split(" ");
            if (parts.size() != 3)
            {
                qDebug() << "*** ERROR: [ifequal] needs parameters: $keyword test";
                return  "*** ERROR: [ifequal] needs parameters: $keyword test";
            }
            bCondition = (expandContents(parts[1], indexVars) == parts[2]);
            //qDebug() << "ifequal: "<< expandContents(parts[1], indexVars) << parts[1] << parts[2] << bCondition;
            if (bCondition) iPassLevel++;
        }
        else if ((*n)->m_szCommand.startsWith(m_szELSE))
        {
            if (!bCondition) iPassLevel++;
            else iPassLevel--;
        }
        else if (bFollow && (*n)->m_szCommand.startsWith(m_szEVENODDBEGIN))
        {
            m_bEvenOddHandling = true;
            m_nextEvenOddIsOdd = true;
        }
        else if (bFollow && (*n)->m_szCommand.startsWith(m_szEVENODDEND))
        {
            m_bEvenOddHandling = false;
        }
        else if (bFollow && (*n)->m_szCommand == "showblock(__files__)")
        {
            for (int ifile = 0; ifile < m_iFilesInSection; ifile++)
            {
                QString szNum = int2str(ifile);
                indexVars.insert("[*]", szNum);
                szHTML.append(generateOneBlock(m_szFILESBLOCKNAME, indexVars));
                qDebug() << "szHTML.length()" << szHTML.length();
            }
        }
        else if (bFollow && (*n)->m_szCommand == "showblock(__userattributes__)")
        {
            for (int i = 0; i < m_numStepUserAttrs; ++i)
            {
                QString szNum = int2str(i);
                indexVars.insert("[**]", szNum);
                szHTML.append(generateOneBlock(m_szUSERATTRBLOCKNAME, indexVars, true));
            }
        }
        else if (bFollow && (*n)->m_szCommand == "showblock(__userfileattributes__)")
        {
            QString szNum = indexVars.value("[*]", "-1");
            int curFile = szNum.toInt();

            for (int i = 0; i < m_LnumFileUserAttrs.at(curFile); ++i)
            {
                QString szNum = int2str(i);
                indexVars.insert("[**]", szNum);
                szHTML.append(generateOneBlock(m_szUSERFILEATTRBLOCKNAME, indexVars, true));
            }
        }
        else if (bFollow && (*n)->m_szCommand == "showblock(__previewsWS__)")
        {
            for (int i = 0; i < m_numStepPreviews; ++i)
            {
                QString szNum = int2str(i);
                indexVars.insert("[**]", szNum);
                szHTML.append(generateOneBlock(m_szPREVIEWWSBLOCKNAME, indexVars, true));
            }
        }
        else if (bFollow && (*n)->m_szCommand == "showblock(__previewsFile__)")
        {
            QString szNum = indexVars.value("[*]", "-1");
            int curFile = szNum.toInt();
            for (int i = 0; i < m_LnumFilePreviews.at(curFile); ++i)
            {
                QString szNum = int2str(i);
                indexVars.insert("[**]", szNum);
                szHTML.append(generateOneBlock(m_szPREVIEWFILEBLOCKNAME, indexVars, true));
            }
        }

        //else qDebug() << (*n)->m_szCommand;
        
    } // for node iterator

    return szHTML;
}


QString ParseTemplate::generateHtml(const QString szFileName,
                                    const QString szTemplateFile,
                                    const Summaries &summaries,
                                    bool bGeneratePDF)
{
    unsigned int iStartClock = clock();
    
    QString szHTML;
    qDebug() << "[generateHtml]" << " bGeneratePDF:" << bGeneratePDF;

    m_HOneSection.clear();
    QFileInfo fi(szFileName);
    qDebug() << "template name" << szFileName;


    // generate header variables
    if (bGeneratePDF) m_HOneSection.insert("$ISPDFGEN", "TRUE");

    m_HOneSection.insert("$VHISTFILE", fi.fileName());
    m_HOneSection.insert("$VHISTFILELASTMODIFIED",
                         getDateTimeAsString(fi.lastModified(), false));
    m_HOneSection.insert("$VHISTFILESIZEINBYTES",
                         QString("%1").arg(fi.size()));
    m_HOneSection.insert("$VHISTFILESIZEINMB",
                         QString("%1").arg(fi.size() / 1024.0 / 1024.0,
                         0, 'f', 2));
    m_HOneSection.insert("$VHISTFILE", fi.fileName());
    
    m_HOneSection.insert("$TEMPLATEFILE", QFileInfo(szTemplateFile).fileName());
    m_HOneSection.insert("$UNVHISTTIME", getCurrentDateTimeAsString());
    m_HOneSection.insert("$VHISTZARDVERSION", VHISTZARD_VERSION);    
    qDebug() << "[generateHtml] TOPBLOCKNAME";
    
    szHTML += generateOneBlock(m_szTOPBLOCKNAME, QHash<QString, QString>());

    Summaries::const_iterator it;
    int iSection = 0;
    for (it = summaries.begin(); it != summaries.end(); ++it){
        Summary s = *it;
        m_HOneSection.clear();
        m_LnumFilePreviews.clear();
        m_LnumFileUserAttrs.clear();
        m_numStepPreviews = 0;
        
        if (bGeneratePDF)
            m_HOneSection.insert("$ISPDFGEN", "TRUE");
        if (iSection == 0)
            m_HOneSection.insert("$ISFIRSTSECTION", "TRUE"); 
        else if (iSection == summaries.count()-1)
            m_HOneSection.insert("$ISLASTSECTION", "TRUE");
        
        m_HOneSection.insert("$SECTIONNUM", QString("%1").arg(iSection));

        m_HOneSection.insert("$VHISTFILE", QFileInfo(szFileName).fileName());
        m_HOneSection.insert("$TEMPLATEFILE", QFileInfo(szTemplateFile).fileName());
        m_HOneSection.insert("$UNVHISTTIME", getCurrentDateTimeAsString());
        m_HOneSection.insert("$VHISTZARDVERSION", VHISTZARD_VERSION);
    
        m_HOneSection.insert("$SECTIONID",
                             QString("%1").arg(s.step.tagID));

        qDebug() << "[generateHtml] Section:" << s.step.tagID;
    
        m_HOneSection.insert("$TITLE", s.step.title);
        m_HOneSection.insert("$SECTION", s.step.title);
        m_HOneSection.insert("$DESCRIPTION", s.step.description);
        m_HOneSection.insert("$TIMESTAMP", s.step.timestamp);
        m_HOneSection.insert("$COMMENT", s.step.comment);
        m_HOneSection.insert("$COMMAND", s.step.command);
        m_HOneSection.insert("$TOOL", s.step.tool);
        m_HOneSection.insert("$TOOLPATH", s.step.toolPath);
        m_HOneSection.insert("$HOST", s.step.host);
        m_HOneSection.insert("$USER", s.step.user);
        m_HOneSection.insert("$CREATOR", s.step.creator);
        m_HOneSection.insert("$PDFTITLE", s.step.pdfTitle);
        m_HOneSection.insert("$PDFAUTHOR", s.step.pdfAuthor);
        m_HOneSection.insert("$PDFSUBJECT", s.step.pdfSubject);
        m_HOneSection.insert("$PDFKEYWORDS", s.step.pdfKeywords);
        m_HOneSection.insert("$PDFPRODUCER", s.step.pdfProducer);
        m_HOneSection.insert("$PDFCREATOR", s.step.pdfCreator);


        m_numStepUserAttrs = s.step.userAttrs.count();
        if (m_numStepUserAttrs > 0) {
            m_HOneSection.insert("$WSHASUSRATTRIBUTES", "TRUE");
            for (int i = 0; i < m_numStepUserAttrs; ++i)
            {
                QString key = s.step.userAttrOrder[i];
                QString value = s.step.userAttrs[key];
                QString unit = s.step.userAttrUnits.value(key, "");
                QString keyname = QString("$USRKEY[%1]").arg(i);
                QString valuename = QString("$USRVALUE[%1]").arg(i);
                QString unitname = QString("$USRUNIT[%1]").arg(i);
                m_HOneSection.insert(keyname, key);
                m_HOneSection.insert(valuename, value);
                m_HOneSection.insert(unitname, unit);
            }
        }
        
        QString szNum;
        int lastNonPreviewImage = -1;
        if (s.step.files.count() > 0)
            m_HOneSection.insert("$WSHASFILES", "TRUE");

        for (int i = 0; i < s.step.files.count(); i++) {
            m_LnumFilePreviews.append(0);

            const FileSummary& f = s.step.files.at(i);
            szNum = QString("[%1]").arg(i);

            qDebug() << "[generateHtml] File:" << i;


            if (!f.preview && !f.previewWS && !f.thumbnail && !f.thumbnailOnly)
                lastNonPreviewImage = i;

            if (f.preview)
            {
                if (lastNonPreviewImage >= 0)
                {
                    int numPreviews = m_LnumFilePreviews[lastNonPreviewImage];
                    m_LnumFilePreviews[lastNonPreviewImage] = numPreviews + 1;
                    QString key = QString("$PREVIEWID[%1][%2]").arg(lastNonPreviewImage).arg(numPreviews);
                    QString value = int2str(f.tagID);
                    m_HOneSection.insert(key, value);

                    QString key2 = QString("$HASPREVIEWS[%1]").arg(lastNonPreviewImage);
                    m_HOneSection.insert(key2, "TRUE");
                }
            }

            if (f.previewWS)
            {
                m_numStepPreviews++;
                QString key = QString("$WSPREVIEWID[%1]").arg(m_numStepPreviews - 1);
                QString value = int2str(f.tagID);
                m_HOneSection.insert(key, value);
                m_HOneSection.insert("$HASWSPREVIEWS", "TRUE");
            }

            if (f.thumbnail)
                m_HOneSection.insert("$THUMBNAILID" + szNum, int2str(f.tagID));
            
            if (f.thumbnailOnly) {
                m_HOneSection.insert("$THUMBNAILONLY" + szNum, "TRUE");
                // CAVEAT: it refers to the NEXT image
                QString szRNum = QString("[%1]").arg(i + 1);
                m_HOneSection.insert("$THUMBNAILID" + szRNum, int2str(f.tagID));
            }
    
            QString szAutoThumbnail = "icon-file-32.png";            
            if (f.filetype.toLower().contains("txt"))
                szAutoThumbnail = "icon-txt-32.png";
            if (f.filetype.toLower().contains("ecat"))
                szAutoThumbnail = "icon-pet-32.png";
            if (f.filetype.toLower().contains("pdf"))
                szAutoThumbnail = "icon-pdf-32.png";
            if (f.filetype.toLower().contains("xml"))
                szAutoThumbnail = "icon-txt-32.png";
            m_HOneSection.insert("$AUTOTHUMBNAIL32" + szNum, szAutoThumbnail);
            
            m_HOneSection.insert("$POSINSECTION" + szNum, int2str(i));
            
            m_HOneSection.insert("$FILENAME" + szNum, f.filename);
            m_HOneSection.insert("$FILEPATH" + szNum, f.filepath);
            m_HOneSection.insert("$FILETYPE" + szNum, f.filetype);
            m_HOneSection.insert("$FILEPURPOSE" + szNum, f.filePurpose);
            m_HOneSection.insert("$DESCRIPTION" + szNum, f.description);
            m_HOneSection.insert("$COMMENT" + szNum, f.comment);
            m_HOneSection.insert("$LASTMODIFIED" + szNum, f.lastModified);
            m_HOneSection.insert("$GENERATEMD5" + szNum, f.generateMD5 ? "yes": "no");
            m_HOneSection.insert("$EMBEDDED" + szNum, f.embedded ? "yes": "no");
            m_HOneSection.insert("$UNCOMPRESSEDSIZE" + szNum,
                                 int2str(f.uncompressedSize));
            m_HOneSection.insert("$UNCOMPRESSEDMD5" + szNum, f.uncompressedMD5);
            m_HOneSection.insert("$COMPRESSEDSIZE" + szNum,
                                 int2str(f.compressedSize));
            m_HOneSection.insert("$COMPRESSEDMD5" + szNum, f.compressedMD5);
            m_HOneSection.insert("$COMPRESSIONMETHOD" + szNum, f.compressedMethod);
            

            m_LnumFileUserAttrs.append(f.userAttrs.count());
            if (f.userAttrs.count() > 0)
            {
                m_HOneSection.insert("$FILEHASUSRATTRIBUTES" + szNum, "TRUE");

                for (int j = 0; j < f.userAttrOrder.size(); ++j)
                {
                    QString key = f.userAttrOrder[j];
                    QString value = f.userAttrs[key];
                    QString unit = s.step.userAttrUnits.value(key, "");
                    QString keyname = QString("$USRKEY[%1][%2]").arg(i).arg(j);
                    QString valuename = QString("$USRVALUE[%1][%2]").arg(i).arg(j);
                    QString unitname = QString("$USRUNIT[%1][%2]").arg(i).arg(j);
                    m_HOneSection.insert(keyname, key);
                    m_HOneSection.insert(valuename, value);
                    m_HOneSection.insert(unitname, unit);
                }
            }
        } // for file
        
        m_iFilesInSection = s.step.files.count();
        szHTML += generateOneBlock(m_szMAINBLOCKNAME, QHash<QString, QString>());

        // special treatment for dump-all template
        if (m_HNamedBlocks.contains(m_szDUMPALLBLOCKNAME)) {
            qDebug() << "[generateHtml] DUMPALL";
            m_HOneSection.insert("$DUMPTABLECONTENTS", "");
            QString szTableContents;
            
            QMap<QString, QString>* sortedmap =
                convertHashToMap<QString, QString> (&m_HOneSection);
            QMap<QString, QString>::const_iterator is = sortedmap->constBegin();

            while (is != sortedmap->constEnd()) {            
                m_HOneSection.insert("$KEY", is.key());
                m_HOneSection.insert("$VALUE", is.value());
                szTableContents += generateOneBlock(m_szDUMPENTRYBLOCKNAME,
                                                    QHash<QString, QString>(), true);
                ++is;
            }
            m_HOneSection.insert("$DUMPTABLECONTENTS", szTableContents);
            szHTML += generateOneBlock(m_szDUMPALLBLOCKNAME, QHash<QString, QString>());

            delete sortedmap;
        }
        
        iSection++;
    } // for section
    
    qDebug() << "[generateHtml] BOTTOM";    
    szHTML += generateOneBlock(m_szBOTTOMBLOCKNAME, QHash<QString, QString>());
    qDebug() << "[generateHtml] Done";

    double tGenerateHtml = 1.0 *  (clock() - iStartClock) / CLOCKS_PER_SEC;
    emit log(QString("[generateHtml] finished in: %1s")
        .arg(tGenerateHtml, 0, 'f', 2));    
    qDebug() << "End of generateHTML";
    return szHTML;
}

/**
 * Loads the contents of a template file and stores it in m_szTemplate.
 * Then, the content of the template file is parsed using parser().
 */
void ParseTemplate::loadTemplateFile(QString szTemplateFile)
{
    unsigned int iStartClock = clock();
    QFile file(szTemplateFile);
    if (file.open(QFile::ReadOnly)) {
        qDebug() << "reading template file: " << szTemplateFile;
        m_szTemplate = QString::fromUtf8(file.readAll());
    }
    else {
        qDebug() << "*** ERROR: template file: " << szTemplateFile;
        qDebug() << "  could not be opened for reading.";
        return;
    }
    file.close();
    
    m_maxpos = m_szTemplate.size();
    if (!parser()) cleanup();
    double tTemplateFile = 1.0 * (clock() - iStartClock) / CLOCKS_PER_SEC;
    emit log(QString("[loadTemplateFile] finished in: %1s")
             .arg(tTemplateFile, 0, 'f', 2));
}
