/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: preferences.cpp 1456 2008-10-22 10:16:59Z jnock $
 */
#include "preferences.h"

#include "../shared/vsettings.h"

#include <QDir>
#include <QSettings>

#define QL1(x) QLatin1String (x)

// Log
QString Preferences::logFontFamily ()
{
	// Font	
#ifdef Q_OS_MAC
	QString defaultfont = QL1 ("Lucida Grande");
#else
	QString defaultfont = QL1 ("Courier New");
#endif

	return (getString (QL1 ("Log/FontFamily"), defaultfont));
}

void Preferences::setLogFontFamily (const QString& newfamily)
{
	setValue (QL1 ("Log/FontFamily"), newfamily);
}

qreal Preferences::logFontPointSize ()
{
	return (getReal (QL1 ("Log/FontPointSize"), 12.0f));
}

void Preferences::setLogFontPointSize (qreal newsize)
{
	setValue (QL1 ("Log/FontPointSize"), newsize);
}

unsigned Preferences::logShow ()
{
	return (getInt (QL1 ("Log/Show"), 0));
}

void Preferences::setLogShow (unsigned state)
{
	setValue (QL1 ("Log/Show"), state);
}

// Browser
QString Preferences::browTemplate ()
{
	return (getString (QL1 ("Browser/Template"), QL1 ("orange.htm")));
}

void Preferences::setBrowTemplate (const QString& value)
{
	setValue (QL1 ("Browser/Template"), value);
}

int Preferences::browZoomLevel ()
{
	return (getInt (QL1 ("Browser/ZoomLevel"), 0));
}

void Preferences::setBrowZoomLevel (int value)
{
	setValue (QL1 ("Browser/ZoomLevel"), value);
}

// Automatic validation
SizeChoice Preferences::valPerformance ()
{
	int perf = getInt (QL1 ("Validation/Performance"), MEDIUM);

	return ((SizeChoice)perf);
}

void Preferences::setValPerformance (SizeChoice value)
{
	setValue (QL1 ("Validation/Performance"), value);
}

ActionChoice Preferences::valAutomaticFileValidation ()
{
	int afv = getInt (QL1 ("Validation/AutomaticFileValidation"), ALWAYS);

	return ((ActionChoice)afv);
}

void Preferences::setValAutomaticFileValidation (ActionChoice value)
{
	setValue (QL1 ("Validation/AutomaticFileValidation"), value);
}

unsigned Preferences::valAFVLimit ()
{
	return (getInt (QL1 ("Validation/AFVLimit"), 8));
}

void Preferences::setValAFVLimit (unsigned value)
{
	setValue (QL1 ("Validation/AFVLimit"), value);
}

ActionChoice Preferences::valAutomaticSectionValidation ()
{
	int asv = getInt (QL1 ("Validation/AutomaticSectionValidation"), ALWAYS);

	return ((ActionChoice)asv);
}

void Preferences::setValAutomaticSectionValidation (ActionChoice value)
{
	setValue (QL1 ("Validation/AutomaticSectionValidation"), value);
}

unsigned Preferences::valASVLimit ()
{
	return (getInt (QL1 ("Validation/ASVLimit"), 8));
}

void Preferences::setValASVLimit (unsigned value)
{
	setValue (QL1 ("Validation/ASVLimit"), value);
}

// Extraction
unsigned Preferences::extrAskBeforeOverwrite ()
{
	return (getInt (QL1 ("Extraction/AskBeforeOverwrite"), 1));
}

void Preferences::setExtrAskBeforeOverwrite (unsigned value)
{
	setValue (QL1 ("Extraction/AskBeforeOverwrite"), value);
}

unsigned Preferences::extrValidateFiles ()
{
	return (getInt (QL1 ("Extraction/ValidateExtractedFiles"), 1));
}

void Preferences::setExtrValidateFiles (unsigned value)
{
	setValue (QL1 ("Extraction/ValidateExtractedFiles"), value);
}

unsigned Preferences::extrNumberFiles ()
{
	return (getInt (QL1 ("Extraction/NumberFiles"), 1));
}

void Preferences::setExtrNumberFiles (unsigned value)
{
	setValue (QL1 ("Extraction/NumberFiles"), value);
}

ActionChoice Preferences::extrValidateAfter ()
{
	int vae = getInt (QL1 ("Extraction/ValidateAfterExtraction"), ALWAYS);

	return ((ActionChoice)vae);
}

void Preferences::setExtrValidateAfter (ActionChoice value)
{
	setValue (QL1 ("Extraction/ValidateAfterExtraction"), value);
}

unsigned Preferences::extrVAELimit ()
{
	return (getInt (QL1 ("Extraction/VAELimit"), 8));
}

void Preferences::setExtrVAELimit (unsigned value)
{
	setValue (QL1 ("Extraction/VAELimit"), value);
}

QString Preferences::extrLastDirectory ()
{
	return (getString (QL1 ("Extraction/LastDirectory"), QDir::toNativeSeparators (QDir::homePath ())));
}

void Preferences::setExtrLastDirectory (const QString& value)
{
	setValue (QL1 ("Extraction/LastDirectory"), value);
}

unsigned Preferences::extrDefaultDirectory ()
{
	return (getInt (QL1 ("Extraction/UseDefaultDirectory"), 0));
}

void Preferences::setExtrDefaultDirectory (unsigned value)
{
	setValue (QL1 ("Extraction/UseDefaultDirectory"), value);
}

// META
QString Preferences::extrDirectory ()
{
	if (extrDefaultDirectory () == (unsigned int)true) {
		return (QDir::toNativeSeparators (QDir::homePath ())); 
	} else {
		return (extrLastDirectory ());
	}
}

//--- PRIVATE ---
QString Preferences::getString (const QString& item, const QString& defaultvalue)
{
    QSettings settings (VSettings::getOrganization (),
    VSettings::getApplication ());

    QString result = settings.value (item, defaultvalue).toString ();

    return (result);
}

qreal Preferences::getReal (const QString& item, const qreal defaultvalue)
{
	QSettings settings (VSettings::getOrganization (),
						VSettings::getApplication ());

	qreal result = settings.value (item, defaultvalue).toDouble ();

	return (result);
}

unsigned Preferences::getInt (const QString& item, const unsigned defaultvalue)
{
	QSettings settings (VSettings::getOrganization (),
						VSettings::getApplication ());

	unsigned result = settings.value (item, defaultvalue).toInt ();

	return (result);
}

void Preferences::setValue (const QString& item, const QVariant& value)
{
	QSettings settings (VSettings::getOrganization (),
						VSettings::getApplication ());

	settings.setValue (item, value);
}
