/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: unvhistpreferenceswindow.cpp 1456 2008-10-12 10:16:59Z jnock $
 */
#include "unvhistpreferenceswindow.h"
#include "preferences.h"
#include "../shared/vsettings.h"
#include "../shared/utilities.h"

#include <QDir>
#include <QFontDialog>
#include <QMessageBox>
#include <QFileDialog>

UnVHISTPreferencesWindow::UnVHISTPreferencesWindow (QWidget* parent) : QDialog (parent)
{
	setupUi (this);
	initSignalSlots ();
}

void UnVHISTPreferencesWindow::initWidgets ()
{
	// Now we take the current settings and write them to the widget
	m_logfontfamily = Preferences::logFontFamily ();
	m_logfontsize = Preferences::logFontPointSize ();
	updateCurrentFontLabel ();

	unsigned showlog = Preferences::logShow ();
	if (showlog == (unsigned int)true) {
		m_showLog->setChecked (true);
	} else {
		m_hideLog->setChecked (true);
	}

	// Zoom level
	int zoomlevel = Preferences::browZoomLevel ();
	m_zoomLevelBox->setCurrentIndex (-zoomlevel + 3);
	// As we can use -3 to 3, and -3 is index 0 in the box

	// Template
	QString templa = getTemplatesDir() + QDir::separator() + Preferences::browTemplate ();

	m_templateLineEdit->setText (templa); 

	// Validation Performance // TODO: Spaghetti!
	SizeChoice bufsize = Preferences::valPerformance ();

	switch (bufsize) {
		case SMALL:
			m_smallValidationPerformance->setChecked (true);
			break;

		case MEDIUM:
			m_mediumValidationPerformance->setChecked (true);
			break;

		case LARGE:
			m_largeValidationPerformance->setChecked (true);
			break;
	}

	// Automatic file validation
	ActionChoice afv = Preferences::valAutomaticFileValidation ();

	switch (afv) {
		case ALWAYS:
			m_alwaysAutomaticFileValidation->setChecked (true);
			break;

		case NEVER:
			m_neverAutomaticFileValidation->setChecked (true);
			break;

		case DEPENDS:
			m_dependsAutomaticFileValidation->setChecked (true);
			break;
	}

	m_AFVLimit->setValue (Preferences::valAFVLimit ());

	// Automatic section validation
	ActionChoice asv = Preferences::valAutomaticSectionValidation ();

	switch (asv) {
		case ALWAYS:
			m_alwaysAutomaticSectionValidation->setChecked (true);
			break;

		case NEVER:
			m_neverAutomaticSectionValidation->setChecked (true);
			break;

		case DEPENDS:
			m_dependsAutomaticSectionValidation->setChecked (true);
			break;
	}

	m_ASVLimit->setValue (Preferences::valASVLimit ());

	// Extraction
	m_askBeforeOverwrite->setChecked (Preferences::extrAskBeforeOverwrite ());
	m_validateExtractedFiles->setChecked (Preferences::extrValidateFiles ());
	m_numberFiles->setChecked (Preferences::extrNumberFiles ());

	ActionChoice vae = Preferences::extrValidateAfter ();

	switch (vae) {
		case ALWAYS:
			m_alwaysValidateAfterExtraction->setChecked (true);
			break;

		case NEVER:
			m_neverValidateAfterExtraction->setChecked (true);
			break;

		case DEPENDS:
			m_dependsValidateAfterExtraction->setChecked (true);
			break;
	}	

	m_VAELimit->setText (QString::number (Preferences::extrVAELimit ()));
	m_defaultExtractionDirectory->setText (Preferences::extrLastDirectory ());
	m_defaultExtractionDirectoryCheckBox->setChecked (Preferences::extrDefaultDirectory ());
}

void UnVHISTPreferencesWindow::initSignalSlots ()
{
	connect (m_okButton, SIGNAL(clicked ()), 
			 this, SLOT(okButtonPressed ()));
	
	connect (m_changeFontButton, SIGNAL(clicked ()),
			 this, SLOT(changeFontButtonPressed ()));

	connect (m_defaultTemplate, SIGNAL(stateChanged(int)),
			 this, SLOT(defaultTemplateBoxClicked(int)));

	connect (m_defaultExtractionDirectoryCheckBox, SIGNAL(stateChanged(int)),
			this, SLOT(defaultExtractionDirBoxClicked(int)));

	connect (m_setPathButton, SIGNAL(clicked()),
			 this, SLOT(setPathButtonPressed()));

	connect (m_setExtractionPathButton, SIGNAL(clicked()),
			this, SLOT(setExtractionPathButtonPressed()));
}

void UnVHISTPreferencesWindow::updateCurrentFontLabel ()
{
	m_currentFontLabel->setText (m_logfontfamily + ", " + QString::number (m_logfontsize) + " pt");
}

void UnVHISTPreferencesWindow::okButtonPressed ()
{
	// Store the changes
	// Log
	Preferences::setLogFontFamily (m_logfontfamily);
	Preferences::setLogFontPointSize (m_logfontsize);

	if (m_showLog->isChecked ()) {
		Preferences::setLogShow (1);
	} else {
		Preferences::setLogShow (0);
	}

	// Zoom level
	Preferences::setBrowZoomLevel (m_zoomLevelBox->currentText ().toInt ());

	// Template
	if (m_defaultTemplate->isChecked ()) {
		Preferences::setBrowTemplate ("black-and-white.htm");
	} else {
		// Note: As we don't support several directories now, we just cut the directory away.
		// However, as the default directory + the file name don't need to result in a path
		// to an existant file, we need to check this first, too. If this fails, we just use the
		// default.
		QString fullpath = m_templateLineEdit->text ();
		int lastindex = fullpath.lastIndexOf (QDir::separator());

		if (lastindex == -1) { // That isn't a valid input, anyway
			Preferences::setBrowTemplate ("black-and-white.htm");
		} else {
			QString filename = fullpath.right (fullpath.length() - lastindex - 1);

			QDir* dir = new QDir ();

			if (dir->exists (getTemplatesDir() + QDir::separator() + filename)) { // Okay, useable
				Preferences::setBrowTemplate (filename);
			} else {
				Preferences::setBrowTemplate ("black-and-white.htm");
			}

			delete dir;
		}
	} 

	// Automatic Validation
	if (m_smallValidationPerformance->isChecked () == true) {
		Preferences::setValPerformance (SMALL);
	}
	if (m_mediumValidationPerformance->isChecked () == true) {
		Preferences::setValPerformance (MEDIUM);
	}
	if (m_largeValidationPerformance->isChecked () == true) {
		Preferences::setValPerformance (LARGE);
	}

	if (m_alwaysAutomaticFileValidation->isChecked () == true) {
		Preferences::setValAutomaticFileValidation (ALWAYS);
	}
	if (m_neverAutomaticFileValidation->isChecked () == true) {
		Preferences::setValAutomaticFileValidation (NEVER);
	}
	if (m_dependsAutomaticFileValidation->isChecked () == true) {
		Preferences::setValAutomaticFileValidation (DEPENDS);
	}

	Preferences::setValAFVLimit (m_AFVLimit->value ());

	if (m_alwaysAutomaticSectionValidation->isChecked () == true) {
		Preferences::setValAutomaticSectionValidation (ALWAYS);
	}
	if (m_neverAutomaticSectionValidation->isChecked () == true) {
		Preferences::setValAutomaticSectionValidation (NEVER);
	}
	if (m_dependsAutomaticSectionValidation->isChecked () == true) {
		Preferences::setValAutomaticSectionValidation (DEPENDS);
	}

	Preferences::setValASVLimit (m_ASVLimit->value ());

	// Extraction
	if (m_askBeforeOverwrite->isChecked () == true) {
		Preferences::setExtrAskBeforeOverwrite (1);
	} else {
		Preferences::setExtrAskBeforeOverwrite (0);
	}

	if (m_validateExtractedFiles->isChecked () == true) {
		Preferences::setExtrValidateFiles (1);
	} else {
		Preferences::setExtrValidateFiles (0);
	}

	if (m_numberFiles->isChecked () == true) {
		Preferences::setExtrNumberFiles (1);
	} else {
		Preferences::setExtrNumberFiles (0);
	}

	if (m_alwaysValidateAfterExtraction->isChecked () == true) {
		Preferences::setExtrValidateAfter (ALWAYS);
	}
	if (m_neverValidateAfterExtraction->isChecked () == true) {
		Preferences::setExtrValidateAfter (NEVER);
	}
	if (m_dependsValidateAfterExtraction->isChecked () == true) {
		Preferences::setExtrValidateAfter (DEPENDS);
	}	
	
	Preferences::setExtrVAELimit (m_VAELimit->text ().toInt ());

	Preferences::setExtrLastDirectory (m_defaultExtractionDirectory->text ());

	if (m_defaultExtractionDirectoryCheckBox->isChecked () == true) {
		Preferences::setExtrDefaultDirectory (1);	
	} else {
		Preferences::setExtrDefaultDirectory (0);	
	}

	// Accept the dialog
	accept ();
}

void UnVHISTPreferencesWindow::defaultTemplateBoxClicked (int state)
{
	if (Qt::Checked == state) { // Turn the rest to grey
		m_templateLineEdit->setEnabled (false);
		m_setPathButton->setEnabled (false);
	} else { 
		m_templateLineEdit->setEnabled (true);
		m_setPathButton->setEnabled (true);
	}
}


void UnVHISTPreferencesWindow::defaultExtractionDirBoxClicked (int state)
{
	if (Qt::Checked == state) { // Turn the rest to grey
		m_defaultExtractionDirectory->setEnabled (false);
		m_setExtractionPathButton->setEnabled (false);
	} else { 
		m_defaultExtractionDirectory->setEnabled (true);
		m_setExtractionPathButton->setEnabled (true);
	}
}

void UnVHISTPreferencesWindow::changeFontButtonPressed ()
{
	bool ok;

	QFont font = QFontDialog::getFont (&ok, QFont (m_logfontfamily, (int)m_logfontsize), this);

	if (ok) {
		m_logfontfamily = font.family ();
		m_logfontsize = font.pointSizeF ();

		updateCurrentFontLabel ();
	} 
}

void UnVHISTPreferencesWindow::setPathButtonPressed ()
{
	QString newtemplate = QFileDialog::getOpenFileName (this, "Open template...", getTemplatesDir());
	
	if (!newtemplate.isEmpty ()) {
		newtemplate = QDir::toNativeSeparators (newtemplate);
		m_templateLineEdit->setText (newtemplate);
	}
}

void UnVHISTPreferencesWindow::setExtractionPathButtonPressed ()
{
	QString newtemplate = QFileDialog::getExistingDirectory (this, "Set path...", QDir::homePath ());
	
	if (!newtemplate.isEmpty ()) {
		newtemplate = QDir::toNativeSeparators (newtemplate);
		m_defaultExtractionDirectory->setText (newtemplate);
	}
}


int UnVHISTPreferencesWindow::exec ()
{
	// This is called each time when the dialog pops up.
	// As the configuration items may have been changed
	// between the creation of the dialog and the first exec
	// or between two execs, we need to do this each time
	// in order to be up to date.
	
	initWidgets ();
	return (QDialog::exec ());
}
