/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: unvhistwindow.cpp 2740 2013-04-24 08:54:27Z michael $
 */

#include "unvhistwindow.h"
#include "parsetemplate.h"
#include "hellomessage.h"
#include "preferences.h"
#include "extractdialog.h"

#include <unvhistlib/summarycreator.h>
#include <shared/stopwatch.h>
#include <unvhistlib/md5generator.h>

#include "../shared/utilities.h"
#include "../shared/helpviewer.h"
#include "../shared/constants.h"
#include "../shared/vsettings.h" 

#include <QString>
#include <QDropEvent>
#include <QDialog>
#include <QFileDialog>
#include <QFileInfo>
#include <QVector>
#include <QDesktopServices>
#include <QtDebug>
#include <QIcon>
#include <QMessageBox>
#include <QPrinter>
#include <QTextBlock>
#include <QTimer>
#include <QFontDialog>
#include <QList>
#include <QProgressBar>
#include <QCloseEvent>
#include <QMimeData>

using namespace vhist;


UnVHISTWindow::UnVHISTWindow(const QString& filename, QString szTemplatesDir) :
        m_validation(PROGRESS_FINISHED), m_extraction(PROGRESS_FINISHED),
        m_extractContextAction(NULL), m_validateContextAction(NULL),
        m_infoBrowser(NULL)
{
    setupUi(this);

    m_infoBrowser = new VhistBrowser();
    m_infoViewer->setTextBrowser(m_infoBrowser);

    m_szTemplatesDir = szTemplatesDir;
    if (m_szTemplatesDir == "")
        m_szTemplatesDir = getTemplatesDir();

    m_templatemenuready = false;
    
    createLogContextMenu();
    createBrowserContextMenu();
    createOutlineContextMenu();
    createPreferencesWindow();
    initSharedMenu("Unvhist/RecentFiles");
    initSignalSlots();

    initShortcuts();
    initTemplateMenu();
    initMainWidgets();
    initHelloMessage();
    initWhatsThis();
    initIcons();
    
    loadSettings ();
    if (filename != "")
        setVHISTFile(filename);

    setAcceptDrops(true);
    QTimer::singleShot(10, this, SLOT(vneIncrementally()));
}

UnVHISTWindow::~UnVHISTWindow()
{
    // Store zoom level
    // This is not done in VhistBrowser, as VhistBrowser is used twice
    // for the Infobrowser and the Backgroundbrowser
    Preferences::setBrowZoomLevel (m_infoBrowser->zoom ());
}


void UnVHISTWindow::closeEvent(QCloseEvent* event)
{
    if (m_extraction == PROGRESS_RUNNING)
    {
        QMessageBox::StandardButton b = QMessageBox::question(this,
                tr("extraction is running"), tr("VHISTzard is currently "
                "extracting a file. If you close this window now, the "
                "extraction will be canceled and the file may be incomplete. "
                "Do you really want to close the window?"),
                QMessageBox::Yes | QMessageBox::No);
        
        if (b == QMessageBox::Yes)
            event->accept();
        else
            event->ignore();
    }
    else
        event->accept();
}


void UnVHISTWindow::log(const QString szText)
{
    QString trimmedText = szText.trimmed();
    m_log->textBrowser()->append(QString("<small>%1</small>").arg(trimmedText));
}


void UnVHISTWindow::initSignalSlots()
{
    // init
    
    // logging
    m_parser.publishToLogSlot(this);
    m_validationQueue.publishToLogSlot(this);
    m_extractionQueue.publishToLogSlot(this);
    
    connect(m_templateMenu,
            SIGNAL(currentIndexChanged(const QString &)),
            this, SLOT(on_templateMenu_currentIndexChanged(const QString &)));
    connect(m_openAction, SIGNAL(triggered(bool)),
            this, SLOT(selectVHISTFile()));
    connect(actionOpen_Template_Directory, SIGNAL(triggered(bool)),
            this, SLOT(openTemplateDir()));
    connect(actionExport_as_PDF, SIGNAL(triggered(bool)),
            this, SLOT(on_actionExport_as_PDF()));
    connect(actionExport_as_HTML, SIGNAL(triggered(bool)),
            this, SLOT(on_actionExport_as_HTML()));
    connect(actionExport_as_Raw_HTML, SIGNAL(triggered(bool)),
            this, SLOT(on_actionExport_as_Raw_HTML()));
    connect(actionExport_as_Plain_Text, SIGNAL(triggered(bool)),
            this, SLOT(on_actionExport_as_Plain_Text()));
    connect(m_outlineTree, SIGNAL(itemDoubleClicked(QTreeWidgetItem*, int)),
            this, SLOT(outlineDoubleClicked(QTreeWidgetItem*, int)));
    connect(m_outlineTree, SIGNAL(itemSelectionChanged()),
            this, SLOT(outlineSelectionChanged()));
    connect(m_stopButton, SIGNAL(pressed()), this, SLOT(stopCurrentWork()));
    
    // validation
    connect(&m_validationQueue, SIGNAL(fileProcessed(int, bool)),
            this, SLOT(fileValidated(int, bool)));
    connect(&m_validationQueue, SIGNAL(sectionProcessed(int, bool)),
            this, SLOT(sectionValidated(int, bool)));
    connect(&m_validationQueue, SIGNAL(percentFinishedChanged(int, const QString)),
            this, SLOT(updateProgressBar(int, const QString)));
    connect(m_validateAction, SIGNAL(triggered(bool)), this, SLOT(validateSelected()));
    
    // extraction
    connect(&m_extractionQueue, SIGNAL(percentFinishedChanged(int, const QString)),
            this, SLOT(updateProgressBar(int, const QString)));
    connect(m_extractAction, SIGNAL(triggered(bool)), this, SLOT(extractSelected()));
    
    // !sv HTML browser, receive notifications about selected links
    connect(m_infoBrowser, SIGNAL(anchorClicked(const QUrl&)), this,
            SLOT(anchorClicked(const QUrl&)));
    connect(m_infoBrowser, SIGNAL(sourceChanged(const QUrl&)), this,
            SLOT(anchorClicked(const QUrl&)));    
    connect(m_helpAction, SIGNAL(triggered(bool)), this, SLOT(showHelp()));
    connect(m_helpAction_2, SIGNAL(clicked()), this, SLOT(showHelp()));
    
    // find bar controls
    connect(m_findAction, SIGNAL(triggered(bool)),
            this, SLOT(showFindBar()));
    connect(m_findNextAction, SIGNAL(triggered(bool)), this,
            SLOT(findNext()));
    connect(m_findPrevAction, SIGNAL(triggered(bool)), this,
            SLOT(findPrev()));
    
    // Zoom buttons
    connect(m_zoomPlus, SIGNAL(clicked()), m_infoBrowser, SLOT(zoomIn()));
    connect(m_zoomMinus, SIGNAL(clicked()), m_infoBrowser, SLOT(zoomOut()));
    
    connect(actionIncrease_Text_Size_zoom_in, SIGNAL(triggered(bool)),
            m_infoBrowser, SLOT(zoomIn()));
    connect(actionDecrease_Text_Size_zoom_out, SIGNAL(triggered(bool)),
            m_infoBrowser, SLOT(zoomOut()));
    

    // Custom Menu for the Log
    connect (m_log, SIGNAL(customContextMenuRequested(QPoint)),
             this, SLOT (createLogContextMenu(QPoint)));

    // View Menu
    connect (m_actionShow_log, SIGNAL(triggered(bool)),
             this, SLOT(showLog ()));
    connect (m_actionShow_preferences, SIGNAL(triggered(bool)), 
             this, SLOT(showPreferences()));

    // Recent files entries
    connect(this, SIGNAL(recentFileSelected(QString)), this,
           SLOT(setVHISTFile(QString)));
}


void UnVHISTWindow::initIcons()
{
    m_zoomPlus->setIcon(QIcon::fromTheme("zoom-in",
            QIcon(":/resources/zoom-plus.png")));
    m_zoomPlus->setIconSize(QSize(16, 16));
    m_zoomMinus->setIcon(QIcon::fromTheme("zoom-out",
            QIcon(":/resources/zoom-minus.png")));
    m_zoomMinus->setIconSize(QSize(16, 16));
}


void UnVHISTWindow::initShortcuts()
{
    m_openAction->setShortcuts(QKeySequence::Open);
    m_findAction->setShortcuts(QKeySequence::Find);
    m_findNextAction->setShortcuts(QKeySequence::FindNext);
    m_findPrevAction->setShortcuts(QKeySequence::FindPrevious);

    actionIncrease_Text_Size_zoom_in->setShortcuts(QKeySequence::ZoomIn);
    actionDecrease_Text_Size_zoom_out->setShortcuts(QKeySequence::ZoomOut);
}


void UnVHISTWindow::initMainWidgets()
{
    m_outlineTree->setIndentation(10); // default: 20
    
    // make tree outline collapsible, but not the html preview
    m_splitter->setCollapsible(0, true);
    m_splitter->setCollapsible(1, false);
    m_splitter->setStretchFactor(1, 10);
    m_splitter->widget(0)->hide();
    
    // make log under html preview collapsible
    m_splitterLog->setCollapsible(0, false);
    m_splitterLog->setCollapsible(1, true);
    m_splitterLog->setStretchFactor(5, 1);
    m_splitterLog->widget(1)->hide();
    
    // we want to handle any links ourselves
    m_infoBrowser->setOpenExternalLinks(false);
    m_infoBrowser->setOpenLinks(false);
    m_infoBrowser->setFocus(Qt::OtherFocusReason);
    
    m_infoBrowser->setReadOnly(true);
    m_log->textBrowser()->setReadOnly(true);

    // Configure our log to have a custom context menu, as we want more
    // options than "Select all" and "Copy".
    m_log->textBrowser()->setContextMenuPolicy (Qt::CustomContextMenu);
    
    m_progress->setVisible(false);
    m_stopButton->setVisible(false);
    
    outlineSelectionChanged();
}

void toolTip2WhatsThis(QWidget* widget)
{
    widget->setWhatsThis(widget->toolTip());
}

void UnVHISTWindow::initWhatsThis()
{
    // toolTip2WhatsThis is defined above
    toolTip2WhatsThis(m_outlineTree);
    toolTip2WhatsThis(m_templateMenu);
    toolTip2WhatsThis(m_infoBrowser);
    //toolTip2WhatsThis(m_findButton);
    //toolTip2WhatsThis(m_closeFindButton);
    toolTip2WhatsThis(m_zoomPlus);
    toolTip2WhatsThis(m_zoomMinus);
}


void UnVHISTWindow::keyPressEvent(QKeyEvent* event)
{
    if (event->modifiers() & Qt::ControlModifier)
    {
        // !sv mostly for developers, use if the template has changed
        if (event->key() == Qt::Key_U)
        {
            if (!m_lastFileName.isEmpty())
            {
                openFile(m_lastFileName, false);
            }
            event->accept();
            return;
        }
    }
    ManagedWindow::keyPressEvent(event);
}


void UnVHISTWindow::dragEnterEvent(QDragEnterEvent* event)
{
    const QMimeData* mime = event->mimeData();
    if (mime->hasUrls())
        event->acceptProposedAction();
}


void UnVHISTWindow::dropEvent(QDropEvent* event)
{
    const QMimeData* mime = event->mimeData();
    if (mime->hasUrls())
    {
        QList<QUrl> urlList = mime->urls();
        if (!urlList.empty())
            setVHISTFile(urlList.at(0).toLocalFile());
    }
}

void UnVHISTWindow::setVHISTFile(const QString& filename)
{
    openFile(filename, false);
}

void UnVHISTWindow::selectVHISTFile()
{
    QString caption(tr("Select a VHIST File"));
    QString filter(tr("VHIST File (*.vhist *.pdf);;All files (*)"));
    QString result = QFileDialog::getOpenFileName(this, caption, "", filter);
    setVHISTFile(result);
}

void UnVHISTWindow::initHelloMessage()
{
    QString message(getUnVhistHelloMessage());
    message = message.arg(VHISTZARD_FULLNAME).arg(VHISTZARD_COPYRIGHT_DATE);
    m_infoBrowser->setHtml(message);
    m_infoBrowser->show();
    
    QString szT = QString("<b>%1</b><p>started at: %2</p>")
        .arg(VHISTZARD_FULLNAME).arg(getCurrentDateTimeAsString());
    log(szT);
}

// code related to exporting

void UnVHISTWindow::on_actionExport_as_HTML()
{
    QString szFileName = QFileDialog::getSaveFileName
        (this, "Please specify a filename for the new HTML file:", "");
    if (szFileName.isEmpty()) return;
    
    if (!szFileName.endsWith(".htm") && !szFileName.endsWith(".html"))
        szFileName = szFileName + ".htm";
    
    log(QString("[Export as HTML] file:\"%1\"").arg(szFileName));
    StopWatch totalWatch(this);
    totalWatch.start("[Export as HTML] total");
    
    QString szHtml = m_infoBrowser->toHtml();
    log(QString("[Save_as_HTML] file: \"%1\"\n").arg(szFileName));
    QFile f(szFileName);
    if (!f.open(QIODevice::WriteOnly)) {
        log(QString("*** ERROR: file could not be opened for writing."));
        return;
    }
    f.write(szHtml.toLatin1(), szHtml.size());
    f.close();
}

void UnVHISTWindow::on_actionExport_as_Raw_HTML()
{
    QString szFileName = QFileDialog::getSaveFileName
        (this, "Please specify a filename for the new Raw HTML file:", "");
    if (szFileName.isEmpty()) return;
    
    if (!szFileName.endsWith(".htm") && !szFileName.endsWith(".html"))
        szFileName = szFileName + ".htm";
    
    log(QString("[Export as Raw HTML] file:\"%1\"").arg(szFileName));
    StopWatch totalWatch(this);
    totalWatch.start("[Export as Raw HTML] total");
    
    log(QString("[Save_as_HTML] file: \"%1\"\n").arg(szFileName));
    QFile f(szFileName);
    if (!f.open(QIODevice::WriteOnly)) {
        log(QString("*** ERROR: file could not be opened for writing."));
        return;
    }
    f.write(m_RawHtml.toLatin1(), m_RawHtml.size());
    f.close();
}

void UnVHISTWindow::on_actionExport_as_Plain_Text()
{
    QString szFileName = QFileDialog::getSaveFileName
        (this, "Please specify a filename for the new TEXT file:", "");
    if (szFileName.isEmpty()) return;
    
    if (!szFileName.endsWith(".txt"))
        szFileName = szFileName + ".txt";
    
    log(QString("[Export as Plain Text] file:\"%1\"").arg(szFileName));
    StopWatch totalWatch(this);
    totalWatch.start("[Export as Plain Text] total");

    QString szText = m_infoBrowser->toPlainText();
    log(QString("[Save_as_PlainText] file: \"%1\"\n").arg(szFileName));
    QFile f(szFileName);
    if (!f.open(QIODevice::WriteOnly)) {
        log(QString("*** ERROR: file could not be opened for writing."));
        return;
    }
    f.write(szText.toLatin1(), szText.size());
    f.close();
}

void UnVHISTWindow::on_actionExport_as_PDF()
{
    QString szFileName = QFileDialog::getSaveFileName
        (this, "Please specify a filename for the new PDF file:", "");
    if (szFileName.isEmpty()) return;

    if (!szFileName.endsWith(".pdf"))
        szFileName = szFileName + ".pdf";

    log(QString("[Export as PDF] file:\"%1\"").arg(szFileName));
    StopWatch totalWatch(this);
    totalWatch.start("[Export as PDF] total");
    
    if (!m_lastFileName.isEmpty()) {
        openFile(m_lastFileName, true); // optimize for PDF
    } else return;
    
    QPrinter printer(QPrinter::HighResolution);
    printer.setOutputFileName(szFileName);
    m_backgroundBrowser.print(&printer);
}

void UnVHISTWindow::openTemplateDir()
{
//    QUrl qu(QString("file://%1").arg(getTemplatesDir()));
    QUrl qu = QUrl::fromLocalFile (m_szTemplatesDir); //getTemplatesDir ());
    QDesktopServices::openUrl(qu);
}

void UnVHISTWindow::openFile(const QString& filename, bool bGeneratePDF)
{
    log(QString("[open] file: \"%1\"\n").arg(filename));
    
    // init stopwatches
    StopWatch totalWatch(this);
    totalWatch.start("[open] total");
    
    StopWatch partWatch(this);
    
    if ((filename != "") && (m_templateMenu->count() != 0))
    {
        VWaitCursor();
        m_lastFileName = filename;
        
        partWatch.start("parsing VHIST file");
        m_parser.open(filename);
        //qApp->processEvents(); // make sure GUI is not "frozen"
        
        SectionTagList sections = m_parser.getSections();
        FileTagList files = m_parser.getFiles();
        log(QString("found  sections: %1, files: %2\n")
            .arg(sections.count()).arg(files.count()));
        
        if (!bGeneratePDF)
            m_infoBrowser->setParser(&m_parser);
        else
            m_backgroundBrowser.setParser(&m_parser);
        
        //qApp->processEvents(); // make sure GUI is not "frozen"
        partWatch.start("createSummaries");
        SummaryCreator summaryCreator;
        Summaries summaries = summaryCreator.createSummaries(m_parser);
        
        //qApp->processEvents(); // make sure GUI is not "frozen"
        
        if (summaries.count() == 0)
        {
            QMessageBox::warning(NULL, "VHIST", QString("\"%1\"\n"
                    "does not appear to be a VHIST file, sorry.")
                    .arg(filename));
            return;
        }
        else
        {
            partWatch.start("initOutline");
            initOutline(summaries);

            //qApp->processEvents(); // make sure GUI is not "frozen"
            
            partWatch.start("makeHtmlListing");
            QString szTemplateFile = m_szTemplatesDir //getTemplatesDir()
                + QDir::separator() + m_templateMenu->currentText();
            //qApp->processEvents(); // make sure GUI is not "frozen"
            
            ParseTemplate pt(this);
            
            pt.loadTemplateFile(szTemplateFile);
            m_RawHtml = pt.generateHtml(filename,
                                        szTemplateFile,
                                        summaries,
                                        bGeneratePDF);
            //qApp->processEvents(); // make sure GUI is not "frozen"
            
            partWatch.stop();
            
            if (!bGeneratePDF){
                partWatch.start("setHTML");
                m_infoBrowser->setHtml(m_RawHtml);
                //qApp->processEvents(); // make sure GUI is not "frozen"
                
                if (m_splitter->widget(0)->isHidden()) {
                    m_splitter->widget(0)->show();
                    m_splitterLog->widget(1)->show();
                     m_actionShow_log->setChecked (true);
//                     QList<int> ql;
//                     ql << 5 << 1;
//                     m_splitterLog->setSizes(ql);
//                     QListIterator<int> it(m_splitterLog->sizes());
//                     while (it.hasNext())
//                         qDebug() << "splitter size: " << it.next();
                    
                }
                QString title("View VHIST File - %1");
                QFileInfo info(filename);
                setWindowTitle(title.arg(info.fileName()));
                addRecentFile(info.absoluteFilePath());
                
            } else {
                m_backgroundBrowser.setHtml(m_RawHtml);
                m_backgroundBrowser.zoomOut(3);
            }
            
            m_validationQueue.repopulate(&m_parser);
            m_extractionQueue.reset(&m_parser);
            continueValidation();
        }
    }
}

void UnVHISTWindow::initTemplateMenu(void) {
    QDir qd(m_szTemplatesDir); //getTemplatesDir());
    qd.setFilter(QDir::Files);
    QFileInfoList list = qd.entryInfoList();


    for (int i = 0; i < list.size(); ++i) {
        QFileInfo fileInfo = list.at(i);
        m_templateMenu->addItem(fileInfo.fileName());
    }

    if (m_templateMenu->count() == 0){
        QMessageBox::critical(this, "VHIST",
                              QString("No templates have been found in dir:\n"
                                      "\"%1\"").arg(m_szTemplatesDir)); //getTemplatesDir()));
        QCoreApplication::quit();
    }

    m_templatemenuready = true;
}

void UnVHISTWindow::on_templateMenu_currentIndexChanged(const QString &)
{
    if (!m_lastFileName.isEmpty())
        openFile(m_lastFileName, false);
 
    //  Store
    if (m_templatemenuready) { // If we would store each time, we'd store the first entry that gets inserted in initTemplateMenu
        Preferences::setBrowTemplate (m_templateMenu->currentText ());
    }
}

void UnVHISTWindow::anchorClicked (const QUrl & link)
{
    QString szLink = link.toString();
    if (szLink.startsWith(tr("doc://"))) {
        // it is a "pseudo-URL" referring to our documentation
        szLink.remove(0, 6);
        QString szSec;
        int iHash = szLink.indexOf(tr("#"));
        if (iHash > 0) {
            szSec = szLink.mid(iHash + 1, szLink.size() - iHash - 1);
            szLink.remove(iHash, szLink.size() - iHash);
        }

        HelpViewer::globalOpenPage(szLink, szSec);
    }
    else
        QDesktopServices::openUrl(link);
}


void UnVHISTWindow::initOutline(const Summaries& summaries)
{
    m_outlineTree->clear();
    for (int i = 0; i < summaries.count(); i++)
    {
        const Workflowstep step = summaries.at(i).step;
        QString anchor = makeSectionAnchor(step.tagID);
        QTreeWidgetItem* stepItem = makeOutlineItem(step.title,
                SECTION_TAG_TYPE, step.tagID, anchor);
        m_outlineTree->addTopLevelItem(stepItem);
        
        QList<FileData> files;
        for (int j = 0; j < step.files.count(); j++)
        {
            //qDebug() << "outline file:" << i << j;
            const FileSummary& file = step.files.at(j);
            QString anchor = makeFileAnchor(step.tagID, j);

            FileData d;
            d.filename = file.filepath;
            d.id = file.tagID;
            d.anchor = anchor;
            files << d;
            //QTreeWidgetItem* fileItem = makeOutlineItem(file.filename,
            //        FILE_TAG_TYPE, file.tagID, anchor);
            
            //stepItem->addChild(fileItem);
        }
        stepItem->addChild(treeFromFileList(files));
        QIcon stepIcon(":/resources/not-validated.png");
        stepItem->setIcon(0, stepIcon);
        m_outlineTree->expandItem(stepItem);
    }
}

void UnVHISTWindow::outlineDoubleClicked(QTreeWidgetItem* item, int)
{
    qDebug() << item->data(0, TAG_ANCHOR_ROLE).toString();
    QString anchor = item->data(0, TAG_ANCHOR_ROLE).toString();
    m_infoBrowser->scrollToAnchor(anchor);
}

void UnVHISTWindow::extractFiles()
{
    QList<int> toExtract = getSelectedFileIDs();
    QString baseDir(".");
    for (int i = 0; i < toExtract.count(); i++)
    {
        int tagID = toExtract.at(i);
        MaybeFileTag maybe = m_parser.getFile(tagID);
        if (maybe != NULL)
        {
            FileTag& tag = *maybe;
            QString filename = tag.filename;
            QString caption = tr("Extract %1").arg(filename);
            QString fullPath = joinPaths(baseDir, filename);
            QString extractPath = QFileDialog::getSaveFileName(this,
                    caption, fullPath);
            if (extractPath.isEmpty()) break;
            if (extractPath != "")
            {
                QFileInfo finfo(extractPath);
                baseDir = finfo.path();
                m_parser.writeFileToDisk(tagID, extractPath);
            }
        }
    }
}


void UnVHISTWindow::outlineSelectionChanged()
{
    int numSelectedFiles = getSelectedFileIDs().count();
    int numSelectedSections = getSelectedSectionIDs().count();
    int numSelections = numSelectedFiles + numSelectedSections;
    
    // adjust extract options
    bool enabled = numSelectedFiles != 0;
    m_extractAction->setEnabled(enabled);
    m_extractContextAction->setEnabled(enabled);
    
    QString plural = (numSelectedFiles == 1) ? "..." : "s...";
    QString caption = QString("Extract %1 File") + plural;
    caption = caption.arg(numSelectedFiles);
    m_extractAction->setText(caption);
    m_extractContextAction->setText(caption);
    
    // adjust validate options
    enabled = numSelections != 0;
    m_validateAction->setEnabled(enabled);
    m_validateContextAction->setEnabled(enabled);
    
    plural = (numSelections == 1) ? "..." : "s...";
    caption = QString("Validate %1 Item") + plural;
    caption = caption.arg(numSelections);
    m_validateAction->setText(caption);
    m_validateContextAction->setText(caption);
}


QList<int> UnVHISTWindow::getRecursiveIDs(TagType expectedType, QTreeWidgetItem* root)
{
    QList<int> result;

    for (int i = 0;  i < root->childCount(); ++i)
    {
        QTreeWidgetItem* child = root->child(i);
        int type = child->data(0, TAG_TYPE_ROLE).toInt();
        if (type == expectedType)
        {
            int tagID = child->data(0, TAG_ID_ROLE).toInt();
            if (tagID >= 0)
                result.append(tagID);
        }
        else if (type == DIR_TAG_TYPE)
            result += getRecursiveIDs(expectedType, child);
    }
    return result;
}

//QList<int> UnVHISTWindow::get()
//{
//}

QList<int> UnVHISTWindow::getSelectedIDs(TagType expectedType)
{
    QList<QTreeWidgetItem*> items = m_outlineTree->selectedItems();
    QList<int> result;
    
    for (int i = 0; i < items.count(); i++)
    {
        int type = items.at(i)->data(0, TAG_TYPE_ROLE).toInt();
        if (type == expectedType)
        {
            int tagID = items.at(i)->data(0, TAG_ID_ROLE).toInt();
            if (tagID >= 0)
                result.append(tagID);
        }
        else if (type == DIR_TAG_TYPE)
            result += getRecursiveIDs(expectedType, items.at(i));
    }
    return result;
}


QList<int> UnVHISTWindow::getSelectedFileIDs()
{
    return getSelectedIDs(FILE_TAG_TYPE);
}


QList<int> UnVHISTWindow::getSelectedSectionIDs()
{
    return getSelectedIDs(SECTION_TAG_TYPE);
}


QString UnVHISTWindow::getFullpathForTreeNode(const QTreeWidgetItem* item)
{
    if (item == NULL)
        return "";

    int type = item->data(0, TAG_TYPE_ROLE).toInt();
    if (type != FILE_TAG_TYPE && type != DIR_TAG_TYPE)
        return "";
    else
    {
        QString path = getFullpathForTreeNode(item->parent()) + item->text(0);
        if (path.startsWith("<files>"))
            path.remove(0, QString("<files>").length());
        return path;
    }
}


QStringList UnVHISTWindow::getSelectedDirectories()
{
    QList<QTreeWidgetItem*> items = m_outlineTree->selectedItems();
    QStringList result;

    for (int i = 0; i < items.count(); ++i)
    {
        int type = items.at(i)->data(0, TAG_TYPE_ROLE).toInt();
        if (type == DIR_TAG_TYPE)
            result << getFullpathForTreeNode(items.at(i));
    }
    return result;
}


void UnVHISTWindow::showHelp()
{
    //qDebug("UnVHISTWindow::showHelp()");
    HelpViewer::globalOpenPage("unvhist.html", QString());
}


// code related to validation


void UnVHISTWindow::stopCurrentWork()
{
    if (m_extraction != PROGRESS_FINISHED)
        cancelExtraction();
    else if (m_validation != PROGRESS_FINISHED)
        cancelValidation();
}


void UnVHISTWindow::continueValidation()
{
    QTimer::singleShot(0, this, SLOT(vneIncrementally()));
    m_validation = PROGRESS_RUNNING;
    m_progress->setVisible(true);
    m_stopButton->setVisible(true);
    m_currentJobLabel->setVisible(true);
    if (!m_currentJobLabel->text().startsWith("Validating"))
        m_currentJobLabel->setText("Validating...");
}


void UnVHISTWindow::finishValidation()
{
    m_validation = PROGRESS_FINISHED;
    bool visible = m_extraction != PROGRESS_FINISHED;
    m_progress->setVisible(visible);
    m_stopButton->setVisible(visible);
    m_currentJobLabel->setVisible(visible);
}


void UnVHISTWindow::pauseValidation()
{
    m_validation = PROGRESS_PAUSED;
}


void UnVHISTWindow::cancelValidation()
{
    finishValidation();
    m_validationQueue.cancelValidation();
}


/*void UnVHISTWindow::toggleValidation()
{
    qDebug() << m_validation;
    if (m_validation == PROGRESS_RUNNING)
        pauseValidation();
    else if (m_validation == PROGRESS_PAUSED)
        continueValidation();
}*/


void UnVHISTWindow::vneIncrementally()
{
    if (m_extraction == PROGRESS_RUNNING)
    {
        if (m_extractionQueue.process())
        {
            continueExtraction();
        }
        else
            finishExtraction();
    }
    else if (m_validation == PROGRESS_RUNNING)
    {
        if (m_validationQueue.process())
        {
            continueValidation();
        }
        else
            finishValidation();
    }
}


void UnVHISTWindow::validateSelected()
{
    QList<int> files = getSelectedFileIDs();
    QList<int> sections = getSelectedSectionIDs();
    
    m_validationQueue.addValidatees(files, sections);
    continueValidation();
}


QTreeWidgetItem* UnVHISTWindow::findOutlineObjectWithID(int id, QTreeWidgetItem* root)
{
    for (int i = 0; i < root->childCount(); ++i)
    {
        QTreeWidgetItem* child = root->child(i);
        if (child->data(0, TAG_ID_ROLE).toInt() == id)
            return child;
        else
        {
            QTreeWidgetItem* grandChild = findOutlineObjectWithID(id, child);
            if (grandChild != NULL)
                return grandChild;
        }
    }
    return NULL;
}


void UnVHISTWindow::fileValidated(int id, bool success)
{
    for (int i = 0; i < m_outlineTree->topLevelItemCount(); i++)
    {
        QTreeWidgetItem* pitem = m_outlineTree->topLevelItem(i);
        QTreeWidgetItem* subitem = findOutlineObjectWithID(id, pitem);
        if (subitem != NULL)
        {
            QString iconName;
            if (success)
                iconName = ":/resources/file-valid.png";
            else
                iconName = ":/resources/file-invalid.png";
            QIcon icon(iconName);
            subitem->setIcon(0, icon);
        }
    }
}

void UnVHISTWindow::sectionValidated(int id, bool success)
{
    for (int i = 0; i < m_outlineTree->topLevelItemCount(); i++)
    {
        QTreeWidgetItem* item = m_outlineTree->topLevelItem(i);
        int itemID = item->data(0, TAG_ID_ROLE).toInt();
        QString iconName;
        if (itemID == id)
        {
            if (success)
                iconName = ":/resources/section-valid.png";
            else
                iconName = ":/resources/section-invalid.png";
            QIcon icon(iconName);
            item->setIcon(0, icon);
        }
    }
}

void UnVHISTWindow::createLogContextMenu(const QPoint& /*pos*/)
{
    m_logContextMenu->exec (QCursor::pos ());
}

void UnVHISTWindow::createLogContextMenu()
{
    QAction* ac;

    m_logContextMenu = new QMenu (this);
    
    ac = m_logContextMenu->addAction ("Copy");
    connect (ac, SIGNAL(triggered(bool)), m_log->textBrowser(), SLOT(copy()));
    ac = m_logContextMenu->addAction ("Set Font ...");
    connect (ac, SIGNAL(triggered(bool)), this, SLOT(setLogFont()));
    ac = m_logContextMenu->addAction ("Increase Font Size");
    connect (ac, SIGNAL(triggered(bool)), this, SLOT(increaseLogFontSize()));
    ac = m_logContextMenu->addAction ("Decrease Font Size");
    connect (ac, SIGNAL(triggered(bool)), this, SLOT(decreaseLogFontSize()));
}

void UnVHISTWindow::setLogFont ()
{
    bool ok;

    QFont chosenFont = QFontDialog::getFont (&ok, m_log->textBrowser()->font (), this);

    if (ok) { // The user has chosen a new font
        m_log->textBrowser()->setFont (chosenFont);

        Preferences::setLogFontFamily (chosenFont.family ());
        Preferences::setLogFontPointSize (chosenFont.pointSizeF ());
    }
}

void UnVHISTWindow::increaseLogFontSize ()
{
    QFont font = m_log->textBrowser()->font ();
    font.setPointSizeF (font.pointSizeF () * 1.1);
    m_log->textBrowser()->setFont (font);

    Preferences::setLogFontPointSize (font.pointSizeF ());
}

void UnVHISTWindow::decreaseLogFontSize ()
{
    QFont font = m_log->textBrowser()->font ();
    font.setPointSizeF (font.pointSizeF () * 0.9);
    m_log->textBrowser()->setFont (font);

    Preferences::setLogFontPointSize (font.pointSizeF ());
}

void UnVHISTWindow::createBrowserContextMenu ()
{
    QAction* ac;

    QMenu* browsercontextmenu = new QMenu (m_infoBrowser);

    ac = browsercontextmenu->addAction ("Find");
    connect(ac, SIGNAL(triggered(bool)), m_infoViewer, SLOT(showFindBar()));
    ac = browsercontextmenu->addAction ("Find Next");
    connect(ac, SIGNAL(triggered(bool)), m_infoViewer, SLOT(findNext()));
    ac = browsercontextmenu->addAction ("Find Previous");
    connect(ac, SIGNAL(triggered(bool)), m_infoViewer, SLOT(findPrev()));
    browsercontextmenu->addSeparator ();
    ac = browsercontextmenu->addAction ("Increase Font Size");
    connect(ac, SIGNAL(triggered(bool)), m_infoBrowser, SLOT(zoomIn()));
    ac = browsercontextmenu->addAction ("Decrease Font Size");
    connect(ac, SIGNAL(triggered(bool)), m_infoBrowser, SLOT(zoomOut()));
    browsercontextmenu->addSeparator ();
    //ac = browsercontextmenu->addAction ("Validate");
    //connect(ac, SIGNAL(triggered(bool)), this, SLOT(toggleValidation()));

    m_infoBrowser->setContextMenu (browsercontextmenu);
}


void UnVHISTWindow::createOutlineContextMenu()
{
    QMenu* outlineContextMenu = new QMenu(m_outlineTree);
    m_extractContextAction = outlineContextMenu->addAction("Extract...");
    connect(m_extractContextAction, SIGNAL(triggered(bool)),
            m_extractAction, SLOT(trigger()));
    m_validateContextAction = outlineContextMenu->addAction("Validate selected items");
    connect(m_validateContextAction, SIGNAL(triggered(bool)),
            m_validateAction, SLOT(trigger()));
    
    
    m_outlineTree->setContextMenu(outlineContextMenu);
}


void UnVHISTWindow::showLog ()
{
    if (m_splitterLog->widget (1)->isVisible ()) {
        m_splitterLog->widget (1)->hide ();
        m_actionShow_log->setChecked (false);
    } else {
        m_splitterLog->widget (1)->show ();
        m_actionShow_log->setChecked (true);
    }
}

void UnVHISTWindow::createPreferencesWindow ()
{
    m_preferenceswindow = new UnVHISTPreferencesWindow (this);
}

void UnVHISTWindow::showPreferences ()
{
    // Store here again
    Preferences::setBrowZoomLevel (m_infoBrowser->zoom ());

    if (m_preferenceswindow->exec () == QDialog::Accepted) {
        loadSettings ();
    }
}

void UnVHISTWindow::loadSettings ()
{
    // Configure log font
    QString fontfamily = Preferences::logFontFamily ();
    qreal fontsize = Preferences::logFontPointSize ();

    QTextCharFormat format;
    format.setFontFamily (fontfamily);
    format.setFontPointSize (fontsize);
    m_log->textBrowser()->setFont (format.font ());
    
    // Configure zoom level
    int zoomlevel = Preferences::browZoomLevel ();

    m_infoBrowser->setZoom (zoomlevel);

    // Reload template
    QString lasttemplate = Preferences::browTemplate ();

    int lastindex = -1;

    for (int i = 0;i < m_templateMenu->count ();i++) {
        if (m_templateMenu->itemText (i) == lasttemplate) {
            lastindex = i;
        }
    }

    if (lastindex != -1) {
        m_templateMenu->setCurrentIndex (lastindex);
    }
}





// code related to extraction of files

void UnVHISTWindow::continueExtraction()
{
    QTimer::singleShot(0, this, SLOT(vneIncrementally()));
    m_extraction = PROGRESS_RUNNING;
    m_progress->setVisible(true);
    m_stopButton->setVisible(true);
    m_currentJobLabel->setVisible(true);
    if (!m_currentJobLabel->text().startsWith("Extracting"))
        m_currentJobLabel->setText("Extracting...");
}


void UnVHISTWindow::finishExtraction()
{
    m_extraction = PROGRESS_FINISHED;
    bool visible = m_validation != PROGRESS_FINISHED;
    m_progress->setVisible(visible);
    m_stopButton->setVisible(visible);
    m_currentJobLabel->setVisible(visible);
}


void UnVHISTWindow::pauseExtraction()
{
    m_extraction = PROGRESS_PAUSED;
}


void UnVHISTWindow::cancelExtraction()
{
    finishExtraction();
    m_extractionQueue.cancelExtraction();
}


QString UnVHISTWindow::filepathForTagID(Summaries summaries, int tagID)
{
    for (int i = 0; i < summaries.length(); ++i)
    {
        const Summary& cSum = summaries.at(i);
        const QList<FileSummary>& files = cSum.step.files;
        for (int j = 0; j < files.length(); ++j)
        {
            const FileSummary& cFile = files.at(j);
            if (cFile.tagID == tagID)
                return cFile.filepath;
        }
    }
    return "";
}


QString UnVHISTWindow::fixPath(const QString& filename)
{
    QStringList parts = splitPath(filename);

    for (int i = 0; i < parts.length(); ++i)
    {
        if (parts.at(i).startsWith("."))
            parts[i] = "_" + parts.at(i);
        parts[i].replace(":", "_");
        parts[i].replace("\\\\", "/");
        parts[i].replace("\\", "/");
    }
    QString s = parts.join("");
    return QDir::toNativeSeparators(s);
}


void UnVHISTWindow::extractSelected()
{
    // key is id, value is filename
    QHash<int, QString> files;
    QList<int> toExtract = getSelectedFileIDs();
    QList<SavedFile> toValidate;

    // parse summaries to get all filenames including paths
    SummaryCreator creator;
    Summaries summaries = creator.createSummaries(m_parser);

    // get filenames (with paths) of files to extract and find common prefix
    QStringList filenames;
    for (int i = 0; i < toExtract.length(); ++i)
        filenames << filepathForTagID(summaries, toExtract.at(i));

    // strip prefix from all files
    QStringList prefixes = getSelectedDirectories();
    qSort(prefixes);
    //QString prefix = getCommonPrefix(filenames);


    // strip the shortest prefix possible
    for (int i = 0; i < toExtract.length(); ++i)
    {
        bool prefixFound = false;
        for (int j = 0; j < prefixes.length() && !prefixFound; ++j)
        {
            QString prefix = prefixes.at(j);
            // TODO! handle special cases ("..", ".", probably everything with "." at the beginning)
            // handle in full paths
            // 1) split into individual entries
            // 2) fix special entries
            // 3) join elements again
            QString prefixDir = lastPartOfPath(prefix);
            if (prefixDir != "" && filenames[i].startsWith(prefix))
            {
                filenames[i] = joinPaths(prefixDir, filenames[i].mid(prefix.length()));
                prefixFound = true;
            }
        }

        // no directory prefix found => not part of a directory tree
        if (!prefixFound)
            filenames[i] = lastPartOfPath(filenames.at(i));
        filenames[i] = QDir::toNativeSeparators(filenames.at(i));
    }

    for (int i = 0; i < filenames.length(); ++i)
    {
        filenames[i] = fixPath(filenames.at(i));
    }

    // get directories of all files to extract
    QStringList filepaths;
    for (int i = 0; i < filenames.length(); ++i)
    {
        QFileInfo currentFilename(filenames.at(i));
        filepaths << currentFilename.path();
    }


    ExtractDialog* dialog = new ExtractDialog(this, ".", toExtract.count());
    int success = dialog->exec();
    bool askBeforeOverwrite = dialog->getAskBeforeOverwrite();
    bool validate = dialog->getValidateFiles();
    
    if (success == ExtractDialog::Accepted)
    {
        QString baseDir = dialog->getDirectory();
        for (int i = 0; i < toExtract.count(); i++)
        {
            int tagID = toExtract.at(i);
            
            MaybeFileTag maybe = m_parser.getFile(tagID);
            if (maybe != NULL)
            {
                FileTag& tag = *maybe;
                
                QString filename = tag.filename;
                filename = dialog->adjustFilename(filename, toExtract.count(),
                        i + 1);
                QString midPath = makePathValid(filepaths.at(i));
                QString fullPath = joinPaths(baseDir, midPath);
                QDir d(baseDir);
                d.mkpath(midPath);
                QString fullFilename = joinPaths(fullPath, filename);
                QFileInfo finfo(fullFilename);
                if (askBeforeOverwrite && finfo.exists() && finfo.isFile())
                {
                    QString msg = tr("The file %1 does already exist. Do you want to overwrite it?").arg(fullFilename);
                    int b = QMessageBox::question(this, "Overwrite File?", msg, QMessageBox::Yes, QMessageBox::No);
                    if (b == QMessageBox::No)
                        break;
                }
                
                if (finfo.exists() && finfo.isDir())
                {
                    QString msg = tr("A directory with the name %1 does already exist").arg(fullFilename);
                    QMessageBox::critical(this, "Could not write File", msg, QMessageBox::Ok);
                    break;
                }
                
                files.insert(tagID, fullFilename);
                
                if (validate)
                {
                    SavedFile sfile;
                    sfile.filename = fullPath;
                    sfile.filesize = tag.filesize;
                    sfile.md5 = tag.md5file;
                    toValidate.append(sfile);
                }
            }
            m_extractionQueue.addExtractees(files);
            m_validationQueue.addValidatees(toValidate);
            continueExtraction();
            continueValidation();
        }
    }
    delete dialog;
}

void UnVHISTWindow::setJobLabel(const QString& text)
{
    //int w = m_currentJobLabel->width();
    //QString elidedText = m_currentJobLabel->fontMetrics().elidedText(text,
    //        Qt::ElideMiddle, w);
    m_currentJobLabel->setText(text);
}

void UnVHISTWindow::updateProgressBar(int percent, const QString name)
{
    //qDebug() << "foo" << name;
    m_progress->setValue(percent);
    if (m_currentJobLabel->text().startsWith("Validating"))
        setJobLabel(QString("Validating %1...").arg(name));
    else
        setJobLabel(QString("Extracting %1...").arg(name));
}


void UnVHISTWindow::showFindBar()
{
    FindTextBrowser::showFindBarForActiveWidget(m_infoViewer);
}


void UnVHISTWindow::findNext()
{
    FindTextBrowser::findNextForActiveWidget(m_infoViewer);
}


void UnVHISTWindow::findPrev()
{
    FindTextBrowser::findPrevForActiveWidget(m_infoViewer);
}


/*void UnVHISTWindow::toggleExtraction()
{
    if (m_extraction == PROGRESS_RUNNING)
        pauseExtraction();
    else if (m_extraction == PROGRESS_PAUSED)
        continueExtraction();
}*/


ManagedWindow* UnVHISTWindow::createUnVHISTWindow(
        const QString& filename,
        const QVariantHash& parameters)
{
    QString szTemplatesDir;
    if (parameters.contains("TemplateDir"))
        szTemplatesDir=parameters.value("TemplateDir",QVariant()).toString();
    ManagedWindow* win = new UnVHISTWindow(filename,szTemplatesDir);
    return win;
}

QString UnVHISTWindow::windowName()
{
    return "View VHIST File";
}

QString UnVHISTWindow::windowShortcut()
{
    return "ctrl+T";
}
