/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: unvhistwindow.h 2509 2012-08-10 09:37:01Z ahuesgen $
 */

#ifndef UNVHISTWINDOW_INCLUDED
#define UNVHISTWINDOW_INCLUDED

#include "ui_unvhist.h"
#include <QMultiHash>
#include <QPair>
#include <QString>
#include <QTextCursor>

#include "../shared/managedwindow.h"

#include <unvhistlib/summary.h>
#include <unvhistlib/summaryreader.h>
#include <unvhistlib/vhistparser.h>
#include <unvhistlib/validationqueue.h>
#include <unvhistlib/extractionqueue.h>
#include "maketree.h"
#include "vhistbrowser.h"

#include "unvhistpreferenceswindow.h"


enum ProgressStatus
{
    PROGRESS_RUNNING = 1,
    PROGRESS_PAUSED = 2,
    PROGRESS_FINISHED = 3
};


// first: id, second: distance
typedef QPair<int, int> FileDistance;

class QDragEnterEvent;
class QDropEvent;
class QKeyEvent;
class QCloseEvent;

class QTreeWidgetItem;

class UnVHISTWindow: public ManagedWindow, public Ui::UnVHISTWindow
{
    Q_OBJECT
    public: // construction, destruction and assignment
        UnVHISTWindow(const QString& filename, QString szTemplatesDir = QString(""));
        ~UnVHISTWindow();
    
    public slots:
        void selectVHISTFile();
        void setVHISTFile(const QString& filename);
    
    protected: // qt internas
        virtual void dragEnterEvent(QDragEnterEvent* event);
        virtual void dropEvent(QDropEvent* event);
        virtual void keyPressEvent(QKeyEvent* event);
        virtual void closeEvent(QCloseEvent* event);
    
    private:
        void initSignalSlots();
        void initMainWidgets();
        void initShortcuts();
        void initIcons();
        void initOutline(const vhist::Summaries& summaries);
        void initTemplateMenu();
        void initHelloMessage();
        void initWhatsThis();
        
        QString getFullpathForTreeNode(const QTreeWidgetItem* item);
        QList<int> getRecursiveIDs();
        QList<int> getSelectedFileIDs();
        QList<int> getSelectedSectionIDs();
        QStringList getSelectedDirectories();
        QList<int> getSelectedIDs(TagType expectedType);
        QList<int> getRecursiveIDs(TagType expectedType, QTreeWidgetItem* root);
        
        void createLogContextMenu();
        void createBrowserContextMenu();
        void createOutlineContextMenu();
        void createPreferencesWindow();
        void loadSettings();
    
    public slots:
        void log(const QString szT);
    
    private slots:
        void openFile(const QString& filename, bool bGeneratePDF);
        void openTemplateDir();
        void on_actionExport_as_PDF();
        void on_actionExport_as_HTML();
        void on_actionExport_as_Raw_HTML();
        void on_actionExport_as_Plain_Text();        
        void on_templateMenu_currentIndexChanged(const QString &szSelected);
        void outlineDoubleClicked(QTreeWidgetItem* item, int column);
        void anchorClicked (const QUrl & link);
        void extractFiles();
        void outlineSelectionChanged();
        void showHelp();

        void showFindBar();
        void findNext();
        void findPrev();

        void vneIncrementally();
        void stopCurrentWork();
        QTreeWidgetItem* findOutlineObjectWithID(int id, QTreeWidgetItem* root);
        void fileValidated(int id, bool success);
        void sectionValidated(int id, bool success);
        void createLogContextMenu(const QPoint&); 
        void setLogFont ();
        void increaseLogFontSize (); 
        void decreaseLogFontSize ();
        void showLog (); 
        void showPreferences();
        void updateProgressBar(int percent, const QString name);
        void setJobLabel(const QString& text);
    
    private slots:
        void validateSelected();
        QString filepathForTagID(vhist::Summaries summaries, int tagID);
        QString fixPath(const QString& filename);
        void extractSelected();
    
    private:
        void continueValidation();
        void finishValidation();
        void pauseValidation();
        void cancelValidation();
        
    private:
        void continueExtraction();
        void finishExtraction();
        void pauseExtraction();
        void cancelExtraction();
        
    private: // member variables
        vhist::VhistParser m_parser;
        ProgressStatus m_validation;
        ProgressStatus m_extraction;
        QString m_lastFileName;
        QString m_RawHtml;
        VhistBrowser m_backgroundBrowser;
        vhist::ValidationQueue m_validationQueue;
        vhist::ExtractionQueue m_extractionQueue;
        QMenu* m_logContextMenu;
        UnVHISTPreferencesWindow* m_preferenceswindow;
        QAction* m_extractContextAction;
        QAction* m_validateContextAction;
        bool m_templatemenuready;
        QString m_szTemplatesDir;
        VhistBrowser* m_infoBrowser;

    public:
        static ManagedWindow* createUnVHISTWindow(const QString& filename,
            const QVariantHash& parameters);
        static QString windowName();
        static QString windowShortcut();
};

#endif // UNVHISTWINDOW_INCLUDED

