/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: vhistbrowser.cpp 2457 2011-07-29 09:24:21Z ahuesgen $
 */

#include <QtDebug>
#include <QDir>
#include <QTextBlock>
#include <QKeyEvent>
#include <QMenu>

#include "../shared/utilities.h"
#include "vhistbrowser.h"
#include <unvhistlib/vhistparser.h>

using namespace vhist;

VhistBrowser::VhistBrowser(QWidget* parent)
    : QTextBrowser(parent), m_findEnabled(false)
{
	m_contextmenu = NULL;
	m_zoomLevel = 0;
}


VhistBrowser::~VhistBrowser()
{
}

int VhistBrowser::getIDFromUrl(const QUrl& name)
{
    QString str = name.toString();
    if (!str.startsWith("vhisttag://"))
        return -1;
    
    str = str.mid(QString("vhisttag://").length());
    
    bool ok = false;
    int id = str.toInt(&ok);
    if (!ok)
        id = -1;
    return id;
}


QVariant VhistBrowser::loadResource(int type, const QUrl& name)
{
    //qDebug() << type << name;
    if (type == QTextDocument::ImageResource && m_parser != NULL)
    {
        if (name.toString().startsWith("vhisttag://")) {
            int tagID = getIDFromUrl(name);
            MaybeBytes maybe = m_parser->getFileContent(tagID);
            if (maybe != NULL)
            {
                return QVariant(*maybe);
            }
            else
                qDebug() << "Could not load file with tag id" << tagID;
        }
        else if (name.toString().startsWith("template://")) {
            QString szTemplatePath = getTemplatesDir() + "/"; // even on Windows
            QUrl rewrittenName(name.toString()
                               .replace("template://",
                                        szTemplatePath));
            return QTextBrowser::loadResource(type, rewrittenName);
                               
        }
    }
    return QTextBrowser::loadResource(type, name);
}


void VhistBrowser::setParser(VhistParser* parser)
{
    m_parser = parser;
}


void VhistBrowser::enableFind(bool enabled)
{
    m_findEnabled = enabled;
    QString sheet("QTextEdit {selection-background-color: "
            "rgb(255, 210, 108)}");
    if (enabled)
        setStyleSheet(sheet);
    else
        setStyleSheet("");
}


bool VhistBrowser::findEnabled()
{
    return m_findEnabled;
}


int VhistBrowser::zoom ()
{
	return (m_zoomLevel);
}


void VhistBrowser::setZoom (int nz)
{
	if (m_zoomLevel < nz) { // We need to zoom in
		zoomIn (nz - m_zoomLevel);
	} else {
		zoomOut (m_zoomLevel - nz);
	}
}


void VhistBrowser::findStringChanged(const QString& findString)
{
    m_findString = findString;
    QTextCursor oldCursor = textCursor();
    int position = qMin(oldCursor.anchor(), oldCursor.position());
    findNext(position);
}


void VhistBrowser::findNext(int position)
{
    if (!m_findEnabled)
        return;
    
    QTextCursor cursor = textCursor();
    
    if (position == -1)
        position = qMax(0, qMax(cursor.anchor(), cursor.position()));
    
    QTextCursor c = document()->find(m_findString, position);
    
    // if nothing was found, start searching at the beginning of the document
    if (c.selectedText() == "")
        c = document()->find(m_findString, 0);
    
    if (c.isNull())
    {
        c = textCursor();
        c.setPosition(qMin(c.anchor(), c.position()));
    }

    setTextCursor(c);
}


void VhistBrowser::findPrev(int position)
{
    if (!m_findEnabled)
        return;
    
    QTextCursor cursor = textCursor();
    
    if (position == -1)
        position = qMax(0, qMin(cursor.anchor(), cursor.position()));
    
    QTextCursor c = document()->find(m_findString,
            position, QTextDocument::FindBackward);
    
    // if nothing was found, start searching at the end of the document
    if (c.selectedText() == "")
    {
        QTextBlock block = document()->lastBlock(); 
        int endPos = block.position() + block.length();
        c = document()->find(m_findString, endPos,
                QTextDocument::FindBackward);
    }
    
    if (c.isNull())
    {
        c = textCursor();
        c.setPosition(qMin(c.anchor(), c.position()));
    }
    
    setTextCursor(c);
}


void VhistBrowser::zoomIn (int range)
{
	// Note that this makes the assumption
	// that zoomIn (X); and zoomOut (X);
	// cancel each other out
	if ((m_zoomLevel + range) <= 3) {
		m_zoomLevel += range;

		QTextBrowser::zoomIn (range);
	} else {
		QTextBrowser::zoomIn (3-m_zoomLevel);
		
		m_zoomLevel = 3;
	}
}


void VhistBrowser::zoomOut (int range)
{
	if ((m_zoomLevel - range) >= -3) {
		m_zoomLevel -= range;

		QTextBrowser::zoomOut (range);
	} else {
		QTextBrowser::zoomOut (-3-m_zoomLevel);
		
		m_zoomLevel = -3;		
	}
}


void VhistBrowser::keyPressEvent(QKeyEvent* event)
{
    if (event->modifiers() & Qt::ControlModifier)
    {
        if (event->key() == Qt::Key_Plus)
        {
            zoomIn();
            event->accept();
            return;
        }
        
        if (event->key() == Qt::Key_Minus)
        {
            zoomOut();
            event->accept();
            return;
        }
    }
    QTextBrowser::keyPressEvent(event);
}

void VhistBrowser::contextMenuEvent(QContextMenuEvent* event)
{
	if (m_contextmenu != NULL) { // Make sure that we don't hit this situation
		m_contextmenu->exec (QCursor::pos ());
	} else { // Failover, show default menu of the QTextBrowser
		QTextBrowser::contextMenuEvent (event);
	}
}

void VhistBrowser::wheelEvent(QWheelEvent* event)
{
	if (event->modifiers() & Qt::ControlModifier) {
		if (event->delta () > 0) {
			zoomOut ();
		} else {
			zoomIn ();
		}
	} else {
		QTextEdit::wheelEvent(event); // Not doing this breaks scrolling
	}
}

void VhistBrowser::setContextMenu(QMenu* m)
{
	m_contextmenu = m; 
}
