/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 Author: Michael May
 $Id: guiwindow.cpp 2740 2013-04-24 08:54:27Z michael $
 */

#define VERSION "vhistez 1.03"

#include "guiwindow.h"
#include "mimetypes.h"
#include "../shared/managedwindow.h"
#include "../shared/windowlist.h"
#include "../shared/argfilehighlighter.h"
#include "../vhistit/vhistitwindow.h"

#include <arglib/argfilereader.h>

#include <QMenu>
#include <QTreeWidgetItem>
#include <QDir>

#include <QComboBox>
#include <QFileDialog>
#include <QClipboard>
#include <QLineEdit>
#include <QString>
#include <QListView>
#include <QDragMoveEvent>
#include <QDropEvent>
#include <QUrl>
#include <QDateTime>
#include <QTextStream>
#include <QLibraryInfo>
#include <QMessageBox>
#include <QShowEvent>
#include <QMimeData>

#include <QtDebug>
#include <iostream>

using namespace std;

// Constants
const QString GuiWindow::m_userDefinedFileType = "other (User Defined)";
const QString GuiWindow::m_string_InputTextAtFileListWidget = "[Input File] ";
const QString GuiWindow::m_string_OutputTextAtFileListWidget = "[Output File] ";

GuiWindow::GuiWindow(QString szApplicationDir, ManagedWindow* parent)
{
	// Init model
	m_argfile = new vhist::ArgFile (); 

	// Proceed to the rest
	m_szApplicationDir = szApplicationDir;
    m_parent = parent;
	m_szDocDir = m_szApplicationDir + "/../doc";
    // check if there is a vhist subdirectory

    if (QDir(m_szDocDir + "/vhist").exists())
        m_szDocDir = m_szDocDir + "/vhist";

    QDir szTempDocDir(m_szDocDir);
    //QMessageBox::about(NULL, "test", szAdpDir.canonicalPath());
    m_szDocDir = szTempDocDir.canonicalPath();

    //qDebug() << "DocDir: " << m_szDocDir;
	
	// m_preferencesDialog= QDialog(this,0);
	setupUi(this);
        initCommandline();
	setLineEditTextToBlank();
	setUpMimeTypeList();
	setUpVariables_Start();
	setDefaults();
	hideAllButtons();
	setAcceptDrops(true);
   
	initializeWorklfowKeyValuePairs();
	evaluateRadioButtons();

	// Context menus
	m_maintabcm = new MainTabContextMenu (this);
	m_inouttabcm = new InOutTabContextMenu (this);
	m_vhistfilestabcm = new VhistFilesTabContextMenu (this);
	m_exporttabcm = new ExportTabContextMenu (this);

	// Mapping helper
	m_fihelper = new FileItemHelper (this);

	// Now all is ready to connect slots
	connectSlots();
}

GuiWindow::~GuiWindow ()
{
	delete m_exporttabcm;
	delete m_vhistfilestabcm;
	delete m_inouttabcm;
	delete m_maintabcm;
	delete m_argfile;
}

void GuiWindow::connectSlots()
{
	// Slots for "Main" context menu	
	QWidget* maintab = tabWidget->widget (0);
	maintab->setContextMenuPolicy (Qt::CustomContextMenu);
	connect (maintab, SIGNAL (customContextMenuRequested (const QPoint&)), m_maintabcm, SLOT (exec ()));

	// From the menu
	connect (m_maintabcm, SIGNAL(toolBrowserChosen()), this, SLOT(on_toolBrowserButton_clicked()));
	connect (m_maintabcm, SIGNAL(normalChosen()), normalRadioButton, SLOT(click()));
	connect (m_maintabcm, SIGNAL(quietChosen()), quietRadioButton, SLOT(click()));
	connect (m_maintabcm, SIGNAL(verboseChosen()), verboseRadioButton, SLOT(click()));
	connect (m_maintabcm, SIGNAL(pretendChosen()), pretendCheckBox, SLOT(click()));

	// To the menu
	connect (normalRadioButton, SIGNAL(clicked(bool)), m_maintabcm, SLOT(doNormal ()));
	connect (quietRadioButton, SIGNAL(clicked(bool)), m_maintabcm, SLOT(doQuiet ()));
	connect (verboseRadioButton, SIGNAL(clicked(bool)), m_maintabcm, SLOT(doVerbose ()));
	connect (pretendCheckBox, SIGNAL(clicked(bool)), m_maintabcm, SLOT(doPretend ()));

	// Slots for "In/Out" context menu
	QWidget* inouttab = tabWidget->widget (1);
	inouttab->setContextMenuPolicy (Qt::CustomContextMenu);
	connect (inouttab, SIGNAL (customContextMenuRequested (const QPoint&)), m_inouttabcm, SLOT (exec ()));

	// From the menu
	connect (m_inouttabcm, SIGNAL(fileBrowserChosen()),this, SLOT(on_fileBrowserButton_clicked()));
	connect (m_inouttabcm, SIGNAL(addFileChosen()),this, SLOT(on_newFileButton_clicked()));
	connect (m_inouttabcm, SIGNAL(removeFileChosen()), this, SLOT(on_removeFileButton_clicked()));
	connect (m_inouttabcm, SIGNAL(duplicateFileChosen()), this, SLOT(on_duplicateFileButton_clicked()));
	connect (m_inouttabcm, SIGNAL(removeAllFilesChosen()), this, SLOT(on_removeAllFilesButton_clicked()));
	connect (m_inouttabcm, SIGNAL(toOutputChosen()), outputRadioButton, SLOT(click()));
	connect (m_inouttabcm, SIGNAL(toInputChosen()), inputRadioButton, SLOT(click()));
	connect (m_inouttabcm, SIGNAL(calculateChosen()), md5CheckBox, SLOT(click()));
	connect (m_inouttabcm, SIGNAL(embedChosen()), embedFileCheckBox, SLOT(click()));
	connect (m_inouttabcm, SIGNAL(optionalChosen()), optionalCheckBox, SLOT(click()));
	connect (m_inouttabcm, SIGNAL(previewWSChosen()), previewWSCheckBox, SLOT(click()));
	connect (m_inouttabcm, SIGNAL(previewChosen()), previewCheckBox, SLOT(click()));
	connect (m_inouttabcm, SIGNAL(thumbnailChosen()), thumbnailCheckBox, SLOT(click()));

	// To the menu 
	connect (inputRadioButton, SIGNAL(clicked(bool)), m_inouttabcm, SLOT(doInput()));
	connect (outputRadioButton, SIGNAL(clicked(bool)), m_inouttabcm, SLOT(doOutput()));
	connect (md5CheckBox, SIGNAL(clicked(bool)), m_inouttabcm, SLOT(doCalculate ()));
	connect (embedFileCheckBox, SIGNAL(clicked(bool)), m_inouttabcm, SLOT(doEmbed ()));
	connect (optionalCheckBox, SIGNAL(clicked(bool)), m_inouttabcm, SLOT(doOptional ()));
	connect (previewWSCheckBox, SIGNAL(clicked(bool)), m_inouttabcm, SLOT(doPreviewWS ()));
	connect (previewCheckBox, SIGNAL(clicked(bool)), m_inouttabcm, SLOT(doPreview ()));
	connect (thumbnailCheckBox, SIGNAL(clicked(bool)), m_inouttabcm, SLOT(doThumbnail ()));

	// Slots for the "VhistFiles" context menu
	QWidget* filestab = tabWidget->widget (2);
	filestab->setContextMenuPolicy (Qt::CustomContextMenu);
	connect (filestab, SIGNAL (customContextMenuRequested (const QPoint&)), m_vhistfilestabcm, SLOT (exec ()));

	// From the menu
	connect(m_vhistfilestabcm, SIGNAL(appendChosen()), vhistAppendCheckBox, SLOT(click()));
	connect(m_vhistfilestabcm, SIGNAL(noneChosen()), noneCheckBox, SLOT(click()));
	connect(m_vhistfilestabcm, SIGNAL(optionalChosen()), vhistRootOptionalCheckBox, SLOT(click())); 
	connect(m_vhistfilestabcm, SIGNAL(firstInputChosen()), firstInputFileVhistRootFileRadioButton, SLOT(click()));
	connect(m_vhistfilestabcm, SIGNAL(userSpecChosen()), userspecifiedPathAndNameRadioButton, SLOT(click()));
	connect(m_vhistfilestabcm, SIGNAL(firstOutputChosen()), firstOutputFileVhist, SLOT(click()));
	connect(m_vhistfilestabcm, SIGNAL(oneForEachChosen()), one_EachOutputFileRadioButton, SLOT(click()));
	connect(m_vhistfilestabcm, SIGNAL(outputUserSpecChosen()), oneUserSpecificVhistOutFileRadioButton, SLOT(click())); 
	
	// To the menu
	connect (vhistAppendCheckBox, SIGNAL(clicked (bool)), m_vhistfilestabcm, SLOT(doAppend ()));
	connect (noneCheckBox, SIGNAL(clicked (bool)), m_vhistfilestabcm, SLOT(doNone ()));
	connect (vhistRootOptionalCheckBox, SIGNAL(clicked (bool)), m_vhistfilestabcm, SLOT(doOptional ()));
	connect (firstInputFileVhistRootFileRadioButton, SIGNAL(clicked (bool)), m_vhistfilestabcm, SLOT(doFirstInput ()));
	connect (userspecifiedPathAndNameRadioButton, SIGNAL(clicked (bool)), m_vhistfilestabcm, SLOT(doUserSpec ()));
	connect (firstOutputFileVhist, SIGNAL(clicked (bool)), m_vhistfilestabcm, SLOT(doFirstOutput ()));
	connect (one_EachOutputFileRadioButton, SIGNAL(clicked (bool)), m_vhistfilestabcm, SLOT(doOneForEach ()));
	connect (oneUserSpecificVhistOutFileRadioButton, SIGNAL(clicked (bool)), m_vhistfilestabcm, SLOT(doOutputUserSpec ()));

	// Slots for the "Export" context menu
	QWidget* exporttab = tabWidget->widget (3);
	exporttab->setContextMenuPolicy (Qt::CustomContextMenu);
	connect (exporttab, SIGNAL (customContextMenuRequested (const QPoint&)), m_exporttabcm, SLOT (exec ()));

	connect(m_exporttabcm, SIGNAL(copyChosen()), this, SLOT(on_clipBoardButton_clicked()));
	connect(m_exporttabcm, SIGNAL(saveChosen()), this, SLOT(on_saveAsCmdFileButton_clicked()));
	connect(m_exporttabcm, SIGNAL(saveAndOpenChosen()), this, SLOT(on_saveRunButton_clicked()));

	// Rest
	connect(vhistAppendCheckBox, SIGNAL(clicked()), this, SLOT(setAppendMode()));		

	connect(tabWidget, SIGNAL(currentChanged(int)),this,SLOT(on_tabWidget_currentChanged(int)));


    // help buttons
    connect(helpMain, SIGNAL(clicked()), this, SLOT(openVHISTEZHelp()));
    connect(helpInOutFiles, SIGNAL(clicked()), this, SLOT(openVHISTEZHelp()));
    connect(helpVHISTfiles, SIGNAL(clicked()), this, SLOT(openVHISTEZHelp()));
    connect(helpExport, SIGNAL(clicked()), this, SLOT(openVHISTEZHelp()));
}


void GuiWindow::initCommandline()
{
    commandLine->setShortcutsActive(true);
    commandLine->setHighlighterEnabled(false);
    m_highlighter.setDocument(commandLine->textBrowser()->document());
}


void GuiWindow::setUpVariables_Start()
{
	m_number = 0;
	m_row = 0;
	m_lock = false;
	m_optional = false;
	m_noItemInList = true;
	m_lockStrategy = new locknone();
}

void GuiWindow::setUpMimeTypeList()
{
	fileTypeSelector->addItem(m_userDefinedFileType, QRegExp::RegExp);

	for (unsigned i = 0;i < Mimetypes::getTypeCount ();i++) {
		fileTypeSelector->addItem (Mimetypes::getType (i), QRegExp::RegExp);
	}
}

void GuiWindow::setDefaults()
{
	//! starts with main-tab
	tabWidget->setCurrentIndex(0);
	//! default: Root File is not checked
	noneCheckBox->setChecked(false);
	//! default: the first input File is taken as vhist root file
	firstInputFileVhistRootFileRadioButton->setChecked(1);
	//! default: the first output File is taken as storage file
	firstOutputFileVhist->setChecked(true);
	//! default: the normal Button is set at the mainpage
	normalRadioButton->setChecked(true);
	reset_GeneralAttribtuesLabel();

	vhistAppendCheckBox->setChecked(true);
	setAppendMode();
}


void GuiWindow::setAppendMode()
{
	bool bAppend = vhistAppendCheckBox->isChecked();
	vhistStorageFileGroupBox->setEnabled(!bAppend);
	noneCheckBox->setEnabled (!bAppend);
	vhistRootOptionalCheckBox->setEnabled (!bAppend);
}

void GuiWindow::showEvent(QShowEvent * /*event*/) {
        int iUsableWidth = generalAttributesTable->viewport()->width();
	generalAttributesTable->setColumnWidth(0, static_cast<int>(iUsableWidth * 0.3));
	generalAttributesTable->setColumnWidth(1, static_cast<int>(iUsableWidth * 0.7));

	// columns of fileAttributesTable will be resized in setButtonsVisibleAfterNew()
}

void GuiWindow::resizeEvent(QResizeEvent * /*event*/) {
        int iUsableWidth = generalAttributesTable->viewport()->width();
	int iKeyColWidth = generalAttributesTable->columnWidth(0);
	generalAttributesTable->setColumnWidth(1, iUsableWidth - iKeyColWidth);

        iUsableWidth = fileAttributesTable->viewport()->width();
        iKeyColWidth = fileAttributesTable->columnWidth(0);
	fileAttributesTable->setColumnWidth(1, iUsableWidth - iKeyColWidth);
}	

void GuiWindow::openVHISTEZHelp()
{
	switch (tabWidget->currentIndex()) {
		case 0:
		    HelpViewer::globalOpenPage("vhistez.html", QString());
			break;

		case 1:
		    HelpViewer::globalOpenPage("vhistez.html", QString());
			break;

		case 2:
		    HelpViewer::globalOpenPage("vhistez.html", QString());
			break;

		case 3:
		    HelpViewer::globalOpenPage("vhistez.html", QString());
			break;

		default: 
			qDebug ("Problem found: tabWidget->currentIndex () out of range.");
		    HelpViewer::globalOpenPage("vhistez.html", QString());
			break;
	}
}

void GuiWindow::save_cmdfile() {
	tabWidget->setCurrentIndex(3);
	on_saveAsCmdFileButton_clicked();
}

void GuiWindow::reset_AttributesFileLabel()
{
	numberOfAttributesFileLabel->setText(QString("%1 attributes ").arg(0));
}

void GuiWindow::update_AttributesFileLabel() {
	numberOfAttributesFileLabel->setText(QString("%1 attributes ").arg(fileAttributesTable->rowCount()-1));
	if (fileAttributesTable->rowCount()<=1)
		numberOfAttributesFileLabel->setText(QString("%1 attributes ").arg(0));
}

void GuiWindow::reset_GeneralAttribtuesLabel() {
	numberOfAttributesGeneral_Label->setText(QString("%1 attributes ").arg(0));
}

void GuiWindow::update_GeneralAttributesLabel() {
	numberOfAttributesGeneral_Label->setText(QString("%1 attributes ").arg(generalAttributesTable->rowCount()-1));
}

void GuiWindow::addNewRowToGeneralAttributesTable() {
	int row=generalAttributesTable->rowCount();
	generalAttributesTable->insertRow(row);
	QString itemText="";
	QTableWidgetItem* itemLeft = new QTableWidgetItem(itemText,0);
	fileAttributesTable->setItem(row, 0, itemLeft);
	QTableWidgetItem* itemRight = new QTableWidgetItem(itemText,0);
	fileAttributesTable->setItem(row, 1, itemRight);
}

void GuiWindow::initializeWorklfowKeyValuePairs() {
	generalAttributesTable->insertRow(m_number);
	QString itemText="";
	QTableWidgetItem* itemLeft = new QTableWidgetItem(itemText,0);
	generalAttributesTable->setItem(m_number, 0, itemLeft);
	QTableWidgetItem* itemRight = new QTableWidgetItem(itemText,0);
	generalAttributesTable->setItem(m_number, 1, itemRight);
	m_number++;
}

void GuiWindow::setLineEditTextToBlank() {
	vhistStorageFilePathText->setText("");
	vhistStorageFileNameText->setText("");
	vhistRootFilePathText->setText("");
	vhistRootFileNameText->setText("");
}

void GuiWindow::updateNumberOfFilesLabel() {
	numberOfFilesLabel->setText(QString("Number of Files %1").arg(FileListWidget->count()));
}

void GuiWindow::hideAllButtons() {
	removeFileButton->setEnabled(0);
	fileBrowserButton->setEnabled(0);
	inputRadioButton->setEnabled(0);
	outputRadioButton->setEnabled(0);
	fileTypeSelector->setEnabled(0);
	fileTypeLineEdit->setEnabled(0);
	embedFileCheckBox->setEnabled(0);
	md5CheckBox->setEnabled(0);
	embedFileCheckBox->setEnabled(0);
	pathText->setEnabled(0);
	descriptionText->setEnabled(0);
	duplicateFileButton->setEnabled(0);
	fileAttributesTable->setEnabled(0);
	removeFileKeyValuePairs->setEnabled(0);
	removeAllFileValuePairs->setEnabled(0);
	numberOfAttributesFileLabel->setEnabled(0);
	FileListWidget->setEnabled(0);
	fileTypeGroupBox->setEnabled(0);
	fileText->setEnabled(0);
	inOutFilesTab_FileLabel->setEnabled(0);
	descriptionGroupBox->setEnabled(0);
	descriptionText->setEnabled(0);
	fileTypeGroupBox->setEnabled(0);
	listOfFilesGroupBox->setEnabled(0);
	propertiesOfCurrentFileGroupBox->setEnabled(0);
	optionalAttributes->setEnabled(0);
	filenameAndPathGroupBox->setEnabled(0);
	optionsGroupBox->setEnabled(0);
	removeAllFilesButton->setEnabled(0);
	propertiesOfCurrentFileGroupBox->setVisible(0);
	numberOfFilesLabel->setVisible(0);
	//relativePathText->setEnabled(0);
}

void GuiWindow::dragMoveEvent(QDragMoveEvent *event) {
	event->acceptProposedAction();
}

void GuiWindow::dropEvent(QDropEvent *event) {
	QList<QUrl> urls=event->mimeData()->urls();
	if (urls.isEmpty())
		return;
	QString fileName = urls.first().toLocalFile();
	if (fileName.isEmpty())
		return;
	if ((tabWidget->currentIndex())==1)
		changeFileListItemName(fileName);
	if ((tabWidget->currentIndex())==0) {
		processFileData(fileName, "tool");
	}
	if ((tabWidget->currentIndex())==2) {
		if (childAt(event->pos())->objectName()=="rootFileGroupBox" ||childAt(event->pos())->objectName()=="vhistFileNameAndPathGroupBox"
				||childAt(event->pos())->objectName()=="pathLabel"||childAt(event->pos())->objectName()=="fileLabel" ||childAt(event->pos())->objectName()
				=="firstInputFileVhistRootFileRadioButton" ||childAt(event->pos())->objectName()=="userspecifiedPathAndNameRadioButton"
				||childAt(event->pos())->objectName()=="commentLabel" ||childAt(event->pos())->objectName()=="vhistRootOptionalCheckBox"
				||childAt(event->pos())->objectName()=="noneCheckBox" ||childAt(event->pos())->objectName()=="commentLabel2" ||childAt(event->pos())->objectName()=="vhistRootFilePathText" ||childAt(event->pos())->objectName()=="vhistRootFileNameText")
			processFileData(fileName, "vhistRootFile");
		if (childAt(event->pos())->objectName()=="vhistStorageFileGroupBox" ||childAt(event->pos())->objectName()=="vhistRootFileNameText" ||childAt(event->pos())->objectName()== "vhistStorageFilePathText"
				||childAt(event->pos())->objectName()
						=="vhistStorageFileBrowserButton" ||childAt(event->pos())->objectName()=="firstOutputFileVhist" || childAt(event->pos())->objectName()=="one_EachOutputFileRadioButton"
				||childAt(event->pos())->objectName()
						=="oneUserSpecificVhistOutFileRadioButton")
			processFileData(fileName, "vhistStorageFile");
	}
}

void GuiWindow::dragEnterEvent(QDragEnterEvent *event) {
	event->acceptProposedAction();
}

//---------------------------SLOTS-----------------------------------------------------------------

void GuiWindow::on_generalAttributesTable_cellChanged(int row, int colum)
{
	m_argfile->clearUserAttributes ();

	for (int i = 0;i < generalAttributesTable->rowCount ();i++) {
		QTableWidgetItem* ki = generalAttributesTable->item (i, 0);
		QTableWidgetItem* vi = generalAttributesTable->item (i, 1);

		if ((ki == NULL) || (vi == NULL)) {
			continue;
		}

		QString key = ki->text ().trimmed ();
		QString value = vi->text ().trimmed ();

		if ((key == "") || (value == "")) {
			continue;
		}

		m_argfile->addUserAttribute (key, value);
	}

	/* TODO: Check what this is doing */
	if (colum==0) {
		if (m_number==row+1) {
			if (generalAttributesTable->item(row,colum)->text()!=" ")
				initializeWorklfowKeyValuePairs();
		}
	}
	update_GeneralAttributesLabel();
}

void GuiWindow::on_saveAsCmdFileButton_clicked() {
	QString fileName= QFileDialog::getSaveFileName(this, tr("Save as Cmdfile"),
			"", tr("Cmdfile (*.vcmd)"));
	store_cmdFileToDisk(new QFile(fileName));
}


void GuiWindow::on_saveRunButton_clicked()
{
    QString fileName= QFileDialog::getSaveFileName(this,
            tr("Save as Cmdfile"), "", tr("Cmdfile (*.vcmd)"));
	if (fileName.isEmpty ())
		return;
    store_cmdFileToDisk(new QFile(fileName));
    if (m_parent != NULL)
    {
        WindowList* wlist = WindowList::globalInstance();
        wlist->createWindow(VHISTitWindow::windowName(), fileName);
    }
        // TODO add right call
        //m_parent->createVHISTitWindow(fileName);
}

void GuiWindow::store_cmdFileToDisk(QFile* file) {
	tabWidget->setCurrentIndex(3);
	if (!file->open(QFile::WriteOnly| QFile::Text)) {
		cerr << "Cannot  open file for writing: "
				<< qPrintable(file->errorString()) << endl;
		return;
	}
	QString text=buildCommand_File(true);
	QTextStream out(file);
	out<< text << endl;
}

void GuiWindow::on_fileAttributesTable_cellChanged(int row, int colum)
{
	vhist::WSFile* f = m_fihelper->getSelectedFile ();

	QTableWidgetItem* keyitem = fileAttributesTable->item (row, 0);
	QTableWidgetItem* valueitem = fileAttributesTable->item (row, 1);

	if ((keyitem != NULL) && (valueitem != NULL)) {
		if (keyitem->text () != "") {
			f->addUserAttribute (keyitem->text (), valueitem->text ());		
		}
	}

	if (colum==0) {
		if (fileAttributesTable->rowCount () == row+1) {
			if ((fileAttributesTable->item(row,colum)->text()!=" ") &&(fileAttributesTable->item(row,colum)->text()!="")) {
				fileAttributesTable->setRowCount (fileAttributesTable->rowCount () + 1);
				update_AttributesFileLabel();
			}
		}
	}
}


void GuiWindow::on_generalAttributesTable_itemSelectionChanged()
{
	QTableWidgetItem* Item;
	QList <QTableWidgetItem*> liste= generalAttributesTable->selectedItems();
	int r=-1;
	while ( !liste.isEmpty()) {
		Item=liste.takeFirst();
		if (r!=-1) {
			if (Item->row()!=r) {
				generalAttributesTable->item(Item->row(),Item->column())->setSelected(0);
			}
		}
		r=Item->row();
	}
}

void GuiWindow::on_fileAttributesTable_itemSelectionChanged()
{
	QTableWidgetItem* Item;
	QList <QTableWidgetItem*> liste= fileAttributesTable->selectedItems();
	int r=-1;
	while ( !liste.isEmpty()) {
		Item=liste.takeFirst();
		if (r!=-1) {
			if (Item->row()!=r) {
				fileAttributesTable->item(Item->row(),Item->column())->setSelected(0);
			}
		}
		r=Item->row();
	}
}

void GuiWindow::storeToItem()
{
	vhist::WSFile* f = m_fihelper->getSelectedFile ();
	if (f == NULL) { /* Not yet */
		return;
	}

	f->clearUserAttributes ();

	int r=0;
	QString keyValue="";
	while ( (fileAttributesTable->item(r, 0)!=NULL)
			&&(fileAttributesTable->item(r, 1)!=NULL)) {
		if ((fileAttributesTable->item(r,0)->text().trimmed () != "") &&
			(fileAttributesTable->item(r,1)->text().trimmed () != "")) {
			QString key = fileAttributesTable->item(r,0)->text();
			QString value = fileAttributesTable->item(r,1)->text();
			f->addUserAttribute (key, value);
		}
		r++;
	}
}

void GuiWindow::on_fileText_returnPressed() {
	qDebug ("GuiWindow::on_fileText_returnPressed");
	on_fileText_editingFinished();
	guessMimeType ();
}

void GuiWindow::on_fileText_editingFinished() {
	on_pathText_editingFinished();
}

void GuiWindow::on_pathText_returnPressed() {
	on_pathText_editingFinished();
}

void GuiWindow::on_pathText_editingFinished()
{
	guaranteePathTextEndsWithSlashOrBackSlash();
	QDir tmp(pathText->text());
	handleFilePathAndNameEditLineChange();
}

void GuiWindow::guaranteePathTextEndsWithSlashOrBackSlash()
{
	QString trimmedpath = pathText->text ().trimmed ();

	if (trimmedpath != "")
		if (!(trimmedpath.endsWith(("/")) || trimmedpath.endsWith("\\")))
			pathText->setText(pathText->text().append("/"));
}

void GuiWindow::handleFilePathAndNameEditLineChange()
{
	QString m_fileNameAndPath;
	vhist::WSFile* f = m_fihelper->getSelectedFile ();

	/* Store */
	f->setFilename (fileText->text ());
	f->setFilepath (pathText->text ());

	if ((fileText->text()!="") || (pathText->text()!="")) { /* TODO: What is this!? */
		QString fileName = f->filename ();
		if (fileName.trimmed()=="") {
			fileName=".";
		}
		m_fileNameAndPath="";

		guaranteePathTextEndsWithSlashOrBackSlash();
		QDir tmpDir(pathText->text());
		m_fileNameAndPath=tmpDir.absoluteFilePath(fileName);

		QFileInfo fileInfoObject (m_fileNameAndPath);
		QString absoluteFilePathString=fileInfoObject.absoluteFilePath();
		changeFileListItemName(m_fileNameAndPath);
		if (fileText->text()==" ")
			fileText->setText("");
		pathText->setText (fileInfoObject.dir ().absolutePath ());
	}
}

void GuiWindow::on_verboseRadioButton_clicked()
{
	m_argfile->setVerbosity (vhist::ArgFile::VERBOSE);
}

void GuiWindow::on_normalRadioButton_clicked()
{
	m_argfile->setVerbosity (vhist::ArgFile::NORMAL);
}

void GuiWindow::on_quietRadioButton_clicked()
{
	m_argfile->setVerbosity (vhist::ArgFile::QUIET);
}

void GuiWindow::on_pretendCheckBox_clicked()
{
	m_argfile->setPretend (pretendCheckBox->isChecked ());
}

void GuiWindow::on_descriptionText_textChanged()
{
	QListWidgetItem* m_selectedItem = m_fihelper->getSelectedItem ();
	vhist::WSFile* f = m_fihelper->getSelectedFile ();

	if ((f == NULL) || (m_selectedItem == NULL)) { // Not yet!
		return;
	}

	m_selectedItem->setToolTip(QString(descriptionText->toPlainText()));
	f->setDescription (descriptionText->toPlainText());
}

void GuiWindow::on_vhistStorageFileNameText_editingFinished()
{
	setGroupBox(m_fihelper->getSelectedItem ());
	handleStorageFileAndPathChange();
}

void GuiWindow::handleStorageFileAndPathChange() {
	if ((vhistStorageFilePathText->text()!="")
			&&(vhistStorageFileNameText->text()!="")) {
		QDir tmpDir(vhistStorageFilePathText->text());
		QString fileName=vhistStorageFileNameText->text();
		if (fileName.trimmed()=="")
			fileName=".";

		QString storageFileNameAndPath="";
		storageFileNameAndPath.append(tmpDir.absoluteFilePath(fileName));
		processFileData(storageFileNameAndPath, "vhistStorageFile");
		if (vhistStorageFileNameText->text()==" ")
			vhistStorageFileNameText->setText("");
	}
}

void GuiWindow::handleRootFileAndPathChange() {
	if ((vhistRootFilePathText->text()!="") &&(vhistRootFileNameText->text()
			!="")) {
		QDir tmpDir("");
		tmpDir.setPath(vhistRootFilePathText->text());
		QString fileName=vhistRootFileNameText->text();
		if (fileName.trimmed()=="")
			fileName=".";

		QString rootFileNameAndPath="";
		rootFileNameAndPath.append(tmpDir.absoluteFilePath(fileName));
		processFileData(rootFileNameAndPath, "vhistRootFile");
		if (vhistRootFileNameText->text()==" ")
			vhistRootFileNameText->setText("");
	}
}

void GuiWindow::on_vhistStorageFilePathText_editingFinished() {
	handleStorageFileAndPathChange();
}

void GuiWindow::on_vhistRootFilePathText_editingFinished() {
	handleRootFileAndPathChange();
}

void GuiWindow::on_vhistRootFileNameText_editingFinished() {
	handleRootFileAndPathChange();
}

void GuiWindow::on_fileTypeLineEdit_textChanged(const QString& changedText)
{
	fileTypeLineEdit->setText(changedText);
	
	if (m_lock == false) {
		QListWidgetItem* m_selectedItem = m_fihelper->getSelectedItem ();
		m_selectedItem->setStatusTip(changedText);

		vhist::WSFile* f = m_fihelper->getSelectedFile ();
		f->setFiletype (changedText);
	}
}

void GuiWindow::on_fileTypeSelector_activated(const QString& test)
{
	QListWidgetItem* selectedItem= m_fihelper->getSelectedItem ();

	QString help=test;
	if (test==m_userDefinedFileType) {
		fileTypeLineEdit->setEnabled(1);
		fileTypeLineEdit->setText("Please enter the file type here");
	} else {
		fileTypeLineEdit->setEnabled(0);
		fileTypeLineEdit->setText(test);
	}

	setGroupBox(selectedItem);
}

void GuiWindow::on_md5CheckBox_clicked()
{
	vhist::WSFile* f = m_fihelper->getSelectedFile ();
	f->setAutoMD5 (md5CheckBox->checkState ());
}

void GuiWindow::on_embedFileCheckBox_clicked()
{
	vhist::WSFile* f = m_fihelper->getSelectedFile ();
	f->setEmbed (embedFileCheckBox->checkState ());
}

void GuiWindow::on_previewWSCheckBox_clicked ()
{
	vhist::WSFile* f = m_fihelper->getSelectedFile ();
	f->setPreviewWS (previewWSCheckBox->checkState ());
}

void GuiWindow::on_previewCheckBox_clicked ()
{
	vhist::WSFile* f = m_fihelper->getSelectedFile ();
	f->setPreview (previewCheckBox->checkState ());
}

void GuiWindow::on_thumbnailCheckBox_clicked ()
{
	vhist::WSFile* f = m_fihelper->getSelectedFile ();
	f->setThumbnail (thumbnailCheckBox->checkState ());
}

void GuiWindow::on_clipBoardButton_clicked()
{
        QApplication::clipboard()->setText(commandLine->textBrowser()->toPlainText());
}

void GuiWindow::on_tabWidget_currentChanged (int index)
{
	if (index == 3) { /* Do behaviour of old buildCommand, which however is only useful here */
                commandLine->textBrowser()->setHtml(buildCommand_File(false));
		setStatisticsText();
		setHints(returnHints("hintBox"));
	}
}

void GuiWindow::updateToolAttributesTable()
{
	QStringList keys = m_argfile->userAttributeKeys ();

	for (int u = 0;u < keys.size ();u++) {
		QString key = keys.at (u);
		QString value = m_argfile->userAttribute (key);

		QTableWidgetItem* itemKey=new QTableWidgetItem(key, 0);
		QTableWidgetItem* itemValue=new QTableWidgetItem(value, 0);

		generalAttributesTable->setItem(u, 0, itemKey);
		generalAttributesTable->setItem(u, 1, itemValue);
	}
}

void GuiWindow::loadFileAttributesTable ()
{
	vhist::WSFile* selectedfile = m_fihelper->getSelectedFile ();

	clearAllFilesAttributes ();

	QStringList attrs = selectedfile->userAttributeKeys();
	int attrcount = attrs.size();

	fileAttributesTable->setRowCount (attrcount + 1);

	for (int i = 0;i < attrcount;i++) {
		QString key = attrs.at (i);

		if (key == "") { /* Empty key is not right */
			continue;
		}

		QString value = selectedfile->userAttribute (key);

		QTableWidgetItem* item = new QTableWidgetItem(key, 0);
		fileAttributesTable->setItem(i, 0, item);

		QTableWidgetItem* item2 = new QTableWidgetItem(value, 0);
		fileAttributesTable->setItem(i, 1, item2); 
	}
}

void GuiWindow::on_removeAllFileValuePairs_clicked()
{
	clearAllFilesAttributes();
	storeToItem();
	update_AttributesFileLabel();
}

void GuiWindow::clearAllFilesAttributes () 
{
	while (fileAttributesTable->rowCount () > 0) {
		fileAttributesTable->removeRow (0);
	}

	fileAttributesTable->setRowCount (1);
}

void GuiWindow::clearAllToolAttributes() {
	int i=0;
	generalAttributesTable->setEnabled(0); //!Disable the Table to be sure that no item is activated and no item contains an active coursor, otherwise there will be problems during deleting the items
	while ((generalAttributesTable->item(i, 0)!=NULL)
			&&(generalAttributesTable->item(i, 1)!=NULL)) {
		(generalAttributesTable->item( i,1))->setText(" "); //!mark for delete
		(generalAttributesTable->item( i,0))->setText(" "); //!mark for delete
		i++;
	}
	if (i>0)
		for (int j=i; j>0; j--) //!all except one are deleted
		{
			generalAttributesTable->removeRow(j);
			m_number=m_number-1;
		}
	m_number=1; //Debug
	generalAttributesTable->setEnabled(1);
}

void GuiWindow::handle_visibility_InOutFileTab()
{
	if (FileListWidget->count()==0) {
		m_inouttabcm->setFileOperations (false);
		hideAllButtons();
	} else {
		m_inouttabcm->setFileOperations (true);
	}
}

void GuiWindow::checkFreeWidgetItemsfiles(QString /*command*/) {
	int r=0;
	int counter=0;
	while ( (fileAttributesTable->item(r, 0)!=NULL)
			&&(fileAttributesTable->item(r, 1)!=NULL)) {
		if ( ((fileAttributesTable->item(r,0)->text())==(" ")) && (fileAttributesTable->item(r,1)->text()==(" "))) {
			counter=counter+1;
			if (counter>1) {
				fileAttributesTable->removeRow(r);
			}
		}
		r++;
	}
}

void GuiWindow::on_inputRadioButton_clicked()
{
	setInOrOutFile (true);
}

void GuiWindow::on_outputRadioButton_clicked()
{
	setInOrOutFile (false);
}

void GuiWindow::on_removeAllAttributes_clicked() {
	if (m_number!=1)
		clearAllToolAttributes();
	update_GeneralAttributesLabel();
}

void GuiWindow::on_removeFileKeyValuePairs_clicked()
{
	if (fileAttributesTable->rowCount () == 1) { /* No attributes in table */
		return;
	}

	for (int r = 0;r < fileAttributesTable->rowCount ();r++) {
		QTableWidgetItem* key = fileAttributesTable->item (r, 0);
		QTableWidgetItem* value = fileAttributesTable->item (r, 1);

		if (fileAttributesTable->isItemSelected (key) ||
			fileAttributesTable->isItemSelected (value)) {
			fileAttributesTable->removeRow (r);
			break;
		}
	}

	storeToItem ();
	update_AttributesFileLabel();
} 

void GuiWindow::on_removeAttributes_clicked() {
	if (m_number!=1) {
		QList <QTableWidgetItem *> w =(generalAttributesTable->selectedItems());
		QTableWidgetItem* helpItem;
		int r=0;
		while (!(w.isEmpty())&&(m_number>=0)) {
			generalAttributesTable->setEnabled(0);
			helpItem=(w.takeFirst());
			r=helpItem->row();
                        if (! ((generalAttributesTable->item(r,0)->text()==" ") && (generalAttributesTable->item(r,1)->text()==" ") && ((m_number==1)||(m_number==0))))
                        {
				generalAttributesTable->removeRow (r);
				generalAttributesTable->item(0,1)->setSelected(0);
			}
			generalAttributesTable->setEnabled(1);
			if (r!=0) {
				generalAttributesTable->item(r-1,0)->setSelected(1);
			} else {
				if (m_number>r)
					generalAttributesTable->item(r,0)->setSelected(1);
			}
			if (!w.isEmpty())
				w =(generalAttributesTable->selectedItems());
		}
	}
	update_GeneralAttributesLabel();
}

void GuiWindow::checkFreeWidgetItems() {
	int r=0;
	int counter=0;
	while ( (generalAttributesTable->item(r, 0)!=NULL)
			&&(generalAttributesTable->item(r, 1)!=NULL)) {
		if ( ((generalAttributesTable->item(r,0)->text())==(" ")) && (generalAttributesTable->item(r,1)->text()==(" "))) {
			if (counter>0) {
				generalAttributesTable->removeRow(r);
				m_number--;
				counter++;
			}
		}
		r++;
	}

}

void GuiWindow::on_removeFileButton_clicked() {
	QList<QListWidgetItem *> list =FileListWidget->selectedItems();
	while (!list.isEmpty()) {
		QListWidgetItem* lwi = list.takeFirst ();
		
		int index = FileListWidget->row (lwi);
		m_argfile->removeFile (index);
		FileListWidget->takeItem(index);
	}
	(this)->m_row=(this)->m_row-1;
	if (FileListWidget->count()>0)
		FileListWidget->item(FileListWidget->count()-1)->setSelected(1);
	updateNumberOfFilesLabel();
	handle_visibility_InOutFileTab();
}

void GuiWindow::on_removeAllFilesButton_clicked()
{
	FileListWidget->clear ();
	m_row = 0;
	m_argfile->clearFilelist ();
	updateNumberOfFilesLabel();
	handle_visibility_InOutFileTab();
}

void GuiWindow::on_toolNameText_editingFinished()
{
	m_argfile->setTool (toolNameText->text () + versionLineEdit->text ());
}

void GuiWindow::on_versionLineEdit_editingFinished ()
{
	m_argfile->setTool (toolNameText->text () + versionLineEdit->text ());
}

void GuiWindow::on_toolPathText_editingFinished()
{
	m_argfile->setToolpath (toolPathText->text ());
}

void GuiWindow::on_titleLineEdit_editingFinished ()
{
	m_argfile->setTitle (titleLineEdit->text ());
}

void GuiWindow::on_workflowDescriptionTextEdit_textChanged ()
{
	m_argfile->setDescription (workflowDescriptionTextEdit->toPlainText ());
}

void GuiWindow::on_toolBrowserButton_clicked() {
	QString toolFileName= QFileDialog::getOpenFileName(this);
	processFileData(toolFileName, "tool");
	optionalAttributesGroupBox->setVisible(1);
	if (toolNameText->text()==".") {
		toolPathText->setText("");
		toolNameText->setText("");
	}
}

void GuiWindow::on_vhistStorageFileBrowserButton_clicked() {
	QString vhistStorageFileName= QFileDialog::getOpenFileName(this);
	processFileData(vhistStorageFileName, "vhistStorageFile");
	if (vhistStorageFileNameText->text()==".") {
		vhistStorageFileNameText->setText("");
		vhistStorageFilePathText->setText("");
	}
}

void GuiWindow::on_vhistRootFileBrowserButton_clicked() {
	QString vhistRootFileName= QFileDialog::getOpenFileName(this);
	processFileData(vhistRootFileName, "vhistRootFile");
	if (vhistRootFileNameText->text()==".") {
		vhistRootFileNameText->setText("");
		vhistRootFilePathText->setText("");
	}
}

void GuiWindow::setGroupBox(QListWidgetItem* Item) {
	if (Mimetypes::isKnownMimeType (Item->statusTip ())) {
		fileTypeSelector-> setCurrentIndex(fileTypeSelector->findText(
				(Item->statusTip()), Qt::MatchExactly));
		fileTypeLineEdit->setEnabled(0);
		fileTypeLineEdit->setText(Item->statusTip());
	} else {
		fileTypeSelector->setCurrentIndex(0);
		fileTypeLineEdit->setEnabled(1);
		fileTypeLineEdit->setText(Item->statusTip());
	}
	descriptionText->setText(Item->toolTip());
}

void GuiWindow::setInOrOutFile(bool m_inputFile)
{
	QListWidgetItem* item; 
	QList<QListWidgetItem *> list =FileListWidget->selectedItems();

	while (!list.isEmpty()) {
		item=list.takeFirst();
		vhist::WSFile* f = m_fihelper->getFile (item);
		f->setInputFile (m_inputFile);

		if (m_inputFile) {
			item->setIcon(QIcon(QString::fromUtf8(":/resources/input.png"))); // !sv now will be loaded from resources
			item->setText(m_string_InputTextAtFileListWidget+ f->filename());
		} else {
			item->setIcon(QIcon(QString::fromUtf8(":/resources/output.png"))); // !sv now will be loaded from resources
			item->setText(m_string_OutputTextAtFileListWidget+ f->filename());
		}
	}
}

void GuiWindow::setButtonsVisibleAfterNew() {
	hideAllButtons();
	listOfFilesGroupBox->setEnabled(1);
	numberOfFilesLabel->setVisible(1);
	fileBrowserButton->setEnabled(1);
	fileText->setText("");
	pathText->setText("");
	inputRadioButton->setChecked(1);
	fileText->setEnabled(1);
	pathText->setEnabled(1);
	inOutFilesTab_FileLabel->setEnabled(1);
	mainTab_FileLabel->setEnabled(1);
	filenameAndPathGroupBox->setEnabled(1);
	propertiesOfCurrentFileGroupBox->setEnabled(1);
	propertiesOfCurrentFileGroupBox->setVisible(1);
	fileBrowserButton->setEnabled(1);
	inputRadioButton->setEnabled(1);
	outputRadioButton->setEnabled(1);
	fileTypeSelector->setEnabled(1);
	fileTypeLineEdit->setEnabled(1);
	embedFileCheckBox->setEnabled(1);
	md5CheckBox->setEnabled(1);
	embedFileCheckBox->setEnabled(1);
	pathText->setEnabled(1);
	descriptionText->setEnabled(1);
	removeFileButton->setEnabled(1);
	duplicateFileButton->setEnabled(1);
	fileAttributesTable->setEnabled(1);
	removeFileKeyValuePairs->setEnabled(1);
	removeAllFileValuePairs->setEnabled(1);
	numberOfAttributesFileLabel->setEnabled(1);
	FileListWidget->setEnabled(1);
	fileTypeGroupBox->setEnabled(1);
	fileText->setEnabled(1);
	inOutFilesTab_FileLabel->setEnabled(1);
	descriptionGroupBox->setEnabled(1);
	descriptionText->setEnabled(1);
	fileTypeGroupBox->setEnabled(1);
	fileText->setEnabled(1);
	inOutFilesTab_FileLabel->setEnabled(1);
	descriptionGroupBox->setEnabled(1);
	descriptionText->setEnabled(1);
	fileTypeGroupBox->setEnabled(1);
	propertiesOfCurrentFileGroupBox->setEnabled(1);
	optionalAttributes->setEnabled(1);
	filenameAndPathGroupBox->setEnabled(1);
	optionsGroupBox->setEnabled(1);
	removeAllFilesButton->setEnabled(1);
	numberOfFilesLabel->setEnabled(1);
        int iUsableWidth = static_cast<int>(fileAttributesTable->viewport()->width());
	fileAttributesTable->setColumnWidth(0, static_cast<int>(iUsableWidth * 0.3));
	fileAttributesTable->setColumnWidth(1, static_cast<int>(iUsableWidth * 0.7));
}

void GuiWindow::setButtonVisibleAfterBrowse() {
	fileBrowserButton->setEnabled(1);
	inputRadioButton->setEnabled(1);
	outputRadioButton->setEnabled(1);
	fileTypeSelector->setEnabled(1);
	fileTypeLineEdit->setEnabled(1);
	embedFileCheckBox->setEnabled(1);
	md5CheckBox->setEnabled(1);
	embedFileCheckBox->setEnabled(1);
	pathText->setEnabled(1);
	descriptionText->setEnabled(1);
	removeFileButton->setEnabled(1);
	duplicateFileButton->setEnabled(1);
	fileAttributesTable->setEnabled(1);
	removeFileKeyValuePairs->setEnabled(1);
	removeAllFileValuePairs->setEnabled(1);
	numberOfAttributesFileLabel->setEnabled(1);
	FileListWidget->setEnabled(1);
	fileTypeGroupBox->setEnabled(1);
	fileText->setEnabled(1);
	inOutFilesTab_FileLabel->setEnabled(1);
	descriptionGroupBox->setEnabled(1);
	descriptionText->setEnabled(1);
	fileTypeGroupBox->setEnabled(1);
	fileText->setEnabled(1);
	inOutFilesTab_FileLabel->setEnabled(1);
	descriptionGroupBox->setEnabled(1);
	descriptionText->setEnabled(1);
	fileTypeGroupBox->setEnabled(1);
	listOfFilesGroupBox->setEnabled(1);
	propertiesOfCurrentFileGroupBox->setEnabled(1);
	optionalAttributes->setEnabled(1);
	filenameAndPathGroupBox->setEnabled(1);
	optionsGroupBox->setEnabled(1);
	removeAllFilesButton->setEnabled(1);
}

void GuiWindow::setAllVisible() {
	embedFileCheckBox->setCheckState(Qt::Unchecked);
	md5CheckBox->setCheckState(Qt::Unchecked);
	removeFileButton->setEnabled(1);
	fileBrowserButton->setEnabled(1);
	inputRadioButton->setEnabled(1);
	outputRadioButton->setEnabled(1);
	fileTypeSelector->setEnabled(1);
	embedFileCheckBox->setEnabled(1);
	md5CheckBox->setEnabled(1);
	embedFileCheckBox->setEnabled(1);
	pathText->setEnabled(1);
	descriptionText->setEnabled(1);
	descriptionText->setText("");
	fileAttributesTable->setEnabled(1);
	removeFileKeyValuePairs->setEnabled(1);
	removeAllFileValuePairs->setEnabled(1);
	FileListWidget->setEnabled(1);
	fileTypeGroupBox->setEnabled(1);
	fileText->setEnabled(1);
	inOutFilesTab_FileLabel->setEnabled(1);
	descriptionGroupBox->setEnabled(1);
	numberOfAttributesFileLabel->setEnabled(1);
	fileTypeGroupBox->setEnabled(1);
}

QString GuiWindow::buildCommand_File(bool /*withComment*/) 
{
	evaluateRadioButtons (); 

	/* As arglib doesn't do this for us,
	   we need to figure out which input file, if any,
	   and which output files we will use.
	   Currently this is some logic of the model in here.
	   TODO: This might be an indicator for a tiny layer between GuiWindow and ArgFile.
	   This is now where there shouldn't be anything like logic in GuiWindow. */
	if (vhistAppendCheckBox->isChecked ()) {
		/* Append mode is simulated by setting the input file to vhistAppendFilePath + vhistAppendFileName and
		   adding one output file, pointing to the same file. Note that vhistadd supports a -A flag, which would
		   do this, too. */
		QString path = vhistRootFilePathText->text ();
		QString name = vhistRootFileNameText->text ();

		/* <undefined> seems to be translated in arglib, however
		 * the result is empty. This might give you a "-I -O" output
		 * which is nonsense. */
		if (path == "<undefined>") {
			path = "";
		}
		if (name == "<undefined>") {
			name = "";
		}

		m_argfile->setRootfile (path + name);
		m_argfile->clearOutputVhistfiles ();

		if ((path + name) != "") {
			m_argfile->addOutputVhistfile (path + name);
		}
	} else {
		/* In this case, we are going to set the input file and the output file(s) independantly from each other.
		   We will start with the input. */
		m_argfile->setRootfile ("");

		if (!noneCheckBox->isChecked ()) {
			m_argfile->setRootIsOptional (vhistRootOptionalCheckBox->isChecked ());

			if (userspecifiedPathAndNameRadioButton->isChecked ()) {
				m_argfile->setRootfile (vhistRootFilePathText->text () + vhistRootFileNameText->text ());
			} else {
				for (int i = 0;i < m_argfile->numFiles ();i++) {
					vhist::WSFile* file = m_argfile->getFile (i);

					if (file->inputFile ()) {
						m_argfile->setRootfile (file->filename () + ".vhist");
						break;
					}
				}
			}
		}

		/* Now the output */
		m_argfile->clearOutputVhistfiles ();

		/* Note that exactly one of the three if-blocks down here will be executed each time. */
		if (firstOutputFileVhist->isChecked ()) {
			for (int i = 0;i < m_argfile->numFiles ();i++) {
				vhist::WSFile* file = m_argfile->getFile (i);

				if (!file->inputFile ()) {
					m_argfile->addOutputVhistfile (file->filename () + ".vhist");
					break;
				}
			}
		}

		if (one_EachOutputFileRadioButton->isChecked ()) { /* TODO: This is too similiar too the one from above. */
			for (int i = 0;i < m_argfile->numFiles ();i++) {
				vhist::WSFile* file = m_argfile->getFile (i);

				if (!file->inputFile ()) {
					m_argfile->addOutputVhistfile (file->filename () + ".vhist");
				}
			}			
		}

		if (oneUserSpecificVhistOutFileRadioButton->isChecked ()) {
			m_argfile->addOutputVhistfile (vhistStorageFilePathText->text () + vhistStorageFileNameText->text ());
		}
	}

	/* Now build the file */
	if (relativePathsCheckBox->isChecked ()) {
		/* A copy of m_argfile is created here, and in this copy
		 * all absolute paths will be adjusted to be relative.
		 * This is done this way, as we don't want to destroy
		 * the informations in m_argfile. */

		vhist::ArgFile* tmp = new vhist::ArgFile (*m_argfile);

		tmp->setRootfile (adjustRelativity (m_argfile->rootfile ()));

		for (int i = 0;i < tmp->numFiles ();i++) {
			vhist::WSFile* wsf = tmp->getFile (i);

			wsf->setFilenameAndPath (adjustRelativity (wsf->filename ()));
		}

		tmp->clearOutputVhistfiles ();

		for (int i = 0;i < m_argfile->numOutputVhistfiles ();i++) {
			QString n = m_argfile->outputVhistfile (i);

			tmp->addOutputVhistfile (adjustRelativity (n));
		}

		QString ret = tmp->toCommandline (false);

		delete tmp;

		return (ret);
	}

	/* Build file with absolute paths */
	return (m_argfile->toCommandline (false)); // TODO: Correct arglib usage for comments
}

void GuiWindow::setStatisticsText()
{
	/* 3rd reason for a possible layer between GuiWindow and Arglib - the statistics */
	unsigned nif = 0, nia = 0, nof = 0, noa = 0;

	for (int i = 0;i < m_argfile->numFiles ();i++) {
		vhist::WSFile* wsf = m_argfile->getFile (i);

		if (wsf->inputFile ()) {
			++nif;
			nia += wsf->userAttributeKeys ().size ();
		} else {
			++nof;
			noa += wsf->userAttributeKeys ().size ();
		}
	}

	/* Output */
	QString	statisticsText= QString("Workflowstep contains:\n general attribute(s): %1 , ").arg(generalAttributesTable->rowCount() -1);
	statisticsText.append(QString("Input File(s): %1 ").arg(nif));
	statisticsText.append(QString("(file attribute(s): %1), ").arg(nia));
	statisticsText.append(QString("Output file(s): %1 ").arg(nof));
	statisticsText.append(QString("(file attribute(s): %1).").arg(noa));
	statistics_label->setText(statisticsText);
}

void GuiWindow::processFileData(QString fileName, QString fileType) {
	QDir dir(fileName);
	QString name=dir.dirName();
	QString path="";
	QString completeFileName=fileName;
	QString workCopyOfFileName=fileName;

	if (fileType=="vhistRootFile") {
		vhistRootFilePathText->setText(workCopyOfFileName.remove(name));
		vhistRootFileNameText->setText(fileName.remove(workCopyOfFileName));
	}
	if (fileType=="vhistStorageFile") {
		vhistStorageFilePathText->setText(workCopyOfFileName.remove(name));
		vhistStorageFileNameText->setText(fileName.remove(workCopyOfFileName));
	}
}

void GuiWindow::changeFileListItemName(const QString& fileName)
{
	// Update file item
	QListWidgetItem* selectedItem = m_fihelper->getSelectedItem ();
	vhist::WSFile* f = m_fihelper->getSelectedFile ();
	f->setFilenameAndPath (fileName);

	if (f->inputFile ())
		selectedItem->setText(m_string_InputTextAtFileListWidget+fileName);
	else
		selectedItem->setText(m_string_OutputTextAtFileListWidget+fileName);

	selectedItem->setSelected(1);
}

void GuiWindow::on_FileListWidget_itemSelectionChanged()
{
	QList<QListWidgetItem *> l = FileListWidget->selectedItems();
	if (l.size () != 0) {
		m_fihelper->setSelectedItem (l.takeFirst ());
	} else {
		return;
	}

	//!Avoid that files are selected in the box
	vhist::WSFile* selectedfile = m_fihelper->getSelectedFile ();

	m_inouttabcm->adjustCheckStates (selectedfile);

	if (selectedfile->autoMD5()) {
		md5CheckBox->setCheckState(Qt::Checked);
	} else {
		md5CheckBox->setCheckState(Qt::Unchecked);
	}

	if (selectedfile->embed ()) {
		embedFileCheckBox->setCheckState(Qt::Checked);
	} else {
		embedFileCheckBox->setCheckState(Qt::Unchecked);
	}

	if (selectedfile->optional ()) {
		optionalCheckBox->setCheckState(Qt::Checked);
	} else {
		optionalCheckBox->setCheckState(Qt::Unchecked);
	}

	if (selectedfile->previewWS ()) {
		previewWSCheckBox->setCheckState(Qt::Checked);
	} else {
		previewWSCheckBox->setCheckState(Qt::Unchecked);
	}

	if (selectedfile->preview ()) {
		previewCheckBox->setCheckState(Qt::Checked);
	} else {
		previewCheckBox->setCheckState(Qt::Unchecked);
	}

	if (selectedfile->thumbnail ()) {
		thumbnailCheckBox->setCheckState(Qt::Checked);
	} else {
		thumbnailCheckBox->setCheckState(Qt::Unchecked);
	}

	if (selectedfile->inputFile ()) {
		inputRadioButton->setChecked(1);
	} else {
		outputRadioButton->setChecked(1);
	}

	loadIOFileName ();
	loadFileAttributesTable ();
	setGroupBox(m_fihelper->getSelectedItem ());
	QList<QListWidgetItem *> list= FileListWidget->selectedItems();
	descriptionText->setText((list.takeFirst())->toolTip());
}

void GuiWindow::on_duplicateFileButton_clicked()
{
	int index = m_fihelper->getSelectedIndex ();
	vhist::WSFile* f = m_argfile->getFile (index);
	vhist::WSFile* copy = m_argfile->addFile (*f);

	QListWidgetItem* newItem = new QListWidgetItem ();
	newItem->setToolTip (f->description ());
	newItem->setStatusTip (f->filetype ());

	m_fihelper->setUpNewPair (newItem, copy);

	(this)->m_row=(this)->m_row+1;
	FileListWidget->insertItem(m_row, newItem);
	FileListWidget->item((FileListWidget->row(newItem)))->setSelected(1);
	numberOfFilesLabel->setText(QString("#attributes: %1").arg(FileListWidget->count()));

	setInOrOutFile (f->inputFile ());
	setGroupBox(m_fihelper->getSelectedItem ());
	descriptionText->setText(f->description ());
}

void GuiWindow::on_newFileButton_clicked()
{
	vhist::WSFile* newfile = m_argfile->addFile (vhist::WSFile ());
	if (m_argfile->numFiles () == 1) {
		setButtonsVisibleAfterNew();
	}

	/* Set new name */
	newfile->setFilename ("new file " + QString::number (m_argfile->numFiles ()));
	newfile->setFilepath (QDir::rootPath ());

	//!deselect files in the list of file box to avoid two files beeing selected at the same time
	m_fihelper->clearSelections ();

	addFile (newfile);
	m_inouttabcm->setFileOperations (true);
}

void GuiWindow::addFile(vhist::WSFile* file)
{
	QString fileName = file->filename ();

	update_AttributesFileLabel();
	m_lock=true;

	m_fihelper->clearSelections ();

	QDir binDir(file->filename ());
	QString nameOfFile=binDir.dirName();
	fileText->setText(nameOfFile);
	fileName.remove(nameOfFile);
	QString path= (fileName);
	QString trimmedpath = path.trimmed ();

	if (!(trimmedpath.endsWith(("/"))||trimmedpath.endsWith("\\")))
		path.append("/");
	pathText->setText(path);
	setButtonVisibleAfterBrowse();
	QListWidgetItem* newItem = new QListWidgetItem ();
	m_fihelper->setUpNewPair (newItem, file);
	(this)->m_row=(this)->m_row+1;
	descriptionText->setText (file->description ());
	newItem->setToolTip(QString(descriptionText->toPlainText()));
	
	FileListWidget->insertItem(m_row, newItem);
	FileListWidget->item((FileListWidget->row(newItem)))->setSelected(1);
	loadFileAttributesTable ();
	if (m_noItemInList==true) {
		m_noItemInList=false;
		removeFileButton->setEnabled(1);
	}
	
	if (file->filetype () == "") {
		fileTypeLineEdit->setText("Please enter the file type here:");
	} else {
		fileTypeLineEdit->setEnabled (true);

		for (int i = 0;i < fileTypeSelector->count ();i++) {
			QString it = fileTypeSelector->itemText (i);

			if (it == file->filetype ()) {
				fileTypeSelector->setCurrentIndex (i);
				fileTypeLineEdit->setEnabled (false);
				break;
			}
		}

		fileTypeLineEdit->setText (file->filetype ());
	}

	newItem->setStatusTip(fileTypeLineEdit->text());
	propertiesOfCurrentFileGroupBox->setVisible(1);
	m_lock=false;
	updateNumberOfFilesLabel();
	changeFileListItemName (file->filename ());

	if (file->inputFile ()) {
		newItem->setIcon(QIcon(QString::fromUtf8(":/resources/input.png"))); //!sv now will be loaded from resources
		newItem->setText(m_string_InputTextAtFileListWidget + m_fihelper->getSelectedFilename ());
	} else {
		newItem->setIcon(QIcon(QString::fromUtf8(":/resources/output.png"))); //!sv now will be loaded from resources
		newItem->setText(m_string_OutputTextAtFileListWidget + m_fihelper->getSelectedFilename ());
	}

	/* Set checkboxes */
	embedFileCheckBox->setChecked (file->embed ());
	md5CheckBox->setChecked (file->autoMD5 ());
	optionalCheckBox->setChecked (file->optional ());
	previewWSCheckBox->setChecked (file->previewWS ());
	previewCheckBox->setChecked (file->preview ());
	thumbnailCheckBox->setChecked (file->thumbnail ());
}

void GuiWindow::on_vhistRootOptionalCheckBox_stateChanged(int /* state */) 
{
	m_argfile->setRootIsOptional (vhistRootOptionalCheckBox->isChecked ());
	evaluateRadioButtons();
}

void GuiWindow::on_optionalCheckBox_clicked()
{
	vhist::WSFile* f = m_fihelper->getSelectedFile ();

	f->setOptional (optionalCheckBox->checkState ());
}

// TODO: Looks like evaluteRadioButtons would be a good slot
void GuiWindow::on_oneUserSpecificVhistOutFileRadioButton_clicked() {
	evaluateRadioButtons();
}

void GuiWindow::on_firstOutputFileVhist_clicked() {
	evaluateRadioButtons();
}

void GuiWindow::on_one_EachOutputFileRadioButton_clicked() {
	evaluateRadioButtons();
}

void GuiWindow::on_firstInputFileVhistRootFileRadioButton_clicked() {
	evaluateRadioButtons();
}

void GuiWindow::evaluateRadioButtons() { /* TODO: Still necessary? Answer: Yes, until the radio buttons get their own event handlers. */
	if (userspecifiedPathAndNameRadioButton->isChecked()) {
		vhistRootFileBrowserButton->setEnabled(1);
		vhistRootFileNameText->setEnabled(1);
		vhistRootFilePathText->setEnabled(1);
		vhistRootFileBrowserButton->setEnabled(1);
		pathLabel->setEnabled(1);
		fileLabel->setEnabled(1);
		if (vhistRootFilePathText->text()=="<undefined>") {
			vhistRootFilePathText->setText("");
		}
	} else {
		vhistRootFileBrowserButton->setEnabled(0);
		vhistRootFileNameText->setEnabled(0);
		vhistRootFilePathText->setEnabled(0);
		vhistRootFileBrowserButton->setEnabled(0);
		pathLabel->setEnabled(0);
		fileLabel->setEnabled(0);
	}
	if (oneUserSpecificVhistOutFileRadioButton->isChecked()) {
		vhistStorageFileNameText->setEnabled(1);
		vhistStorageFilePathText->setEnabled(1);
		vhistStorageFileBrowserButton->setEnabled(1);
		file2Label->setEnabled(1);
		path2Label->setEnabled(1);

		if (vhistStorageFilePathText->text()=="<undefined>") {
			vhistStorageFilePathText->setText("");
		}
	} else {
		vhistStorageFileNameText->setEnabled(0);
		vhistStorageFilePathText->setEnabled(0);
		vhistStorageFileBrowserButton->setEnabled(0);
		file2Label->setEnabled(0);
		path2Label->setEnabled(0);
	}
	//!Take first Output File as name stup for VHIST File
	if (firstOutputFileVhist->isChecked()) {
		QListWidgetItem* firstOutputItem=0;
		if (m_row>0) { //!go through the list and search for the item
			for (int i=0; i<m_row; i++) {
				vhist::WSFile* firstoutfile = m_fihelper->getFile (FileListWidget->item(i));

				if (!firstoutfile->inputFile ()) {
					firstOutputItem=FileListWidget->item(i);
					break;
				}
			}
		}
		if (firstOutputItem!=0) {
			vhist::WSFile* firstoutfile = m_fihelper->getFile (firstOutputItem);
			QString completeName= firstoutfile->filename ();
			QDir binDir(completeName);
			QString nameOfFile=binDir.dirName();
			completeName.remove(nameOfFile);
			QString path="";
			path= binDir.absoluteFilePath(completeName);
			vhistStorageFilePathText->setText(path);
			vhistStorageFileNameText->setText(nameOfFile+".vhist");
		} else {
			vhistStorageFilePathText->setText("<undefined>");
			vhistStorageFileNameText->setText("<undefined>");
		}
	}

	if (firstInputFileVhistRootFileRadioButton->isChecked()) {
		QListWidgetItem* firstInputItem=0;
		if (m_row>0) { //!walk though the list and search for the item
			for (int i=0; i<m_row; i++) {
				vhist::WSFile* firstinfile = m_fihelper->getFile (FileListWidget->item(i));

				if (firstinfile->inputFile ()) {
					firstInputItem=FileListWidget->item(i);
					break;
				}
			}
		}
		if (firstInputItem!=0) { /* TODO: Still necessary? */
			vhist::WSFile* firstinfile = m_fihelper->getFile (firstInputItem);
			QString completeName= firstinfile->filename ();
			QDir binDir(completeName);
			QString nameOfFile=binDir.dirName();
			completeName.remove(nameOfFile);
			QString path="";
			//ACHTUNG NACH DRY PRINZIP REFACTORN!!!
			path= binDir.absoluteFilePath(completeName);
			vhistRootFilePathText->setText(path);
			vhistRootFileNameText->setText(nameOfFile.append(".vhist"));
		} else {
			if (!noneCheckBox->isChecked()) {
				vhistRootFilePathText->setText("<undefined>");
				vhistRootFileNameText->setText("<undefined>");
			}
		}
	}
	/* No if for the one for each-case, as this has already been successfully refactored. */
}

void GuiWindow::on_userspecifiedPathAndNameRadioButton_clicked()
{
	evaluateRadioButtons();
}

void GuiWindow::on_noneCheckBox_clicked()
{
	if (noneCheckBox->isChecked()) {
		vhistFileNameAndPathGroupBox->setEnabled(0);
	} else {
		vhistFileNameAndPathGroupBox->setEnabled(1);
	}
}

void GuiWindow::on_fileBrowserButton_clicked() {
	QString fileFileName= QFileDialog::getOpenFileName(this, tr("Select File"),
			" ", tr("Files (*)"));
	if (fileFileName != "") { // Something was actually clicked
		changeFileListItemName(fileFileName);
		guessMimeType ();
		loadIOFileName ();
		if (fileText->text()==".") {
			fileText->setText("");
			pathText->setText("");
			QString empty=" ";
			changeFileListItemName(empty);
			pathText->setText("");
		}
	}
}

void GuiWindow::setHints(QString commentText) {
	comment_label->setText(commentText);
}

QString GuiWindow::returnHints(QString purpose)
{
	bool forFile = (purpose=="file");
	bool visible=false;
	QString commentText="";
	
	if (forFile) {
		commentText.append("# Hints \n");
	}

	if (noneCheckBox->isChecked()) {
		if (forFile) {
			commentText.append("#");
		}
		commentText.append("vhistadd will start a new VHIST file from scratch\n");
	}

	if (m_argfile->rootfile ().trimmed () == "") { 
		if (m_argfile->rootIsOptional ()) { // Option "-J"
			if (forFile)
				commentText.append("s");
			commentText.append("Optional flag is set, vhistadd will continue even if Root file can not be read.\n");
		} else {
			if (forFile)
				commentText.append("#");
			commentText.append("Fatal Error:Root file is undefined.\nSolutions: choose a Root file, set optional flag, start a new VHIST-File from scratch\n");
			visible=true;
		} 
	} 
	if (m_argfile->numOutputVhistfiles () == 0) {
		if (forFile)
			commentText.append("#");
		commentText.append("Fatal Error: Output file undefined \n");
		visible=true;
	} 
	if ((toolNameText->text()=="")|| (toolNameText->text()==" ")) {
		if (forFile)
			commentText.append("#");
		commentText.append("You have not entered a tool (program) name.\n");
		visible=true;
	}
	if ((titleLineEdit->text()=="")|| (titleLineEdit->text()==" ")) {
		if (forFile)
			commentText.append("#");
		commentText.append("There is no title for this workflow step.\n");
		visible=true;
	}
	if ((workflowDescriptionTextEdit->toPlainText()=="")
			||(workflowDescriptionTextEdit->toPlainText()==" ")) {
		if (forFile)
			commentText.append("#");
		commentText.append("There is no description for this workflow step (recommended).\n");
		visible=true;
	}
	if (FileListWidget->count()==0) {
		if (forFile)
			commentText.append("#");
		commentText.append("You have not added any input or output files.\n");
		visible=true;
	}

	comment_groupBox->setVisible(visible);
	comment_label->setVisible(visible);
	return commentText;
}

QString GuiWindow::extractXML1(QString szXML, QString szKey) const
{
	int iKeyLength = szKey.length() + 2; // add "<", ">"
	int iKeyStart = szXML.indexOf("<" + szKey + ">");
	int iKeyEnd = szXML.indexOf("</" + szKey + ">", iKeyStart);
	if ((iKeyStart<0) || (iKeyEnd<0))
		return QString("");
	return szXML.mid(iKeyStart + iKeyLength, iKeyEnd - iKeyStart - iKeyLength);
}

QString GuiWindow::getVhistInfo() const {
	QFile file(":/resources/vhist.xml");
	if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
		return QString("");
	return extractXML1(file.readAll(), "Version");
}

void GuiWindow::set_Quiet_Verbose_pretend()
{
	switch (m_argfile->verbosity ()) {
		case vhist::ArgFile::NORMAL:
			normalRadioButton->setChecked(true);
			break;

		case vhist::ArgFile::QUIET:
			quietRadioButton->setChecked(true);
			break;

		case vhist::ArgFile::VERBOSE:
			verboseRadioButton->setChecked(true);
			break;
	}

	if (m_argfile->pretend ()) {
		pretendCheckBox->setChecked(true);
	}
}

void GuiWindow::applyToThisFile_clicked()
{ 
	on_pathText_editingFinished();
}

void GuiWindow::guessMimeType () 
{
	QString type = Mimetypes::guessMimeType (fileText->text ());

	if (type != "") {
		fileTypeSelector->setCurrentIndex(fileTypeSelector->findText(type, Qt::MatchExactly));
		fileTypeLineEdit->setEnabled(0);
		fileTypeLineEdit->setText(type);
	}
}

void GuiWindow::loadIOFileName ()
{
	QString filename = m_fihelper->getSelectedFile ()->filename ();
	QFileInfo fi (filename);
	QString nameOfFile= fi.fileName();
	QString trimmedpath = fi.absolutePath ();

	//TODO: refactoring-> extract method
	if (trimmedpath != "")
		if (!(trimmedpath.endsWith(("/")) || trimmedpath.endsWith("\\")))
			trimmedpath.append("/");

	//                                   ----------
	pathText->setText(trimmedpath);
	fileText->setText(nameOfFile);
}

void GuiWindow::open_cmdfile()
{
    QString filename = QFileDialog::getOpenFileName(this);
    openArgFile(filename);
}

/* This function is used once each by GuiWindow and MainWindow! */
void GuiWindow::openArgFile(const QString& filename)
{
	if (filename == "") { /* Nothing to open */
		return;
	}
        m_parent->addRecentFile(filename);

	/* Parse */
	QFile f (filename);
	f.open (QIODevice::ReadOnly);
	QByteArray arr = f.readAll ();
	f.close ();
	QString content (arr);

	vhist::ArgfileReader* reader = new vhist::ArgfileReader ();

	delete m_argfile;

	m_argfile = new vhist::ArgFile (reader->readArgfile (content));

	delete reader;

	/* Update GUI */

	/* Main tab */
	titleLineEdit->setText (m_argfile->title ());
	workflowDescriptionTextEdit->setText (m_argfile->description ());
	toolNameText->setText (m_argfile->tool ());
	toolPathText->setText (m_argfile->toolpath ());

	QStringList uakeys = m_argfile->userAttributeKeys ();

	generalAttributesTable->clear ();
//	generalAttributesTable->setRowCount (uakeys.size ()+1);

	for (int i = 0;i < uakeys.size ();i++) {
		QString key = uakeys.at (i);
		QString value = m_argfile->userAttribute (key);

		QTableWidgetItem* itemLeft = new QTableWidgetItem(key,0);
		generalAttributesTable->setItem(i, 0, itemLeft);
		QTableWidgetItem* itemRight = new QTableWidgetItem(value,0);
		generalAttributesTable->setItem(i, 1, itemRight);
	}

	switch (m_argfile->verbosity ()) {
		case vhist::ArgFile::QUIET:
			quietRadioButton->setChecked (true);
			break;

		case vhist::ArgFile::NORMAL:
			normalRadioButton->setChecked (true);
			break;

		case vhist::ArgFile::VERBOSE:
			verboseRadioButton->setChecked (true);
			break;
	}

	pretendCheckBox->setChecked (m_argfile->pretend ());

	/* In & Out Tab */
	FileListWidget->clear ();
	m_row = 0;

	for (int i = 0;i < m_argfile->numFiles ();i++) {
		vhist::WSFile* f = m_argfile->getFile (i);

		addFile (f);
	}

	updateNumberOfFilesLabel();

	/* Export tab
	 * Out of order, as the files tab is coded here with a few returns. 
	 * We just need to guess if relative paths are used.
	 * We assume that no relative paths are used, until we find one. */
	bool relative = false;

	QFileInfo fi (m_argfile->rootfile ());
	if (fi.isRelative ()) {
		relative = true;
	}

	for (int i = 0;i < m_argfile->numOutputVhistfiles ();i++) {
		fi = QFileInfo (m_argfile->outputVhistfile (i));

		if (fi.isRelative ()) {
			relative = true;
			break;
		}
	}

	for (int i = 0;i < m_argfile->numFiles ();i++) {
		fi = QFileInfo (m_argfile->getFile (i)->filepath ());

		if (fi.isRelative ()) {
			relative = true;
			break;
		}
	}

	relativePathsCheckBox->setChecked (relative);

	/* Files tab
	 * This is unfortunately not much better then an educated guess.
	 * Either we are appending, a case where we just set the root file
	 * and append mode, or we'll have to decide for the root and the output
	 * file(s) individually. */
	if (m_argfile->numOutputVhistfiles () == 1) {
		if (m_argfile->rootfile () == m_argfile->outputVhistfile (0)) { /* Ok, append mode */
			vhistAppendCheckBox->setChecked (true);
			QFileInfo fi (m_argfile->rootfile ());
			vhistRootFileNameText->setText (fi.fileName ());
			vhistRootFilePathText->setText (fi.absolutePath ());
			return;
		}
	}

	vhistAppendCheckBox->setChecked (false);

	setAppendMode ();

	/* If we got here, start with the root file. */
	userspecifiedPathAndNameRadioButton->setChecked (true); /* TODO: Could be first input mode file, too */
	fi = QFileInfo (m_argfile->rootfile ());		
	vhistRootFileNameText->setText (fi.fileName ());
	vhistRootFilePathText->setText (fi.absolutePath ());

	/* Then the output file */
	if (m_argfile->numOutputVhistfiles () > 1) { /* One for each */
		one_EachOutputFileRadioButton->setChecked (true);
		evaluateRadioButtons ();
		return;
	}

	oneUserSpecificVhistOutFileRadioButton->setChecked (true); /* TODO: Could be first input mode file, too */
	fi = QFileInfo (m_argfile->outputVhistfile (0));		
	vhistStorageFileNameText->setText (fi.fileName ());
	vhistStorageFilePathText->setText (fi.absolutePath ());

	evaluateRadioButtons ();
}

void GuiWindow::go_mainAct()
{
	tabWidget->setCurrentIndex(0);
}

void GuiWindow::go_inoutfilesAct()
{
	tabWidget->setCurrentIndex(1);
}

void GuiWindow::go_vhistfilesAct()
{
	tabWidget->setCurrentIndex(2);
}
void GuiWindow::go_exportAct()
{
	tabWidget->setCurrentIndex(3);
}

QString GuiWindow::adjustRelativity (QString path) /* TODO: 2nd reason for a thin layer between argfile and guiwindow.*/
{
	if (!relativePathsCheckBox->isChecked ()) { /* No adjusting wanted, no adjusting done */
		return (path);
	}

	if (m_argfile->numOutputVhistfiles () == 0) { /* No output to adjust too */
		return (path);
	}

	QFileInfo sf;
	
	sf = QFileInfo (m_argfile->outputVhistfile (0));
	QDir stordir = sf.dir ();

	if (stordir.isRelative ()) { /* Itself relative, so don't anything. TODO: Or do something? */
		return (path);
	}

	return (stordir.relativeFilePath (path));
}

void GuiWindow::on_relativePathsCheckBox_stateChanged (int /* value */)
{
        commandLine->textBrowser()->setHtml(buildCommand_File(false));
}

#ifdef VDEBUG

#include <QXmlStreamWriter>

QString GuiWindow::toolOperationToString ()
{
	if (normalRadioButton->isChecked ()) {
		return ("normal");
	}
	if (quietRadioButton->isChecked ()) {
		return ("quiet");
	}
	if (verboseRadioButton->isChecked ()) {
		return ("verbose");
	}

	return ("bug");
}

QString GuiWindow::inputNameSource ()
{
	if (firstInputFileVhistRootFileRadioButton->isChecked ()) {
		return ("first");
	}
	if (userspecifiedPathAndNameRadioButton->isChecked ()) {
		return ("user");
	}

	return ("bug");
}

QString GuiWindow::outputNameSource ()
{
	if (firstOutputFileVhist->isChecked ()) {
		return ("first");
	}
	if (one_EachOutputFileRadioButton->isChecked ()) {
		return ("oneforeach");
	}
	if (oneUserSpecificVhistOutFileRadioButton->isChecked ()) {
		return ("user");
	}

	return ("bug");
}

void GuiWindow::go_export ()
{
	QString fname = QFileDialog::getSaveFileName (this);

	if (fname == "") {
		return;
	}

	/* Prepare */
	QFile file (fname);

	file.open (QIODevice::WriteOnly);

	QXmlStreamWriter xsw (&file);

	xsw.writeStartDocument ();
	xsw.writeStartElement ("state");

	/* Main tab */
	xsw.writeStartElement ("main");

	xsw.writeTextElement ("workflowtitle", titleLineEdit->text ());
	xsw.writeTextElement ("workflowdescription", workflowDescriptionTextEdit->toPlainText ());
	xsw.writeTextElement ("toolname", toolNameText->text ());
	xsw.writeTextElement ("toolpath", toolPathText->text ());
	xsw.writeTextElement ("toolversion", versionLineEdit->text ());
	xsw.writeStartElement ("generalattributes"); 
	xsw.writeAttribute ("count", QString::number (generalAttributesTable->rowCount () - 1));
	for (int i = 0;i < generalAttributesTable->rowCount()-1;i++) {
		xsw.writeStartElement ("attribute");
		xsw.writeAttribute ("key", generalAttributesTable->item (i, 0)->text ());
		xsw.writeAttribute ("value", generalAttributesTable->item (i, 1)->text ());
		xsw.writeEndElement ();
	}
	xsw.writeEndElement ();
	xsw.writeTextElement ("tooloperation", toolOperationToString ());
	xsw.writeTextElement ("pretend", QString (pretendCheckBox->isChecked ()?"true":"false"));

	xsw.writeEndElement ();

	/* In&Out files tab
	 * This tab is a bit special.
	 * If you have 2 in/out files or more, the data of all files
	 * is no longer visible at one instant.
	 * This means that here the data is taken out of CommandFile. */
	xsw.writeStartElement ("inout");
	xsw.writeAttribute ("count", QString::number (m_argfile->numFiles ()));

	for (int u = 0;u < m_argfile->numFiles ();u++) {
		vhist::WSFile* file = m_argfile->getFile (u);

		xsw.writeStartElement ("file");
		xsw.writeTextElement ("name", file->filename ()); // TODO: Not 100% fit to GUI
		xsw.writeTextElement ("description", file->description ());
		xsw.writeTextElement ("inputfile", QString (file->inputFile ()?"true":"false"));
		xsw.writeTextElement ("mime", file->filetype ());
		xsw.writeTextElement ("embed", QString (file->embed ()?"true":"false"));
		xsw.writeTextElement ("md5", QString (file->autoMD5 ()?"true":"false"));
		xsw.writeTextElement ("fileoptional", QString (file->optional ()?"true":"false"));
		/* Preview WS, Preview and Thumbnail need implementation first  */
		xsw.writeStartElement ("attributes");
		if (file->userAttributeKeys ().size () == 0) { /* TODO: Seems to be a slight problem in File. */
			xsw.writeAttribute ("count", "0");
		} else {
			xsw.writeAttribute ("count", QString::number (file->userAttributeKeys ().size ()-1));
		}
		for (int i = 0;i < file->userAttributeKeys ().size ()-1;i++) {
			QString key = file->userAttributeKeys ().at (i);
			QString value = file->userAttribute (key);
			xsw.writeStartElement ("attribute");
			xsw.writeAttribute ("key", key);
			xsw.writeAttribute ("value", value);
			xsw.writeEndElement ();
		}
		xsw.writeEndElement ();
		xsw.writeEndElement ();
	}

	xsw.writeEndElement ();

	/* VHIST files tab */
	xsw.writeStartElement ("vhistfiles");
	xsw.writeTextElement ("append", QString (vhistAppendCheckBox->isChecked ()?"true":"false"));
	xsw.writeTextElement ("none", QString (noneCheckBox->isChecked ()?"true":"false"));
	xsw.writeTextElement ("optional", QString (vhistRootOptionalCheckBox->isChecked ()?"true":"false"));
	xsw.writeTextElement ("inputnamesource", inputNameSource ());
	xsw.writeTextElement ("inputname", vhistRootFileNameText->text ());
	xsw.writeTextElement ("inputpath", vhistRootFilePathText->text ());
	xsw.writeTextElement ("outputnamesource", outputNameSource ());
	xsw.writeTextElement ("outputname", vhistStorageFileNameText->text ());
	xsw.writeTextElement ("outputpath", vhistStorageFilePathText->text ());
	xsw.writeEndElement ();

	/* Export tab */
        commandLine->textBrowser()->setHtml(buildCommand_File(false)); /* Update the commandLine-Widget, as it might be out of date otherwhise. */
        xsw.writeTextElement ("export", commandLine->textBrowser()->toPlainText ());

	/* Done */
	xsw.writeEndElement ();
	xsw.writeEndDocument ();
	file.close ();
}
#endif
