/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

Author: Michael May
$Id: mainwindow.cpp 2794 2013-05-06 11:26:53Z michael $
*/

#include <QCloseEvent>
#include <QMessageBox>
#include <QtDebug>
#include <QMenuBar>

#include "mainwindow.h"
#include "guiwindow.h"
#include "../shared/utilities.h"

MainWindow::MainWindow (const QString& filename)
{
        m_win = new GuiWindow(getApplicationDir(), this);
        setCentralWidget(m_win);
        createActions();
        createMenus();
        setWindowTitle(tr("Create VHIST Arg File"));
        openArgFile(filename);
}


MainWindow::~MainWindow()
{
}


void MainWindow::createActions()
{
	openAct = new QAction(tr("&Open..."), this);
	openAct->setShortcut(tr("Ctrl+O"));
	openAct->setStatusTip(tr("Open a Cmd file"));
	connect(openAct, SIGNAL(triggered()), m_win, SLOT(open_cmdfile()));
	
        saveAct = new QAction(tr("&Save As..."), this);
	saveAct->setShortcut(tr("Ctrl+S"));
	saveAct->setStatusTip(tr("Save a Cmd file"));
	connect(saveAct, SIGNAL(triggered()), m_win, SLOT(save_cmdfile()));

    cutAct = new QAction(tr("Cu&t"), this);
    cutAct->setShortcut(tr("Ctrl+X"));
    cutAct->setStatusTip(tr("Cut the current selection's contents to the "
                            "clipboard"));
    //connect(cutAct, SIGNAL(triggered()), m_win, SLOT(cut()));
        
    copyAct = new QAction(tr("&Copy"), this);
    copyAct->setShortcut(tr("Ctrl+C"));
    copyAct->setStatusTip(tr("Copy the current selection's contents to the "
                             "clipboard"));
    //connect(copyAct, SIGNAL(triggered()), m_win, SLOT(copy()));
        
    pasteAct = new QAction(tr("&Paste"), this);
    pasteAct->setShortcut(tr("Ctrl+V"));
    pasteAct->setStatusTip(tr("Paste the clipboard's contents into the current "
                              "selection"));
    //connect(pasteAct, SIGNAL(triggered()), m_win, SLOT(paste()));     

    go_mainAct = new QAction(tr("&Main Tab"), this);
    go_mainAct->setShortcut(tr("Ctrl+1"));
    go_mainAct->setStatusTip(tr("open the Main tab"));
    connect(go_mainAct, SIGNAL(triggered()), m_win, SLOT(go_mainAct()));        

    go_inoutfilesAct = new QAction(tr("&In/Out Files Tab"), this);
    go_inoutfilesAct->setShortcut(tr("Ctrl+2"));
    go_inoutfilesAct->setStatusTip(tr("open the In/Out files tab"));
    connect(go_inoutfilesAct, SIGNAL(triggered()), m_win, SLOT(go_inoutfilesAct()));
        
    go_vhistfilesAct = new QAction(tr("&VHIST files Tab"), this);
    go_vhistfilesAct->setShortcut(tr("Ctrl+3"));
    go_vhistfilesAct->setStatusTip(tr("open the VHIST files tab"));
    connect(go_vhistfilesAct, SIGNAL(triggered()), m_win, SLOT(go_vhistfilesAct()));
        
    go_exportAct = new QAction(tr("&Export Tab"), this);
    go_exportAct->setShortcut(tr("Ctrl+4"));
    go_exportAct->setStatusTip(tr("open the Export tab"));
    connect(go_exportAct, SIGNAL(triggered()), m_win, SLOT(go_exportAct()));

#ifdef VDEBUG
	exportAct = new QAction(tr("Export contents"), this);
	connect(exportAct, SIGNAL(triggered()), m_win, SLOT(go_export ()));
#endif
	
    helpAct = new QAction(tr("Help Contents"), this);
    helpAct->setShortcut(tr("F1"));
	connect(helpAct, SIGNAL(triggered()), m_win, SLOT(openVHISTEZHelp()));

	//aboutAct = new QAction(tr("About vhistez"), this);
        //connect(aboutAct, SIGNAL(triggered()), m_win, SLOT(about()));
//	preferencesAct=new QAction(tr("Preferences"), this);
//	connect(preferencesAct, SIGNAL(triggered()), m_win, SLOT(preferences()));
}

void MainWindow::createMenus()
{
	fileMenu = menuBar()->addMenu(tr("&File"));
	fileMenu->addAction(openAct);
	fileMenu->addAction(saveAct);   
	
	editMenu = menuBar()->addMenu(tr("&Edit"));
	editMenu->addAction(cutAct);
	editMenu->addAction(copyAct);
	editMenu->addAction(pasteAct);
	
	viewMenu = menuBar()->addMenu(tr("&View"));
	viewMenu->addAction(go_mainAct);
	viewMenu->addAction(go_inoutfilesAct);
	viewMenu->addAction(go_vhistfilesAct);  
	viewMenu->addAction(go_exportAct);      
#ifdef VDEBUG
	viewMenu->addAction(exportAct);
#endif
    
    // initialize shared menu items
    
    helpMenu = menuBar()->addMenu(tr("&Help"));
    helpMenu->addAction(helpAct);
    helpMenu->addSeparator();
    //helpMenu->addAction(aboutAct);
//    helpMenu->addAction(preferencesAct);
    
    initSharedMenu("Vhistez/RecentFiles");
    connect(this, SIGNAL(recentFileSelected(QString)), this,
            SLOT(openArgFile(QString)));
}

void MainWindow::closeEvent(QCloseEvent *)
{ 
}

void MainWindow::openArgFile(const QString& filename)
{
    m_win->openArgFile(filename);
}


//! returns the directory in which vhistez resides
QString MainWindow::getApplicationDir() const
{
	QString szApplicationDir=QCoreApplication::applicationDirPath();
    
#ifdef Q_OS_MAC         
	// fix for debugging
	if (szApplicationDir.contains(QString("/vhistez/"))){
		int iStartDPath = szApplicationDir.indexOf(QString("/vhistez/"));
		QString szT = szApplicationDir.left(iStartDPath);
		
		szApplicationDir = szT + QString("/bin_macx");
        }
	if (szApplicationDir.endsWith(QString("/Contents/MacOS")))
		szApplicationDir = szApplicationDir.append("/../../..");
#endif // end of Mac specific code      
	
	return szApplicationDir;
}


ManagedWindow* MainWindow::createVHISTezWindow(
        const QString& filename,
        const QVariantHash&)
{
    ManagedWindow* win = new MainWindow(filename);
    return win;
}


QString MainWindow::windowName()
{
    return "Create VHIST Arg-File";
}


QString MainWindow::windowShortcut()
{
    return "ctrl+E";
}
