/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: quote.cpp 1597 2009-02-27 13:13:25Z ahuesgen $
 */

#include "quote.h"

// cp.: http://www.mpi-inf.mpg.de/~uwe/lehre/unixffb/quoting-guide.html
// for a detailed discussion on how quoting in a shell works


/**
 * Quotes the string using the unix shell quoting rules.
 *
 * @param[in] str The string which should be quoted.
 * @param[in] supportVariables If \c true, the $-signs of shell variables
 * are not escaped.
 * 
 * @return The quoted string.
 */
QString quoteString(const QString& str, bool supportVariables)
{
    // use " to quote and escape special characters ", \, ` (and $)
    QString result = str;
    // escape backslashes first
    result.replace('\\', "\\\\");
    result.replace('\"', "\\\"");
    result.replace('`', "\\`");
    if (!supportVariables)
        result.replace('$', "\\$");
    return QString("\"" + result + "\"");
}


/**
 * Quotes all strings in the list using the unix shell quoting rules.
 * 
 * @param[in] list A list of strings which should be quoted.
 * @param[in] supportVariables If \c true, the $-signs of shell variables
 * are not escaped.
 * 
 * @return The list of quoted strings.
 */
QStringList quoteStringList(const QStringList& list, bool supportVariables)
{
    QStringList result;
    for (int i = 0; i < list.size(); i++)
        result.append(quoteString(list.at(i), supportVariables));
    return result;
}

