/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: queryhighlighter.cpp 2167 2011-05-30 09:06:57Z ahuesgen $
 */

#include "queryhighlighter.h"
#include "vhistfinder.h"


int findError(const vhist::QueryErrors& errors)
/**
 * Finds the index of the first error in \a errors. If \a errors does not
 * contain any error messages, -1 is returned.
 */
{
    for (int i = 0; i < errors.length(); ++i)
    {
        const vhist::QueryError& e = errors.at(i);
        if (e.isError())
            return i;
    }
    return -1;
}


int findWarning(const vhist::QueryErrors& errors)
/**
 * Finds the index of the first warning in \a errors. If \a errors does not
 * contain any error messages, -1 is returned.
 */
{
    for (int i = 0; i < errors.length(); ++i)
    {
        const vhist::QueryError& e = errors.at(i);
        if (e.isWarning())
            return i;
    }
    return -1;
}


QueryHighlighter::QueryHighlighter() : QSyntaxHighlighter((QObject*) NULL)
{
}


QueryHighlighter::~QueryHighlighter()
{
}


void QueryHighlighter::hightlightText(const QString& text,
        const QString& errorMessage, int pos, int length, const QColor& color)
{
    // make sure error marker is at least one character long
    int l = qMax(1, length);

    QTextCharFormat formatError;
    formatError.setUnderlineStyle(QTextCharFormat::WaveUnderline);
    formatError.setUnderlineColor(color);
    formatError.setToolTip(errorMessage);

    QTextCharFormat formatNormal;
    formatNormal.setUnderlineStyle(QTextCharFormat::NoUnderline);
    formatNormal.setUnderlineColor(QColor(0, 0, 0));
    formatNormal.setToolTip("");

    setFormat(0, pos, formatNormal);
    setFormat(pos, pos+l, formatError);
    setFormat(pos+l, text.length(), formatNormal);
}


void QueryHighlighter::highlightBlock(const QString& text)
{
    VhistFinder finder;
    vhist::QueryErrors errors = finder.checkQuery(text);

    int firstError = findError(errors);
    int firstWarning = findWarning(errors);

    if (firstError >= 0)
    {
        const vhist::QueryError& e = errors.at(firstError);
        int errorPos = e.position();
        int errorLength = e.length();
        QString errorMessage = e.errorMessage();
        QColor errorColor(255, 0, 0);
        hightlightText(text, errorMessage, errorPos, errorLength, errorColor);
        emit errorOccurred(errorMessage, true);
    }
    else if (firstWarning >= 0)
    {
        const vhist::QueryError& e = errors.at(firstWarning);
        int errorPos = e.position();
        int errorLength = e.length();
        QString errorMessage = e.errorMessage();
        QColor errorColor(50, 60, 230);
        hightlightText(text, errorMessage, errorPos, errorLength, errorColor);
        emit errorOccurred(errorMessage, false);
    }
    else
    {
        QTextCharFormat formatNormal;
        formatNormal.setUnderlineStyle(QTextCharFormat::NoUnderline);
        setFormat(0, text.length(), formatNormal);
        emit errorOccurred("", false);
    }
}
