/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: vhistfinder.cpp 2167 2011-05-30 09:06:57Z ahuesgen $
 */

#include "vhistfinder.h"

#include <unvhistlib/vhistparser.h>
#include <unvhistlib/summarycreator.h>

#include <vhistfindlib/vhistfindquery.h>

#include <QDir>
#include <QFileInfo>


EvaluationResult::EvaluationResult() : matches(false)
{
}

VhistFinder::VhistFinder() : m_nextFile(0), m_invertResult(false)
{
}


VhistFinder::~VhistFinder()
{
}


vhist::QueryErrors VhistFinder::checkQuery(const QString& query)
{
    vhist::VhistFindQuery validator(query);
    return validator.getErrors();
}


vhist::QueryErrors VhistFinder::init(const QString &findQuery,
        const QString &dir, bool invertResult)
{
    m_findQuery = findQuery;
    m_dir = dir;
    m_files = gatherVhistFiles(m_dir, true);
    m_nextFile = 0;
    m_invertResult = invertResult;

    // create a parse tree for the search expression
    m_query.setQuery(m_findQuery);
    return m_query.getErrors();
}


bool VhistFinder::initialized()
{
    return m_query.isValid();
}


int VhistFinder::nextFileNumber()
{
    return m_nextFile;
}


int VhistFinder::numberOfFiles()
{
    return m_files.length();
}


EvaluationResult VhistFinder::findNext()
{
    EvaluationResult result;

    // invalid call?
    if (nextFileNumber() >= numberOfFiles())
        return result;

    vhist::VhistParser parser(vhist::MODE_ROBUST);

    result.filename = m_files.at(m_nextFile);

    if (parser.open(result.filename))
    {
        vhist::SummaryCreator creator;
        vhist::Summaries summaries = creator.createSummaries(parser);

        result.matches = m_query.matches(summaries, &result.associations);
        if (m_invertResult)
            result.matches = !result.matches;
    }
    m_nextFile++;
    return result;
}


QStringList VhistFinder::gatherVhistFiles(const QString& dir,
        bool followSymlinks)
{
    QStringList results;
    QDir directory(dir);

    QDir::Filters filters = QDir::NoDotAndDotDot | QDir::Readable
            | QDir::AllDirs | QDir::Files;
    if (followSymlinks) filters |= QDir::NoSymLinks;

    QStringList patterns; patterns << "*.vhist" << "*.pdf";

    QStringList filesAndDirs = directory.entryList(patterns, filters);

    for (int i = 0; i < filesAndDirs.length(); ++i)
    {
        QFileInfo info(directory.filePath(filesAndDirs.at(i)));
        if (info.isDir())
            results.append(gatherVhistFiles(info.filePath(), followSymlinks));
        else
            results.append(info.filePath());
    }
    return results;
}
