/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: vhistfindwindow.cpp 2458 2011-08-03 09:21:52Z ahuesgen $
 */

#include "vhistfindhellomessage.h"
#include "vhistfindwindow.h"
#include "vhistfinder.h"
#include "../shared/constants.h"
#include "../shared/helpviewer.h"

#include "../shared/windowlist.h"

#include <QFileInfo>
#include <QDir>
#include <QFileDialog>
#include <QCoreApplication>
#include <QListWidget>

#include <QtDebug>



VhistFindWindow::VhistFindWindow(const QString& query, const QString& dir)
{
    setupUi(this);
    initSharedMenu();

    m_resultsBrowser->textBrowser()->setAcceptRichText(true);
    m_resultsBrowser->textBrowser()->setOpenLinks(false);
    m_resultsBrowser->textBrowser()->setOpenExternalLinks(false);
    m_resultsBrowser->setShortcutsActive(true);
    QString helloMessage = getVhistFindHelloMessage().arg(VHISTZARD_FULLNAME)
            .arg(VHISTZARD_COPYRIGHT_DATE);
    m_resultsBrowser->textBrowser()->setHtml(helloMessage);
    m_dirCombo->setDuplicatesEnabled(false);

    connect(m_findButton, SIGNAL(clicked()), this, SLOT(findClicked()));
    connect(m_browseDirsButton, SIGNAL(clicked()), this, SLOT(browseForDir()));
    connect(m_resultsBrowser->textBrowser(), SIGNAL(anchorClicked(QUrl)), this,
            SLOT(openVhistFile(QUrl)));
    connect(m_queryEditor, SIGNAL(enterPressed()), this,
            SLOT(handleEnterPress()));
    connect(m_helpContent, SIGNAL(triggered()), this, SLOT(showHelp()));
    connect(this, SIGNAL(processMoreFiles()), this, SLOT(processSomeFiles()),
            Qt::QueuedConnection); // important, keeps UI responsive
    connect(m_queryEditor, SIGNAL(errorOccured(QString,bool)), this,
            SLOT(handleError(QString,bool)));

    if (query != "")
        m_queryEditor->setQuery(query);
    if (dir != "")
        m_dirCombo->setEditText(dir);
}


VhistFindWindow::~VhistFindWindow()
{

}


void VhistFindWindow::browseForDir()
{
    QString currentDir = m_dirCombo->currentText();
    QString newDir = QFileDialog::getExistingDirectory(this,
            "Select a Directory", currentDir);
    if (newDir != "")
    {
        m_dirCombo->addItem(newDir);
        m_dirCombo->setEditText(newDir);
    }
}


void VhistFindWindow::handleError(QString, bool isError)
{
    m_findButton->setEnabled(!isError);
}


void VhistFindWindow::handleEnterPress()
{
    if (m_findButton->isEnabled())
        findClicked();
}


void VhistFindWindow::findClicked()
{
    QString findQuery = m_queryEditor->query();
    QString dir = m_dirCombo->currentText();
    bool invertResult = m_invertQueryBox->isChecked();

    m_dirCombo->addItem(dir);
    m_statusLabel->setText("initializing query and browsing directory...");
    m_resultsBrowser->textBrowser()->clear();
    repaint(); // ah: does not seem to work with every style

    // start checking files (using timer hack to keep gui responsive)
    vhist::QueryErrors errors = m_finder.init(findQuery, dir, invertResult);
    if (!vhist::containsErrors(errors))
        emit processMoreFiles(); // Queued connection, UI has a chance to update
}


void VhistFindWindow::processSomeFiles()
{
    //qDebug() << "beginning of processSomeFiles" << QTime::currentTime();
    if (!m_finder.initialized())
        return;

    QDateTime time = QDateTime::currentDateTime();
    bool more = true;
    while (m_finder.nextFileNumber() < m_finder.numberOfFiles() && more)
    {
        EvaluationResult result = m_finder.findNext();
        handleResult(result, m_finder.nextFileNumber(), m_finder.numberOfFiles());

        // process files for 100 milliseconds
        more = time.msecsTo(QDateTime::currentDateTime()) < 100;
    }

    if (m_finder.nextFileNumber() < m_finder.numberOfFiles())
        emit processMoreFiles(); // Queued connection, UI has a chance to update...
}


void VhistFindWindow::openVhistFile(const QUrl& link)
{
    QString szLink = link.toString();
    if (szLink.startsWith(tr("doc://")))
    {
        // it is a "pseudo-URL" referring to our documentation
        szLink.remove(0, 6);
        QString szSec;
        int iHash = szLink.indexOf(tr("#"));
        if (iHash > 0)
        {
            szSec = szLink.mid(iHash + 1, szLink.size() - iHash - 1);
            szLink.remove(iHash, szLink.size() - iHash);
        }
        HelpViewer::globalOpenPage(szLink, szSec);
    }
    else
    {
        WindowList* wlist = WindowList::globalInstance();
        if (wlist)
        {
            if (link.scheme() == "vhistfile")
                wlist->openFile(link.path());
        }
    }
}


void VhistFindWindow::handleResult(EvaluationResult& result,
        int testedSoFar, int total)
{
    if (result.matches)
    {
        QString filename = result.filename;
        QHash<QString, QString> variables = result.associations;
        QList<QString> keys = variables.uniqueKeys();

        QString htmlCode;
        htmlCode = QString("<a href=\"vhistfile://%1\">%1</a>\n").arg(filename);

        for (int i = 0; i < keys.length(); ++i)
        {
            QString key = keys.at(i);
            QString value = variables.value(key);
            QString listItem = QString("<div><b>%1</b> = %2</div>\n").arg(key).arg(value);
            htmlCode += listItem;
        }
        m_resultsBrowser->textBrowser()->append(htmlCode);
    }

    //qDebug() << m_resultsBrowser->toHtml();
    m_statusLabel->setText(QString("%1/%2 files checked").arg(testedSoFar).arg(total));
}


void VhistFindWindow::showHelp()
{
    //qDebug("VhistFindWindow::showHelp()");
    HelpViewer::globalOpenPage("vhistfind.html", QString());
}


ManagedWindow* VhistFindWindow::createVHISTFindWindow(
        const QString&, const QVariantHash& parameters)
{
    QString query = "";
    QString dir = "";

    if (parameters.contains("Query"))
        query = parameters.value("Query").toString();
    if (parameters.contains("Directory"))
        dir = parameters.value("Directory").toString();

    ManagedWindow* win = new VhistFindWindow(query, dir);
    return win;
}


QString VhistFindWindow::windowName()
{
    return "Find VHIST Files";
}


QString VhistFindWindow::windowShortcut()
{
    return "ctrl+D";
}
