/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: vhistaddthread.cpp 2794 2013-05-06 11:26:53Z michael $
 */

#include "vhistaddthread.h"

#include <QString>
#include <QStringList>
#include <QProcess>
#include <QDir>
#include <QRegExp>

#include <QtDebug>

#include "../shared/utilities.h"

void VhistaddThread::init(const QString& argfile, const QString& workingDir)
{
    m_argfile = argfile;
    m_workingDir = workingDir;
}


QString VhistaddThread::parseOutput(const QString& text)
{
    QString result("");
    QString indent("&nbsp;&nbsp;&nbsp;&nbsp;");
    // parse output
    QStringList lines = text.split(QRegExp("(\r\n)|[\n\r]"));
    QStringList::iterator i = lines.begin();
    while (i != lines.end())
    {
        QString curLine(*i);
        ++i;
        if (curLine.startsWith("***ERROR"))
            curLine = makeRed(curLine);
        else if (curLine.startsWith("***WARNING"))
            curLine = makeOrange(curLine);
        else if (curLine.startsWith("        "))
            curLine = indent + indent + escapeHtml(curLine);
        else if (curLine.startsWith("    "))
            curLine = indent + escapeHtml(curLine);
        
        // only output last line if it is not empty
        if (i != lines.end() || curLine != QString(""))
            result += curLine  + escapeHtml("\n");
    }
    return result;
}


void VhistaddThread::run()
{
    runProcess();
    emit vhistaddFinished();
}


void VhistaddThread::runProcess()
{
    // create process
    QProcess proc;
    
    // init output buffer
    QString outStr = "";
    
    // get application and binary directories
    QDir vhistDir(getApplicationDir());
#ifdef VHAVE_VHISTLIB
    QDir vhistMacDir(getBinaryDir() + "/vhist/bin/");
#else
    QDir vhistMacDir(getBinaryDir() + "/../vhist-src/bin/");
#endif
    // generate vhist executable name
    QString executable("");
    QStringList args;
    
    #ifdef Q_OS_MAC
    // search in mac os x bundle
    if (vhistMacDir.exists("vhistadd.py"))
    {
        executable = "python";
        args << vhistMacDir.filePath("vhistadd.py");
    }
    else
    #endif
    #ifdef Q_OS_WIN32
    if (vhistDir.exists("vhistadd.exe"))
        executable = vhistDir.filePath("vhistadd.exe");
    else
    #endif
    if (vhistDir.exists("vhistadd.py"))
    {
        #ifdef Q_OS_WIN32
        executable = "python.exe";
        #else
        executable = "python";
        #endif
        args << vhistDir.filePath("vhistadd.py");
    }
    else
    {
        outStr += makeRed(QObject::tr("could not find any vhistadd executable"));
        emit outputUpdated(outStr);
        return;
    }
    //qDebug() << "vhistadd version: " << vhistadd;
    
    // init status messages
    QString cwdMessage(QObject::tr("Setting working directory to \"%1\"...\n"));
    cwdMessage = cwdMessage.arg(m_workingDir);

    QString runMessage(QObject::tr("running vhistadd...\n"));
    
    //QString call("%1 -c \"%2\"");
    //call = call.arg(vhistadd, m_argfile);
    //qDebug() << call;
    
    QString errMessage(QObject::tr("Programm exited with error code: %1\n"));
    
    outStr += makeBold(cwdMessage);
    proc.setWorkingDirectory(m_workingDir);
    outStr += makeBold(runMessage);
    //outStr += escapeHtml(call + "\n\n");
    
    // create vhistadd call
    args << "-c" << m_argfile;
    proc.setProcessChannelMode(QProcess::MergedChannels);

    // start process
    proc.start(executable, args);
    
    // print message on error
    if (!proc.waitForStarted())
    {
        if (proc.error() == QProcess::FailedToStart)
        {
            // generate error message
            QString mail("<a href=\"mailto:vhist@nf.mpg.de\">"
                    "vhist@nf.mpg.de</a>");
            outStr += makeRed(QObject::tr("vhistadd could not be started "
                    "and failed with errorcode: %1. "
                    "Please contact ").arg(proc.error()));
            outStr += mail;
            // send output, abort and return
            emit outputUpdated(outStr);
            return;
        }
    }
    
    emit outputUpdated(outStr);
    while (proc.state() != QProcess::NotRunning)
    {
        if (proc.waitForReadyRead(1000))
        {
            // get new data
            QString result = QString(proc.readAllStandardOutput());
            outStr += parseOutput(result);
            emit outputUpdated(outStr);
        }
    }
    
    QString rest = QString(proc.readAllStandardOutput());
    outStr += parseOutput(rest);
    
    outStr += makeBold(errMessage.arg(proc.exitCode()));
    emit outputUpdated(outStr);
}

