/* 
This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

 $Id: vhistitwindow.cpp 2459 2011-08-03 11:09:53Z ahuesgen $
 */

#include "vhistitwindow.h"
#include "../shared/utilities.h"
#include "../shared/windowlist.h"
#include "../shared/helpviewer.h"
#include "../shared/constants.h"
#include "../vhistez/mainwindow.h"

#include <QFileInfo>
#include <QDir>
#include <QDialog>
#include <QFileDialog>
#include <QDragEnterEvent>
#include <QMimeData>
#include <QtDebug>
#include <QProcess>
#include <QMessageBox>
#include <QLabel>
#include <QScrollBar>
#include <QCloseEvent>
#include <QMessageBox>
#include <QDesktopServices>

/**
 * Constructor. Creates a new VHISTit Window.
 *
 * @param[in] argfile The argfile, which should be selected on
 * creation. If \a argfile is \c NULL, no file is selected.
 */
VHISTitWindow::VHISTitWindow(const QFileInfo* argfile)
    : m_jobRunning(false)
{
    // setup ui and slots and everything
    setupUi(this);
    initSignalSlots();

    // we want to handle any links ourselves
    m_outputBrowser->textBrowser()->setOpenExternalLinks(false);
    m_outputBrowser->textBrowser()->setOpenLinks(false);

    m_outputBrowser->setShortcutsActive(true);
    m_argfileBrowser->setShortcutsActive(true);
    m_argfileBrowser->setHighlighterEnabled(false);
    m_highlighter.setDocument(m_argfileBrowser->textBrowser()->document());
    
    initHelloMessage();
    setAcceptDrops(true);
    m_syncDirCheck->setChecked(true);
    
    // set argfile if one was selected
    if (argfile != NULL)
    {
        QString filepath = argfile->canonicalFilePath();
        setArgfile(filepath);
        emit argfileUpdated(filepath);
    }
    else
        // propagate argfile setting to all parts of the application
        emit argfileUpdated("");
}


VHISTitWindow::~VHISTitWindow()
{
}


/**
 * Initializes all static signal-slot connections of this window.
 */
void VHISTitWindow::initSignalSlots()
{
    // file and directory selection buttons
    connect(m_argfileButton, SIGNAL(clicked()), this, SLOT(selectFile()));
    connect(m_cwdButton, SIGNAL(clicked()), this, SLOT(selectCwd()));
    
    // filename changes
    connect(m_argfileEdit, SIGNAL(textChanged(const QString&)),
            this, SIGNAL(argfileUpdated(const QString&)));
    
    connect(this, SIGNAL(argfileUpdated(const QString&)),
            this, SLOT(autoAdjustCwd(const QString&)));
    connect(this, SIGNAL(argfileUpdated(const QString&)),
            this, SLOT(updateRunOption()));
    connect(this, SIGNAL(argfileUpdated(const QString&)),
            this, SLOT(showArgfile(const QString&)));
    connect(this, SIGNAL(argfileUpdated(const QString&)),
            this, SLOT(modifyTitle(const QString&)));
    
    // cwd changes
    connect(m_syncDirCheck, SIGNAL(toggled(bool)),
            this, SLOT(disableCwdUi(bool)));
    
    // text edit field
    connect(&m_runner, SIGNAL(outputUpdated(const QString&)),
            this, SLOT(updateCommandlineOutput(const QString&)));
    
    // run and close buttons
    connect(m_runButton, SIGNAL(clicked()), this, SLOT(runVhistadd()));
    connect(m_editButton, SIGNAL(clicked()), this, SLOT(editArgFile()));
    
    // vhistadd connections
    connect(&m_runner, SIGNAL(vhistaddFinished()), this, SLOT(finishJob()));
    
    // menu actions
    // window creation entries and close/quit
    initSharedMenu("Vhistit/RecentFiles");
    
    // Recent files entries
    connect(this, SIGNAL(recentFileSelected(QString)), this,
           SLOT(setArgfile(QString)));

    // other menu actions
    connect(m_openAction, SIGNAL(triggered(bool)), this, SLOT(selectFile()));
    connect(m_runAction, SIGNAL(triggered(bool)), this, SLOT(runVhistadd()));
    connect(m_examplesAction, SIGNAL(triggered(bool)), 
            this, SLOT(installExamples()));
    connect(m_helpAction, SIGNAL(triggered(bool)), this, SLOT(showHelp()));

    // !sv HTML browser, receive notifications about selected links
    connect(m_outputBrowser, SIGNAL(anchorClicked(const QUrl&)), this,
            SLOT(anchorClicked(const QUrl&)));
    connect(m_outputBrowser, SIGNAL(sourceChanged(const QUrl&)), this,
            SLOT(anchorClicked(const QUrl&)));
}


/**
 * Enables dragging of files onto the dialog.
 * 
 * @param event The drag event generated by Qt.
 */
void VHISTitWindow::dragEnterEvent(QDragEnterEvent* event)
{
    event->acceptProposedAction();
}


/**
 * Sets the argfile in case that a url is dragged onto the
 * dialog.
 *
 * @param event The drop event generated by Qt.
 */
void VHISTitWindow::dropEvent(QDropEvent* event)
{
    const QMimeData* mime = event->mimeData();
    if (mime->hasUrls())
    {
        QList<QUrl> urlList = mime->urls(); 
        if (!urlList.empty())
            setArgfile(urlList.at(0).toLocalFile());
    }
}

void VHISTitWindow::closeEvent(QCloseEvent* event)
{
    if (isJobRunning())
    {
        QMessageBox::StandardButton b = QMessageBox::question(this,
                tr("vhistadd is running"), tr("A vhistadd job is currently "
                "running. Closing this window will kill the job. Do you "
                "really want to close the window?"),
                QMessageBox::Yes | QMessageBox::No);

        if (b == QMessageBox::Yes)
            event->accept();
        else
            event->ignore();
    }
    else
        event->accept();

}


void VHISTitWindow::showArgfile(const QString& text)
{
    m_argfileBrowser->textBrowser()->setText("");
    QFile f(text);
    if (text != "" && f.open(QIODevice::ReadOnly | QIODevice::Text))
    {
        // avoid reading large files (since they probably are no cmd files
        if (f.size() < 500 * 1024)
        {
            QString data = QString::fromUtf8(f.readAll());
            m_argfileBrowser->textBrowser()->setText(data);
        }
        else
        {
            QString errMessage(tr("Probably no Argfile"
                    "- File too large: %1 MB"));
            errMessage = errMessage.arg(float(f.size()) / (1024*1024));
            m_argfileBrowser->textBrowser()->setText(errMessage);
        }
        f.close();
    }
}


void VHISTitWindow::editArgFile()
{
    // TODO: add right call
    //createVHISTezWindow(getArgfile());
    WindowList* wlist = WindowList::globalInstance();
    wlist->createWindow(MainWindow::windowName(), m_argfileEdit->text());
}


/**
 * Executes vhistadd with the given vhistfile.
 */
void VHISTitWindow::runVhistadd()
{
    m_infoTabs->setCurrentIndex(0);
    m_outputBrowser->textBrowser()->clear();
    QString vhistadd = "/usr/local/bin/vhistadd";
    QString argfile = m_argfileEdit->text();
    QString cwd = m_cwdEdit->text();

    addRecentFile(argfile);
    startJob();
    m_runner.runVhistadd(argfile, cwd);
}


void VHISTitWindow::startJob()
{
    setJobRunning(true);
    statusBar()->showMessage(tr("Job running..."));
}


void VHISTitWindow::finishJob()
{
    setJobRunning(false);
    statusBar()->showMessage(tr("job finished"));
}


void VHISTitWindow::updateCommandlineOutput(const QString& output)
{
    m_outputBrowser->textBrowser()->setHtml(output);
    QScrollBar* bar = m_outputBrowser->textBrowser()->verticalScrollBar();
    if (bar != NULL)
        bar->setValue(bar->maximum());
}


/**
 * Disables or enables the user interface for current working directory
 * selection. The content of the cwd ui is updated automatically.
 * 
 * @param[in] disabled If \c true, the UI is disabled, else it is enabled.
 */
void VHISTitWindow::disableCwdUi(bool disabled)
{
    m_cwdEdit->setEnabled(!disabled);
    m_cwdButton->setEnabled(!disabled);
    // if cwd selection is disabled, the cwd is choosen automatically
    if (disabled)
        autoAdjustCwd(m_argfileEdit->text());
}


void VHISTitWindow::selectFile()
{
    QString filter(tr("VHIST Argfile (*.vcmd);;All files (*)"));
    QString caption(tr("select a VHIST Argfile..."));
    
    QString result = QFileDialog::getOpenFileName(this, caption,
            getSelectionDirectory(), filter);
    
    if (result != "")
        setArgfile(result);
}


void VHISTitWindow::selectCwd()
{
    QString caption(tr("select a working directory..."));
    QString result = QFileDialog::getExistingDirectory(this, caption);
    if (result != "")
        setCwd(result);
}


void VHISTitWindow::autoAdjustCwd(const QString& filename)
{
    if (m_syncDirCheck->isChecked())
    {
        QFileInfo cfile(filename);
        QString dirname = cfile.path();
        setCwd(dirname);
    }
}


void VHISTitWindow::setArgfile(const QString& filename)
{
    setFilename(filename);
    setSelectionDirectory(filename);
}


void VHISTitWindow::setCwd(const QString& dirname)
{
    m_cwdEdit->setText(dirname);
    m_cwdEdit->setToolTip(dirname);
}


void VHISTitWindow::modifyTitle(const QString& argfile)
{
    QFileInfo info(argfile);
    setWindowTitle(tr("Run VHIST Arg File - ") + info.fileName());
    updateWindowMenu();
}


bool VHISTitWindow::isJobRunning()
{
    return m_jobRunning;
}


void VHISTitWindow::setJobRunning(bool running)
{
    m_jobRunning = running;
    updateRunOption();
}


QString VHISTitWindow::getArgfile()
{
    return m_argfileEdit->text();
}


void VHISTitWindow::updateRunOption()
{
    bool enabled = !isJobRunning() && (getArgfile() != "");
    m_runAction->setEnabled(enabled);
    m_runButton->setEnabled(enabled);
    m_editButton->setEnabled(enabled);
}


void VHISTitWindow::showHelp()
{
    HelpViewer::globalOpenPage("vhistit.html", QString());
}


void VHISTitWindow::setFilename(const QString& filename)
{
    m_argfileEdit->setText(filename);
    m_argfileEdit->setToolTip(filename);
}


void VHISTitWindow::installExamples()
{
    QString infoMessage("In the following dialog, select a directory into "
            "which the examples should be installed. A directory "
            "\"examples\" will be created inside the selected directory and "
            "will contain all the example files. Afterwards, a dialog for "
            "selecting one of the examples will be opened.");
    
    QMessageBox::information(this, "How to install examples",
            infoMessage);
    
    QString dirStr = QFileDialog::getExistingDirectory(this,
            tr("Select a directory into which the examples should"
            "be installed"), "", QFileDialog::ShowDirsOnly |
            QFileDialog::DontResolveSymlinks);
    
    if (dirStr != "")
    {
        QDir srcDir(getExamplesDir());
        QDir destDir(dirStr);
        QString fullDest = destDir.absoluteFilePath("examples");
        
        bool success = true;
        if (destDir.exists("examples"))
        {
            QString question = tr("The selected directory (%1) does already "
                    "exist. Do you want to overwrite it?").arg(fullDest);
            QMessageBox::StandardButtons buttons = QMessageBox::Yes |
                    QMessageBox::No;
            QMessageBox::StandardButton selectedButton = 
                    QMessageBox::question(this,
                    tr("Directory already exists"), question, buttons);
            if (selectedButton != QMessageBox::Yes)
                success = false;
            else
                success = removeDirectory(fullDest);
        }

        if (!success)
            return;
        
        success = destDir.mkdir("examples") && destDir.cd("examples");
        if (!success)
        {
            QString errorMsg = tr("Could not create directory (%1)");
            errorMsg = errorMsg.arg(fullDest);
            QMessageBox::critical(this, tr("Could not create directory"),
                    errorMsg);
            return;
        }
        
        success = copyDirectory(srcDir, destDir);
        
        if (!success)
        {
            QString errorMsg = tr("Could not copy directory (%1) to (%2)");
            errorMsg = errorMsg.arg(srcDir.absolutePath()).
                    arg(destDir.absolutePath());
            QMessageBox::critical(this, tr("Could not copy directory"),
                    errorMsg);
            return;
        }
        
        QString infoMessage("You can now select an example in the following "
                "dialog. The output of the examples will be in this "
                "examples folder, as well.");
        
        QMessageBox::information(this, "How to run examples",
                infoMessage);
        
        setSelectionDirectory(destDir.absolutePath());
        selectFile();
    }
}


QString VHISTitWindow::getSelectionDirectory()
{
    return m_selectionDirectory;
}


void VHISTitWindow::setSelectionDirectory(const QString& dir)
{
    m_selectionDirectory = dir;
}


void VHISTitWindow::initHelloMessage()
{
//     QString message(tr("This is %1 /> <br /> To install and run the "
//             "examples, select the \"Install Examples\" option in the "
//             "\"Extras\" menu. After selecting one example, press the "
//             "\"Run\" button."));

    QString message(tr(
                       "<table>"
                       "<tr class=\"simpletable\">"
                       "<td class=\"simpletable\" vertical-align=\"middle\">"
                       "<a href=\"http://www.nf.mpg.de/vhist\" >"
                       "<img border=\"0\" width=\"64\" height=\"64\" src=\":/resources/vhistzard1-64.png\"/>"
                       "</a> &nbsp;"
                       "</td>"
                       "<td class=\"simpletable\" vertical-align=\"middle\">"
                       "%1. <br />"
                       "<small>Copyright (c) %2</small><br/>"
                       "<small>Max Planck Institute for Neurological Research</small><br/>"
                       "<small>Cologne, Germany &nbsp;"
					   "<a href=\"http://www.nf.mpg.de/vhist\">http://www.nf.mpg.de/vhist</a></small><br/>"
                       "<small>Email: <a href=\"mailto:vhist@nf.mpg.de\">vhist@nf.mpg.de</a>"
                       "</td>"
                       "</tr>"
                       "</table>"
                       "<br /> <br />"
                       "To install and run the "
                       "<b>examples</b>, select the \"Install Examples\" option in the "
                       "\"Extras\" menu. After selecting one example, press the "
                       "\"Run\" button."
                       "<br /> <br />"
                       "Please see the "
                       "<a href=\"doc://theindex.html\" >"
                       "<b>Local Help</b>"
                       "</a>"
                       " (F1) for further "
                       "information, also for "
                       "<a href=\"doc://disclaimer.html\" >"
                       "<b>DISCLAIMER</b>"
                       "</a>"
                       " and "
                       "<a href=\"doc://license.html\" >"
                       "<b>license conditions</b>"
                       "</a>"
                       "."
                       ));
    message = message.arg(VHISTZARD_FULLNAME).arg(VHISTZARD_COPYRIGHT_DATE);
    m_outputBrowser->textBrowser()->setHtml(message);
    m_outputBrowser->textBrowser()->show();
}

void VHISTitWindow::anchorClicked (const QUrl & link)
{
    QString szLink = link.toString();
    if (szLink.startsWith(tr("doc://")))
    {
        // it is a "pseudo-URL" referring to our documentation
        szLink.remove(0, 6);
        QString szSec;
        int iHash = szLink.indexOf(tr("#"));
        if (iHash > 0)
        {
            szSec = szLink.mid(iHash + 1, szLink.size() - iHash - 1);
            szLink.remove(iHash, szLink.size() - iHash);
        }

        HelpViewer::globalOpenPage(szLink, szSec);
        //qDebug() << szLink.toUtf8();
        //qDebug() << szSec.toUtf8();
    }
    else
        QDesktopServices::openUrl(link);
}


ManagedWindow* VHISTitWindow::createVHISTitWindow(
        const QString& filename,
        const QVariantHash&)
{
    QFileInfo finfo(filename);
    ManagedWindow* win = new VHISTitWindow(&finfo);
    return win;
}


QString VHISTitWindow::windowName()
{
    return "Run VHIST Arg-File";
}


QString VHISTitWindow::windowShortcut()
{
    return "ctrl+R";
}

