#!/usr/bin/env python

"""

This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This script is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: setup.py 2303 2011-07-11 12:49:15Z ahuesgen $
"""

import shutil
import sys, os
import platform

UNIX_INSTALL_DIR_SYSTEM = "/usr/local/vhist/vhist-183"
UNIX_INSTALL_DIR_USER = "~/vhist/vhist-183"
UNIX_SYMLINK_DIR_SYSTEM = "/usr/local/bin"
UNIX_SYMLINK_DIR_USER = "~/vhist/bin"
SHORT_VERSION = "183"
LONG_VERSION = "VHIST 1.83.0, Jun 21 2013"

def findDefaultDir(SYS_DIR, USER_DIR):
    curDir = SYS_DIR
    continueSearch = True
    while curDir != "/" and continueSearch:
        if os.access(curDir, os.R_OK | os.W_OK):
            return SYS_DIR
        else:
            if not os.path.exists(curDir):
                curDir = os.path.abspath(os.path.join(curDir, ".."))
            else:
                continueSearch = False
    return USER_DIR


def removeAny(pathname):
    """
    Removes the given path regardless of type (directory or file)
    """
    if os.path.isdir(pathname):
        shutil.rmtree(pathname)
    elif os.path.isfile(pathname):
        os.remove(pathname)

def removeIfExists(filename):
    if os.path.exists(filename):
        os.remove(filename)

def askYesNo(question):
    question = question + " [y/n]\n"
    validAnswer = False
    while not validAnswer:
        answer = raw_input(question)
        if answer.lower().startswith("y"):
            return True
        elif answer.lower().startswith("n"):
            return False
        else:
            print "invalid answer"


def isUnixSystem():
    return os.name in ["posix", "mac"]


def copyOnUnix(oldDir, newDir):
    import stat
    print "copying VHIST files from %s to %s..." % (oldDir, newDir)
    allowCopy = True
    if os.path.exists(newDir):
        allowCopy = askYesNo(("\n%s does already exist. Do you want\nto "
            "overwrite it") % newDir)
        if allowCopy:
            removeAny(newDir)
    if allowCopy:
        newDirRoot = os.path.abspath(os.path.join(newDir, ".."))
        try:
            os.makedirs(newDirRoot, 0755)
        except Exception, E:
            pass

        # copytree will create missing directories (ah: not on python2.4...)
        shutil.copytree(oldDir, newDir)
        join = os.path.join
        vhistadd = join(newDir, "bin/vhistadd.py")
        vhistxl = join(newDir, "bin/vhistxl.py")
        vhistxs = join(newDir, "bin/vhistxs.py")
        flags = (stat.S_IRWXU | stat.S_IRGRP | stat.S_IROTH | stat.S_IXGRP | 
                stat.S_IXOTH)
        # set flags of executables
        os.chmod(vhistadd, flags)
        os.chmod(vhistxl, flags)
        os.chmod(vhistxs, flags)
    else:
        print "installation terminated"
        sys.exit(0)

def createSymlinksOnUnix(appDir, linkDir, makeDefault):
    print "creating symlinks into directory %s..." % linkDir
    join = os.path.join
    # pathes to applications
    vhistaddApp = join(appDir, "bin/vhistadd.py")
    vhistxlApp = join(appDir, "bin/vhistxl.py")
    vhistxsApp = join(appDir, "bin/vhistxs.py")
    vhistzardApp = join(appDir, "bin/vhistzard")
    # pathes to symlinks containing version numbers
    vhistaddLink = join(linkDir, "vhistadd-%s" % SHORT_VERSION)
    vhistxlLink = join(linkDir, "vhistxl-%s" % SHORT_VERSION)
    vhistxsLink = join(linkDir, "vhistxs-%s" % SHORT_VERSION)
    vhistzardLink = join(linkDir, "vhistzard-%s" % SHORT_VERSION)
    # pathes to default symlinks
    vhistaddDefault = join(linkDir, "vhistadd")
    vhistxlDefault = join(linkDir, "vhistxl")
    vhistxsDefault = join(linkDir, "vhistxs")
    vhistzardDefault = join(linkDir, "vhistzard")
    
    try:
        os.makedirs(linkDir, 0755)
    except Exception, E:
        pass
        
    
    # make links with version numbers
    removeIfExists(vhistaddLink)
    os.symlink(vhistaddApp, vhistaddLink)
    removeIfExists(vhistxlLink)
    os.symlink(vhistxlApp, vhistxlLink)
    removeIfExists(vhistxsLink)
    os.symlink(vhistxsApp, vhistxsLink)
    if os.path.exists(vhistzardApp):
        removeIfExists(vhistzardLink)
        os.symlink(vhistzardApp, vhistzardLink)
    
    
    if makeDefault:
        print "making this version of vhist the default one..."
        # make default links
        removeIfExists(vhistaddDefault)
        os.symlink(vhistaddApp, vhistaddDefault)
        removeIfExists(vhistxlDefault)
        os.symlink(vhistxlApp, vhistxlDefault)
        removeIfExists(vhistxsDefault)
        os.symlink(vhistxsApp, vhistxsDefault)
        if os.path.exists(vhistzardApp):
            removeIfExists(vhistzardDefault)
            os.symlink(vhistzardApp, vhistzardDefault)


def installOnUnix(appDir):
    print "A unix system was detected..."
    print 
    print "In order to install VHIST for all users on the system, "
    print "root privileges are probably required. To install VHIST in "
    print "the current user's home directory, user privileges are "
    print "sufficient. Running VHIST does not require any special rights."
    print
    
    answer = askYesNo("\nDo you want to install VHIST into another directory?")
    
    if answer == True:
        defaultDir = findDefaultDir(UNIX_INSTALL_DIR_SYSTEM,
                UNIX_INSTALL_DIR_USER)
        installDir = raw_input("\nPlease insert the name of the directory, "
                "into which you want VHIST\nto be installed [default: "
                "%s]\n" % defaultDir)
        if installDir.strip() == "":
            installDir = defaultDir
        installDir = os.path.expanduser(installDir)
        copyOnUnix(appDir, installDir)
        #appDir = os.path.abspath(installDir)
    
    answer = askYesNo("\nDo you want to create symlinks to the VHIST "
            "executables?")
    if answer == True:
        defaultDir = findDefaultDir(UNIX_SYMLINK_DIR_SYSTEM,
                UNIX_SYMLINK_DIR_USER)
        linkDir = raw_input("\nPlease insert the name of the directory, "
                "in which the symlinks\nshould be located [default: %s]\n"
                % defaultDir)
        if linkDir.strip() == "":
            linkDir = defaultDir
        linkDir = os.path.expanduser(linkDir)
        #linkDir = os.path.abspath(linkDir)
        
        defaultVhist = askYesNo("\nDo you want this version of vhist to "
                "be the default one?")
        createSymlinksOnUnix(installDir, linkDir, defaultVhist)
    
    print "installation successful"


def isWindowsSystem():
    return os.name == "nt"

def installOnWindows():
    print "A Microsoft Windows system was detected..."
    print
    print "Windows is currently not supported"
    print "Please use the Windows Installer package from"
    print "http://www.nf.mpg.de/vhist instead"
    sys.exit(-1)

################################################################

appname = os.path.abspath(sys.argv[0])
appdir = os.path.split(appname)[0]

print "This is the installer of %s" % LONG_VERSION
try:
    if isUnixSystem():
        installOnUnix(appdir)
    elif isWindowsSystem():
        installOnWindows(appdir)
    else:
        print "unknown system type"
        sys.exit(-1)
except KeyboardInterrupt:
    print "setup terminated by user"
    sys.exit(-2)

except Exception, e:
    print "an error occured during installation: %s" % str(e)

